import os
import sys
from collections import namedtuple


_root = os.path.dirname(os.path.abspath(__file__))


class NB301:
    """Lightweight wrapper for NB301"""
    def __init__(self, repo_path=os.path.sep.join([_root, "nasbench301-v09"]),
                 perf_model_path=os.path.sep.join([_root, "nasbench301-v09", "nb_models", "xgb_v0.9"]),
                 runtime_model_path=os.path.sep.join([_root, "nasbench301-v09", "nb_models", "lgb_runtime_v0.9"]),
                 config_space_json_path=os.path.sep.join([_root, "nasbench301-v09", "configspace.json"]),
                 version="0.9", log_f=print):
        self.repo_path = repo_path
        self.perf_model_path = perf_model_path
        self.runtime_model_path = runtime_model_path
        self.config_space_json_path = config_space_json_path
        self.version = version
        self.log_f = log_f
        self.performance_model = None
        self.runtime_model = None
        self._build()

    def _build(self):
        sys.path.insert(0, self.repo_path)
        from nasbench301.api import load_ensemble
        self.log_f("Specified NB301 repo path: {}".format(self.repo_path))
        self.log_f("Specified NB301 perf model path: {}".format(self.perf_model_path))
        self.log_f("Specified NB301 runtime model path: {}".format(self.runtime_model_path))
        self.log_f("Specified NB301 config space json path: {}".format(self.config_space_json_path))
        self.log_f("Specified NB301 version: {}".format(self.version))
        self.log_f("Loading performance surrogate model for NB301...")
        self.performance_model = load_ensemble(self.perf_model_path, self.config_space_json_path)
        self.log_f("Loading runtime surrogate model for NB301...")
        self.runtime_model = load_ensemble(self.runtime_model_path, self.config_space_json_path)

    def run_example(self, with_noise=False):
        self.log_f("Creating NB301 example genotype...")
        Genotype = namedtuple('Genotype', 'normal normal_concat reduce reduce_concat')
        genotype_config = Genotype(
            normal=[('sep_conv_3x3', 0), ('sep_conv_3x3', 1), ('sep_conv_3x3', 0), ('sep_conv_3x3', 1),
                    ('sep_conv_3x3', 1), ('skip_connect', 0), ('skip_connect', 0), ('dil_conv_3x3', 2)],
            normal_concat=[2, 3, 4, 5],
            reduce=[('max_pool_3x3', 0), ('max_pool_3x3', 1), ('skip_connect', 2), ('max_pool_3x3', 1),
                    ('max_pool_3x3', 0), ('skip_connect', 2), ('skip_connect', 2), ('max_pool_3x3', 1)],
            reduce_concat=[2, 3, 4, 5]
        )
        prediction_genotype = self.performance_model.predict(config=genotype_config, representation="genotype",
                                                             with_noise=with_noise)
        runtime_genotype = self.runtime_model.predict(config=genotype_config, representation="genotype")
        self.log_f("Example genotype performance: %f, runtime %f" % (prediction_genotype, runtime_genotype))

    def get_geno_acc(self, genotype, with_noise=False):
        return self.performance_model.predict(config=genotype, representation="genotype",
                                              with_noise=with_noise)

    def get_geno_runtime(self, genotype):
        return self.runtime_model.predict(config=genotype, representation="genotype")

    def get_geno_acc_runtime(self, genotype, with_noise=False):
        return self.get_geno_acc(genotype, with_noise), self.get_geno_runtime(genotype)


if __name__ == "__main__":
    NB301().run_example()
    print("done")
