import copy
import torch
import numpy as np
from utils.model_utils import device, to_sorted_tensor, to_original_tensor, init_weights
from utils.nas_utils import hash_module, prune_self_edge_from_edge_list


class SimpleRNN(torch.nn.Module):

    def __init__(self, input_size, hidden_size,
                 n_layers=1, rnn_dir=2, dropout_prob=0.0, rnn_type="gru",
                 return_aggr_vector_only=False, return_output_vector_only=False):
        super(SimpleRNN, self).__init__()
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.n_layers = n_layers
        self.rnn_dir = rnn_dir
        self.rnn_type = rnn_type
        self.dropout_prob = dropout_prob
        self.return_aggr_vector_only = return_aggr_vector_only
        self.return_output_vector_only = return_output_vector_only
        if self.rnn_type.lower() == "lstm":
            self.rnn = torch.nn.LSTM(input_size=self.input_size,
                                     hidden_size=self.hidden_size,
                                     num_layers=self.n_layers,
                                     dropout=self.dropout_prob if self.n_layers > 1 else 0,
                                     batch_first=True,
                                     bidirectional=self.rnn_dir == 2)
        else:
            self.rnn = torch.nn.GRU(input_size=self.input_size,
                                    hidden_size=self.hidden_size,
                                    num_layers=self.n_layers,
                                    dropout=self.dropout_prob if self.n_layers > 1 else 0,
                                    batch_first=True,
                                    bidirectional=self.rnn_dir == 2)
        init_weights(self, base_model_type="rnn")

    def forward(self, embedded, hidden=None, input_seq_length_tsr=None, input_mask_tsr=None):
        if input_mask_tsr is not None and input_seq_length_tsr is None:
            input_seq_length_tsr = input_mask_tsr.sum(dim=-1).squeeze()
            if len(input_seq_length_tsr.shape) == 0:
                input_seq_length_tsr = input_seq_length_tsr.unsqueeze(0)
        self.rnn.flatten_parameters()
        if input_seq_length_tsr is None:
            outputs, hidden = self.rnn(embedded, hidden)
        else:
            embedded, sorted_lens, s_idx = to_sorted_tensor(embedded, input_seq_length_tsr, sort_dim=0)
            packed = torch.nn.utils.rnn.pack_padded_sequence(embedded, sorted_lens, batch_first=True)
            outputs, hidden = self.rnn(packed, hidden)
            outputs, sorted_lens = torch.nn.utils.rnn.pad_packed_sequence(outputs, batch_first=True)
            outputs = to_original_tensor(outputs, s_idx, sort_dim=0)

        if self.return_aggr_vector_only:
            if input_seq_length_tsr is not None:
                input_seq_length_tsr = input_seq_length_tsr.to(device())
                input_seq_length_tsr = input_seq_length_tsr - 1
                rv = torch.gather(outputs, 1, input_seq_length_tsr.view(-1, 1).unsqueeze(2).repeat(1, 1, outputs.size(-1)))
            else:
                rv = outputs
            return rv
        elif self.return_output_vector_only:
            return outputs
        else:
            return outputs, hidden


class SimpleBlock:

    def __init__(self, node_label_list, node_feature_tsr, edge_list):
        self.curr_node_label_list = node_label_list
        self.curr_node_feature_tsr = node_feature_tsr
        self.curr_edge_list = copy.deepcopy(edge_list)

    def __len__(self):
        return len(self.curr_node_label_list)

    def __str__(self):
        return "SimpleBlock{}".format(str(self.curr_node_label_list)) + ", Edges: " + str(self.curr_edge_list)

    def __repr__(self):
        return str(self)

    def __deepcopy__(self, memodict={}):
        _state = SimpleBlock(
            node_label_list=[v for v in self.curr_node_label_list],
            node_feature_tsr=copy.deepcopy(self.curr_node_feature_tsr),
            edge_list=copy.deepcopy(self.curr_edge_list),
        )
        return _state

    def to(self, dev):
        self.curr_node_feature_tsr = self.curr_node_feature_tsr.to(dev)
        return self

    def get_str_id(self):
        from utils.nas_utils import edge_list_to_edge_matrix
        if isinstance(self.curr_node_feature_tsr, torch.Tensor):
            node_indices = self.curr_node_feature_tsr.tolist()
        else:
            node_indices = self.curr_node_feature_tsr
        src_list, dst_list = prune_self_edge_from_edge_list(self.curr_edge_list)
        matrix = edge_list_to_edge_matrix([src_list, dst_list], len(node_indices))
        fingerprint = hash_module(matrix, node_indices)
        return fingerprint


class RNNIndexEncoder(torch.nn.Module):

    def __init__(self, n_unique_inds, input_size, hidden_size,
                 n_layers=1, rnn_dir=2, dropout_prob=0.0, rnn_type="gru",
                 return_aggr_vector_only=False, return_output_vector_only=False,
                 add_normal_prior=False):
        super(RNNIndexEncoder, self).__init__()
        self.pre_embedding = torch.nn.Embedding(n_unique_inds, input_size)
        self.rnn = SimpleRNN(input_size, hidden_size,
                             n_layers, rnn_dir=rnn_dir, dropout_prob=dropout_prob, rnn_type=rnn_type,
                             return_aggr_vector_only=return_aggr_vector_only,
                             return_output_vector_only=return_output_vector_only)
        self.add_normal_prior = add_normal_prior
        self.init_ff = torch.nn.Linear(input_size * 2, input_size)

    def add_prior(self, embedding):
        prior = np.random.normal(size=embedding.shape)
        prior = torch.from_numpy(prior).float().to(device())
        embedding = torch.cat([embedding, prior], dim=-1)
        return self.init_ff(embedding)

    def forward(self, x):
        embedding = self.pre_embedding(x.to(device()))
        if self.add_normal_prior and embedding.shape[1] == 1:
            embedding = self.add_prior(embedding)
        return self.rnn(embedding)


class MultiInputLinears(torch.nn.Module):

    def __init__(self, input_size_list, output_size, bias=True):
        super(MultiInputLinears, self).__init__()
        self.nets = torch.nn.ModuleList()
        self.output_size = output_size
        self.input_size_list = input_size_list
        self.size2idx = {}
        for size in input_size_list:
            assert size not in self.size2idx
            self.size2idx[size] = len(self.nets)
            self.nets.append(torch.nn.Linear(size, self.output_size, bias=bias))

    def forward(self, x):
        size = x.shape[-1]
        idx = self.size2idx[size]
        return self.nets[idx](x)

    def forward_all(self, x):
        return [net(x) for net in self.nets]


class MultiOutputLinears(torch.nn.Module):

    def __init__(self, input_size, output_size_list, bias=True):
        super(MultiOutputLinears, self).__init__()
        self.nets = torch.nn.ModuleList()
        self.input_size = input_size
        self.output_size_list = output_size_list
        for size in output_size_list:
            self.nets.append(torch.nn.Linear(self.input_size, size, bias=bias))

    def forward(self, x, idx):
        return self.nets[idx](x)

    def forward_all(self, x):
        return [net(x) for net in self.nets]

