import os
import re
import time
import copy
import random
import subprocess
from tqdm import tqdm
from model_nas_base import BlockPerfEvaluator
from utils.efficientnet_utils import *
from utils.model_utils import device, can_parallel, ExponentialMovingAverage
from utils.darts_utils import AverageMeter, accuracy
from utils.debug_utils import IntervalInfo as interval_info
from utils.misc_utils import rmdir


class EfficientNet(nn.Module):
    """
    EfficientNet model.
    References:
        [1] https://arxiv.org/abs/1905.11946 (EfficientNet)
    Example:
        >>> import torch
        >>> from model_efficientnet import EfficientNet
        >>> inputs = torch.rand(1, 3, 224, 224)
        >>> model = EfficientNet.from_pretrained('efficientnet-b0')
        >>> model.eval()
        >>> outputs = model(inputs)
    """

    def __init__(self, blocks_args, global_params):
        super().__init__()
        assert isinstance(blocks_args, list), 'blocks_args should be a list'
        assert len(blocks_args) > 0, 'block args must be greater than 0'
        self._global_params = global_params
        self._blocks_args = blocks_args

        # Batch norm parameters
        bn_mom = 1 - self._global_params.batch_norm_momentum
        bn_eps = self._global_params.batch_norm_epsilon

        # Get stem static or dynamic convolution depending on image size
        image_size = global_params.image_size
        Conv2d = get_same_padding_conv2d(image_size=image_size)

        # Stem
        in_channels = 3  # rgb
        out_channels = round_filters(32, self._global_params)  # number of output channels
        self._conv_stem = Conv2d(in_channels, out_channels, kernel_size=3, stride=2, bias=False)
        self._bn0 = nn.BatchNorm2d(num_features=out_channels, momentum=bn_mom, eps=bn_eps)
        image_size = calculate_output_image_size(image_size, 2)

        # Build blocks
        self._blocks = nn.ModuleList([])
        for block_args in self._blocks_args:

            # Update block input and output filters based on depth multiplier.
            block_args = block_args._replace(
                input_filters=round_filters(block_args.input_filters, self._global_params),
                output_filters=round_filters(block_args.output_filters, self._global_params),
                num_repeat=round_repeats(block_args.num_repeat, self._global_params)
            )

            # The first block needs to take care of stride and filter size increase.
            block = MBConvBlock(block_args.se_ratio, block_args.id_skip, block_args.input_filters,
                                block_args.expand_ratio, block_args.kernel_size, block_args.stride,
                                block_args.output_filters, global_params.batch_norm_momentum,
                                global_params.batch_norm_epsilon, image_size=image_size)
            self._blocks.append(block)
            image_size = calculate_output_image_size(image_size, block_args.stride)
            if block_args.num_repeat > 1:  # modify block_args to keep same output size
                block_args = block_args._replace(input_filters=block_args.output_filters, stride=1)
            for _ in range(block_args.num_repeat - 1):
                block = MBConvBlock(block_args.se_ratio, block_args.id_skip, block_args.input_filters,
                                    block_args.expand_ratio, block_args.kernel_size, block_args.stride,
                                    block_args.output_filters, global_params.batch_norm_momentum,
                                    global_params.batch_norm_epsilon, image_size=image_size)
                self._blocks.append(block)
                # image_size = calculate_output_image_size(image_size, block_args.stride)  # stride = 1

        # Head
        in_channels = block_args.output_filters  # output of final block
        out_channels = round_filters(1280, self._global_params)
        Conv2d = get_same_padding_conv2d(image_size=image_size)
        self._conv_head = Conv2d(in_channels, out_channels, kernel_size=1, bias=False)
        self._bn1 = nn.BatchNorm2d(num_features=out_channels, momentum=bn_mom, eps=bn_eps)

        # Final linear layer
        self._avg_pooling = nn.AdaptiveAvgPool2d(1)
        self._dropout = nn.Dropout(self._global_params.dropout_rate)
        self._fc = nn.Linear(out_channels, self._global_params.num_classes)
        self._swish = MemoryEfficientSwish()

    def set_swish(self, memory_efficient=True):
        """Sets swish function as memory efficient (for training) or standard (for export).
        Args:
            memory_efficient (bool): Whether to use memory-efficient version of swish.
        """
        self._swish = MemoryEfficientSwish() if memory_efficient else Swish()
        for block in self._blocks:
            block.set_swish(memory_efficient)

    def extract_endpoints(self, inputs):
        """Use convolution layer to extract features
        from reduction levels i in [1, 2, 3, 4, 5].
        Args:
            inputs (tensor): Input tensor.
        Returns:
            Dictionary of last intermediate features
            with reduction levels i in [1, 2, 3, 4, 5].
            Example:
                >>> import torch
                >>> from model_efficientnet import EfficientNet
                >>> inputs = torch.rand(1, 3, 224, 224)
                >>> model = EfficientNet.from_pretrained('efficientnet-b0')
                >>> endpoints = model.extract_endpoints(inputs)
                >>> print(endpoints['reduction_1'].shape)  # torch.Size([1, 16, 112, 112])
                >>> print(endpoints['reduction_2'].shape)  # torch.Size([1, 24, 56, 56])
                >>> print(endpoints['reduction_3'].shape)  # torch.Size([1, 40, 28, 28])
                >>> print(endpoints['reduction_4'].shape)  # torch.Size([1, 112, 14, 14])
                >>> print(endpoints['reduction_5'].shape)  # torch.Size([1, 1280, 7, 7])
        """
        endpoints = dict()

        # Stem
        x = self._swish(self._bn0(self._conv_stem(inputs)))
        prev_x = x

        # Blocks
        for idx, block in enumerate(self._blocks):
            drop_connect_rate = self._global_params.drop_connect_rate
            if drop_connect_rate:
                drop_connect_rate *= float(idx) / len(self._blocks)  # scale drop connect_rate
            x = block(x, drop_connect_rate=drop_connect_rate)
            if prev_x.size(2) > x.size(2):
                endpoints[f'reduction_{len(endpoints) + 1}'] = prev_x
            prev_x = x

        # Head
        x = self._swish(self._bn1(self._conv_head(x)))
        endpoints[f'reduction_{len(endpoints) + 1}'] = x

        return endpoints

    def extract_features(self, inputs):
        """use convolution layer to extract feature .
        Args:
            inputs (tensor): Input tensor.
        Returns:
            Output of the final convolution
            layer in the efficientnet model.
        """
        # Stem
        x = self._swish(self._bn0(self._conv_stem(inputs)))

        # Blocks
        for idx, block in enumerate(self._blocks):
            drop_connect_rate = self._global_params.drop_connect_rate
            if drop_connect_rate:
                drop_connect_rate *= float(idx) / len(self._blocks)  # scale drop connect_rate
            x = block(x, drop_connect_rate=drop_connect_rate)

        # Head
        x = self._swish(self._bn1(self._conv_head(x)))

        return x

    def forward(self, inputs):
        """EfficientNet's forward function.
           Calls extract_features to extract features, applies final linear layer, and returns logits.
        Args:
            inputs (tensor): Input tensor.
        Returns:
            Output of this model after processing.
        """
        # Convolution layers
        x = self.extract_features(inputs)

        # Pooling and final linear layer
        x = self._avg_pooling(x)
        x = x.flatten(start_dim=1)
        x = self._dropout(x)
        x = self._fc(x)

        return x

    @classmethod
    def from_name(cls, model_name, in_channels=3, **override_params):
        """create an efficientnet model according to name.
        Args:
            model_name (str): Name for efficientnet.
            in_channels (int): Input data's channel number.
            override_params (other key word params):
                Params to override model's global_params.
                Optional key:
                    'width_coefficient', 'depth_coefficient',
                    'image_size', 'dropout_rate',
                    'num_classes', 'batch_norm_momentum',
                    'batch_norm_epsilon', 'drop_connect_rate',
                    'depth_divisor', 'min_depth'
        Returns:
            An efficientnet model.
        """
        blocks_args, global_params = get_imagenet_params(model_name, override_params)
        model = cls(blocks_args, global_params)
        model._change_in_channels(in_channels)
        return model

    @classmethod
    def from_pretrained(cls, model_name, weights_path=None, advprop=False,
                        in_channels=3, num_classes=1000, **override_params):
        """create an efficientnet model according to name.
        Args:
            model_name (str): Name for efficientnet.
            weights_path (None or str):
                str: path to pretrained weights file on the local disk.
                None: use pretrained weights downloaded from the Internet.
            advprop (bool):
                Whether to load pretrained weights
                trained with advprop (valid when weights_path is None).
            in_channels (int): Input data's channel number.
            num_classes (int):
                Number of categories for classification.
                It controls the output size for final linear layer.
            override_params (other key word params):
                Params to override model's global_params.
                Optional key:
                    'width_coefficient', 'depth_coefficient',
                    'image_size', 'dropout_rate',
                    'num_classes', 'batch_norm_momentum',
                    'batch_norm_epsilon', 'drop_connect_rate',
                    'depth_divisor', 'min_depth'
        Returns:
            A pretrained efficientnet model.
        """
        model = cls.from_name(model_name, num_classes=num_classes, **override_params)
        load_pretrained_weights(model, model_name, weights_path=weights_path, load_fc=(num_classes == 1000),
                                advprop=advprop)
        model._change_in_channels(in_channels)
        return model

    @classmethod
    def get_image_size(cls, model_name):
        """Get the input image size for a given efficientnet model.
        Args:
            model_name (str): Name for efficientnet.
        Returns:
            Input image size (resolution).
        """
        _, _, res, _ = efficientnet_imagenet_params(model_name)
        return res

    def _change_in_channels(self, in_channels):
        """Adjust model's first convolution layer to in_channels, if in_channels not equals 3.
        Args:
            in_channels (int): Input data's channel number.
        """
        if in_channels != 3:
            Conv2d = get_same_padding_conv2d(image_size=self._global_params.image_size)
            out_channels = round_filters(32, self._global_params)
            self._conv_stem = Conv2d(in_channels, out_channels, kernel_size=3, stride=2, bias=False)


def run_efficientnet_epoch(model, loader, criterion, optimizer=None, max_grad_norm=5.0,
                           verbose=True, desc="Train"):
    model = model.to(device())
    avg_top_1_acc = AverageMeter()
    avg_top_5_acc = AverageMeter()
    bar = None
    if verbose:
        bar = tqdm(total=len(loader), desc=desc, ascii=True)
    start_time = time.time()
    for feature, target in loader:
        feature = feature.to(device())
        target = target.to(device())
        logits = model(feature)
        loss = criterion(logits, target)
        if optimizer is not None:
            optimizer.zero_grad()
            loss.backward()
            nn.utils.clip_grad_norm_(model.parameters(), max_grad_norm)
            optimizer.step()
        acc1, acc5 = accuracy(logits, target, topk=[1, 5])
        avg_top_1_acc.update(acc1.item(), feature.shape[0])
        avg_top_5_acc.update(acc5.item(), feature.shape[0])
        if bar is not None:
            bar.update(1)
    lat = time.time() - start_time
    if bar is not None: bar.close()
    return avg_top_1_acc.avg, avg_top_5_acc.avg, lat


def epoch_train_efficientnet(n_epochs, model, train_loader, criterion, optimizer, book_keeper,
                             dev_loader=None, max_grad_norm=5.0, completed_epochs=0,
                             verbose=True, allow_parallel=False, checkpoint=True):
    model = model.to(device())
    criterion = criterion.to(device())
    for epoch in range(n_epochs):
        report_epoch = epoch + completed_epochs + 1
        model.train()
        if allow_parallel and can_parallel():
            model = torch.nn.DataParallel(model)
        run_efficientnet_epoch(model, train_loader, criterion, optimizer=optimizer, max_grad_norm=max_grad_norm,
                               verbose=verbose)
        if isinstance(model, torch.nn.DataParallel):
            model = model.module
        if hasattr(optimizer, "update"):
            optimizer.update()
        if hasattr(optimizer, "schedule_lr"):
            optimizer.schedule_lr(epoch)
        if checkpoint:
            book_keeper.checkpoint_model("_latest.pt", report_epoch, model, optimizer)
        if dev_loader is not None:
            with torch.no_grad():
                model.eval()
                dev_top_1_acc, _, _ = run_efficientnet_epoch(model, dev_loader, criterion, optimizer=None,
                                                             max_grad_norm=max_grad_norm, verbose=verbose)
                if checkpoint:
                    book_keeper.checkpoint_model("_best.pt", report_epoch, model, optimizer,
                                                 eval_perf=dev_top_1_acc)
        book_keeper.log("")


def step_train_efficientnet(n_steps, model, train_loader, criterion, optimizer, book_keeper,
                            dev_loader=None, max_grad_norm=5.0, ema=None,
                            allow_parallel=True, checkpoint=True):
    model = model.to(device())
    criterion = criterion.to(device())
    n_curr_steps = 0
    bar = tqdm(total=n_steps, desc="Train", ascii=True, leave=False)
    if allow_parallel and can_parallel():
        model = torch.nn.DataParallel(model)
    model.train()

    def _dev():
        book_keeper.log("Dev at step: {}".format(n_curr_steps))
        if isinstance(model, torch.nn.DataParallel):
            dev_model = model.module
        else:
            dev_model = model
        with torch.no_grad():
            dev_model.eval()
            dev_top_1_acc, dev_top_5_acc, _ = run_efficientnet_epoch(dev_model, dev_loader, criterion, optimizer=None,
                                                                     max_grad_norm=max_grad_norm, desc="Dev")
            book_keeper.log("Dev top 1 acc: {}".format(dev_top_1_acc))
            book_keeper.log("Dev top 5 acc: {}".format(dev_top_5_acc))
            if checkpoint:
                book_keeper.checkpoint_model("_best.pt", n_curr_steps, dev_model, optimizer,
                                             eval_perf=dev_top_1_acc)
            book_keeper.report_curr_best()
            book_keeper.log("")
        model.train()
        return dev_top_1_acc

    while n_curr_steps < n_steps:
        avg_top_1_acc = AverageMeter()
        avg_loss= AverageMeter()
        for feature, target in train_loader:
            if n_curr_steps >= n_steps:
                break
            feature = feature.to(device())
            target = target.to(device())
            logits = model(feature)
            loss = criterion(logits, target)
            optimizer.zero_grad()
            loss.backward()
            nn.utils.clip_grad_norm_(model.parameters(), max_grad_norm)
            optimizer.step()
            if ema is not None:
                ema.update(model.parameters())
            acc1 = accuracy(logits, target, topk=[1])[0]
            avg_top_1_acc.update(acc1.item(), feature.shape[0])
            avg_loss.update(loss.item(), 1)
            n_curr_steps += 1
            bar.desc = "Train, loss: {:.3f}, acc: {:.2f}".format(avg_loss.avg, avg_top_1_acc.avg)
            bar.update(1)
            # Optimizer update
            if hasattr(optimizer, "update"):
                optimizer.update()
        # Train checkpoint
        if checkpoint:
            bar.close()
            if isinstance(model, torch.nn.DataParallel):
                chkpt_model = model.module
            else:
                chkpt_model = model
            book_keeper.checkpoint_model("_latest.pt", n_curr_steps, chkpt_model, optimizer)
            bar = tqdm(total=n_steps, desc="Train, loss: {:.3f}, acc: {:.2f}".format(avg_loss.avg, avg_top_1_acc.avg),
                       ascii=True, leave=False)
            bar.update(n_curr_steps)
        # Dev
        if dev_loader is not None:
            bar.close()
            book_keeper.log("Train avg acc at step {}: {}".format(n_curr_steps, avg_top_1_acc.avg))
            if hasattr(optimizer, "get_lr"):
                book_keeper.log("Optimizer lr at step {}: {}".format(n_curr_steps, optimizer.get_lr()))
            _dev()
            bar = tqdm(total=n_steps, desc="Train, loss: {:.3f}, acc: {:.2f}".format(avg_loss.avg, avg_top_1_acc.avg),
                       ascii=True, leave=False)
            bar.update(n_curr_steps)
    bar.close()
    return _dev()


def make_sp_searched_efficientnet(k_e_list, n_classes):
    if n_classes >= 1000:
        blocks_args, global_params = get_sp_searchable_imagenet_params(k_e_list, num_classes=n_classes)
    else:
        blocks_args, global_params = get_sp_searchable_cifar_params(k_e_list, num_classes=n_classes)
    model = EfficientNet(blocks_args, global_params)
    return model


def build_k_e_pairs(node_labels):
    k_e_list = []
    for label in  node_labels:
        assert "MBConv" in label
        label_list = [s for s in label.split("-") if len(s.strip()) > 0]
        header, e_str, k_str = label_list
        e_val = int(e_str.replace("e", ""))
        k_val = int(k_str.replace("k", ""))
        k_e_list.append((k_val, e_val))
    return k_e_list


def get_sp_searched_eb0_imagenet_n_params(node_labels, n_classes=1000):
    k_e_pairs = build_k_e_pairs(node_labels)
    model = make_sp_searched_efficientnet(k_e_pairs, n_classes)
    n_params = sum(p.numel() for p in model.parameters() if p.requires_grad)
    return n_params


def get_sp_searched_eb0_cifar10_n_params(node_labels, n_classes=10):
    k_e_pairs = build_k_e_pairs(node_labels)
    model = make_sp_searched_efficientnet(k_e_pairs, n_classes)
    n_params = sum(p.numel() for p in model.parameters() if p.requires_grad)
    return n_params


class EB0SPPTEvaluator(BlockPerfEvaluator):

    def __init__(self, model_maker, optim_maker, criterion, train_loader, test_loader,
                 lat_predictor=None, allow_parallel=True, steps=10, default_clue=1.0,
                 enable_perf_memo=True):
        super(EB0SPPTEvaluator, self).__init__(default_clue=default_clue, enable_perf_memo=enable_perf_memo)
        self.steps = steps
        self.model_maker = model_maker
        self.optim_maker = optim_maker
        self.criterion = criterion.to(device())
        self.allow_parallel = allow_parallel
        self.train_loader = train_loader
        self.test_loader = test_loader
        self.lat_predictor = lat_predictor

    def get_perf_reward(self, state):
        raise NotImplementedError

    def get_perf_values(self, block):
        key = block.get_str_id()
        if self.enable_perf_memo and key in self.perf_memo:
            return self.perf_memo[key]
        labels = block.curr_node_label_list[1:-1] # Chop input and output labels
        model = self.model_maker(labels).to(device())
        optimizer = self.optim_maker(model)
        ema = ExponentialMovingAverage(model.parameters(), decay=0.9999)
        pred_acc = self._step_wise_train_test(model, optimizer, ema)
        pred_acc /= 100.
        pred_lat = self.lat_predictor(block) if self.lat_predictor is not None else 1.0
        self.perf_memo[key] = (pred_acc, pred_lat)
        interval_info.num_block_eval += 1
        return pred_acc, pred_lat

    def _step_wise_train_test(self, model, optimizer, ema):
        model = model.to(device())
        n_curr_steps = 0
        if self.allow_parallel and can_parallel():
            model = torch.nn.DataParallel(model)
        model.train()

        def _dev(_loader):
            if isinstance(model, torch.nn.DataParallel):
                dev_model = model.module
            else:
                dev_model = model
            with torch.no_grad():
                dev_model.eval()
                dev_top_1_acc, _, _ = run_efficientnet_epoch(dev_model, _loader, self.criterion,
                                                             optimizer=None, desc="Dev", verbose=False)
            model.train()
            return dev_top_1_acc

        while n_curr_steps < self.steps:
            for feature, target in self.train_loader:
                if n_curr_steps >= self.steps:
                    break
                feature = feature.to(device())
                target = target.to(device())
                logits = model(feature)
                loss = self.criterion(logits, target)
                optimizer.zero_grad()
                loss.backward()
                nn.utils.clip_grad_norm_(model.parameters(), 5.0)
                optimizer.step()
                if ema is not None:
                    ema.update(model.parameters())
                n_curr_steps += 1
                # Optimizer update
                if hasattr(optimizer, "update"):
                    optimizer.update()

        return _dev(self.test_loader)


class SuperMBConvBlock(nn.Module):
    """Supports varying K and E"""
    def __init__(self, k_e_pairs,
                 se_ratio, id_skip, input_filters, stride, output_filters,
                 batch_norm_momentum, batch_norm_epsilon, image_size=None):
        super().__init__()
        self.se_ratio = se_ratio
        self.id_skip = id_skip
        self.input_filters = input_filters
        self.k_e_pairs = k_e_pairs
        self.stride = stride
        self.output_filters = output_filters
        self._bn_mom = batch_norm_momentum
        self._bn_eps = batch_norm_epsilon
        self.has_se = (self.se_ratio is not None) and (0 < self.se_ratio <= 1)
        self.id_skip = id_skip  # whether to use skip connection and drop connect
        self.op_label2idx = {}
        self.ops = nn.ModuleList()
        for kernel_size, expand_ratio in self.k_e_pairs:
            block = MBConvBlock(self.se_ratio, self.id_skip, self.input_filters, expand_ratio,
                                kernel_size, self.stride, self.output_filters, self._bn_mom, self._bn_eps,
                                image_size=image_size)
            self.ops.append(block)
            self.op_label2idx[self.get_op_label(kernel_size, expand_ratio)] = len(self.ops) - 1

    @staticmethod
    def get_op_label(k, e):
        return "MBConv-e{}-k{}".format(e, k)

    def forward(self, inputs, k_e_pair, drop_connect_rate=None):
        k, e = k_e_pair
        op_label = self.get_op_label(k, e)
        op = self.ops[self.op_label2idx[op_label]]
        x = op(inputs, drop_connect_rate=drop_connect_rate)
        return x


class EB0SPSupernet(nn.Module):

    def __init__(self, k_e_pairs, blocks_args, global_params):
        super().__init__()
        assert isinstance(blocks_args, list), 'blocks_args should be a list'
        assert len(blocks_args) > 0, 'block args must be greater than 0'
        self._global_params = global_params
        self._blocks_args = blocks_args

        # Batch norm parameters
        bn_mom = 1 - self._global_params.batch_norm_momentum
        bn_eps = self._global_params.batch_norm_epsilon

        # Get stem static or dynamic convolution depending on image size
        image_size = global_params.image_size
        Conv2d = get_same_padding_conv2d(image_size=image_size)

        # Stem
        in_channels = 3  # rgb
        out_channels = round_filters(32, self._global_params)  # number of output channels
        self._conv_stem = Conv2d(in_channels, out_channels, kernel_size=3, stride=2, bias=False)
        self._bn0 = nn.BatchNorm2d(num_features=out_channels, momentum=bn_mom, eps=bn_eps)
        image_size = calculate_output_image_size(image_size, 2)

        # Build blocks
        self._blocks = nn.ModuleList([])
        for block_args in self._blocks_args:

            # Update block input and output filters based on depth multiplier.
            block_args = block_args._replace(
                input_filters=round_filters(block_args.input_filters, self._global_params),
                output_filters=round_filters(block_args.output_filters, self._global_params),
                num_repeat=round_repeats(block_args.num_repeat, self._global_params)
            )

            # The first block needs to take care of stride and filter size increase.
            block = SuperMBConvBlock(k_e_pairs,
                                     block_args.se_ratio, block_args.id_skip, block_args.input_filters,
                                     block_args.stride, block_args.output_filters, global_params.batch_norm_momentum,
                                     global_params.batch_norm_epsilon, image_size=image_size)
            self._blocks.append(block)
            image_size = calculate_output_image_size(image_size, block_args.stride)
            if block_args.num_repeat > 1:  # modify block_args to keep same output size
                block_args = block_args._replace(input_filters=block_args.output_filters, stride=1)
            for _ in range(block_args.num_repeat - 1):
                block = SuperMBConvBlock(k_e_pairs,
                                         block_args.se_ratio, block_args.id_skip, block_args.input_filters,
                                         block_args.stride,
                                         block_args.output_filters, global_params.batch_norm_momentum,
                                         global_params.batch_norm_epsilon, image_size=image_size)
                self._blocks.append(block)
                # image_size = calculate_output_image_size(image_size, block_args.stride)  # stride = 1

        # Head
        in_channels = block_args.output_filters  # output of final block
        out_channels = round_filters(1280, self._global_params)
        Conv2d = get_same_padding_conv2d(image_size=image_size)
        self._conv_head = Conv2d(in_channels, out_channels, kernel_size=1, bias=False)
        self._bn1 = nn.BatchNorm2d(num_features=out_channels, momentum=bn_mom, eps=bn_eps)

        # Final linear layer
        self._avg_pooling = nn.AdaptiveAvgPool2d(1)
        self._dropout = nn.Dropout(self._global_params.dropout_rate)
        self._fc = nn.Linear(out_channels, self._global_params.num_classes)
        self._swish = MemoryEfficientSwish()

    def set_swish(self, memory_efficient=True):
        """Sets swish function as memory efficient (for training) or standard (for export).
        Args:
            memory_efficient (bool): Whether to use memory-efficient version of swish.
        """
        self._swish = MemoryEfficientSwish() if memory_efficient else Swish()
        for block in self._blocks:
            block.set_swish(memory_efficient)

    def extract_features(self, inputs, k_e_pairs):
        # Stem
        x = self._swish(self._bn0(self._conv_stem(inputs)))

        # Blocks
        for idx, block in enumerate(self._blocks):
            drop_connect_rate = self._global_params.drop_connect_rate
            if drop_connect_rate:
                drop_connect_rate *= float(idx) / len(self._blocks)  # scale drop connect_rate
            x = block(x, k_e_pairs[idx], drop_connect_rate)

        # Head
        x = self._swish(self._bn1(self._conv_head(x)))

        return x

    def forward(self, inputs, k_e_pairs):
        assert len(k_e_pairs) == len(self._blocks), "k_e_pairs len: {}, self._blocks_len: {}".format(len(k_e_pairs),
                                                                                                     len(self._blocks))
        # Convolution layers
        x = self.extract_features(inputs, k_e_pairs)

        # Pooling and final linear layer
        x = self._avg_pooling(x)
        x = x.flatten(start_dim=1)
        x = self._dropout(x)
        x = self._fc(x)

        return x

    @classmethod
    def get_image_size(cls, model_name):
        """Get the input image size for a given efficientnet model.
        Args:
            model_name (str): Name for efficientnet.
        Returns:
            Input image size (resolution).
        """
        _, _, res, _ = efficientnet_imagenet_params(model_name)
        return res

    def _change_in_channels(self, in_channels):
        """Adjust model's first convolution layer to in_channels, if in_channels not equals 3.
        Args:
            in_channels (int): Input data's channel number.
        """
        if in_channels != 3:
            Conv2d = get_same_padding_conv2d(image_size=self._global_params.image_size)
            out_channels = round_filters(32, self._global_params)
            self._conv_stem = Conv2d(in_channels, out_channels, kernel_size=3, stride=2, bias=False)


def eb0_sp_supernet_complete_params(width_coefficient=None, depth_coefficient=None, image_size=None,
                                    dropout_rate=0.2, drop_connect_rate=0.2, num_classes=1000,
                                    blocks_args_template=None):

    if blocks_args_template is None:
        blocks_args_template = (
            'r1_k9_s11_e9_i32_o16_se0.25',  # NOTE: value for k and e here does not matter
            'r2_k9_s22_e9_i16_o24_se0.25',
            'r2_k9_s22_e9_i24_o40_se0.25',
            'r3_k9_s22_e9_i40_o80_se0.25',
            'r3_k9_s11_e9_i80_o112_se0.25',
            'r4_k9_s22_e9_i112_o192_se0.25',
            'r1_k9_s11_e9_i192_o320_se0.25',

            # 'r1_k9_s11_e9_i32_o16_se0',  # NOTE: value for k and e here does not matter
            # 'r2_k9_s22_e9_i16_o24_se0',
            # 'r2_k9_s22_e9_i24_o40_se0',
            # 'r3_k9_s22_e9_i40_o80_se0',
            # 'r3_k9_s11_e9_i80_o112_se0',
            # 'r4_k9_s22_e9_i112_o192_se0',
            # 'r1_k9_s11_e9_i192_o320_se0',
        )

    blocks_args = BlockDecoder.decode(blocks_args_template)

    global_params = GlobalParams(
        width_coefficient=width_coefficient,
        depth_coefficient=depth_coefficient,
        image_size=image_size,
        dropout_rate=dropout_rate,
        num_classes=num_classes,
        batch_norm_momentum=0.99,
        batch_norm_epsilon=1e-3,
        drop_connect_rate=drop_connect_rate,
        depth_divisor=8,
        min_depth=None,
    )

    return blocks_args, global_params


def eb0_sp_supernet_local_params(model_name):
    """Map EfficientNet model name to parameter coefficients.
    Args:
        model_name (str): Model name to be queried.
    Returns:
        params_dict[model_name]: A (width,depth,res,dropout) tuple.
    """
    params_dict = {
        # Coefficients:   width,depth,res,dropout
        'efficientnet-b0': (1.0, 1.0, 224, 0.2),
    }
    return params_dict[model_name]


def get_eb0_sp_supernet_params(model_name="efficientnet-b0", override_params=None, num_classes=1000):
    if model_name.startswith('efficientnet'):
        w, d, s, p = eb0_sp_supernet_local_params(model_name)
        blocks_args, global_params = eb0_sp_supernet_complete_params(
            num_classes=num_classes, width_coefficient=w, depth_coefficient=d, dropout_rate=p, image_size=s)
    else:
        raise NotImplementedError('model name is not pre-defined: %s' % model_name)
    if override_params:
        # ValueError will be raised here if override_params has fields not included in global_params.
        global_params = global_params._replace(**override_params)
    return blocks_args, global_params


def make_eb0_sp_supernet(k_e_pairs, n_classes):
    blocks_args, global_params = get_eb0_sp_supernet_params(num_classes=n_classes)
    model = EB0SPSupernet(k_e_pairs, blocks_args, global_params)
    return model


def run_eb0_sp_supernet_epoch(i, model, loader, criterion, n_ops, cand_k_e_pairs,
                              block_repeats=(1,2,2,3,3,4,1), ema=None,
                              optimizer=None, log_f=print, desc="Train",
                              max_grad_norm=5.0, global_unique_archs=None):
    assert n_ops == len(block_repeats)
    avg_loss = AverageMeter()
    avg_top_1_acc = AverageMeter()
    avg_top_5_acc = AverageMeter()
    unique_archs = set()
    for feature, target in tqdm(loader, desc=desc, ascii=True):
        feature = feature.to(device())
        target = target.to(device())
        # Sample with replacement
        discrete_k_e_pairs = random.choices(cand_k_e_pairs, k=n_ops)
        # # EB0 setup
        # discrete_k_e_pairs = [(3, 1), (3, 6), (5, 6), (3, 6), (5, 6), (5, 6), (3, 6)]
        repeated_k_e_pairs = []
        for pi, repeat in enumerate(block_repeats):
            for n in range(repeat):
                repeated_k_e_pairs.append(copy.deepcopy(discrete_k_e_pairs[pi]))
        logits = model(feature, repeated_k_e_pairs)
        loss = criterion(logits, target)
        unique_archs.add(str(discrete_k_e_pairs))
        if optimizer is not None:
            optimizer.zero_grad()
            loss.backward()
            nn.utils.clip_grad_norm_(model.parameters(), max_grad_norm)
            optimizer.step()
            if ema is not None:
                ema.update(model.parameters())
            if hasattr(optimizer, "update"):
                optimizer.update()
        p1, p5 = accuracy(logits, target, topk=(1, 5))
        avg_loss.update(loss.item(), feature.shape[0])
        avg_top_1_acc.update(p1.item(), feature.shape[0])
        avg_top_5_acc.update(p5.item(), feature.shape[0])
    msg = "{} epoch {} loss: {}, top-1 acc: {}, top-5 acc: {}".format(desc, i, avg_loss.avg,
                                                                      avg_top_1_acc.avg, avg_top_5_acc.avg)
    if global_unique_archs is not None:
        global_unique_archs.update(unique_archs)
    if log_f is not None:
        log_f(msg)
        log_f("Num unique arch sampled: {}".format(len(unique_archs)))
    return avg_loss.avg, avg_top_1_acc.avg


def train_eb0_sp_supernet(num_epochs, model, train_loader, criterion, optimizer, n_ops, cand_k_e_pairs, book_keeper,
                          max_grad_norm=5.0, completed_epochs=0, dev_loader=None, ema=None,
                          allow_parallel=True, checkpoint=True):
    model = model.to(device())
    criterion = criterion.to(device())
    unique_archs = set()
    for epoch in range(num_epochs):
        report_epoch = epoch + completed_epochs + 1
        model.train()
        if allow_parallel and can_parallel():
            model = torch.nn.DataParallel(model)
        run_eb0_sp_supernet_epoch(report_epoch, model, train_loader, criterion, n_ops, cand_k_e_pairs,
                                  optimizer=optimizer, log_f=book_keeper.log, ema=ema,
                                  desc="Train", max_grad_norm=max_grad_norm,
                                  global_unique_archs=unique_archs)
        if hasattr(optimizer, "get_lr"):
            book_keeper.log("Optimizer lr at epoch {}: {}".format(report_epoch, optimizer.get_lr()))
        if isinstance(model, torch.nn.DataParallel):
            model = model.module
        book_keeper.log("Trained on {} unique archs so far".format(len(unique_archs)))

        if checkpoint:
            book_keeper.checkpoint_model("_latest.pt", report_epoch, model, optimizer)

        if dev_loader is not None:
            model.eval()
            _, dev_acc = run_eb0_sp_supernet_epoch(report_epoch, model, dev_loader, criterion, n_ops, cand_k_e_pairs,
                                                   log_f=book_keeper.log, desc="Dev")
            if checkpoint:
                book_keeper.checkpoint_model("_best.pt", report_epoch, model, optimizer, eval_perf=dev_acc)
                book_keeper.report_curr_best()
        book_keeper.log("")


def run_eb0_sp_supernet_infer(model, loader, criterion, discrete_k_e_pairs,
                              block_repeats=(1,2,2,3,3,4,1), optimizer=None,
                              max_grad_norm=5.0):
    avg_top_1_acc = AverageMeter()
    avg_top_5_acc = AverageMeter()
    repeated_k_e_pairs = []
    for pi, repeat in enumerate(block_repeats):
        for n in range(repeat):
            repeated_k_e_pairs.append(copy.deepcopy(discrete_k_e_pairs[pi]))
    for feature, target in loader:
        feature = feature.to(device())
        target = target.to(device())
        logits = model(feature, repeated_k_e_pairs)
        loss = criterion(logits, target)
        if optimizer is not None:
            optimizer.zero_grad()
            loss.backward()
            nn.utils.clip_grad_norm_(model.parameters(), max_grad_norm)
            optimizer.step()
            if hasattr(optimizer, "update"):
                optimizer.update()
        p1, p5 = accuracy(logits, target, topk=(1, 5))
        avg_top_1_acc.update(p1.item(), feature.shape[0])
        avg_top_5_acc.update(p5.item(), feature.shape[0])
    return avg_top_1_acc.avg, avg_top_5_acc.avg


class EB0SPSupernetEvaluator(BlockPerfEvaluator):

    def __init__(self, model, train_loader, test_loader, lat_predictor, optim_maker,
                 criterion=torch.nn.CrossEntropyLoss(), default_clue=1.0,
                 enable_perf_memo=True, allow_parallel=False):
        super(EB0SPSupernetEvaluator, self).__init__(default_clue=default_clue, enable_perf_memo=enable_perf_memo)
        self.model = model
        self.train_loader = train_loader
        self.test_loader = test_loader
        self.criterion = criterion.to(device())
        self.lat_predictor = lat_predictor
        self.optim_maker = optim_maker
        self.allow_parallel = allow_parallel

    def get_perf_reward(self, state, model=None):
        raise NotImplementedError

    def get_perf_values(self, block, model=None):
        if model is None:
            model = self.model
        key = block.get_str_id()
        if self.enable_perf_memo and key in self.perf_memo:
            return self.perf_memo[key]
        if self.allow_parallel:
            model = torch.nn.DataParallel(model)
        model = model.to(device())
        k_e_pairs = build_k_e_pairs(block.curr_node_label_list[1:-1])
        with torch.no_grad():
            model.eval()
            pred_acc, _ = run_eb0_sp_supernet_infer(model, self.test_loader, self.criterion, k_e_pairs)
        pred_acc /= 100.
        pred_lat = self.lat_predictor(block) if self.lat_predictor is not None else 1.0
        self.perf_memo[key] = (pred_acc, pred_lat)
        interval_info.num_block_eval += 1
        return pred_acc, pred_lat


class EB0ExtPTEvaluator(BlockPerfEvaluator):

    def __init__(self, trainer_dir, trainer_logs_dir, data_dir,
                 train_batch_size, eval_batch_size,
                 train_steps, steps_per_eval,
                 default_clue=1.0, enable_perf_memo=True):
        super(EB0ExtPTEvaluator, self).__init__(default_clue=default_clue, enable_perf_memo=enable_perf_memo)
        self.trainer_dir = trainer_dir
        self.trainer_logs_dir = trainer_logs_dir
        self.data_dir = data_dir
        self.train_batch_size = train_batch_size
        self.eval_batch_size = eval_batch_size
        self.train_steps = train_steps
        self.steps_per_eval = steps_per_eval
        self.blocks_args_template = [
            'r1_k{}_s11_e{}_i32_o16_se0.25',  # MBConv{}, k{}x{}
            'r2_k{}_s22_e{}_i16_o24_se0.25',  # MBConv{}, k{}x{}
            'r2_k{}_s22_e{}_i24_o40_se0.25',  # MBConv{}, k{}x{}
            'r3_k{}_s22_e{}_i40_o80_se0.25',  # MBConv{}, k{}x{}
            'r3_k{}_s11_e{}_i80_o112_se0.25',  # MBConv{}, k{}x{}
            'r4_k{}_s22_e{}_i112_o192_se0.25',  # MBConv{}, k{}x{}
            'r1_k{}_s11_e{}_i192_o320_se0.25',  # MBConv{}, k{}x{}
        ]

    def get_perf_reward(self, state):
        raise NotImplementedError

    def get_perf_values(self, block):
        key = block.get_str_id()
        if self.enable_perf_memo and key in self.perf_memo:
            return self.perf_memo[key]
        labels = block.curr_node_label_list[1:-1] # Chop input and output labels
        k_e_pairs = build_k_e_pairs(labels)
        block_args_strs = []
        for i, (k, e) in enumerate(k_e_pairs):
            template = self.blocks_args_template[i]
            block_args_str = template.format(k, e)
            block_args_strs.append(block_args_str)
        arg_str = "|".join(block_args_strs)
        if os.path.exists(self.trainer_logs_dir):
            rmdir(self.trainer_logs_dir)
        os.mkdir(self.trainer_logs_dir)
        command = "export PYTHONPATH={}".format(self.trainer_dir)
        command += " && "
        main_file = self.trainer_dir + "main_searched_gpu.py"
        command += "python3 -u {} -searched_model_str \"{}\" -data_dir {} -model_dir {} " \
                   "-train_batch_size {} -eval_batch_size {} " \
                   "-train_steps {} -steps_per_eval {}".format(main_file, arg_str, self.data_dir, self.trainer_logs_dir,
                                                               self.train_batch_size, self.eval_batch_size,
                                                               self.train_steps, self.steps_per_eval)
        print("Issuing command: {}".format(command))
        result = subprocess.run(command, shell=True, stdout=subprocess.PIPE)
        status = result.returncode
        output = result.stdout
        output = output.decode("utf-8")
        print("Return code: {}".format(status))
        keyword = "Result: max_eval_top_1_acc: "
        assert keyword in output
        res = re.compile(r'(?<=Result: max_eval_top_1_acc: )\d+\.?\d*')
        vals = res.findall(output)
        pred_acc = float(vals[0])
        print("Got acc: {}".format(pred_acc))
        pred_lat = 1.0
        self.perf_memo[key] = (pred_acc, pred_lat)
        interval_info.num_block_eval += 1
        return pred_acc, pred_lat

