from model_nas_base import *
from sklearn.metrics import f1_score
from model_helpers import report_blocks_info, step_wise_check
from model_helpers import random_mutate_blocks
from model_supernet import SupernetEvaluator
from utils.darts_utils import AverageMeter


class BlockDiscriminator(torch.nn.Module):
    """ Pairwise Siamese discriminator """
    def __init__(self, in_channels, hidden_size, out_channels, gnn_constructor, node_pre_embedding,
                 activ=torch.nn.Tanh(), num_layers=3, dropout_prob=0.1):
        super(BlockDiscriminator, self).__init__()
        self.node_pre_embedding = node_pre_embedding
        self.in_channels = in_channels
        self.hidden_size = hidden_size
        self.out_channels = out_channels
        self.gnn_layers = torch.nn.ModuleList()
        for i in range(num_layers):
            input_size, output_size = hidden_size, hidden_size
            if i == 0:
                input_size = in_channels
            if i == num_layers - 1:
                output_size = out_channels
            gnn_layer = gnn_constructor(input_size, output_size)
            self.gnn_layers.append(gnn_layer)
        self.activ = activ
        self.predictor = torch.nn.Linear(2 * self.out_channels, 2)
        self.dropout = torch.nn.Dropout(dropout_prob)

    def forward(self, data_pair):
        ctx = []
        for data in data_pair:
            node_feature_tsr_list, edge_index_list, batch_last_node_idx_list = data
            graph_embedding = self._forward(node_feature_tsr_list, edge_index_list, batch_last_node_idx_list)
            ctx.append(graph_embedding)
        ctx = torch.cat([ctx[0], ctx[1]], dim=-1)
        batch_logits = self.predictor(ctx)
        return batch_logits

    def _forward(self, node_tsr_list, edge_index_list, batch_last_node_idx_list):
        node_embedding_list = []
        for node_tsr in node_tsr_list:
            node_tsr = self.node_pre_embedding(node_tsr.to(device()))
            node_embedding_list.append(node_tsr)
        data_list = [
            Data(x=node_embedding, edge_index=edge_index_list[i].to(device()), edge_attr=None)
            for i, node_embedding in enumerate(node_embedding_list)
        ]
        torch_geo_batch = Batch.from_data_list(data_list)
        edge_index_tsr = torch_geo_batch.edge_index
        curr_gnn_output = torch_geo_batch.x
        for li, gnn_layer in enumerate(self.gnn_layers):
            node_embedding_tsr = gnn_layer(curr_gnn_output, edge_index_tsr)
            curr_gnn_output = self.activ(node_embedding_tsr)
        curr_gnn_output = self.dropout(curr_gnn_output)
        batch_embedding_list = torch_geo_batch_to_data_list(curr_gnn_output, batch_last_node_idx_list)
        batch_graph_embedding = []
        for i, node_embedding in enumerate(batch_embedding_list):
            graph_embedding = node_embedding[-1, :].unsqueeze(0)
            batch_graph_embedding.append(graph_embedding)
        return torch.cat(batch_graph_embedding, dim=0)


def build_d_pair_input(b1, b2):
    node_tsr_list_0, edge_list_0, _, idx_list_0 = b1.get_gpn_compatible_input(include_scaffold=False)
    node_tsr_list_1, edge_list_1, _, idx_list_1 = b2.get_gpn_compatible_input(include_scaffold=False)
    block_0 = (node_tsr_list_0, edge_list_0, idx_list_0)
    block_1 = (node_tsr_list_1, edge_list_1, idx_list_1)
    return block_0, block_1


class DisRewardModule:

    def __init__(self, d_trainer, evaluator, output_node_label, n_random_eval,
                 node_label2feature, rm_cand_node_labels,
                 n_truth_per_src=100, n_fake_total=400, aggr_fake=True, rm_depth=2, n_mutations=1000,
                 max_n_nodes=7, enable_mutation=False, input_node_idx=0, pos_reward=0.0, neg_reward=-0.1,
                 lat_weight=0.5, max_d_reward=1.0, valid_block_keys=None):
        self.d_trainer = d_trainer
        self.model = self.d_trainer.get_model()
        self.max_d_reward = max_d_reward
        self.evaluator = evaluator
        self.output_node_label = output_node_label
        self.pos_reward = pos_reward
        self.neg_reward = neg_reward
        self.input_node_idx = input_node_idx
        self.n_truth_per_src = n_truth_per_src
        self.n_fake_total = n_fake_total
        self.src_block2perfs = {}
        self.src_blocks = []
        self.truth_block2perfs = {}
        self.truth_blocks = []
        self.src2truth = {}
        self.n_random_eval = n_random_eval
        self.node_label2feature = node_label2feature
        self.rm_cand_node_labels = rm_cand_node_labels
        self.rm_depth = rm_depth
        self.n_mutations = n_mutations
        self.max_n_nodes = max_n_nodes
        self.lat_weight = lat_weight
        self.valid_block_keys = valid_block_keys
        self.enable_mutation = enable_mutation
        self.aggr_fake = aggr_fake
        self.past_fake_blocks = []

    def train(self):
        self.model.train()

    def eval(self):
        self.model.eval()

    def to(self, dev):
        self.d_trainer.to(dev)
        return self

    def get_model(self):
        return self.model

    def share_memory(self):
        self.model.share_memory()

    def clear(self):
        self.src_block2perfs = {}
        self.src_blocks = []
        self.truth_block2perfs = {}
        self.truth_blocks = []
        self.src2truth = {}
        self.past_fake_blocks = []

    def get_fake_data_size(self):
        if self.enable_mutation:
            return self.n_fake_total
        else:
            return len(self.truth_blocks)

    def build_truth_blocks(self, src_block2perfs, log_f=print):
        self.clear()
        self.src_blocks = [v[0] for k, v in src_block2perfs.items()]
        self.src_block2perfs = src_block2perfs
        if self.enable_mutation:
            for src_b in self.src_blocks:
                mutated_blocks, _, _ = random_mutate_blocks(self.rm_depth, [src_b], self.n_mutations,
                                                            self.node_label2feature, self.max_n_nodes,
                                                            self.rm_cand_node_labels, self.output_node_label,
                                                            valid_block_keys=self.valid_block_keys,
                                                            add_node_depth_limit=0, verbose=False)
                blocks = mutated_blocks[:self.n_truth_per_src]
                self.truth_blocks.extend(blocks)
                self.src2truth[src_b.get_str_id()] = blocks
        else:
            self.truth_blocks = [b for b in self.src_blocks]
            for src_b in self.src_blocks:
                self.src2truth[src_b.get_str_id()] = self.truth_blocks
        log_f("{} truth blocks built".format(len(self.truth_blocks)))

    def random_eval_blocks(self):
        random.shuffle(self.truth_blocks)
        cand_blocks = self.truth_blocks[:self.n_random_eval]
        for block in cand_blocks:
            key = block.get_str_id()
            if key not in self.truth_block2perfs:
                acc, lat = self.evaluator.get_perf_values(block)
                self.truth_block2perfs[key] = (block, (acc, lat))
        return self.truth_block2perfs

    def d_train(self, fake_blocks, log_f=None):
        if self.aggr_fake:
            unique_fake_blocks = UniqueList(key_func=lambda b: b.get_str_id())
            unique_fake_blocks.extend(fake_blocks)
            unique_fake_blocks.extend(self.past_fake_blocks)
            unique_fake_blocks = unique_fake_blocks.tolist()
        else:
            unique_fake_blocks = fake_blocks
        # Build truth pairs
        truth_block_pairs = []
        truth_keys = set()
        for src_key, truth_blocks in self.src2truth.items():
            src_block = self.src_block2perfs[src_key][0]
            for b in truth_blocks:
                truth_block_pairs.append((src_block, b))
                truth_keys.add(b.get_str_id())
        # Build fake pairs
        fake_block_pairs = []
        fake_keys = set()
        for src_block in self.src_blocks:
            for b in unique_fake_blocks:
                k = b.get_str_id()
                fake_keys.add(k)
                fake_block_pairs.append((src_block, b))
        common_keys = truth_keys.intersection(fake_keys)
        log_f("Truth/fake key overlap: {}%".format(100. * len(common_keys) / (len(truth_keys.union(fake_keys)))))
        self.d_trainer.train(truth_block_pairs, fake_block_pairs, log_f)
        # Update past fake blocks
        unique_fake_blocks = UniqueList(key_func=lambda _b: _b.get_str_id())
        unique_fake_blocks.extend(fake_blocks)
        unique_fake_blocks.extend(self.past_fake_blocks)
        self.past_fake_blocks = unique_fake_blocks.tolist()

    def _get_d_reward(self, block):
        d_model = self.model
        reward = 0.

        with torch.no_grad():
            d_model.eval()
            for p_block in self.src_blocks:
                p_block = copy.deepcopy(p_block)
                gen_block_score = d_model(build_d_pair_input(p_block, block))
                gen_block_score = torch.softmax(gen_block_score, dim=-1).squeeze()
                reward = max(reward, get_tsr_value(gen_block_score[1]))

        return reward

    def _get_validity_reward(self, state, encourage_length=False):
        n_violations = get_num_block_violations(state, self.output_node_label, self.input_node_idx)
        if n_violations == 0:
            if encourage_length:
                reward = self.pos_reward * max(0, len(state.curr_node_label_list) - 2)
            else:
                reward = self.pos_reward
            return reward
        else:
            return n_violations * self.neg_reward

    def get_perf_diff(self, acc, lat):
        assert 0 <= acc <= 1
        assert 0 <= lat <= 1
        # Find the block that is closest to the input block in terms of L1-distance
        src_d = []
        for _, (s_block, perf) in self.src_block2perfs.items():
            s_acc, s_lat = perf
            src_d.append( (s_block, abs(s_acc - acc) + abs(s_lat - lat), s_acc, s_lat) )
        src_d.sort(key=lambda t: t[1])
        s_acc, s_lat = src_d[0][2], src_d[0][3]
        acc_diff = acc - s_acc
        lat_diff = (1. - lat) - (1. - s_lat)
        return acc_diff, lat_diff

    def get_reward(self, state):
        assert self.src_block2perfs is not None
        key = state.get_str_id()
        validity_reward = self._get_validity_reward(state)
        if validity_reward < 0.:  # No other reward if block is invalid
            rv = validity_reward
        elif key in self.src_block2perfs:
            rv = validity_reward  # No additional reward if generated the same src blocks!
        else:
            rv = self._get_d_reward(state)
            if key in self.truth_block2perfs:
                # Check if this state is in block2perfs
                # If yes, the reward depends on the dominance over the src perfs
                # If complete dominance, d_reward reward is max, then the perf improvements are added.
                # If partial/no dominance, reward = improvements + d_reward
                _, (acc, lat) = self.truth_block2perfs[key]
                acc_diff, lat_diff = self.get_perf_diff(acc, lat)
                lat_diff *= self.lat_weight  # Reduce the impact of easy-to-improve metric
                perf_diff = acc_diff + lat_diff
                if perf_diff > 0:
                    rv = self.max_d_reward + perf_diff
                else:
                    rv += perf_diff
        return rv


class EvaluatorRewardModule:

    def __init__(self, evaluator:BlockPerfEvaluator, reward_weight=1.0):
        super(EvaluatorRewardModule, self).__init__()
        self.evaluator = evaluator
        self.reward_weight = reward_weight

    def share_memory(self):
        pass

    def train(self):
        pass

    def eval(self):
        pass

    def to(self, *args):
        return self

    def get_reward(self, state):
        acc, lat = self.evaluator.get_perf_values(state)
        assert 0 <= acc <= 1, "Got invalid acc: {}".format(acc)
        assert 0 <= lat <= 1, "Got invalid lat: {}".format(lat)
        lat = 1.0 - lat
        reward = self.reward_weight * acc + (1.0 - self.reward_weight) * lat
        return reward


class BlockTopAccSet:

    def __init__(self, init_block2acc, top_percentile, max_cand_set_size, only_keep_top=True,
                 ignore_init_update=True):
        self.top_percentile = top_percentile
        self.max_cand_set_size = max_cand_set_size
        if ignore_init_update:
            self._block2acc = init_block2acc
            self._curr_top_blocks = [b for _, (b, _) in self._block2acc.items()]
            self._curr_top_block_keys = set(b.get_str_id() for b in self._curr_top_blocks)
        else:
            self._curr_top_blocks = []
            self._curr_top_block_keys = set()
            self._block2acc = {}
            self.update(init_block2acc)
            if only_keep_top:
                self._block2acc = {b.get_str_id():self._block2acc[b.get_str_id()] for b in self._curr_top_blocks}

    def set_block2acc(self, new_block2acc):
        self._block2acc = new_block2acc

    def top_block2acc(self):
        return {b.get_str_id():self._block2acc[b.get_str_id()] for b in self._curr_top_blocks}

    def get_complete_block_with_acc(self):
        rv = []
        for k, (block, acc) in self._block2acc.items():
            rv.append((block, acc))
        return rv

    def state_dict(self):
        return {
            "top_percentile": self.top_percentile,
            "max_cand_set_size": self.max_cand_set_size,
            "_curr_top_blocks": self._curr_top_blocks,
            "_curr_top_block_keys": self._curr_top_block_keys,
            "_block2acc": self._block2acc,
        }

    def load_state_dict(self, sd):
        self.top_percentile = sd["top_percentile"]
        self.max_cand_set_size = sd["max_cand_set_size"]
        self._curr_top_blocks = sd["_curr_top_blocks"]
        self._curr_top_block_keys = sd["_curr_top_block_keys"]
        self._block2acc = sd["_block2acc"]

    def get_top_blocks_with_acc(self):
        rv = []
        for block in self._curr_top_blocks:
            rv.append((block, self._block2acc[block.get_str_id()][1]))
        return rv

    def get_curr_blocks(self):
        return [b for b in self._curr_top_blocks]

    def get_pareto_neighborhood_blocks(self, max_runs=0, log_f=print):
        return [b for b in self._curr_top_blocks]

    def get_top_acc_values(self):
        rv = []
        for block in self._curr_top_blocks:
            rv.append(self._block2acc[block.get_str_id()][1])
        return rv

    def update(self, new_block2acc, log_f=print):
        # new_block2acc expected format: {block_key: (block_object, acc)}
        log_f("Updating Acc front")
        old_top_keys = self._curr_top_block_keys
        # Merge new blocks with current front, overwrite existing block's perf if duplicated
        self._block2acc = {**self._block2acc, **new_block2acc}
        _blocks = [v for _, v in self._block2acc.items()]
        _blocks.sort(key=lambda t: t[1], reverse=True)
        _blocks = _blocks[:self.max_cand_set_size]
        acc_list = [v[1] for v in _blocks]
        acc_threshold = np.percentile(acc_list, 100. - self.top_percentile).item()
        log_f("Number of candidate blocks: {}".format(len(_blocks)))
        log_f("New top {} percentile acc: {}".format(self.top_percentile, acc_threshold))
        new_top_blocks = [v[0] for v in _blocks if v[1] >= acc_threshold]
        new_top_acc = [v[1] for v in _blocks if v[1] >= acc_threshold]
        assert len(new_top_blocks) > 0
        new_top_keys = set(b.get_str_id() for b in new_top_blocks)
        self._curr_top_blocks = new_top_blocks
        self._curr_top_block_keys = set(b.get_str_id() for b in new_top_blocks)
        log_f("New number of top acc blocks: {}".format(len(self._curr_top_blocks)))
        log_f("{} blocks from previous front are kept".format(len(new_top_keys.intersection(old_top_keys))))
        log_f("{} new blocks added".format(len(new_top_keys - old_top_keys)))
        aggr_acc = mean(new_top_acc)
        log_f("New avg acc: {} ".format(aggr_acc))
        log_f("New top-3 acc: {} ".format(new_top_acc[:3]))
        return aggr_acc

    def re_map_performance(self, perf_dict, log_f=print):
        log_f("Re-mapping acc")
        missing_blocks = set()
        new_block2acc = {}
        _curr_block2acc = self.top_block2acc()
        for key, (block, acc) in _curr_block2acc.items():
            if key not in perf_dict:
                missing_blocks.add(key)
            else:
                acc, lat = perf_dict[key]
                new_block2acc[key] = (block, acc)
        self._curr_top_blocks = []
        self._curr_top_block_keys = set()
        self._block2acc = {}
        self.update(new_block2acc, log_f=log_f)
        log_f("Number of missing blocks: {}".format(len(missing_blocks)))

    def __contains__(self, block):
        return block.get_str_id() in self._curr_top_block_keys

    def __str__(self):
        return "BlockTopAccSet[size: {}]".format(len(self._curr_top_blocks))

    def __repr__(self):
        return str(self)

    def __len__(self):
        return len(self._curr_top_blocks)


class BlockTopKAccSet:

    def __init__(self, init_block2acc, top_k, ignore_init_update=False, log_f=print):
        self.top_k = top_k
        if ignore_init_update:
            self._block2acc = init_block2acc
            self._curr_top_blocks = [b for _, (b, _) in self._block2acc.items()]
            self._curr_top_block_keys = set(b.get_str_id() for b in self._curr_top_blocks)
        else:
            self._curr_top_blocks = []
            self._curr_top_block_keys = set()
            self._block2acc = {}
            self.update(init_block2acc, log_f=log_f)

    def set_block2acc(self, new_block2acc, log_f=print):
        log_f("Setting new global block2acc")
        self._block2acc = new_block2acc
        self._curr_top_blocks = [b for _, (b, _) in self._block2acc.items()]
        self._curr_top_block_keys = set(b.get_str_id() for b in self._curr_top_blocks)
        new_top_acc = [v[1] for k, v in self._block2acc.items()]
        new_top_acc.sort(reverse=True)
        aggr_acc = mean(new_top_acc)
        log_f("New avg acc: {} ".format(aggr_acc))
        log_f("New top-3 acc: {} ".format(new_top_acc[:3]))
        log_f("New bottom-3 acc: {} ".format(new_top_acc[-3:]))

    def top_block2acc(self):
        return {b.get_str_id():self._block2acc[b.get_str_id()] for b in self._curr_top_blocks}

    def get_complete_block_with_acc(self):
        rv = []
        for k, (block, acc) in self._block2acc.items():
            rv.append((block, acc))
        return rv

    def state_dict(self):
        return {"top_k": self.top_k,
                "_curr_top_blocks": self._curr_top_blocks,
                "_curr_top_block_keys": self._curr_top_block_keys,
                "_block2acc": self._block2acc}

    def load_state_dict(self, sd):
        self.top_k = sd["top_k"]
        self._curr_top_blocks = sd["_curr_top_blocks"]
        self._curr_top_block_keys = sd["_curr_top_block_keys"]
        self._block2acc = sd["_block2acc"]

    def get_top_blocks_with_acc(self):
        rv = []
        for block in self._curr_top_blocks:
            rv.append((block, self._block2acc[block.get_str_id()][1]))
        return rv

    def get_curr_blocks(self):
        return [b for b in self._curr_top_blocks]

    def get_pareto_neighborhood_blocks(self, max_runs=0, log_f=print):
        return [b for b in self._curr_top_blocks]

    def get_top_acc_values(self):
        rv = []
        for block in self._curr_top_blocks:
            rv.append(self._block2acc[block.get_str_id()][1])
        return rv

    def update(self, new_block2acc, log_f=print):
        # new_block2acc expected format: {block_key: (block_object, acc)}
        log_f("Updating acc set")
        old_top_keys = self._curr_top_block_keys
        # Merge new blocks with current front, overwrite existing block's perf if duplicated
        self._block2acc = {**self._block2acc, **new_block2acc}
        _blocks = [v for _, v in self._block2acc.items()]
        _blocks.sort(key=lambda t: t[1], reverse=True)
        _blocks = _blocks[:self.top_k]
        new_top_blocks = [v[0] for v in _blocks]
        new_top_acc = [v[1] for v in _blocks]
        assert len(new_top_blocks) > 0
        new_top_keys = set(b.get_str_id() for b in new_top_blocks)
        self._curr_top_blocks = new_top_blocks
        self._curr_top_block_keys = set(b.get_str_id() for b in new_top_blocks)
        log_f("New number of top acc blocks: {}".format(len(self._curr_top_blocks)))
        log_f("{} blocks from previous front are kept".format(len(new_top_keys.intersection(old_top_keys))))
        log_f("{} new blocks added".format(len(new_top_keys - old_top_keys)))
        aggr_acc = mean(new_top_acc)
        log_f("New avg acc: {} ".format(aggr_acc))
        log_f("New top-3 acc: {} ".format(new_top_acc[:3]))
        log_f("New bottom-3 acc: {} ".format(new_top_acc[-3:]))
        return aggr_acc

    def re_map_performance(self, perf_dict, log_f=print):
        log_f("Re-mapping acc")
        missing_blocks = set()
        new_block2acc = {}
        _curr_block2acc = self.top_block2acc()
        for key, (block, acc) in _curr_block2acc.items():
            if key not in perf_dict:
                missing_blocks.add(key)
            else:
                acc, lat = perf_dict[key]
                new_block2acc[key] = (block, acc)
        self._curr_top_blocks = []
        self._curr_top_block_keys = set()
        self._block2acc = {}
        self.update(new_block2acc, log_f=log_f)
        log_f("Number of missing blocks: {}".format(len(missing_blocks)))

    def __contains__(self, block):
        return block.get_str_id() in self._curr_top_block_keys

    def __str__(self):
        return "BlockTopKAccSet[size: {}]".format(len(self._curr_top_blocks))

    def __repr__(self):
        return str(self)

    def __len__(self):
        return len(self._curr_top_blocks)


class DisTrainer:

    def __init__(self, n_epochs, model, init_lr, batch_size, max_grad_norm=5.0, early_stop_train_f1=0.95):
        self.criterion = torch.nn.CrossEntropyLoss()
        self.init_lr = init_lr
        self.original_model = model
        self.curr_model = model
        self.n_epochs = n_epochs
        self.batch_size = batch_size
        self.max_grad_norm = max_grad_norm
        self.early_stop_train_f1 = early_stop_train_f1

    def to(self, dev):
        self.curr_model = self.curr_model.to(dev)
        self.criterion = self.criterion.to(dev)
        return self

    def _make_optimizer(self, model):
        return torch.optim.Adam(model.parameters(), lr=self.init_lr)

    def get_model(self):
        return self.curr_model

    def train(self, truth_block_pairs, fake_block_pairs, log_f=None):
        self.curr_model.train()
        optimizer = self._make_optimizer(self.curr_model)
        truth_data = [(p[0], p[1], torch.LongTensor([1]).to(device())) for p in truth_block_pairs]
        fake_data = [(p[0], p[1], torch.LongTensor([0]).to(device())) for p in fake_block_pairs]
        if log_f is not None:
            log_f("Training discriminator with {} truth, {} fake instances".format(len(truth_data), len(fake_data)))
        assert len(truth_data) > 0, "DisTrainer received 0 truth data"
        assert len(fake_data) > 0, "DisTrainer received 0 fake data"
        data = truth_data + fake_data
        for epoch in range(self.n_epochs):
            random.shuffle(data)
            total_loss, n_insts = 0, 0
            y_pred, y_true = [], []
            batch_loss = torch.FloatTensor([0.]).to(device())
            for b1, b2, target in tqdm(data, desc="Training on {} instances".format(len(data)),
                                      leave=False, ascii=True):
                target.requires_grad = False
                block_score = self.curr_model(build_d_pair_input(b1, b2))
                inst_loss = self.criterion(block_score, target)
                batch_loss += inst_loss
                total_loss += get_tsr_value(inst_loss)
                if (n_insts + 1) % self.batch_size == 0 or n_insts == len(data) - 1:
                    optimizer.zero_grad()
                    batch_loss.backward()
                    torch.nn.utils.clip_grad_norm_(self.curr_model.parameters(), self.max_grad_norm)
                    optimizer.step()
                    batch_loss = torch.FloatTensor([0.]).to(device())
                n_insts += 1
                _, pred_ind = block_score.max(1)
                pred_ind = get_tsr_value(pred_ind)
                truth_ind = get_tsr_value(target)
                y_pred.append(pred_ind)
                y_true.append(truth_ind)
            rv_loss = total_loss / n_insts
            if len(set(y_pred)) <= 1:
                train_f1 = 0.
            else:
                train_f1 = f1_score(y_true=y_true, y_pred=y_pred)
            if log_f is not None: log_f("D-train epoch {} of {} loss: {:.3f}, f1: {:.3f}".format(epoch+1, self.n_epochs,
                                                                                                 rv_loss, train_f1))
            if train_f1 >= self.early_stop_train_f1:
                if log_f is not None: log_f("D-train early stop at epoch {}".format(epoch + 1))
                break


def make_block_discriminator(num_unique_nodes, dis_in_channels, dis_hidden_size, dis_out_channels, dis_gnn_constructor,
                             dis_num_layers, dis_dropout_prob):
    pre_embedding = torch.nn.Embedding(num_unique_nodes, dis_in_channels)
    discriminator = BlockDiscriminator(dis_in_channels, dis_hidden_size, dis_out_channels, dis_gnn_constructor,
                                       node_pre_embedding=pre_embedding,
                                       activ=torch.nn.Tanh(),
                                       num_layers=dis_num_layers, dropout_prob=dis_dropout_prob)
    return discriminator


def init_nas_bench_acc_set(data_list, perf_key_list, perf_smaller_better_list, top_percentile, max_cand_set_size,
                           scaffold_node_labels, scaffold_node_feature_tsr, book_keeper,
                           init_perf_rank=0.5, allow_self_edge_for_non_input=False, perf_func=None,
                           only_keep_top=False, random_init_set_size=None):
    # DEPRECATED
    assert len(perf_key_list) == len(perf_smaller_better_list)
    book_keeper.log("Candidate data size for init acc set: {}".format(len(data_list)))
    idx2data = {idx: data_dict for idx, data_dict in enumerate(data_list)}
    data_list = [(i, d) for i, d in enumerate(data_list)]
    if random_init_set_size is not None:
        indices = [t[0] for t in data_list]
        selected_indices = random.sample(indices, k=random_init_set_size)
        book_keeper.log("Randomly selected {} blocks as initial set".format(len(selected_indices)))
    else:
        book_keeper.log("init_perf_rank: {}".format(init_perf_rank))
        keep_length = int(len(data_list) * init_perf_rank)
        key2sorted_indices = {}
        for ki, key in enumerate(perf_key_list):
            assert key not in key2sorted_indices
            key2sorted_indices[key] = sorted(data_list, key=lambda t: t[1][key], reverse=not perf_smaller_better_list[ki])
            key2sorted_indices[key] = key2sorted_indices[key][-keep_length:]
            key2sorted_indices[key] = [t[0] for t in key2sorted_indices[key]]
        common_set = None
        for key, sorted_indices in key2sorted_indices.items():
            indices = set(sorted_indices)
            if common_set is None:
                common_set = indices
            else:
                common_set = common_set.intersection(indices)
        selected_indices = list(common_set)
    pareto_data = [idx2data[i] for i in selected_indices]
    # Need to build state from data dict, then convert to block2perfs format
    block2perfs = {}
    for data_dict in pareto_data:
        state = data_dict_to_state(data_dict, scaffold_node_labels, scaffold_node_feature_tsr,
                                   allow_self_edge_for_non_input)
        state_key = state.get_str_id()
        perfs = [data_dict[perf_key] for perf_key in perf_key_list]
        block2perfs[state_key] = (state, perfs)
    block2acc = {k: (v[0], v[1][0]) for k, v in block2perfs.items()}
    book_keeper.log("Initial block2acc size: {}".format(len(block2acc)))
    top_acc_set = BlockTopAccSet(block2acc, top_percentile, max_cand_set_size, only_keep_top=only_keep_top,
                                 ignore_init_update=random_init_set_size is not None)
    book_keeper.log("Initial top acc set size: {}".format(len(top_acc_set.top_block2acc())))
    # Cannot do this the same way as the Pareto front because of the number of candidate blocks
    if perf_func is None:
        book_keeper.log("Initialized acc using truth values!")
    else:
        book_keeper.log("Re-initialized acc using provided function!")
        new_block2acc = {}
        for b in tqdm(top_acc_set.get_curr_blocks(), desc="Getting new acc for init blocks", ascii=True):
            mapped_acc, _ = perf_func(b)
            new_block2acc[b.get_str_id()] = (b, mapped_acc)
        top_acc_set.set_block2acc(new_block2acc)
    return top_acc_set


def init_nas_bench_top_k_acc_set(data_list, perf_key_list, perf_smaller_better_list, top_k,
                                 scaffold_node_labels, scaffold_node_feature_tsr, book_keeper,
                                 random_init_set_size=None,
                                 init_perf_rank=0.5, allow_self_edge_for_non_input=False, perf_func=None):
    assert len(perf_key_list) == len(perf_smaller_better_list)
    book_keeper.log("Candidate data size for init acc set: {}".format(len(data_list)))
    idx2data = {idx: data_dict for idx, data_dict in enumerate(data_list)}
    data_list = [(i, d) for i, d in enumerate(data_list)]
    if random_init_set_size is not None:
        indices = [t[0] for t in data_list]
        selected_indices = random.sample(indices, k=random_init_set_size)
        book_keeper.log("Randomly selected {} blocks as initial set".format(len(selected_indices)))
    else:
        book_keeper.log("init_perf_rank: {}".format(init_perf_rank))
        keep_length = int(len(data_list) * init_perf_rank)
        key2sorted_indices = {}
        for ki, key in enumerate(perf_key_list):
            assert key not in key2sorted_indices
            key2sorted_indices[key] = sorted(data_list, key=lambda t: t[1][key], reverse=not perf_smaller_better_list[ki])
            key2sorted_indices[key] = key2sorted_indices[key][-keep_length:]
            key2sorted_indices[key] = [t[0] for t in key2sorted_indices[key]]
        common_set = None
        for key, sorted_indices in key2sorted_indices.items():
            indices = set(sorted_indices)
            if common_set is None:
                common_set = indices
            else:
                common_set = common_set.intersection(indices)
        selected_indices = list(common_set)
    pareto_data = [idx2data[i] for i in selected_indices]
    # need to build state from data dict, then convert to block2perfs format
    block2perfs = {}
    for data_dict in pareto_data:
        state = data_dict_to_state(data_dict, scaffold_node_labels, scaffold_node_feature_tsr,
                                   allow_self_edge_for_non_input)
        state_key = state.get_str_id()
        perfs = [data_dict[perf_key] for perf_key in perf_key_list]
        block2perfs[state_key] = (state, perfs)
    block2acc = {k: (v[0], v[1][0]) for k, v in block2perfs.items()}
    book_keeper.log("Initial block2acc size: {}".format(len(block2acc)))
    top_acc_set = BlockTopKAccSet(block2acc, top_k, log_f=book_keeper.log)
    book_keeper.log("Initial top acc set size: {}".format(len(top_acc_set.top_block2acc())))
    # Cannot do this the same way as the Pareto front because of the number of candidate blocks
    if perf_func is None:
        book_keeper.log("Initialized acc using truth values!")
    else:
        book_keeper.log("Re-initialized acc using provided function!")
        new_block2acc = {}
        for b in tqdm(top_acc_set.get_curr_blocks(), desc="Getting new acc for init blocks", ascii=True):
            mapped_acc, _ = perf_func(b)
            new_block2acc[b.get_str_id()] = (b, mapped_acc)
        top_acc_set.set_block2acc(new_block2acc)
    return top_acc_set


def ga_nas_pareto_search(env_manager, ppo_optimizer, value_criterion, pareto_front:BlockParetoFront,
                         evaluator:BlockPerfEvaluator, n_train_iterations, n_train_rounds,
                         k_epochs, t_timesteps, batch_size, max_actor_steps, n_desired_blocks_eval,
                         output_node_label, book_keeper, parallel_explore=False,
                         m_repeats=2, reward_discount_factor=1.0, debug=True,
                         gae_lambda=0.97, clip_epsilon=0.2, c1=0.5, c2=0.1, c3=0.1,
                         actor_update_episode_interval=500, n_desired_blocks_eval_inc=0,
                         n_block_samples_eval=None, n_block_samples_dis=None,
                         pareto_update_callback=None, valid_block_keys=None):
    env_manager = env_manager.to(device())
    value_criterion = value_criterion.to(device())
    reward_module = env_manager.final_reward_module
    g_start = time.time()
    # PPO training iterations
    for i in range(n_train_iterations):
        report_iteration = i + 1
        book_keeper.log("Running iteration {} of {}".format(report_iteration, n_train_iterations))
        env, actor = env_manager.get_env_actor_rep()
        book_keeper.log("Current c2: {}".format(c2))
        book_keeper.log("Current c3: {}".format(c3))
        src_block2perfs = pareto_front.get_pareto_neighborhood_block2perfs(max_runs=3, log_f=book_keeper.log)
        reward_module.build_truth_blocks(src_block2perfs, log_f=book_keeper.log)
        reward_module.random_eval_blocks()

        blocks_from_train = UniqueList(key_func=lambda t: t[0].get_str_id())
        for mi in range(m_repeats):
            book_keeper.log("Inner training loop {} of {}".format(mi + 1, m_repeats))
            # Train discriminator first
            n_desired_blocks_dis = reward_module.get_fake_data_size()
            cand_blocks = gpn_generate_blocks(actor, env, max_actor_steps, n_desired_blocks_dis, output_node_label,
                                              max_n_samples=n_block_samples_dis,
                                              desc="Gen blocks for discriminator training")
            book_keeper.log("{} unique blocks for training discriminator".format(len(cand_blocks)))
            reward_module.d_train(cand_blocks, log_f=book_keeper.log)

            # PPO train actor (generator)
            actor, block_w_rewards = ppo_train(env_manager, ppo_optimizer, value_criterion, n_train_rounds,
                                               k_epochs, t_timesteps, batch_size, reward_discount_factor, clip_epsilon,
                                               gae_lambda, c1, c2, c3, book_keeper, parallel_explore=parallel_explore,
                                               actor_update_episode_interval=actor_update_episode_interval)
            book_keeper.log("{} unique valid blocks generated during training".format(len(block_w_rewards)))
            blocks_from_train.extend(block_w_rewards)
        blocks_from_train = blocks_from_train.tolist()
        blocks_from_train.sort(key=lambda t: t[1], reverse=True)

        book_keeper.log("Eval at iteration {}".format(report_iteration))
        pareto_blocks = pareto_front.get_curr_blocks()
        truth_blocks = [b for b in reward_module.truth_blocks]

        gen_blocks = eval_acc_lat(report_iteration, actor, env, evaluator, pareto_front,
                                  n_desired_blocks_eval + i * n_desired_blocks_eval_inc,
                                  max_actor_steps, output_node_label, book_keeper,
                                  n_block_samples_eval=n_block_samples_eval,
                                  valid_block_keys=valid_block_keys,
                                  pareto_update_callback=pareto_update_callback)

        # gen_blocks = eval_ext_acc_lat([t[0] for t in blocks_from_train], report_iteration, evaluator,
        #                               pareto_front, n_desired_blocks_eval, output_node_label, book_keeper,
        #                               valid_block_keys=valid_block_keys,
        #                               pareto_update_callback=pareto_update_callback)

        if debug:
            book_keeper.log("")
            book_keeper.log("====== Start of debug info ======")
            pair_dis_pred_check(truth_blocks, pareto_blocks, reward_module.get_model(),
                                log_f=book_keeper.log, desc="truth")
            pair_dis_pred_check(gen_blocks, pareto_blocks, reward_module.get_model(),
                                log_f=book_keeper.log, desc="fake")
            book_keeper.log("------")
            truth_keys = set(b.get_str_id() for b in truth_blocks)
            train_block_keys = set([b.get_str_id() for b, _ in blocks_from_train])
            common_keys = train_block_keys.intersection(truth_keys)
            book_keeper.log("{} unique blocks from train".format(len(train_block_keys)))
            book_keeper.log("{}% blocks from train are from truth".format(100. * float(len(common_keys)) / len(train_block_keys)))
            truth_keys = set(b.get_str_id() for b in truth_blocks)
            gen_keys = set(b.get_str_id() for b in gen_blocks)
            common_keys = gen_keys.intersection(truth_keys)
            book_keeper.log("{} unique gen blocks".format(len(gen_keys)))
            book_keeper.log("{}% gen blocks are from truth".format(100. * float(len(common_keys)) / len(gen_keys)))
            book_keeper.log("------")
            report_blocks_info(truth_blocks, log_f=book_keeper.log, desc="truth")
            report_blocks_info(gen_blocks, log_f=book_keeper.log, desc="fake")
            book_keeper.log("------")
            truth_blocks.sort(key=lambda b: len(b))
            book_keeper.log("Scaffold node labels: {}".format(truth_blocks[0].scaffold_node_label_list))
            step_wise_check(truth_blocks[0], actor, env, log_f=book_keeper.log)
            step_wise_check(truth_blocks[-1], actor, env, log_f=book_keeper.log)
            book_keeper.log("====== End of debug info ======")
            book_keeper.log("")
        book_keeper.log("Num blocks evaluated so far: {}".format(interval_info.num_block_eval))
    book_keeper.log("Search seconds: {}".format(time.time() - g_start))


def eval_acc_lat(iteration, actor, env, evaluator:BlockPerfEvaluator, pareto_front:BlockParetoFront,
                 n_desired_blocks, max_steps, output_node_label, book_keeper,
                 n_block_samples_eval=None, pareto_update_callback=None, valid_block_keys=None):
    new_block2perfs, new_block2validity = {}, {}
    new_block2count = collections.defaultdict(int)
    new_blocks = gpn_generate_blocks(actor, env, max_steps, n_desired_blocks, output_node_label,
                                     max_n_samples=n_block_samples_eval,
                                     desc="Eval", valid_block_keys=valid_block_keys)
    assert len(new_blocks) > 0, "No new blocks sampled"
    for block in tqdm(new_blocks, desc="Getting performance", ascii=True):
        key = block.get_str_id()
        new_block2count[key] += 1
        if key in new_block2validity: continue
        validity = get_validity_metrics(block, output_node_label)
        new_block2validity[key] = (block, validity)
        acc, lat = evaluator.get_perf_values(block)
        new_block2perfs[key] = (block, (acc, lat))
    book_keeper.log("Sampled {} blocks, {} unique ones".format(len(new_blocks), len(new_block2perfs)))

    valid_counts = collections.defaultdict(float)
    for _, tup in new_block2validity.items():
        state, validity = tup
        for k, v in validity.items():
            valid_counts[k] += 1. if v else 0.
    valid_acc = {k: v / len(new_block2validity) for k, v in valid_counts.items()}
    book_keeper.log("Validity report at iteration {}: {}, num unique blocks: {}".format(iteration, valid_acc,
                                                                                        len(new_block2validity)))
    if len(new_block2perfs) == 0:
        book_keeper.log("No valid blocks with perf generated")
    else:
        node_lens = [len(b) for _, (b, v) in new_block2perfs.items()]
        edge_counts = [len(b.curr_edge_list[0]) for _, (b, v) in new_block2perfs.items()]
        book_keeper.log("Avg number of nodes in generated blocks: {}".format(mean(node_lens)))
        book_keeper.log("Max number of nodes in generated blocks: {}".format(max(node_lens)))
        book_keeper.log("Min number of nodes in generated blocks: {}".format(min(node_lens)))
        book_keeper.log("Avg number of edges in generated blocks: {}".format(mean(edge_counts)))
        book_keeper.log("Max number of edges in generated blocks: {}".format(max(edge_counts)))
        book_keeper.log("Min number of edges in generated blocks: {}".format(min(edge_counts)))
        gen_acc_values = [v[0] for _, (b, v) in new_block2perfs.items()]
        book_keeper.log("Max acc in generated blocks: {}".format(max(gen_acc_values)))
        book_keeper.log("Min acc in generated blocks: {}".format(min(gen_acc_values)))
        pareto_front.update_pareto_front(new_block2perfs, log_f=book_keeper.log)
        if pareto_update_callback is not None:
            pareto_update_callback(pareto_front, iteration, book_keeper.model_name)
        book_keeper.checkpoint_state_dict(pareto_front.state_dict(),
                                          "{}_pareto_front.pkl".format(book_keeper.model_name))
        pareto_perf_values = pareto_front.get_pareto_block_perf_values()
        p_acc_values = [t[0] for t in pareto_perf_values]
        p_lat_values = [t[1] for t in pareto_perf_values]
        book_keeper.log("Max acc in current Pareto: {}".format(max(p_acc_values)))
        book_keeper.log("Min acc in current Pareto: {}".format(min(p_acc_values)))
        book_keeper.log("Max lat in current Pareto: {}".format(max(p_lat_values)))
        book_keeper.log("Min lat in current Pareto: {}".format(min(p_lat_values)))
        book_keeper.save_acc_lat_pareto_values(p_acc_values, p_lat_values, "iter {}".format(iteration))
        book_keeper.save_block2perfs_values(new_block2perfs, iteration)
        book_keeper.log("New front size: {}, new_block2perfs len: {}".format(len(pareto_front),
                                                                             len(new_block2perfs)))
    return [b for _, (b, _) in new_block2perfs.items()]


def eval_ext_acc_lat(new_blocks, iteration, evaluator:BlockPerfEvaluator, pareto_front:BlockParetoFront,
                     n_desired_blocks, output_node_label, book_keeper,
                     pareto_update_callback=None, valid_block_keys=None):
    filtered_blocks = []
    n_excluded = 0
    for block in new_blocks:
        if valid_block_keys is not None and block.get_str_id() not in valid_block_keys:
            n_excluded += 1
            continue
        filtered_blocks.append(block)
        if len(filtered_blocks) >= n_desired_blocks:
            break
    book_keeper.log("{} blocks not in ext valid set".format(n_excluded))
    book_keeper.log("{} blocks submitted for eval".format(len(filtered_blocks)))
    new_blocks = filtered_blocks
    new_block2perfs, new_block2validity = {}, {}
    new_block2count = collections.defaultdict(int)
    assert len(new_blocks) > 0, "No new blocks sampled"
    for block in tqdm(new_blocks, desc="Getting performance", ascii=True):
        key = block.get_str_id()
        new_block2count[key] += 1
        if key in new_block2validity: continue
        validity = get_validity_metrics(block, output_node_label)
        new_block2validity[key] = (block, validity)
        acc, lat = evaluator.get_perf_values(block)
        new_block2perfs[key] = (block, (acc, lat))
    book_keeper.log("Sampled {} blocks, {} unique ones".format(len(new_blocks), len(new_block2perfs)))

    valid_counts = collections.defaultdict(float)
    for _, tup in new_block2validity.items():
        state, validity = tup
        for k, v in validity.items():
            valid_counts[k] += 1. if v else 0.
    valid_acc = {k: v / len(new_block2validity) for k, v in valid_counts.items()}
    book_keeper.log("Validity report at iteration {}: {}, num unique blocks: {}".format(iteration, valid_acc,
                                                                                        len(new_block2validity)))
    if len(new_block2perfs) == 0:
        book_keeper.log("No valid blocks with perf generated")
    else:
        node_lens = [len(b) for _, (b, v) in new_block2perfs.items()]
        edge_counts = [len(b.curr_edge_list[0]) for _, (b, v) in new_block2perfs.items()]
        book_keeper.log("Avg number of nodes in generated blocks: {}".format(mean(node_lens)))
        book_keeper.log("Max number of nodes in generated blocks: {}".format(max(node_lens)))
        book_keeper.log("Min number of nodes in generated blocks: {}".format(min(node_lens)))
        book_keeper.log("Avg number of edges in generated blocks: {}".format(mean(edge_counts)))
        book_keeper.log("Max number of edges in generated blocks: {}".format(max(edge_counts)))
        book_keeper.log("Min number of edges in generated blocks: {}".format(min(edge_counts)))
        gen_acc_values = [v[0] for _, (b, v) in new_block2perfs.items()]
        book_keeper.log("Max acc in generated blocks: {}".format(max(gen_acc_values)))
        book_keeper.log("Min acc in generated blocks: {}".format(min(gen_acc_values)))
        pareto_front.update_pareto_front(new_block2perfs, log_f=book_keeper.log)
        if pareto_update_callback is not None:
            pareto_update_callback(pareto_front, iteration, book_keeper.model_name)
        book_keeper.checkpoint_state_dict(pareto_front.state_dict(),
                                          "{}_pareto_front.pkl".format(book_keeper.model_name))
        pareto_perf_values = pareto_front.get_pareto_block_perf_values()
        p_acc_values = [t[0] for t in pareto_perf_values]
        p_lat_values = [t[1] for t in pareto_perf_values]
        book_keeper.log("Max acc in current Pareto: {}".format(max(p_acc_values)))
        book_keeper.log("Min acc in current Pareto: {}".format(min(p_acc_values)))
        book_keeper.log("Max lat in current Pareto: {}".format(max(p_lat_values)))
        book_keeper.log("Min lat in current Pareto: {}".format(min(p_lat_values)))
        book_keeper.save_acc_lat_pareto_values(p_acc_values, p_lat_values, "iter {}".format(iteration))
        book_keeper.save_block2perfs_values(new_block2perfs, iteration)
        book_keeper.log("New front size: {}, new_block2perfs len: {}".format(len(pareto_front),
                                                                             len(new_block2perfs)))
    return [b for _, (b, _) in new_block2perfs.items()]


def ga_nas_acc_search(env_manager, ppo_optimizer, value_criterion, top_acc_set,
                      evaluator:BlockPerfEvaluator, n_train_iterations, n_train_rounds,
                      k_epochs, t_timesteps, batch_size, max_actor_steps, n_desired_blocks_eval,
                      output_node_label, book_keeper, parallel_explore=False,
                      n_desired_blocks_eval_inc=0,
                      m_repeats=1, reward_discount_factor=0.99, debug=True,
                      gae_lambda=0.97, clip_epsilon=0.2, c1=0.5, c2=0.1, c3=0.1,
                      actor_update_episode_interval=500, num_examples_to_print=10,
                      n_block_samples_eval=None, n_block_samples_dis=None,
                      acc_update_callback=None, valid_block_keys=None):
    env_manager = env_manager.to(device())
    value_criterion = value_criterion.to(device())
    reward_module = env_manager.final_reward_module
    g_start = time.time()
    # PPO training iterations
    for i in range(n_train_iterations):
        report_iteration = i + 1
        book_keeper.log("Running iteration {} of {}".format(report_iteration, n_train_iterations))
        env, actor = env_manager.get_env_actor_rep()
        book_keeper.log("Current c2: {}".format(c2))
        book_keeper.log("Current c3: {}".format(c3))
        src_block2perfs = top_acc_set.top_block2acc()
        reward_module.build_truth_blocks(src_block2perfs, log_f=book_keeper.log)

        blocks_from_train = UniqueList(key_func=lambda b: b.get_str_id())
        for mi in range(m_repeats):
            book_keeper.log("Inner training loop {} of {}".format(mi + 1, m_repeats))
            # Train discriminator first
            n_desired_blocks_dis = reward_module.get_fake_data_size()
            cand_blocks = gpn_generate_blocks(actor, env, max_actor_steps, n_desired_blocks_dis, output_node_label,
                                              max_n_samples=n_block_samples_dis,
                                              desc="Gen blocks for discriminator training")
            book_keeper.log("{} unique blocks for training discriminator".format(len(cand_blocks)))
            reward_module.d_train(cand_blocks, log_f=book_keeper.log)

            # PPO train actor (generator)
            actor, block_w_rewards = ppo_train(env_manager, ppo_optimizer, value_criterion, n_train_rounds,
                                               k_epochs, t_timesteps, batch_size, reward_discount_factor, clip_epsilon,
                                               gae_lambda, c1, c2, c3, book_keeper, parallel_explore=parallel_explore,
                                               actor_update_episode_interval=actor_update_episode_interval)
            book_keeper.log("{} unique valid blocks generated during training".format(len(block_w_rewards)))
            blocks_from_train.extend([t[0] for t in block_w_rewards])

        book_keeper.log("Eval at iteration {}".format(report_iteration))
        top_blocks = top_acc_set.get_curr_blocks()
        truth_blocks = top_acc_set.get_curr_blocks()
        gen_blocks = eval_acc(report_iteration, actor, env, evaluator, top_acc_set,
                              n_desired_blocks_eval + i * n_desired_blocks_eval_inc,
                              max_actor_steps, output_node_label, book_keeper, num_examples_to_print,
                              n_block_samples_eval=n_block_samples_eval,
                              valid_block_keys=valid_block_keys, update_callback=acc_update_callback)
        if debug:
            book_keeper.log("")
            book_keeper.log("====== Start of debug info ======")
            pair_dis_pred_check(truth_blocks, top_blocks, reward_module.get_model(),
                                log_f=book_keeper.log, desc="truth")
            pair_dis_pred_check(gen_blocks, top_blocks, reward_module.get_model(),
                                log_f=book_keeper.log, desc="fake")
            book_keeper.log("------")
            truth_keys = set(b.get_str_id() for b in truth_blocks)
            train_block_keys = set(b.get_str_id() for b in blocks_from_train.tolist())
            common_keys = train_block_keys.intersection(truth_keys)
            book_keeper.log("{} unique blocks from train".format(len(train_block_keys)))
            book_keeper.log("{}% blocks from train are from truth".format(100. * float(len(common_keys)) / len(train_block_keys)))
            truth_keys = set(b.get_str_id() for b in truth_blocks)
            gen_keys = set(b.get_str_id() for b in gen_blocks)
            common_keys = gen_keys.intersection(truth_keys)
            book_keeper.log("{} unique gen blocks".format(len(gen_keys)))
            book_keeper.log("{}% gen blocks are from truth".format(100. * float(len(common_keys)) / len(gen_keys)))
            book_keeper.log("------")
            report_blocks_info(truth_blocks, log_f=book_keeper.log, desc="truth")
            report_blocks_info(gen_blocks, log_f=book_keeper.log, desc="fake")
            book_keeper.log("------")
            truth_blocks.sort(key=lambda b: len(b))
            book_keeper.log("Scaffold node labels: {}".format(top_blocks[0].scaffold_node_label_list))
            step_wise_check(truth_blocks[0], actor, env, log_f=book_keeper.log)
            step_wise_check(truth_blocks[-1], actor, env, log_f=book_keeper.log)
            book_keeper.log("====== End of debug info ======")
            book_keeper.log("")
        book_keeper.log("Num blocks evaluated so far: {}".format(interval_info.num_block_eval))
    book_keeper.log("Search seconds: {}".format(time.time() - g_start))


def eval_acc(iteration, actor, env, evaluator:BlockPerfEvaluator, acc_set,
             n_desired_blocks, max_steps, output_node_label, book_keeper, num_examples_to_print,
             n_block_samples_eval=None, update_callback=None, valid_block_keys=None):
    new_block2acc, new_block2validity = {}, {}
    book_keeper.log("Number of desired blocks eval: {}".format(n_desired_blocks))
    new_blocks = gpn_generate_blocks(actor, env, max_steps, n_desired_blocks, output_node_label,
                                     max_n_samples=n_block_samples_eval,
                                     desc="Eval", valid_block_keys=valid_block_keys)
    for block in tqdm(new_blocks, desc="Getting performance", ascii=True):
        key = block.get_str_id()
        validity = get_validity_metrics(block, output_node_label)
        new_block2validity[key] = (block, validity)
        acc, lat = evaluator.get_perf_values(block)
        new_block2acc[key] = (block, acc)

    if len(new_block2acc) < n_desired_blocks:
        book_keeper.log("Only generated {} unique blocks".format(len(new_block2acc)))
    assert len(new_block2validity) > 0, "No new blocks generated"
    candidates = [(k, v) for k, v in new_block2validity.items()]
    if len(candidates) > num_examples_to_print:
        samples = random.sample(candidates, k=num_examples_to_print)
    else:
        samples = candidates
    for _, tup in samples:
        state, _ = tup
        book_keeper.log("Generated examples at iteration {}:".format(iteration))
        book_keeper.log(str(state))

    valid_counts = collections.defaultdict(float)
    for _, tup in new_block2validity.items():
        state, validity = tup
        for k, v in validity.items():
            valid_counts[k] += 1. if v else 0.
    valid_acc = {k: v / len(new_block2validity) for k, v in valid_counts.items()}
    book_keeper.log("Validity report at iteration {}: {}, num unique blocks: {}".format(iteration, valid_acc,
                                                                                        len(new_block2validity)))
    if len(new_block2acc) == 0:
        book_keeper.log("No blocks with acc generated")
    else:
        node_lens = [len(b) for _, (b, v) in new_block2acc.items()]
        edge_counts = [len(b.curr_edge_list[0]) for _, (b, v) in new_block2acc.items()]
        book_keeper.log("Avg number of nodes in generated blocks: {}".format(mean(node_lens)))
        book_keeper.log("Max number of nodes in generated blocks: {}".format(max(node_lens)))
        book_keeper.log("Min number of nodes in generated blocks: {}".format(min(node_lens)))
        book_keeper.log("Avg number of edges in generated blocks: {}".format(mean(edge_counts)))
        book_keeper.log("Max number of edges in generated blocks: {}".format(max(edge_counts)))
        book_keeper.log("Min number of edges in generated blocks: {}".format(min(edge_counts)))
        gen_acc_values = [acc for _, (b, acc) in new_block2acc.items()]
        book_keeper.log("Max acc in generated blocks: {}".format(max(gen_acc_values)))
        book_keeper.log("Min acc in generated blocks: {}".format(min(gen_acc_values)))
        acc_set.update(new_block2acc, log_f=book_keeper.log)
        if update_callback is not None:
            update_callback(acc_set, iteration, book_keeper.model_name)
        book_keeper.checkpoint_state_dict(acc_set.state_dict(),
                                          "{}_top_acc_set.pkl".format(book_keeper.model_name))
        acc_vals = acc_set.get_top_acc_values()
        book_keeper.log("Best acc so far: {}".format(max(acc_vals)))
        book_keeper.log("New top acc set size: {}, new_block2acc len: {}".format(len(acc_set),
                                                                                 len(new_block2acc)))
    return [b for _, (b, _) in new_block2acc.items()]


def pair_dis_pred_check(gen_blocks, pareto_blocks, d_reward_module, log_f=print, desc="fake"):
    avg_fake_prob = AverageMeter()
    avg_truth_prob = AverageMeter()
    avg_pred_truth_prob = AverageMeter()
    pred_truth_blocks = []
    for block in tqdm(gen_blocks, desc="Discriminator pred check", ascii=True, leave=False):
        max_true_prob = 0.
        for p_block in pareto_blocks:
            logits = d_reward_module(build_d_pair_input(p_block, block))
            prob = torch.softmax(logits, dim=-1).squeeze()
            prob = get_tsr_value(prob)
            max_true_prob = max(max_true_prob, prob[1])
        avg_fake_prob.update(1. - max_true_prob)
        avg_truth_prob.update(max_true_prob)
        if max_true_prob >= 0.5:
            pred_truth_blocks.append(block)
            avg_pred_truth_prob.update(max_true_prob)
    log_f("Discriminator pred check on {} {} blocks".format(len(gen_blocks), desc))
    log_f("Avg fake prob: {}".format(avg_fake_prob.avg))
    log_f("Avg truth prob: {}".format(avg_truth_prob.avg))
    log_f("Avg pred truth prob: {}".format(avg_pred_truth_prob.avg))
    log_f("Number of pred truth blocks: {}".format(len(pred_truth_blocks)))


def ga_nas_ws_acc_search(env_manager, ppo_optimizer, value_criterion, top_acc_set,
                         evaluator:SupernetEvaluator, n_train_iterations, n_train_rounds,
                         k_epochs, t_timesteps, batch_size, max_actor_steps, n_desired_blocks_eval,
                         output_node_label, book_keeper, parallel_explore=False,
                         m_repeats=10, reward_discount_factor=0.99, debug=True,
                         gae_lambda=0.95, clip_epsilon=0.2, c1=0.5, c2=0.01, c3=0.01,
                         n_desired_blocks_eval_inc=0,
                         actor_update_episode_interval=500, num_examples_to_print=10,
                         n_block_samples_eval=None, n_block_samples_dis=None,
                         acc_update_callback=None, valid_block_keys=None):
    env_manager = env_manager.to(device())
    value_criterion = value_criterion.to(device())
    reward_module = env_manager.final_reward_module
    g_start = time.time()
    # PPO training iterations
    for i in range(n_train_iterations):
        report_iteration = i + 1
        book_keeper.log("Running iteration {} of {}".format(report_iteration, n_train_iterations))
        env, actor = env_manager.get_env_actor_rep()
        book_keeper.log("Current c2: {}".format(c2))
        book_keeper.log("Current c3: {}".format(c3))
        src_block2perfs = top_acc_set.top_block2acc()
        reward_module.build_truth_blocks(src_block2perfs, log_f=book_keeper.log)

        blocks_from_train = UniqueList(key_func=lambda b: b.get_str_id())
        for mi in range(m_repeats):
            book_keeper.log("Inner training loop {} of {}".format(mi + 1, m_repeats))
            # Train discriminator first
            n_desired_blocks_dis = reward_module.get_fake_data_size()
            cand_blocks = gpn_generate_blocks(actor, env, max_actor_steps, n_desired_blocks_dis, output_node_label,
                                              max_n_samples=n_block_samples_dis,
                                              desc="Gen blocks for discriminator training")
            book_keeper.log("{} unique blocks for training discriminator".format(len(cand_blocks)))
            reward_module.d_train(cand_blocks, log_f=book_keeper.log)

            # PPO train actor (generator)
            actor, block_w_rewards = ppo_train(env_manager, ppo_optimizer, value_criterion, n_train_rounds,
                                               k_epochs, t_timesteps, batch_size, reward_discount_factor, clip_epsilon,
                                               gae_lambda, c1, c2, c3, book_keeper, parallel_explore=parallel_explore,
                                               actor_update_episode_interval=actor_update_episode_interval)
            book_keeper.log("{} unique valid blocks generated during training".format(len(block_w_rewards)))
            blocks_from_train.extend([t[0] for t in block_w_rewards])

        book_keeper.log("Eval at iteration {}".format(report_iteration))
        top_blocks = top_acc_set.get_curr_blocks()
        truth_blocks = top_acc_set.get_curr_blocks()
        gen_blocks = eval_ws_acc(report_iteration, actor, env, evaluator, top_acc_set,
                                 n_desired_blocks_eval + i * n_desired_blocks_eval_inc,
                                 max_actor_steps, output_node_label,
                                 book_keeper, num_examples_to_print,
                                 n_block_samples_eval=n_block_samples_eval,
                                 valid_block_keys=valid_block_keys, update_callback=acc_update_callback)
        if debug:
            book_keeper.log("")
            book_keeper.log("====== Start of debug info ======")
            pair_dis_pred_check(truth_blocks, top_blocks, reward_module.get_model(),
                                log_f=book_keeper.log, desc="truth")
            pair_dis_pred_check(gen_blocks, top_blocks, reward_module.get_model(),
                                log_f=book_keeper.log, desc="fake")
            book_keeper.log("------")
            truth_keys = set(b.get_str_id() for b in truth_blocks)
            train_block_keys = set(b.get_str_id() for b in blocks_from_train.tolist())
            common_keys = train_block_keys.intersection(truth_keys)
            book_keeper.log("{} unique blocks from train".format(len(train_block_keys)))
            book_keeper.log("{}% blocks from train are from truth".format(100. * float(len(common_keys)) / len(train_block_keys)))
            truth_keys = set(b.get_str_id() for b in truth_blocks)
            gen_keys = set(b.get_str_id() for b in gen_blocks)
            common_keys = gen_keys.intersection(truth_keys)
            book_keeper.log("{} unique gen blocks".format(len(gen_keys)))
            book_keeper.log("{}% gen blocks are from truth".format(100. * float(len(common_keys)) / len(gen_keys)))
            book_keeper.log("------")
            report_blocks_info(truth_blocks, log_f=book_keeper.log, desc="truth")
            report_blocks_info(gen_blocks, log_f=book_keeper.log, desc="fake")
            book_keeper.log("------")
            truth_blocks.sort(key=lambda b: len(b))
            book_keeper.log("Scaffold node labels: {}".format(top_blocks[0].scaffold_node_label_list))
            step_wise_check(truth_blocks[0], actor, env, log_f=book_keeper.log)
            step_wise_check(truth_blocks[-1], actor, env, log_f=book_keeper.log)
            book_keeper.log("====== End of debug info ======")
            book_keeper.log("")
        book_keeper.log("Num blocks evaluated so far: {}".format(interval_info.num_block_eval))
    book_keeper.log("Search seconds: {}".format(time.time() - g_start))


def eval_ws_acc(iteration, actor, env, evaluator:SupernetEvaluator, acc_set,
                n_desired_blocks, max_steps, output_node_label, book_keeper, num_examples_to_print,
                n_block_samples_eval=None, update_callback=None, valid_block_keys=None):
    new_block2acc, new_block2validity = {}, {}
    new_blocks = gpn_generate_blocks(actor, env, max_steps, n_desired_blocks, output_node_label,
                                     max_n_samples=n_block_samples_eval,
                                     desc="Eval", valid_block_keys=valid_block_keys)

    def _optim_maker(_p):
        return torch.optim.Adam(_p, lr=0.0001)
    evaluator.train_on_family(new_blocks, epochs=50, optim_maker=_optim_maker)

    for block in tqdm(new_blocks, desc="Getting performance", ascii=True):
        key = block.get_str_id()
        validity = get_validity_metrics(block, output_node_label)
        new_block2validity[key] = (block, validity)
        acc, lat = evaluator.get_perf_values(block)
        new_block2acc[key] = (block, acc)

    if len(new_block2acc) < n_desired_blocks:
        book_keeper.log("Only generated {} unique blocks".format(len(new_block2acc)))
    assert len(new_block2validity) > 0, "No new blocks generated"
    candidates = [(k, v) for k, v in new_block2validity.items()]
    if len(candidates) > num_examples_to_print:
        samples = random.sample(candidates, k=num_examples_to_print)
    else:
        samples = candidates
    for _, tup in samples:
        state, _ = tup
        book_keeper.log("Generated examples at iteration {}:".format(iteration))
        book_keeper.log(str(state))

    valid_counts = collections.defaultdict(float)
    for _, tup in new_block2validity.items():
        state, validity = tup
        for k, v in validity.items():
            valid_counts[k] += 1. if v else 0.
    valid_acc = {k: v / len(new_block2validity) for k, v in valid_counts.items()}
    book_keeper.log("Validity report at iteration {}: {}, num unique blocks: {}".format(iteration, valid_acc,
                                                                                        len(new_block2validity)))
    if len(new_block2acc) == 0:
        book_keeper.log("No blocks with acc generated")
    else:
        node_lens = [len(b) for _, (b, v) in new_block2acc.items()]
        edge_counts = [len(b.curr_edge_list[0]) for _, (b, v) in new_block2acc.items()]
        book_keeper.log("Avg number of nodes in generated blocks: {}".format(mean(node_lens)))
        book_keeper.log("Max number of nodes in generated blocks: {}".format(max(node_lens)))
        book_keeper.log("Min number of nodes in generated blocks: {}".format(min(node_lens)))
        book_keeper.log("Avg number of edges in generated blocks: {}".format(mean(edge_counts)))
        book_keeper.log("Max number of edges in generated blocks: {}".format(max(edge_counts)))
        book_keeper.log("Min number of edges in generated blocks: {}".format(min(edge_counts)))
        gen_acc_values = [acc for _, (b, acc) in new_block2acc.items()]
        book_keeper.log("Max acc in generated blocks: {}".format(max(gen_acc_values)))
        book_keeper.log("Min acc in generated blocks: {}".format(min(gen_acc_values)))
        acc_set.update(new_block2acc, log_f=book_keeper.log)
        if update_callback is not None:
            update_callback(acc_set, iteration, book_keeper.model_name)
        book_keeper.checkpoint_state_dict(acc_set.state_dict(),
                                          "{}_top_acc_set.pkl".format(book_keeper.model_name))
        acc_vals = acc_set.get_top_acc_values()
        book_keeper.log("Best acc so far: {}".format(max(acc_vals)))
        book_keeper.log("New top acc set size: {}, new_block2acc len: {}".format(len(acc_set),
                                                                                 len(new_block2acc)))
    return [b for _, (b, _) in new_block2acc.items()]


def ga_nas_cons_acc_search(env_manager, ppo_optimizer, value_criterion, top_acc_set,
                           evaluator:BlockPerfEvaluator, n_train_iterations, n_train_rounds,
                           k_epochs, t_timesteps, batch_size, max_actor_steps, n_desired_blocks_eval,
                           output_node_label, max_train_time, max_n_params, cons_perf_dict,
                           book_keeper,
                           parallel_explore=False, n_desired_blocks_eval_inc=0,
                           m_repeats=1, reward_discount_factor=0.99, debug=True,
                           gae_lambda=0.97, clip_epsilon=0.2, c1=0.5, c2=0.1, c3=0.1,
                           actor_update_episode_interval=500,
                           n_block_samples_eval=None, n_block_samples_dis=None,
                           acc_update_callback=None, valid_block_keys=None):
    env_manager = env_manager.to(device())
    value_criterion = value_criterion.to(device())
    reward_module = env_manager.final_reward_module
    g_start = time.time()
    for i in range(n_train_iterations):
        report_iteration = i + 1
        book_keeper.log("Running iteration {} of {}".format(report_iteration, n_train_iterations))
        env, actor = env_manager.get_env_actor_rep()
        book_keeper.log("Current c2: {}".format(c2))
        book_keeper.log("Current c3: {}".format(c3))
        src_block2perfs = top_acc_set.top_block2acc()
        reward_module.build_truth_blocks(src_block2perfs, log_f=book_keeper.log)

        blocks_from_train = UniqueList(key_func=lambda b: b.get_str_id())
        for mi in range(m_repeats):
            book_keeper.log("Inner training loop {} of {}".format(mi + 1, m_repeats))
            # Train discriminator first
            n_desired_blocks_dis = reward_module.get_fake_data_size()
            cand_blocks = gpn_generate_blocks(actor, env, max_actor_steps, n_desired_blocks_dis, output_node_label,
                                              max_n_samples=n_block_samples_dis,
                                              desc="Gen blocks for discriminator training")
            book_keeper.log("{} unique blocks for training discriminator".format(len(cand_blocks)))
            reward_module.d_train(cand_blocks, log_f=book_keeper.log)

            # PPO train actor (generator)
            actor, block_w_rewards = ppo_train(env_manager, ppo_optimizer, value_criterion, n_train_rounds,
                                               k_epochs, t_timesteps, batch_size, reward_discount_factor, clip_epsilon,
                                               gae_lambda, c1, c2, c3, book_keeper, parallel_explore=parallel_explore,
                                               actor_update_episode_interval=actor_update_episode_interval)
            book_keeper.log("{} unique valid blocks generated during training".format(len(block_w_rewards)))
            blocks_from_train.extend([t[0] for t in block_w_rewards])

        book_keeper.log("Eval at iteration {}".format(report_iteration))
        top_blocks = top_acc_set.get_curr_blocks()
        truth_blocks = top_acc_set.get_curr_blocks()
        gen_blocks = eval_cons_acc(report_iteration, actor, env, evaluator, top_acc_set,
                                   n_desired_blocks_eval + i * n_desired_blocks_eval_inc,
                                   max_actor_steps, output_node_label,
                                   max_train_time, max_n_params, cons_perf_dict,
                                   book_keeper,
                                   n_block_samples_eval=n_block_samples_eval,
                                   valid_block_keys=valid_block_keys, update_callback=acc_update_callback)
        if debug:
            book_keeper.log("")
            book_keeper.log("====== Start of debug info ======")
            pair_dis_pred_check(truth_blocks, top_blocks, reward_module.get_model(),
                                log_f=book_keeper.log, desc="truth")
            pair_dis_pred_check(gen_blocks, top_blocks, reward_module.get_model(),
                                log_f=book_keeper.log, desc="fake")
            book_keeper.log("------")
            truth_keys = set(b.get_str_id() for b in truth_blocks)
            train_block_keys = set(b.get_str_id() for b in blocks_from_train.tolist())
            common_keys = train_block_keys.intersection(truth_keys)
            book_keeper.log("{} unique blocks from train".format(len(train_block_keys)))
            book_keeper.log("{}% blocks from train are from truth".format(100. * float(len(common_keys)) / len(train_block_keys)))
            truth_keys = set(b.get_str_id() for b in truth_blocks)
            gen_keys = set(b.get_str_id() for b in gen_blocks)
            common_keys = gen_keys.intersection(truth_keys)
            book_keeper.log("{} unique gen blocks".format(len(gen_keys)))
            book_keeper.log("{}% gen blocks are from truth".format(100. * float(len(common_keys)) / len(gen_keys)))
            book_keeper.log("------")
            report_blocks_info(truth_blocks, log_f=book_keeper.log, desc="truth")
            report_blocks_info(gen_blocks, log_f=book_keeper.log, desc="fake")
            book_keeper.log("------")
            truth_blocks.sort(key=lambda b: len(b))
            book_keeper.log("Scaffold node labels: {}".format(top_blocks[0].scaffold_node_label_list))
            step_wise_check(truth_blocks[0], actor, env, log_f=book_keeper.log)
            step_wise_check(truth_blocks[-1], actor, env, log_f=book_keeper.log)
            book_keeper.log("====== End of debug info ======")
            book_keeper.log("")
        book_keeper.log("Num blocks evaluated so far: {}".format(interval_info.num_block_eval))
    book_keeper.log("Search seconds: {}".format(time.time() - g_start))


def eval_cons_acc(iteration, actor, env, evaluator:BlockPerfEvaluator, acc_set,
                  n_desired_blocks, max_steps, output_node_label,
                  max_train_time, max_n_params, cons_perf_dict,
                  book_keeper,
                  n_block_samples_eval=None, update_callback=None, valid_block_keys=None):
    book_keeper.log("Number of desired blocks eval: {}".format(n_desired_blocks))
    new_blocks = gpn_generate_blocks(actor, env, max_steps, n_desired_blocks, output_node_label,
                                     max_n_samples=n_block_samples_eval,
                                     desc="Eval", valid_block_keys=valid_block_keys)

    # Remove any blocks in truth set that do not satisfy the constraints
    filtered_block2acc = {}
    block_w_acc = acc_set.get_complete_block_with_acc()
    book_keeper.log("block2acc size before filtering: {}".format(len(block_w_acc)))
    for b, acc in tqdm(block_w_acc, desc="Filtering curr top acc by constraints", ascii=True):
        key = b.get_str_id()
        train_time, n_params = cons_perf_dict[key]
        if train_time > max_train_time or n_params > max_n_params:
            continue
        filtered_block2acc[b.get_str_id()] = (b, acc)
    acc_set.set_block2acc(filtered_block2acc)
    block_w_acc = acc_set.get_complete_block_with_acc()
    book_keeper.log("block2acc size after filtering: {}".format(len(block_w_acc)))

    new_block2acc = {}
    n_not_found, n_not_match = 0, 0
    for block in tqdm(new_blocks, desc="Getting performance", ascii=True):
        key = block.get_str_id()
        if key not in cons_perf_dict:
            n_not_found += 1
            continue
        train_time, n_params = cons_perf_dict[key]
        if train_time > max_train_time or n_params > max_n_params:
            n_not_match += 1
            continue
        acc, lat = evaluator.get_perf_values(block)
        new_block2acc[key] = (block, acc)

    book_keeper.log("{} generated blocks not in cons_perf_dict".format(n_not_found))
    book_keeper.log("{} generated blocks not matching constraints".format(n_not_match))
    book_keeper.log("{}% success rate".format(100. * len(new_block2acc) / len(new_blocks)))
    if len(new_block2acc) < n_desired_blocks:
        book_keeper.log("Only able to generate {} unique blocks".format(len(new_block2acc)))
    if len(new_block2acc) == 0:
        book_keeper.log("No blocks with acc generated")
    else:
        node_lens = [len(b) for _, (b, v) in new_block2acc.items()]
        edge_counts = [len(b.curr_edge_list[0]) for _, (b, v) in new_block2acc.items()]
        book_keeper.log("Avg number of nodes in generated blocks: {}".format(mean(node_lens)))
        book_keeper.log("Max number of nodes in generated blocks: {}".format(max(node_lens)))
        book_keeper.log("Min number of nodes in generated blocks: {}".format(min(node_lens)))
        book_keeper.log("Avg number of edges in generated blocks: {}".format(mean(edge_counts)))
        book_keeper.log("Max number of edges in generated blocks: {}".format(max(edge_counts)))
        book_keeper.log("Min number of edges in generated blocks: {}".format(min(edge_counts)))
        gen_acc_values = [acc for _, (b, acc) in new_block2acc.items()]
        book_keeper.log("Max acc in generated blocks: {}".format(max(gen_acc_values)))
        book_keeper.log("Min acc in generated blocks: {}".format(min(gen_acc_values)))
        acc_set.update(new_block2acc, log_f=book_keeper.log)
        if update_callback is not None:
            update_callback(acc_set, iteration, book_keeper.model_name)
        book_keeper.checkpoint_state_dict(acc_set.state_dict(),
                                          "{}_top_acc_set.pkl".format(book_keeper.model_name))
        acc_vals = acc_set.get_top_acc_values()
        book_keeper.log("Best acc so far: {}".format(max(acc_vals)))
        book_keeper.log("New top acc set size: {}, new_block2acc len: {}".format(len(acc_set),
                                                                                 len(new_block2acc)))
    return [b for _, (b, _) in new_block2acc.items()]


def ga_nas_acc_search_ablation(env_manager, ppo_optimizer, value_criterion, top_acc_set,
                               evaluator:BlockPerfEvaluator, n_train_iterations, n_train_rounds,
                               k_epochs, t_timesteps, batch_size,
                               output_node_label, book_keeper, parallel_explore=False,
                               m_repeats=1, reward_discount_factor=0.99,
                               gae_lambda=0.97, clip_epsilon=0.2, c1=0.5, c2=0.1, c3=0.1,
                               actor_update_episode_interval=500, num_examples_to_print=10,
                               acc_update_callback=None,):
    env_manager = env_manager.to(device())
    value_criterion = value_criterion.to(device())
    g_start = time.time()
    # PPO training iterations
    for i in range(n_train_iterations):
        report_iteration = i + 1
        book_keeper.log("Running iteration {} of {}".format(report_iteration, n_train_iterations))
        book_keeper.log("Current c2: {}".format(c2))
        book_keeper.log("Current c3: {}".format(c3))

        blocks_from_train = UniqueList(key_func=lambda b: b.get_str_id())
        for mi in range(m_repeats):
            book_keeper.log("Inner training loop {} of {}".format(mi + 1, m_repeats))

            # PPO train actor (generator)
            actor, block_w_rewards = ppo_train(env_manager, ppo_optimizer, value_criterion, n_train_rounds,
                                               k_epochs, t_timesteps, batch_size, reward_discount_factor, clip_epsilon,
                                               gae_lambda, c1, c2, c3, book_keeper, parallel_explore=parallel_explore,
                                               actor_update_episode_interval=actor_update_episode_interval)
            book_keeper.log("{} unique valid blocks generated during training".format(len(block_w_rewards)))
            blocks_from_train.extend([t[0] for t in block_w_rewards])

        new_blocks = blocks_from_train.tolist()
        book_keeper.log("Eval at iteration {}".format(report_iteration))
        eval_acc_ablation(report_iteration, new_blocks, evaluator, top_acc_set,
                          output_node_label, book_keeper, num_examples_to_print,
                          update_callback=acc_update_callback)
        book_keeper.log("Num blocks evaluated so far: {}".format(interval_info.num_block_eval))
    book_keeper.log("Search seconds: {}".format(time.time() - g_start))


def eval_acc_ablation(iteration, new_blocks, evaluator:BlockPerfEvaluator, acc_set,
                      output_node_label, book_keeper, num_examples_to_print,
                      update_callback=None):
    new_block2acc, new_block2validity = {}, {}
    for block in tqdm(new_blocks, desc="Getting performance", ascii=True):
        key = block.get_str_id()
        validity = get_validity_metrics(block, output_node_label)
        new_block2validity[key] = (block, validity)
        acc, lat = evaluator.get_perf_values(block)
        new_block2acc[key] = (block, acc)

    assert len(new_block2validity) > 0, "No new blocks generated"
    candidates = [(k, v) for k, v in new_block2validity.items()]
    if len(candidates) > num_examples_to_print:
        samples = random.sample(candidates, k=num_examples_to_print)
    else:
        samples = candidates
    for _, tup in samples:
        state, _ = tup
        book_keeper.log("Generated examples at iteration {}:".format(iteration))
        book_keeper.log(str(state))

    valid_counts = collections.defaultdict(float)
    for _, tup in new_block2validity.items():
        state, validity = tup
        for k, v in validity.items():
            valid_counts[k] += 1. if v else 0.
    valid_acc = {k: v / len(new_block2validity) for k, v in valid_counts.items()}
    book_keeper.log("Validity report at iteration {}: {}, num unique blocks: {}".format(iteration, valid_acc,
                                                                                        len(new_block2validity)))
    if len(new_block2acc) == 0:
        book_keeper.log("No blocks with acc generated")
    else:
        node_lens = [len(b) for _, (b, v) in new_block2acc.items()]
        edge_counts = [len(b.curr_edge_list[0]) for _, (b, v) in new_block2acc.items()]
        book_keeper.log("Avg number of nodes in generated blocks: {}".format(mean(node_lens)))
        book_keeper.log("Max number of nodes in generated blocks: {}".format(max(node_lens)))
        book_keeper.log("Min number of nodes in generated blocks: {}".format(min(node_lens)))
        book_keeper.log("Avg number of edges in generated blocks: {}".format(mean(edge_counts)))
        book_keeper.log("Max number of edges in generated blocks: {}".format(max(edge_counts)))
        book_keeper.log("Min number of edges in generated blocks: {}".format(min(edge_counts)))
        gen_acc_values = [acc for _, (b, acc) in new_block2acc.items()]
        book_keeper.log("Max acc in generated blocks: {}".format(max(gen_acc_values)))
        book_keeper.log("Min acc in generated blocks: {}".format(min(gen_acc_values)))
        acc_set.update(new_block2acc, log_f=book_keeper.log)
        if update_callback is not None:
            update_callback(acc_set, iteration, book_keeper.model_name)
        book_keeper.checkpoint_state_dict(acc_set.state_dict(),
                                          "{}_top_acc_set.pkl".format(book_keeper.model_name))
        acc_vals = acc_set.get_top_acc_values()
        book_keeper.log("Best acc so far: {}".format(max(acc_vals)))
        book_keeper.log("New top acc set size: {}, new_block2acc len: {}".format(len(acc_set),
                                                                                 len(new_block2acc)))
    return [b for _, (b, _) in new_block2acc.items()]

