from tqdm import tqdm
from model_nas_base import compute_t_step_gae, NodeAggr, run_seq_exploration, \
    run_parallel_exploration, BlockPerfEvaluator
from model_ga_nas import BlockTopKAccSet
from torch.distributions import Categorical
from utils.model_utils import device
from torch_geometric.data import Data, Batch
from utils.nas_utils import *
from utils.debug_utils import IntervalInfo as interval_info
from model_darts import infer, alphas_to_genotype, DARTS_PRIMITIVES, full_train_darts_v2
from sklearn.metrics import f1_score
from model_components import MultiInputLinears, MultiOutputLinears, SimpleRNN
from utils.darts_utils import *


DARTS_INPUT_NODE_LABEL = "input"
DARTS_OUTPUT_NODE_LABEL = "output"


DARTS_OP2IDX = {
    'input': 0,
    'output': 1,
    'max_pool_3x3': 2,
    'avg_pool_3x3': 3,
    'skip_connect': 4,
    'sep_conv_3x3': 5,
    'sep_conv_5x5': 6,
    'dil_conv_3x3': 7,
    'dil_conv_5x5': 8
}


DARTS_IDX2OP = {v: k for k, v in DARTS_OP2IDX.items()}
assert len(DARTS_IDX2OP) == len(DARTS_OP2IDX)


# DARTS_SCAFFOLD_LABELS = copy.deepcopy(DARTS_PRIMITIVES)
DARTS_SCAFFOLD_LABELS = copy.deepcopy(DARTS_PRIMITIVES[1:])


def build_d_pair_input(b1, b2):
    normal_tsr_list_0, normal_edge_list_0, normal_idx_list_0, \
        reduce_tsr_list_0, reduce_edge_list_0, reduce_idx_list_0, _ = b1.get_gpn_compatible_input()
    normal_tsr_list_1, normal_edge_list_1, normal_idx_list_1, \
        reduce_tsr_list_1, reduce_edge_list_1, reduce_idx_list_1, _ = b2.get_gpn_compatible_input()
    block_0 = (normal_tsr_list_0, normal_edge_list_0, normal_idx_list_0,
               reduce_tsr_list_0, reduce_edge_list_0, reduce_idx_list_0)
    block_1 = (normal_tsr_list_1, normal_edge_list_1, normal_idx_list_1,
               reduce_tsr_list_1, reduce_edge_list_1, reduce_idx_list_1)
    return block_0, block_1


def get_curr_cell_node_idx(n_curr_ops, n_input_cell_nodes=2, n_input_edges=2):
    assert n_input_edges > 0, "n_input_edges={}".format(n_input_edges)
    n_curr_ops -= n_input_cell_nodes
    assert n_curr_ops >= 0, "n_curr_ops={}".format(n_curr_ops)
    curr_cell_node_idx = n_curr_ops // n_input_edges
    return curr_cell_node_idx # Note: this index does not count the input cell nodes


def _masked_disjoint_sample(probs:torch.Tensor):
    assert probs.shape[-2] == 2, "Only supports two disjoint samples now"
    # if probs.shape[-1] == 2:
    #     vals = Categorical(probs=probs).sample()
    #     vals_list = get_tsr_value(vals)
    #     if vals_list[0] == vals_list[1]:
    #         vals[1] = 1 if vals_list[0] == 0 else 0
    #     return vals, probs
    # else:
    vals = Categorical(probs=probs[0, :]).sample()
    sample_idx_1 = get_tsr_value(vals)
    probs_2 = probs[1, :]
    mask = torch.zeros_like(probs_2)
    mask[sample_idx_1] = 1
    mask_probs_2 = probs_2.masked_fill(mask == 1, -1e10)
    mask_probs_2 = torch.softmax(mask_probs_2, dim=-1)
    sample_idx_2 = get_tsr_value(Categorical(probs=mask_probs_2).sample())
    return torch.LongTensor([sample_idx_1, sample_idx_2]).to(device())


def get_output_cell_node_input_edges(n_input_nodes=2, n_internal_nodes=4, n_concat_nodes=4):
    # By default DARTS concats the last 4 cell nodes
    input_node_inds = list(range(n_input_nodes + n_internal_nodes))
    output_cell_node_ind = len(input_node_inds)
    edge_pairs = []
    for src in input_node_inds[-n_concat_nodes:]:
        edge_pairs.append((src, output_cell_node_ind))
    return edge_pairs


class GraphEncoder(torch.nn.Module):

    def __init__(self, in_channels, hidden_size, out_channels, gnn_constructor, node_pre_embedding,
                 activ=torch.nn.ReLU(), num_layers=2, dropout_prob=0.0, add_noise_prior=False):
        super(GraphEncoder, self).__init__()
        self.node_pre_embedding = node_pre_embedding
        self.in_channels = in_channels
        self.hidden_size = hidden_size
        self.out_channels = out_channels
        self.gnn_layers = torch.nn.ModuleList()
        for i in range(num_layers):
            input_size, output_size = hidden_size, hidden_size
            if i == 0:
                input_size = in_channels
            if i == num_layers - 1:
                output_size = out_channels
            gnn_layer = gnn_constructor(input_size, output_size)
            self.gnn_layers.append(gnn_layer)
        self.activ = activ
        self.dropout = torch.nn.Dropout(dropout_prob)
        self.init_ff = torch.nn.Linear(2 * in_channels, in_channels)
        self.add_noise_prior = add_noise_prior

    def init_node_embedding_prior(self, node_embedding):
        prior = np.random.normal(size=node_embedding.shape)
        prior = torch.from_numpy(prior).float().to(device())
        node_embedding = torch.cat([node_embedding, prior], dim=-1)
        return self.init_ff(node_embedding)

    def forward(self, batch_node_tsr, edge_tsr_list, batch_last_node_idx_list):
        node_embedding = self.node_pre_embedding(batch_node_tsr.to(device()))
        if self.add_noise_prior and node_embedding.shape[1] <= 2:
            node_embedding = self.init_node_embedding_prior(node_embedding)
        data_list = [
            Data(x=node_embedding[i,:], edge_index=edge_tsr_list[i].to(device()), edge_attr=None)
                for i in range(node_embedding.shape[0])
        ]
        torch_geo_batch = Batch.from_data_list(data_list)
        edge_index_tsr = torch_geo_batch.edge_index
        curr_gnn_output = torch_geo_batch.x
        for li, gnn_layer in enumerate(self.gnn_layers):
            node_embedding_tsr = gnn_layer(curr_gnn_output, edge_index_tsr)
            curr_gnn_output = self.activ(node_embedding_tsr)
        curr_gnn_output = self.dropout(curr_gnn_output)
        batch_embedding_list = torch_geo_batch_to_data_list(curr_gnn_output, batch_last_node_idx_list)
        batch_embedding = torch.cat([t.unsqueeze(0) for t in batch_embedding_list], dim=0)
        return batch_embedding


class DartsBlockGPN(torch.nn.Module):
    """Wraps the GraphEncoder for normal and reduce cells"""
    def __init__(self, normal_cell_encoder, reduce_cell_encoder, action_predictor):
        super(DartsBlockGPN, self).__init__()
        self.normal_cell_encoder = normal_cell_encoder
        self.reduce_cell_encoder = reduce_cell_encoder
        self.action_predictor = action_predictor

    def forward(self, b_normal_tsr, normal_edge_tsr_list, normal_last_node_idx_list,
                b_reduce_tsr, reduce_edge_tsr_list, reduce_last_node_idx_list,
                curr_cell_node_i):
        normal_node_embedding = self.normal_cell_encoder(b_normal_tsr, normal_edge_tsr_list, normal_last_node_idx_list)
        reduce_node_embedding = self.reduce_cell_encoder(b_reduce_tsr, reduce_edge_tsr_list, reduce_last_node_idx_list)
        normal_op_probs, normal_input_node_probs, \
            reduce_op_probs, reduce_input_node_probs = \
                self.action_predictor(normal_node_embedding, reduce_node_embedding, curr_cell_node_i)
        batch_embedding = torch.cat([normal_node_embedding, reduce_node_embedding], dim=-1)
        return batch_embedding, \
               (normal_op_probs, normal_input_node_probs, reduce_op_probs, reduce_input_node_probs)


class RNNCellPredictor(torch.nn.Module):

    def __init__(self, node_embed_size, hidden_size, n_cand_ops,
                 n_input_nodes=2, n_cell_internal_nodes=4, n_input_edges=2):
        super(RNNCellPredictor, self).__init__()
        self.n_cand_ops = n_cand_ops
        self.n_input_nodes = n_input_nodes
        self.n_input_edges = n_input_edges
        self.n_cell_internal_nodes = n_cell_internal_nodes
        self.graph_ff = torch.nn.Linear(node_embed_size, hidden_size)
        self.op_selectors = torch.nn.ModuleList(
            [torch.nn.Linear(hidden_size, n_cand_ops) for _ in range(n_input_edges)]
        )
        input_nodes_cand_size_list = [n_input_nodes + i for i in range(n_cell_internal_nodes)]
        self.input_node_selectors = torch.nn.ModuleList(
            [MultiOutputLinears(hidden_size, input_nodes_cand_size_list) for _ in range(n_input_edges)]
        )
        input_node_rnn_input_size_list = [n_input_nodes + j for j in range(n_cell_internal_nodes)]
        self.input_node_rnn_input_ffs = torch.nn.ModuleList(
            [MultiInputLinears(input_node_rnn_input_size_list, hidden_size) for _ in range(n_input_edges)]
        )
        self.op_rnn = SimpleRNN(n_cand_ops, hidden_size, rnn_dir=1)
        self.input_node_rnn = SimpleRNN(hidden_size, hidden_size, rnn_dir=1)

    def forward(self, node_embedding, curr_cell_node_i):
        node_embedding = node_embedding.to(device())
        graph_embedding = self.graph_ff(node_embedding.sum(dim=1))
        curr_op_hidden = graph_embedding.unsqueeze(0)
        curr_input_node_hidden = graph_embedding.unsqueeze(0)
        op_rnn_input = torch.ones(graph_embedding.shape[0], 1, self.n_cand_ops).to(device())
        op_rnn_input = op_rnn_input / op_rnn_input.sum(-1, keepdim=True)
        input_node_rnn_input = torch.ones(graph_embedding.shape[0], 1, curr_cell_node_i + 2).to(device())
        input_node_rnn_input /= input_node_rnn_input.sum(-1, keepdim=True)
        op_probs_list = []
        input_node_probs_list = []
        for si in range(self.n_input_edges):
            # Op
            curr_op_ctx, curr_op_hidden = self.op_rnn(op_rnn_input, curr_op_hidden)
            op_logits = self.op_selectors[si](curr_op_ctx)
            op_probs = torch.softmax(op_logits, dim=-1)
            op_probs_list.append(op_probs)
            op_rnn_input = op_probs.detach()
            # Edge
            input_node_rnn_input = self.input_node_rnn_input_ffs[si](input_node_rnn_input)
            input_node_ctx, curr_input_node_hidden = self.input_node_rnn(input_node_rnn_input, curr_input_node_hidden)
            input_node_logits = self.input_node_selectors[si](input_node_ctx, curr_cell_node_i)
            input_node_probs = torch.softmax(input_node_logits, dim=-1)
            input_node_probs_list.append(input_node_probs)
            input_node_rnn_input = input_node_probs.detach()
        new_op_probs = torch.cat(op_probs_list, dim=1)
        new_input_node_probs = torch.cat(input_node_probs_list, dim=1)
        return new_op_probs, new_input_node_probs


class DartsActionPredictor(torch.nn.Module):

    def __init__(self, normal_cell_predictor: RNNCellPredictor,
                 reduce_cell_predictor: RNNCellPredictor):
        super(DartsActionPredictor, self).__init__()
        self.normal_cell_predictor = normal_cell_predictor
        self.reduce_cell_predictor = reduce_cell_predictor

    def forward(self, normal_node_embedding, reduce_node_embedding, curr_cell_node_i):
        normal_op_probs, normal_input_node_probs = \
            self.normal_cell_predictor(normal_node_embedding, curr_cell_node_i)
        reduce_op_probs, reduce_input_node_probs = \
            self.reduce_cell_predictor(reduce_node_embedding, curr_cell_node_i)
        return normal_op_probs, normal_input_node_probs, \
               reduce_op_probs, reduce_input_node_probs


class DartsBlockDiscriminator(torch.nn.Module):
    """ Pairwise Siamese discriminator """
    def __init__(self, normal_cell_encoder, reduce_cell_encoder,
                 normal_predictor, reduce_predictor):
        super(DartsBlockDiscriminator, self).__init__()
        self.normal_cell_encoder = normal_cell_encoder
        self.reduce_cell_encoder = reduce_cell_encoder
        self.normal_predictor = normal_predictor
        self.reduce_predictor = reduce_predictor

    def forward(self, data_list):
        normal_ctx = []
        reduce_ctx = []
        for data in data_list:
            normal_tsr, normal_edge_list, normal_last_idx_list, \
                reduce_tsr, reduce_edge_list, reduce_last_idx_list = data
            normal_embedding = self.normal_cell_encoder(normal_tsr, normal_edge_list, normal_last_idx_list)
            reduce_embedding = self.reduce_cell_encoder(reduce_tsr, reduce_edge_list, reduce_last_idx_list)
            normal_embedding = normal_embedding[:, -1, :]
            reduce_embedding = reduce_embedding[:, -1, :]
            normal_ctx.append(normal_embedding)
            reduce_ctx.append(reduce_embedding)
        normal_ctx = torch.cat([normal_ctx[0], normal_ctx[1]], dim=-1)
        reduce_ctx = torch.cat([reduce_ctx[0], reduce_ctx[1]], dim=-1)
        batch_normal_logits = self.normal_predictor(normal_ctx)
        batch_reduce_logits = self.reduce_predictor(reduce_ctx)
        batch_logits = batch_normal_logits + batch_reduce_logits
        return batch_logits


class DartsBlockState:

    def __init__(self, normal_op_labels, normal_op_inds, normal_op_edges,
                 reduce_op_labels, reduce_op_inds, reduce_op_edges,
                 de2ni_map, complete_nb101_edge_list):
        self.normal_op_labels = normal_op_labels # list of op strings
        self.normal_op_inds = normal_op_inds # list of op indices
        self.normal_op_edges = normal_op_edges # list of edge pairs: (src_cell_node_idx, dst_cell_node_idx), ...
        self.reduce_op_labels = reduce_op_labels
        self.reduce_op_inds = reduce_op_inds
        self.reduce_op_edges = reduce_op_edges
        self.de2ni_map = de2ni_map
        self.complete_nb101_edge_list = complete_nb101_edge_list
        # The following are cached lazy state variables
        # Using these means that we should not modify a state object in-place and re-use it
        self.str_id = None
        self.normal_nb101_edges = None
        self.reduce_nb101_edges = None

    def de_tensorize(self):
        b = copy.deepcopy(self)
        return b

    def re_tensorize(self):
        return self

    def __len__(self):
        return len(self.normal_op_labels)

    def __str__(self):
        return "DartsBlockState[{}] Normal: {} | reduce: {}".format(len(self.normal_op_labels),
                ",".join(self.normal_op_labels), ",".join(self.reduce_op_labels))

    def __repr__(self):
        return str(self)

    def __deepcopy__(self, memodict={}):
        _state = DartsBlockState(copy.deepcopy(self.normal_op_labels),
                                 copy.deepcopy(self.normal_op_inds),
                                 copy.deepcopy(self.normal_op_edges),
                                 copy.deepcopy(self.reduce_op_labels),
                                 copy.deepcopy(self.reduce_op_inds),
                                 copy.deepcopy(self.reduce_op_edges),
                                 self.de2ni_map, self.complete_nb101_edge_list)
        return _state

    def to(self, dev):
        return self

    def build_nb101_edges(self):
        if self.normal_nb101_edges is None:
            self.normal_nb101_edges = get_nb101_edge_list_from_darts_cell(self.normal_op_labels, self.normal_op_edges,
                                                                          self.de2ni_map, self.complete_nb101_edge_list,
                                                                          output_node_label=DARTS_OUTPUT_NODE_LABEL)
        if self.reduce_nb101_edges is None:
            self.reduce_nb101_edges = get_nb101_edge_list_from_darts_cell(self.reduce_op_labels, self.reduce_op_edges,
                                                                          self.de2ni_map, self.complete_nb101_edge_list,
                                                                          output_node_label=DARTS_OUTPUT_NODE_LABEL)

    def get_gpn_compatible_input(self):
        self.build_nb101_edges()
        curr_cell_node_i = get_curr_cell_node_idx(len(self.normal_op_labels))
        normal_edge_tsr = torch.LongTensor(self.normal_nb101_edges).to(device())
        reduce_edge_tsr = torch.LongTensor(self.reduce_nb101_edges).to(device())
        normal_node_tsr = torch.LongTensor(self.normal_op_inds).unsqueeze(0).to(device())
        reduce_node_tsr = torch.LongTensor(self.reduce_op_inds).unsqueeze(0).to(device())
        return normal_node_tsr, [normal_edge_tsr], [normal_node_tsr.shape[1] - 1], \
               reduce_node_tsr, [reduce_edge_tsr], [reduce_node_tsr.shape[1] - 1], \
               curr_cell_node_i

    def gen_block_state_with_new_node(self, new_normal_label, new_normal_ind, new_normal_op_edges,
                                      new_reduce_label, new_reduce_ind, new_reduce_op_edges):
        new_normal_labels = copy.deepcopy(self.normal_op_labels) + [new_normal_label]
        new_normal_inds = copy.deepcopy(self.normal_op_inds) + [new_normal_ind]
        new_normal_op_edges = copy.deepcopy(self.normal_op_edges) + new_normal_op_edges
        new_reduce_labels = copy.deepcopy(self.reduce_op_labels) + [new_reduce_label]
        new_reduce_inds = copy.deepcopy(self.reduce_op_inds) + [new_reduce_ind]
        new_reduce_op_edges = copy.deepcopy(self.reduce_op_edges) + new_reduce_op_edges
        new_state = DartsBlockState(new_normal_labels, new_normal_inds, new_normal_op_edges,
                                    new_reduce_labels, new_reduce_inds, new_reduce_op_edges,
                                    self.de2ni_map, self.complete_nb101_edge_list)
        return new_state

    def get_str_id(self):
        if self.str_id is not None:
            return self.str_id
        from utils.nas_utils import edge_list_to_edge_matrix
        self.build_nb101_edges()
        normal_src_list, normal_dst_list = prune_self_edge_from_edge_list(self.normal_nb101_edges)
        normal_matrix = edge_list_to_edge_matrix([normal_src_list, normal_dst_list], len(self.normal_op_inds))
        fingerprint_normal = hash_module(normal_matrix, self.normal_op_inds)
        reduce_src_list, reduce_dst_list = prune_self_edge_from_edge_list(self.reduce_nb101_edges)
        reduce_matrix = edge_list_to_edge_matrix([reduce_src_list, reduce_dst_list], len(self.reduce_op_inds))
        fingerprint_reduce = hash_module(reduce_matrix, self.reduce_op_inds)
        self.str_id = fingerprint_normal + "|" + fingerprint_reduce
        return self.str_id


class BatchDartsBlockStates:
    """Only used during k epoch PPO updates"""
    def __init__(self, states):
        self.states = states
        self.normal_node_tsr = None
        self.reduce_node_tsr = None
        self.normal_edge_tsr_list = []
        self.reduce_edge_tsr_list = []
        self.normal_last_idx_list = []
        self.reduce_last_idx_list = []
        self._build_gpn_compatible_input()

    def __len__(self):
        return len(self.states)

    def __str__(self):
        return "BatchDartsBlockStates of size {}".format(len(self))

    def __repr__(self):
        return str(self)

    def _build_gpn_compatible_input(self, remove_self_edges=False):
        normal_tsr_list = []
        reduce_tsr_list = []
        self.normal_edge_tsr_list = []
        self.reduce_edge_tsr_list = []
        self.normal_last_idx_list = []
        self.reduce_last_idx_list = []
        aggr_normal_size = 0
        aggr_reduce_size = 0
        for state in self.states:
            assert isinstance(state, DartsBlockState)
            state.build_nb101_edges()
            normal_node_tsr = torch.LongTensor(state.normal_op_inds).unsqueeze(0)
            normal_tsr_list.append(normal_node_tsr)
            if remove_self_edges:
                normal_edge_list = prune_self_edge_from_edge_list(state.normal_nb101_edges)
            else:
                normal_edge_list = state.normal_nb101_edges
            normal_edge_tsr = torch.LongTensor(normal_edge_list)
            self.normal_edge_tsr_list.append(normal_edge_tsr)
            self.normal_last_idx_list.append(aggr_normal_size + normal_node_tsr.shape[1] - 1)
            aggr_normal_size += normal_node_tsr.shape[1]
            reduce_node_tsr = torch.LongTensor(state.reduce_op_inds).unsqueeze(0)
            reduce_tsr_list.append(reduce_node_tsr)
            if remove_self_edges:
                reduce_edge_list = prune_self_edge_from_edge_list(state.reduce_nb101_edges)
            else:
                reduce_edge_list = state.reduce_nb101_edges
            reduce_edge_tsr = torch.LongTensor(reduce_edge_list)
            self.reduce_edge_tsr_list.append(reduce_edge_tsr)
            self.reduce_last_idx_list.append(aggr_reduce_size + reduce_node_tsr.shape[1] - 1)
            aggr_reduce_size += reduce_node_tsr.shape[1]
        self.normal_node_tsr = torch.cat(normal_tsr_list, dim=0)
        self.reduce_node_tsr = torch.cat(reduce_tsr_list, dim=0)

    def get_gpn_compatible_input(self):
        curr_cell_node_i = get_curr_cell_node_idx(self.normal_node_tsr.shape[1])
        return self.normal_node_tsr, self.normal_edge_tsr_list, self.normal_last_idx_list, \
               self.reduce_node_tsr, self.reduce_edge_tsr_list, self.reduce_last_idx_list, \
               curr_cell_node_i


class DartsAction:

    def __init__(self, actor_name, timestep,
                 normal_op_probs, normal_input_node_probs,
                 reduce_op_probs, reduce_input_node_probs):
        self.actor_name = actor_name
        self.timestep = timestep
        self.normal_op_probs = normal_op_probs
        self.normal_input_node_probs = normal_input_node_probs
        self.reduce_op_probs = reduce_op_probs
        self.reduce_input_node_probs = reduce_input_node_probs
        self.normal_op_inds = Categorical(probs=self.normal_op_probs).sample()
        self.normal_input_node_inds = _masked_disjoint_sample(probs=self.normal_input_node_probs)
        self.reduce_op_inds = Categorical(probs=self.reduce_op_probs).sample()
        self.reduce_input_node_inds = _masked_disjoint_sample(probs=self.reduce_input_node_probs)
        self.normal_op_entropy = Categorical(probs=self.normal_op_probs).entropy()
        self.normal_op_entropy = self.normal_op_entropy.mean()
        self.reduce_op_entropy = Categorical(probs=self.reduce_op_probs).entropy()
        self.reduce_op_entropy = self.reduce_op_entropy.mean()
        self.normal_input_node_entropy = Categorical(probs=self.normal_input_node_probs).entropy()
        self.normal_input_node_entropy = self.normal_input_node_entropy.mean()
        self.reduce_input_node_entropy = Categorical(probs=self.reduce_input_node_probs).entropy()
        self.reduce_input_node_entropy = self.reduce_input_node_entropy.mean()
        self.op_entropy = torch.cat([self.normal_op_entropy.unsqueeze(0),
                                     self.reduce_op_entropy.unsqueeze(0)]).mean()
        self.input_node_entropy = torch.cat([self.normal_input_node_entropy.unsqueeze(0),
                                             self.reduce_input_node_entropy.unsqueeze(0)]).mean()
        self.entropy = torch.cat([self.op_entropy.unsqueeze(0),
                                  self.input_node_entropy.unsqueeze(0)]).mean()

    def __deepcopy__(self, memodict={}):
        a = DartsAction(self.actor_name, self.timestep,
                        copy.deepcopy(self.normal_op_probs), copy.deepcopy(self.normal_input_node_probs),
                        copy.deepcopy(self.reduce_op_probs), copy.deepcopy(self.reduce_input_node_probs))
        a.normal_op_inds = copy.deepcopy(self.normal_op_inds)
        a.normal_input_node_inds = copy.deepcopy(self.normal_input_node_inds)
        a.reduce_op_inds = copy.deepcopy(self.reduce_op_inds)
        a.reduce_input_node_inds = copy.deepcopy(self.reduce_input_node_inds)
        return a

    def de_tensorize(self):
        a = copy.deepcopy(self)
        members = [attr for attr in dir(a) if not callable(getattr(a, attr)) and not attr.startswith("__")]
        for attr in members:
            v = a.__getattribute__(attr)
            if isinstance(v, torch.Tensor):
                a.__setattr__(attr, get_tsr_value(v))
        return a

    def re_tensorize(self):
        members = [attr for attr in dir(self) if not callable(getattr(self, attr)) and not attr.startswith("__")]
        for attr in members:
            v = self.__getattribute__(attr)
            if v == "timestep": continue
            if isinstance(v, list):
                if "probs" in attr or "entropy" in attr:
                    self.__setattr__(attr, torch.FloatTensor(v))
                else:
                    self.__setattr__(attr, torch.LongTensor(v))
            elif isinstance(v, int):
                self.__setattr__(attr, torch.LongTensor([v]).squeeze())
            elif isinstance(v, float):
                self.__setattr__(attr, torch.FloatTensor([v]).squeeze())
        return self

    def to(self, dev):
        members = [attr for attr in dir(self) if not callable(getattr(self, attr)) and not attr.startswith("__")]
        for attr in members:
            v = self.__getattribute__(attr)
            if isinstance(v, torch.Tensor):
                self.__setattr__(attr, v.to(dev))
        return self

    def get_discrete_action(self):
        normal_op_inds = get_tsr_value(self.normal_op_inds)
        normal_input_node_inds = get_tsr_value(self.normal_input_node_inds)
        reduce_op_inds = get_tsr_value(self.reduce_op_inds)
        reduce_input_node_inds = get_tsr_value(self.reduce_input_node_inds)
        return normal_op_inds, normal_input_node_inds, \
               reduce_op_inds, reduce_input_node_inds

    def get_entropy(self):
        return self.entropy

    def __str__(self):
        s = "DartsAction at time {} by actor {}".format(self.timestep, self.actor_name)
        s += ", normal op inds: {}".format(str(get_tsr_value(self.normal_op_inds)))
        s += ", normal op probs: {}".format(str(get_tsr_value(self.normal_op_probs)))
        s += ", reduce op inds: {}".format(str(get_tsr_value(self.reduce_op_inds)))
        s += ", reduce op probs: {}".format(str(get_tsr_value(self.reduce_op_probs)))
        return s

    def __repr__(self):
        return str(self)


class BatchDartsActions:
    """Only used during k epoch PPO updates"""
    def __init__(self, actions):
        self.actions = actions
        batch_normal_op_probs = []
        batch_normal_op_inds = []
        batch_normal_input_node_probs = []
        batch_normal_input_node_inds = []
        batch_reduce_op_probs = []
        batch_reduce_op_inds = []
        batch_reduce_input_node_probs = []
        batch_reduce_input_node_inds = []
        batch_op_entropy = []
        batch_edge_entropy = []
        for action in self.actions:
            assert isinstance(action, DartsAction)
            batch_normal_op_probs.append(action.normal_op_probs.unsqueeze(0))
            batch_normal_op_inds.append(action.normal_op_inds.unsqueeze(0))
            batch_normal_input_node_probs.append(action.normal_input_node_probs.unsqueeze(0))
            batch_normal_input_node_inds.append(action.normal_input_node_inds.unsqueeze(0))
            batch_reduce_op_probs.append(action.reduce_op_probs.unsqueeze(0))
            batch_reduce_op_inds.append(action.reduce_op_inds.unsqueeze(0))
            batch_reduce_input_node_probs.append(action.reduce_input_node_probs.unsqueeze(0))
            batch_reduce_input_node_inds.append(action.reduce_input_node_inds.unsqueeze(0))
            batch_op_entropy.append(action.op_entropy.unsqueeze(0))
            batch_edge_entropy.append(action.input_node_entropy.unsqueeze(0))
        self.batch_normal_op_probs = torch.cat(batch_normal_op_probs, dim=0)
        self.batch_normal_op_inds = torch.cat(batch_normal_op_inds, dim=0)
        self.batch_normal_input_node_probs = torch.cat(batch_normal_input_node_probs, dim=0)
        self.batch_normal_input_node_inds = torch.cat(batch_normal_input_node_inds, dim=0)
        self.batch_reduce_op_probs = torch.cat(batch_reduce_op_probs, dim=0)
        self.batch_reduce_op_inds = torch.cat(batch_reduce_op_inds, dim=0)
        self.batch_reduce_input_node_probs = torch.cat(batch_reduce_input_node_probs, dim=0)
        self.batch_reduce_input_node_inds = torch.cat(batch_reduce_input_node_inds, dim=0)
        self.batch_op_entropy = torch.cat(batch_op_entropy, dim=0).mean()
        self.batch_edge_entropy = torch.cat(batch_edge_entropy, dim=0).mean()
        self.batch_normal_op_log_probs = \
            Categorical(probs=self.batch_normal_op_probs).log_prob(self.batch_normal_op_inds)
        self.batch_reduce_op_log_probs = \
            Categorical(probs=self.batch_reduce_op_probs).log_prob(self.batch_reduce_op_inds)
        self.batch_normal_input_node_log_probs = \
            Categorical(probs=self.batch_normal_input_node_probs).log_prob(self.batch_normal_input_node_inds)
        self.batch_reduce_input_node_log_probs = \
            Categorical(probs=self.batch_reduce_input_node_probs).log_prob(self.batch_reduce_input_node_inds)

    def __len__(self):
        return len(self.actions)

    def __str__(self):
        return "BatchDartsActions of size {}".format(len(self))

    def __repr__(self):
        return str(self)

    def to(self, dev):
        for action in self.actions:
            action.to(dev)
        return self

    def get_op_entropy(self):
        return self.batch_op_entropy

    def get_edge_entropy(self):
        return self.batch_edge_entropy

    def get_action_log_probs(self, batch_normal_op_inds=None, batch_normal_input_node_inds=None,
                             batch_reduce_op_inds=None, batch_reduce_input_node_inds=None):
        if batch_normal_op_inds is None:
            batch_normal_op_log_probs = self.batch_normal_op_log_probs
        else:
            batch_normal_op_log_probs = \
                Categorical(probs=self.batch_normal_op_probs).log_prob(batch_normal_op_inds)
        if batch_normal_input_node_inds is None:
            batch_normal_input_node_log_probs = self.batch_normal_input_node_log_probs
        else:
            batch_normal_input_node_log_probs = \
                Categorical(probs=self.batch_normal_input_node_probs).log_prob(batch_normal_input_node_inds)
        if batch_reduce_op_inds is None:
            batch_reduce_op_log_probs = self.batch_reduce_op_log_probs
        else:
            batch_reduce_op_log_probs = \
                Categorical(probs=self.batch_reduce_op_probs).log_prob(batch_reduce_op_inds)
        if batch_reduce_input_node_inds is None:
            batch_reduce_input_node_log_probs = self.batch_reduce_input_node_log_probs
        else:
            batch_reduce_input_node_log_probs = \
                Categorical(probs=self.batch_reduce_input_node_probs).log_prob(batch_reduce_input_node_inds)
        return batch_normal_op_log_probs, batch_normal_input_node_log_probs, \
               batch_reduce_op_log_probs, batch_reduce_input_node_log_probs


class DartsEnv:

    def __init__(self, actor, init_state, op2idx, reward_module, max_allowed_steps, output_node_label,
                 scaffold_node_labels, pos_step_reward=0., neg_step_reward=-0.1):
        self._actor = actor
        self._output_node_label = output_node_label
        self._op2idx = op2idx
        self._reward_module = reward_module
        self._curr_state = init_state
        self._init_state = init_state
        self._max_allowed_steps = max_allowed_steps
        self._num_completed_steps = 0
        self._pos_step_reward = pos_step_reward
        self._neg_step_reward = neg_step_reward
        self.scaffold_node_labels = scaffold_node_labels

    @property
    def final_reward_module(self):
        return self._reward_module

    def load_state_dict(self, actor_state_dict):
        if actor_state_dict is not None:
            self._actor.load_state_dict(actor_state_dict)

    def get_init_state(self):
        return self._init_state

    def set_actor(self, actor):
        self._actor = actor

    def get_actor(self):
        return self._actor

    def get_curr_state(self):
        return self._curr_state

    def set_curr_state(self, state):
        self._curr_state = state

    def get_num_steps(self):
        return self._num_completed_steps

    def eval(self):
        self._reward_module.eval()
        self._actor.eval()

    def train(self):
        self._reward_module.train()
        self._actor.train()

    def to(self, dev):
        self._reward_module = self._reward_module.to(dev)
        self._actor = self._actor.to(dev)
        return self

    def reset(self):
        self._curr_state = self._init_state
        self._num_completed_steps = 0

    def get_reward_for_state(self, state):
        return self._reward_module.get_reward(state)

    def step(self, action:DartsAction):
        input_state = self._curr_state
        action_valid = is_darts_action_valid(action)
        new_state = self.gen_new_state_from_action(self._curr_state, action, action_valid)
        assert isinstance(new_state, DartsBlockState)
        self._num_completed_steps += 1
        if self._num_completed_steps >= self._max_allowed_steps and action_valid:
            # When _num_completed_steps == 4, we have a cell with 6 nodes, 4 internal nodes, 4x2 = 8 op edges
            final_state_reached = True
            # Manually add the output node
            input_cell_node_edges = get_output_cell_node_input_edges()
            new_state = new_state.gen_block_state_with_new_node(self._output_node_label,
                                                                self._op2idx[self._output_node_label],
                                                                input_cell_node_edges,
                                                                self._output_node_label,
                                                                self._op2idx[self._output_node_label],
                                                                input_cell_node_edges)
            final_reward = self.get_reward_for_state(new_state)
            reward = final_reward
        elif not action_valid:
            final_state_reached = True
            reward = self._neg_step_reward
        else:
            final_state_reached = False
            reward = self._pos_step_reward
        self._curr_state = new_state
        return final_state_reached, input_state, reward, new_state

    def gen_new_state_from_action(self, input_state, action:DartsAction, action_valid):
        # If the action is invalid, the we do not modify the input_state
        # This is ok because in that case the new_state will not be given to the encoder to get a state value, it's 0
        if not action_valid:
            return input_state
        new_normal_op_inds, new_normal_input_node_inds, \
            new_reduce_op_inds, new_reduce_input_node_inds = action.get_discrete_action()
        assert len(new_normal_op_inds) == len(new_normal_input_node_inds) == \
            len(new_reduce_op_inds) == len(new_reduce_input_node_inds)
        curr_state = input_state
        curr_cell_node_i = get_curr_cell_node_idx(len(input_state))
        dst_cell_node_i = curr_cell_node_i + len(self._init_state)
        for i in range(len(new_normal_op_inds)):
            new_normal_label = self.scaffold_node_labels[new_normal_op_inds[i]]
            new_reduce_label = self.scaffold_node_labels[new_reduce_op_inds[i]]
            normal_cell_edges = [(new_normal_input_node_inds[i], dst_cell_node_i)]
            reduce_cell_edges = [(new_reduce_input_node_inds[i], dst_cell_node_i)]
            curr_state = curr_state.gen_block_state_with_new_node(new_normal_label,
                                                                  self._op2idx[new_normal_label],
                                                                  normal_cell_edges,
                                                                  new_reduce_label,
                                                                  self._op2idx[new_reduce_label],
                                                                  reduce_cell_edges)

        return curr_state

    def __str__(self):
        return "DartsEnv for actor {}, num_steps: {}".format(self._actor.get_name(), self._num_completed_steps)

    def __repr__(self):
        return str(self)


class DartsEnvManager:

    def __init__(self, parallel_actors, n_parallel_actors, init_state, op2idx, scaffold_node_labels,
                 reward_module, max_allowed_steps, output_node_label):
        self.init_state = init_state
        self.max_allowed_steps = max_allowed_steps
        self.op2idx = op2idx
        self.reward_module = reward_module
        self.output_node_label = output_node_label
        self.n_parallel_actors = n_parallel_actors
        self.parallel_actors = parallel_actors
        self.curr_env_idx = 0
        for i, actor in enumerate(self.parallel_actors):
            actor.name += "_{}".format(i)
        self.parallel_envs = [DartsEnv(actor, copy.deepcopy(init_state), op2idx, reward_module,
                                       max_allowed_steps, output_node_label, scaffold_node_labels)
                              for actor in self.parallel_actors]

    def share_memory(self):
        self.reward_module.share_memory()

    def load_checkpoint(self, actor_state_dict):
        for env in self.parallel_envs:
            env.load_state_dict(actor_state_dict)

    def get_actor_rep(self):
        return self.parallel_envs[0].get_actor()

    def get_env_rep(self):
        return self.parallel_envs[0]

    def get_env_actor_rep(self):
        return self.get_env_rep(), self.get_actor_rep()

    def re_clone_actors(self, new_actor):
        for i, actor in enumerate(self.parallel_actors):
            if actor == new_actor: continue
            actor.load_state_dict(new_actor.state_dict())
            self.parallel_envs[i].set_actor(actor)

    def to(self, dev):
        for env in self.parallel_envs:
            env.to(dev)
        return self

    def reset(self):
        for env in self.parallel_envs:
            env.reset()

    def eval(self):
        for env in self.parallel_envs:
            env.eval()

    def train(self):
        for env in self.parallel_envs:
            env.train()

    def __len__(self):
        return len(self.parallel_envs)

    def __str__(self):
        return "DartsEnvManager with {} parallel environments".format(len(self.parallel_envs))

    def __repr__(self):
        return str(self)

    def __iter__(self):
        self.curr_env_idx = 0
        return self

    def __next__(self):
        if self.curr_env_idx < len(self.parallel_envs):
            env = self.parallel_envs[self.curr_env_idx]
            self.curr_env_idx += 1
            return env
        else:
            self.curr_env_idx = 0
            raise StopIteration


class DartsActorCritic(torch.nn.Module):

    def __init__(self, name, policy_net, value_net):
        super(DartsActorCritic, self).__init__()
        self.name = name
        self.policy_net = policy_net
        self.value_net = value_net

    def forward(self, *args):
        return self.policy_net(*args)

    def get_value_for_state(self, state:DartsBlockState):
        node_embedding, _ = self.policy_net(*state.get_gpn_compatible_input())
        value = self.value_net(node_embedding).squeeze()
        return value

    def get_batch_values_for_states(self, batch_states:BatchDartsBlockStates):
        node_embedding, _ = self.policy_net(*batch_states.get_gpn_compatible_input())
        values = self.value_net(node_embedding).squeeze()
        return values

    def get_action_for_state(self, state:DartsBlockState, timestep=0):
        _, (normal_op_probs, normal_input_node_probs, reduce_op_probs, reduce_input_node_probs) = \
            self.policy_net(*state.get_gpn_compatible_input())
        assert normal_op_probs.shape[0] == 1
        assert normal_input_node_probs.shape[0] == 1
        assert reduce_op_probs.shape[0] == 1
        assert reduce_input_node_probs.shape[0] == 1
        normal_op_probs = normal_op_probs.squeeze(0)
        normal_input_node_probs = normal_input_node_probs.squeeze(0)
        reduce_op_probs = reduce_op_probs.squeeze(0)
        reduce_input_node_probs = reduce_input_node_probs.squeeze(0)
        action = DartsAction(self.name, timestep,
                             normal_op_probs, normal_input_node_probs,
                             reduce_op_probs, reduce_input_node_probs)
        return action

    def get_batch_actions_for_states(self, batch_states:BatchDartsBlockStates, timestep=0):
        _, (b_normal_op_probs, b_normal_input_node_probs, b_reduce_op_probs, b_reduce_input_node_probs) = \
            self.policy_net(*batch_states.get_gpn_compatible_input())
        actions = []
        for si, state in enumerate(batch_states.states):
            normal_op_probs = b_normal_op_probs[si, :]
            normal_input_node_probs = b_normal_input_node_probs[si, :]
            reduce_op_probs = b_reduce_op_probs[si, :]
            reduce_input_node_probs = b_reduce_input_node_probs[si, :]
            action = DartsAction(self.name, timestep,
                                 normal_op_probs, normal_input_node_probs,
                                 reduce_op_probs, reduce_input_node_probs).to(device())
            actions.append(action)
        batch_actions = BatchDartsActions(actions)
        return batch_actions

    def get_name(self):
        return self.name


class DartsPPOBatchDataLoader:

    def __init__(self, batch_size, scale_reward=False, reward_discount_factor=1.0, gae_lambda=1.0,
                 min_reward=0., max_reward=None):
        self.batch_size = batch_size
        self.curr_batch_idx = 0
        self.trajectories = []
        self.batches = []
        self.n_instances = 0
        self.scale_reward = scale_reward
        self.data_built = False
        self.reward_discount_factor = reward_discount_factor
        self.gae_lambda = gae_lambda
        self.min_reward = min_reward
        self.max_reward = max_reward

    def clear(self):
        self.curr_batch_idx = 0
        self.trajectories = []
        self.batches = []
        self.data_built = False

    def add_trajectory(self, state_pairs, actions, rewards, input_state_values, end_state_value):
        assert len(state_pairs) == len(actions) == len(rewards) == len(input_state_values)
        self.trajectories.append( (state_pairs, actions, rewards, input_state_values, end_state_value) )

    def get_next_batch(self):
        batch = self.batches[self.curr_batch_idx]
        self.curr_batch_idx += 1
        return batch

    def build_batches(self):
        min_reward, max_reward = None, None
        if self.scale_reward:
            max_reward = self.max_reward if self.max_reward is not None else global_max([t[2] for t in self.trajectories])
            min_reward = self.min_reward if self.min_reward is not None else global_min([t[2] for t in self.trajectories])
        instances = []
        for state_pairs, actions, rewards, input_state_values, end_state_value in tqdm(self.trajectories,
                                                                                       desc="Building instances from trajectories",
                                                                                       ascii=True, leave=False):
            if self.scale_reward:
                scaled_rewards = [
                    exp_scale(r, old_range_min=min_reward, old_range_max=max_reward) if min_reward <= r <= max_reward else r
                        for r in rewards
                ]
                advantages = compute_t_step_gae(scaled_rewards, input_state_values, end_state_value,
                                                self.reward_discount_factor, self.gae_lambda)
            else:
                advantages = compute_t_step_gae(rewards, input_state_values, end_state_value,
                                                self.reward_discount_factor, self.gae_lambda)
            for (input_state, output_state), action, reward, value, adv in zip(state_pairs, actions, rewards,
                                                                               input_state_values, advantages):
                instances.append( (input_state, action, reward, value, adv) )
        random.shuffle(instances)
        self.n_instances = len(instances)
        print("Building batches")
        # Group instances by state dimensions
        size2instances = collections.defaultdict(list)
        for inst in instances:
            partial_graph_size = len(inst[0])
            size2instances[partial_graph_size].append(inst)
        for size, instances in size2instances.items():
            inst_i = 0
            while inst_i < len(instances):
                batch_instances = instances[inst_i: inst_i + self.batch_size]
                batch_states = BatchDartsBlockStates([t[0] for t in batch_instances])
                batch_actions = BatchDartsActions([t[1] for t in batch_instances])
                batch_rewards = torch.FloatTensor([t[2] for t in batch_instances])
                batch_values = torch.FloatTensor([t[3] for t in batch_instances])
                batch_advantages = torch.FloatTensor([t[4] for t in batch_instances])
                self.batches.append((batch_states, batch_actions, batch_rewards, batch_values, batch_advantages))
                inst_i += self.batch_size
        self.data_built = True
        self.shuffle()

    def shuffle(self):
        random.shuffle(self.batches)

    @property
    def n_trajectories(self):
        return len(self.trajectories)

    @property
    def n_batches(self):
        return len(self)

    def __len__(self):
        return len(self.batches)

    def __iter__(self):
        self.curr_batch_idx = 0
        return self

    def __next__(self):
        if not self.data_built:
            self.build_batches()
        if self.curr_batch_idx < self.n_batches:
            batch = self.get_next_batch()
            return batch
        else:
            self.curr_batch_idx = 0
            self.shuffle()
            raise StopIteration


class DartsDisRewardModule:

    def __init__(self, d_trainer, output_node_label, n_random_eval,
                 node_label2feature, rm_cand_node_labels,
                 n_truth_per_src=100, n_fake_total=400, aggr_fake=True, rm_depth=3, n_mutations=10,
                 max_n_nodes=7, enable_mutation=False, pos_reward=0.0, neg_reward=-0.1,
                 lat_weight=0.5, max_d_reward=1.0, valid_block_keys=None):
        self.d_trainer = d_trainer
        self.model = self.d_trainer.get_model()
        self.max_d_reward = max_d_reward
        self.output_node_label = output_node_label
        self.pos_reward = pos_reward
        self.neg_reward = neg_reward
        self.n_truth_per_src = n_truth_per_src
        self.n_fake_total = n_fake_total
        self.src_block2perfs = {}
        self.src_blocks = []
        self.truth_block2perfs = {}
        self.truth_blocks = []
        self.src2truth = {}
        self.n_random_eval = n_random_eval
        self.node_label2feature = node_label2feature
        self.rm_cand_node_labels = rm_cand_node_labels
        self.rm_depth = rm_depth
        self.n_mutations = n_mutations
        self.max_n_nodes = max_n_nodes
        self.lat_weight = lat_weight
        self.valid_block_keys = valid_block_keys
        self.enable_mutation = enable_mutation
        self.aggr_fake = aggr_fake
        self.past_fake_blocks = []

    def train(self):
        self.model.train()

    def eval(self):
        self.model.eval()

    def to(self, dev):
        self.d_trainer.to(dev)
        return self

    def get_model(self):
        return self.model

    def share_memory(self):
        self.model.share_memory()

    def clear(self):
        self.src_block2perfs = {}
        self.src_blocks = []
        self.truth_block2perfs = {}
        self.truth_blocks = []
        self.src2truth = {}
        self.past_fake_blocks = []

    def get_fake_data_size(self):
        if self.enable_mutation:
            return self.n_fake_total
        else:
            return len(self.truth_blocks)

    def build_truth_blocks(self, src_block2perfs, log_f=print):
        self.clear()
        self.src_blocks = [v[0] for k, v in src_block2perfs.items()]
        self.src_block2perfs = src_block2perfs
        if self.enable_mutation:
            raise NotImplementedError
        else:
            self.truth_blocks = [b for b in self.src_blocks]
            for src_b in self.src_blocks:
                self.src2truth[src_b.get_str_id()] = self.truth_blocks
        log_f("{} truth blocks built".format(len(self.truth_blocks)))

    def d_train(self, fake_blocks, log_f=None):
        log_f("Building data pairs")
        if self.aggr_fake:
            unique_fake_blocks = UniqueList(key_func=lambda _b: _b.get_str_id())
            unique_fake_blocks.extend(fake_blocks)
            unique_fake_blocks.extend(self.past_fake_blocks)
            unique_fake_blocks = unique_fake_blocks.tolist()
        else:
            unique_fake_blocks = fake_blocks
        # Build truth pairs
        truth_block_pairs = []
        truth_keys = set()
        for src_key, truth_blocks in self.src2truth.items():
            src_block = self.src_block2perfs[src_key][0]
            for b in truth_blocks:
                truth_block_pairs.append((src_block, b))
                truth_keys.add(b.get_str_id())
        # Build fake pairs
        fake_block_pairs = []
        fake_keys = set()
        for src_block in self.src_blocks:
            for b in unique_fake_blocks:
                k = b.get_str_id()
                fake_keys.add(k)
                # if k in truth_keys:
                #     # We do not let the discriminator to get confused, for the most accurate reward
                #     continue
                fake_block_pairs.append((src_block, b))
        common_keys = truth_keys.intersection(fake_keys)
        log_f("Truth/fake key overlap: {}%".format(100. * len(common_keys) / (len(truth_keys.union(fake_keys)))))
        self.d_trainer.train(truth_block_pairs, fake_block_pairs, log_f)
        # Update past fake blocks
        unique_fake_blocks = UniqueList(key_func=lambda _b: _b.get_str_id())
        unique_fake_blocks.extend(fake_blocks)
        unique_fake_blocks.extend(self.past_fake_blocks)
        self.past_fake_blocks = unique_fake_blocks.tolist()

    def _get_d_reward(self, block):
        d_model = self.model
        reward = 0.
        with torch.no_grad():
            d_model.eval()
            for p_block in self.src_blocks:
                p_block = copy.deepcopy(p_block)
                gen_block_score = d_model(build_d_pair_input(p_block, block))
                gen_block_score = torch.softmax(gen_block_score, dim=-1).squeeze()
                reward = max(reward, get_tsr_value(gen_block_score[1]))
        return reward

    def get_reward(self, state):
        assert self.src_block2perfs is not None
        key = state.get_str_id()
        if key in self.src_block2perfs:
            rv = 0.  # No additional reward if generated the same src blocks!
        else:
            rv = self._get_d_reward(state)
        return rv


class DartsDisTrainer:

    def __init__(self, n_epochs, model, init_lr, batch_size, max_grad_norm=5.0, early_stop_train_f1=0.9):
        self.criterion = torch.nn.CrossEntropyLoss()
        # self.criterion = LabelSmoothingCrossEntropy()
        self.init_lr = init_lr
        self.original_model = model
        self.curr_model = model
        self.n_epochs = n_epochs
        self.batch_size = batch_size
        self.max_grad_norm = max_grad_norm
        self.early_stop_train_f1 = early_stop_train_f1

    def to(self, dev):
        self.curr_model = self.curr_model.to(dev)
        self.criterion = self.criterion.to(dev)
        return self

    def _make_optimizer(self, model):
        return torch.optim.Adam(model.parameters(), lr=self.init_lr)

    def get_model(self):
        return self.curr_model

    def train(self, truth_block_pairs, fake_block_pairs, log_f=None):
        self.curr_model.train()
        optimizer = self._make_optimizer(self.curr_model)
        truth_data = [(p[0], p[1], torch.LongTensor([1]).to(device())) for p in truth_block_pairs]
        fake_data = [(p[0], p[1], torch.LongTensor([0]).to(device())) for p in fake_block_pairs]
        if log_f is not None:
            log_f("Training discriminator with {} truth, {} fake instances".format(len(truth_data), len(fake_data)))
        assert len(truth_data) > 0, "DisTrainer received 0 truth data"
        assert len(fake_data) > 0, "DisTrainer received 0 fake data"
        data = truth_data + fake_data
        for epoch in range(self.n_epochs):
            random.shuffle(data)
            total_loss, n_insts = 0, 0
            y_pred, y_true = [], []
            batch_loss = torch.FloatTensor([0.]).to(device())
            for b1, b2, target in tqdm(data, desc="Training on {} instances".format(len(data)),
                                      leave=False, ascii=True):
                target.requires_grad = False
                block_score = self.curr_model(build_d_pair_input(b1, b2))
                inst_loss = self.criterion(block_score, target)
                batch_loss += inst_loss
                total_loss += get_tsr_value(inst_loss)
                if (n_insts + 1) % self.batch_size == 0 or n_insts == len(data) - 1:
                    optimizer.zero_grad()
                    batch_loss.backward()
                    torch.nn.utils.clip_grad_norm_(self.curr_model.parameters(), self.max_grad_norm)
                    optimizer.step()
                    batch_loss = torch.FloatTensor([0.]).to(device())
                n_insts += 1
                _, pred_ind = block_score.max(1)
                pred_ind = get_tsr_value(pred_ind)
                truth_ind = get_tsr_value(target)
                y_pred.append(pred_ind)
                y_true.append(truth_ind)
            rv_loss = total_loss / n_insts
            if len(set(y_pred)) <= 1:
                train_f1 = 0.
            else:
                train_f1 = f1_score(y_true=y_true, y_pred=y_pred)
            if log_f is not None: log_f("D-train epoch {} of {} loss: {:.3f}, f1: {:.3f}".format(epoch+1, self.n_epochs,
                                                                                                 rv_loss, train_f1))
            if train_f1 >= self.early_stop_train_f1:
                if log_f is not None: log_f("D-train early stop at epoch {}".format(epoch + 1))
                break


def is_darts_cell_valid(block:DartsBlockState):
    # As long as there is no more than one edge between two cell nodes, the cell is valid
    def _check(edges):
        unique_edges = set()
        for src, dst in edges:
            key = "{}->{}".format(src, dst)
            if key in unique_edges:
                return False
            unique_edges.add(key)
        return True
    return _check(block.normal_op_edges) and _check(block.reduce_op_edges)


def is_darts_action_valid(action:DartsAction):
    # Check no duplicated input node indices
    def _check(input_node_inds):
        if isinstance(input_node_inds, torch.Tensor):
            input_node_inds = input_node_inds.tolist()
        unique_edges = set()
        for ind in input_node_inds:
            if ind in unique_edges:
                return False
            unique_edges.add(ind)
        return True
    return _check(action.normal_input_node_inds) and _check(action.reduce_input_node_inds)


def make_darts_block_discriminator(num_unique_nodes, dis_in_channels, dis_hidden_size, dis_out_channels,
                                   dis_gnn_constructor, dis_num_layers, dis_dropout_prob):
    pre_embedding = torch.nn.Embedding(num_unique_nodes, dis_in_channels)
    normal_encoder = GraphEncoder(dis_in_channels, dis_hidden_size, dis_out_channels, dis_gnn_constructor,
                                  node_pre_embedding=pre_embedding, activ=torch.nn.Tanh(),
                                  num_layers=dis_num_layers, dropout_prob=dis_dropout_prob)
    reduce_encoder = GraphEncoder(dis_in_channels, dis_hidden_size, dis_out_channels, dis_gnn_constructor,
                                  node_pre_embedding=pre_embedding, activ=torch.nn.Tanh(),
                                  num_layers=dis_num_layers, dropout_prob=dis_dropout_prob)
    discriminator = DartsBlockDiscriminator(normal_encoder, reduce_encoder,
                                            normal_predictor=torch.nn.Linear(2 * dis_out_channels, 2),
                                            reduce_predictor=torch.nn.Linear(2 * dis_out_channels, 2))
    return discriminator


def make_darts_gpn(num_unique_nodes, gpn_in_channels, gpn_hidden_size, gpn_out_channels, gpn_gnn_constructor,
                   gpn_num_layers, gpn_dropout_prob):
    pre_embedding = torch.nn.Embedding(num_unique_nodes, gpn_in_channels)
    normal_encoder = GraphEncoder(gpn_in_channels, gpn_hidden_size, gpn_out_channels, gpn_gnn_constructor,
                                  node_pre_embedding=pre_embedding, activ=torch.nn.Tanh(), add_noise_prior=True,
                                  num_layers=gpn_num_layers, dropout_prob=gpn_dropout_prob)
    reduce_encoder = GraphEncoder(gpn_in_channels, gpn_hidden_size, gpn_out_channels, gpn_gnn_constructor,
                                     node_pre_embedding=pre_embedding, activ=torch.nn.Tanh(), add_noise_prior=True,
                                     num_layers=gpn_num_layers, dropout_prob=gpn_dropout_prob)
    normal_cell_predictor = RNNCellPredictor(gpn_out_channels, gpn_hidden_size, num_unique_nodes - 2)
    reduce_cell_predictor = RNNCellPredictor(gpn_out_channels, gpn_hidden_size, num_unique_nodes - 2)
    action_predictor = DartsActionPredictor(normal_cell_predictor, reduce_cell_predictor)
    model = DartsBlockGPN(normal_encoder, reduce_encoder, action_predictor)
    return model


def make_darts_actor_critic(policy_net, gpn_out_channels):
    value_net = torch.nn.Sequential(
        NodeAggr(2 * gpn_out_channels, aggr_dim=1, method="sum", enable_aggr_net=False),
        torch.nn.Linear(2 * gpn_out_channels, 1)
    )
    actor_critic = DartsActorCritic("DartsActorCritic", policy_net, value_net)
    return actor_critic


def make_darts_env_manager(actor_critic, n_parallel_actors, init_state, op2idx,
                           reward_module, max_allowed_steps, output_node_label, scaffold_node_labels):
    parallel_actors = [actor_critic] + [copy.deepcopy(actor_critic) for _ in range(n_parallel_actors - 1)]
    env_manager = DartsEnvManager(parallel_actors, n_parallel_actors, init_state, op2idx, scaffold_node_labels,
                                  reward_module, max_allowed_steps, output_node_label)
    return env_manager


def darts_ppo_train(env_manager, optimizer, value_criterion, n_train_rounds, k_epochs, t_timesteps, batch_size,
                    reward_discount_factor, clip_epsilon, gae_lambda, c1, c2, c3, book_keeper,
                    actor_update_episode_interval, parallel_explore=False):
    env_manager.train()
    ppo_data_loader = DartsPPOBatchDataLoader(batch_size, reward_discount_factor=reward_discount_factor,
                                              gae_lambda=gae_lambda)
    block_w_rewards = UniqueList(key_func=lambda t: t[0].get_str_id())
    for i in range(n_train_rounds):
        book_keeper.log("Training round {} of {}".format(i + 1, n_train_rounds))
        book_keeper.log("Running {} actor(s), each for {} steps".format(len(env_manager), t_timesteps))
        book_keeper.log("Episodes per actor: {}".format(actor_update_episode_interval))
        env_manager.train()
        # Exploration
        if parallel_explore:
            br_list = run_parallel_exploration(env_manager, ppo_data_loader, actor_update_episode_interval, t_timesteps,
                                               log_f=book_keeper.log)
        else:
            br_list = run_seq_exploration(env_manager, ppo_data_loader, actor_update_episode_interval, t_timesteps,
                                          log_f=book_keeper.log)
        block_w_rewards.extend(br_list)
        # Train generator(actor)
        updated_actor = env_manager.get_actor_rep()
        ppo_loss = run_darts_k_ppo_batch_update_epoch(updated_actor, optimizer, ppo_data_loader, value_criterion,
                                                      k_epochs, clip_epsilon, c1, c2, c3)
        book_keeper.log("Avg PPO loss: {}".format(ppo_loss))
        env_manager.re_clone_actors(updated_actor)
        ppo_data_loader.clear()
    return env_manager.get_actor_rep(), block_w_rewards


def run_darts_k_ppo_batch_update_epoch(actor, optimizer, data_loader, value_criterion, k_epochs,
                                       clip_epsilon, c1, c2, c3):
    actor.train()
    avg_loss, num_updates = 0., 0
    for e in tqdm(range(k_epochs), desc="PPO train", ascii=True, leave=False):
        for i, batch in enumerate(data_loader):
            b_input_states, b_old_actions, _, old_state_values, advantages = batch
            b_old_actions = b_old_actions.to(device())
            advantages = advantages.to(device())
            old_state_values = old_state_values.to(device())
            b_new_actions = actor.get_batch_actions_for_states(b_input_states)
            new_state_values = actor.get_batch_values_for_states(b_input_states)
            assert isinstance(b_old_actions, BatchDartsActions)
            assert isinstance(b_new_actions, BatchDartsActions)
            old_normal_op_logp, old_normal_input_node_logp, \
                old_reduce_op_logp, old_reduce_input_node_logp = b_old_actions.get_action_log_probs()
            new_normal_op_logp, new_normal_input_node_logp, \
                new_reduce_op_logp, new_reduce_input_node_logp = \
                    b_new_actions.get_action_log_probs(
                        batch_normal_op_inds=b_old_actions.batch_normal_op_inds,
                        batch_normal_input_node_inds=b_old_actions.batch_normal_input_node_inds,
                        batch_reduce_op_inds=b_old_actions.batch_reduce_op_inds,
                        batch_reduce_input_node_inds=b_old_actions.batch_reduce_input_node_inds)
            normal_op_1_ratio = torch.exp(new_normal_op_logp[:, 0] - old_normal_op_logp[:, 0].detach())
            normal_op_2_ratio = torch.exp(new_normal_op_logp[:, 1] - old_normal_op_logp[:, 1].detach())
            reduce_op_1_ratio = torch.exp(new_reduce_op_logp[:, 0] - old_reduce_op_logp[:, 0].detach())
            reduce_op_2_ratio = torch.exp(new_reduce_op_logp[:, 1] - old_reduce_op_logp[:, 1].detach())
            normal_input_node_1_ratio = torch.exp(
                new_normal_input_node_logp[:, 0] - old_normal_input_node_logp[:, 0].detach())
            normal_input_node_2_ratio = torch.exp(
                new_normal_input_node_logp[:, 1] - old_normal_input_node_logp[:, 1].detach())
            reduce_input_node_1_ratio = torch.exp(
                new_reduce_input_node_logp[:, 0] - old_reduce_input_node_logp[:, 0].detach())
            reduce_input_node_2_ratio = torch.exp(
                new_reduce_input_node_logp[:, 1] - old_reduce_input_node_logp[:, 1].detach())

            def _get_policy_loss(b_ratios):
                _surr1 = b_ratios * advantages
                _surr2 = torch.clamp(b_ratios, 1 - clip_epsilon, 1 + clip_epsilon) * advantages
                _policy_loss = -1. * torch.min(_surr1, _surr2)
                return _policy_loss.mean()

            normal_op_1_loss = _get_policy_loss(normal_op_1_ratio)
            normal_op_2_loss = _get_policy_loss(normal_op_2_ratio)
            reduce_op_1_loss = _get_policy_loss(reduce_op_1_ratio)
            reduce_op_2_loss = _get_policy_loss(reduce_op_2_ratio)
            normal_input_node_1_loss = _get_policy_loss(normal_input_node_1_ratio)
            normal_input_node_2_loss = _get_policy_loss(normal_input_node_2_ratio)
            reduce_input_node_1_loss = _get_policy_loss(reduce_input_node_1_ratio)
            reduce_input_node_2_loss = _get_policy_loss(reduce_input_node_2_ratio)
            policy_loss = normal_op_1_loss + normal_op_2_loss + \
                          reduce_op_1_loss + reduce_op_2_loss + \
                          normal_input_node_1_loss + normal_input_node_2_loss + \
                          reduce_input_node_1_loss + reduce_input_node_2_loss
            target_values = old_state_values + advantages
            value_loss = c1 * value_criterion(new_state_values, target_values).mean()
            entropy_loss = -1. * c2 * b_new_actions.get_op_entropy() + -1. * c3 * b_new_actions.get_edge_entropy()
            batch_loss = policy_loss + value_loss + entropy_loss
            optimizer.zero_grad()
            batch_loss.backward()
            optimizer.step()
            avg_loss += batch_loss.item()
            num_updates += 1
        data_loader.shuffle()
    assert num_updates > 0, "data_loader len: {}".format(len(data_loader))
    return avg_loss / num_updates


def _generate_darts_blocks(actor, env, n_desired_blocks,
                           max_n_attempts=1000, verbose=True, desc="Generating blocks",
                           valid_block_keys=None):
    blocks = UniqueList(key_func=lambda b: b.get_str_id())
    n_attempts = 0
    bar = None
    if verbose: bar = tqdm(total=n_desired_blocks, desc=desc, ascii=True)
    while n_attempts < max_n_attempts and len(blocks) < n_desired_blocks:
        env.reset()
        with torch.no_grad():
            final_state_reached = False
            while not final_state_reached:
                action = actor.get_action_for_state(env.get_curr_state(), env.get_num_steps())
                final_state_reached, input_state, reward, new_state = env.step(action)
            valid = is_darts_action_valid(action) and is_darts_cell_valid(new_state)
            if valid_block_keys is not None:
                key = new_state.get_str_id()
                valid = valid and key in valid_block_keys
            if valid and blocks.append(new_state):
                n_attempts = 0
                if bar is not None: bar.update(1)
        n_attempts += 1
    env.reset()
    if bar is not None: bar.close()
    blocks = blocks.tolist()
    return blocks


def eval_darts_cons_acc(iteration, actor, env, evaluator:BlockPerfEvaluator, acc_set,
                        n_desired_blocks, max_cons_score, cons_predictor, book_keeper,
                        update_callback=None, valid_block_keys=None):
    new_block2acc = {}
    new_blocks = _generate_darts_blocks(actor, env, n_desired_blocks,
                                        desc="Eval", valid_block_keys=valid_block_keys)

    report_darts_blocks_info(new_blocks, log_f=book_keeper.log, desc="Pre-evaluation")

    n_not_match = 0.
    for block in tqdm(new_blocks, desc="Getting performance", ascii=True):
        key = block.get_str_id()
        if cons_predictor is not None:
            cons_score = cons_predictor(block)
            if cons_score > max_cons_score:
                n_not_match += 1.
                continue
        acc, _ = evaluator.get_perf_values(block)
        new_block2acc[key] = (block, acc)

    book_keeper.log("{}% generated blocks not matching constraints".format(100 * n_not_match / len(new_blocks)))

    if len(new_block2acc) < n_desired_blocks:
        book_keeper.log("Only generated {} unique blocks".format(len(new_block2acc)))

    if len(new_block2acc) == 0:
        book_keeper.log("No blocks with acc generated")
    else:
        node_lens = [len(b) for _, (b, v) in new_block2acc.items()]
        book_keeper.log("Avg number of nodes in generated blocks: {}".format(mean(node_lens)))
        book_keeper.log("Max number of nodes in generated blocks: {}".format(max(node_lens)))
        book_keeper.log("Min number of nodes in generated blocks: {}".format(min(node_lens)))
        gen_acc_values = [acc for _, (b, acc) in new_block2acc.items()]
        book_keeper.log("Max acc in generated blocks: {}".format(max(gen_acc_values)))
        book_keeper.log("Min acc in generated blocks: {}".format(min(gen_acc_values)))
        acc_set.update(new_block2acc, log_f=book_keeper.log)
        if update_callback is not None:
            update_callback(acc_set, iteration, book_keeper.model_name)
        book_keeper.checkpoint_state_dict(acc_set.state_dict(),
                                          "{}_top_acc_set.pkl".format(book_keeper.model_name))
        acc_vals = acc_set.get_top_acc_values()
        book_keeper.log("Best acc so far: {}".format(max(acc_vals)))
        book_keeper.log("New top acc set size: {}, new_block2acc len: {}".format(len(acc_set),
                                                                                 len(new_block2acc)))
    return [b for _, (b, _) in new_block2acc.items()]


class DartsSupernetEvaluator(BlockPerfEvaluator):

    def __init__(self, model, train_loader, test_loader, op2idx, lat_predictor,
                 criterion=torch.nn.CrossEntropyLoss(), enable_perf_memo=True):
        super(DartsSupernetEvaluator, self).__init__(enable_perf_memo=enable_perf_memo)
        self.model = model.to(device())
        self.train_loader = train_loader
        self.test_loader = test_loader
        self.criterion = criterion.to(device())
        self.lat_predictor = lat_predictor
        self.op2idx = op2idx
        self.model.setOneshot(False)

    def get_perf_reward(self, state):
        raise NotImplementedError

    def get_perf_values(self, block:DartsBlockState):
        key = block.get_str_id()
        if self.enable_perf_memo and key in self.perf_memo:
            pred_acc, pred_lat = self.perf_memo[key]
        else:
            # Create discrete alpha from state
            normal_d_alphas = darts_cell_to_discrete_alphas(block.normal_op_labels, block.normal_op_edges,
                                                            block.de2ni_map, self.op2idx, fill_alpha_val=0).to(device())
            reduce_d_alphas = darts_cell_to_discrete_alphas(block.reduce_op_labels, block.reduce_op_edges,
                                                            block.de2ni_map, self.op2idx, fill_alpha_val=0).to(device())
            # # Discretize again according to the supernet's rule
            # normal_d_alphas, reduce_d_alphas = discretize_alpha(normal_d_alphas), discretize_alpha(reduce_d_alphas)
            self.model.setAlphas(normal_d_alphas, reduce_d_alphas)
            with torch.no_grad():
                pred_acc, _, _ = infer(self.test_loader, self.model, self.criterion, log_f=None)
            pred_acc /= 100.
            pred_lat = self.lat_predictor(block) if self.lat_predictor is not None else 1.0
            self.perf_memo[key] = (pred_acc, pred_lat)
            interval_info.num_block_eval += 1
        return pred_acc, pred_lat


def ga_nas_darts_cons_acc_search(env_manager, ppo_optimizer, value_criterion, top_acc_set,
                                  evaluator, n_train_iterations, n_train_rounds,
                                  k_epochs, t_timesteps, batch_size, n_desired_blocks_eval,
                                  max_cons_score, cons_predictor, book_keeper,
                                  parallel_explore=False,
                                  m_repeats=2, reward_discount_factor=1.0, debug=True,
                                  gae_lambda=0.97, clip_epsilon=0.2, c1=0.5, c2=0.01, c3=0.01,
                                  n_desired_blocks_inc=0,
                                  actor_update_episode_interval=500, early_stop=False,
                                  acc_update_callback=None, valid_block_keys=None):
    env_manager = env_manager.to(device())
    value_criterion = value_criterion.to(device())
    reward_module = env_manager.reward_module
    g_start = time.time()
    # PPO training iterations
    for i in range(n_train_iterations):
        report_iteration = i + 1
        book_keeper.log("Running iteration {} of {}".format(report_iteration, n_train_iterations))
        env, actor = env_manager.get_env_actor_rep()
        src_block2perfs = top_acc_set.top_block2acc()
        reward_module.build_truth_blocks(src_block2perfs, log_f=book_keeper.log)

        blocks_from_train = UniqueList(key_func=lambda b: b.get_str_id())
        for mi in range(m_repeats):
            book_keeper.log("Inner training loop {} of {}".format(mi + 1, m_repeats))
            # Train discriminator first
            n_desired_blocks_dis = reward_module.get_fake_data_size()
            cand_blocks = _generate_darts_blocks(actor, env, n_desired_blocks_dis,
                                                 desc="Gen blocks for discriminator training")
            book_keeper.log("{} unique blocks for training discriminator".format(len(cand_blocks)))
            reward_module.d_train(cand_blocks, log_f=book_keeper.log)

            actor, block_w_rewards = darts_ppo_train(env_manager, ppo_optimizer, value_criterion, n_train_rounds,
                                                     k_epochs, t_timesteps, batch_size, reward_discount_factor,
                                                     clip_epsilon, gae_lambda, c1, c2, c3, book_keeper,
                                                     parallel_explore=parallel_explore,
                                                     actor_update_episode_interval=actor_update_episode_interval)
            book_keeper.log("{} unique valid blocks generated during training".format(len(block_w_rewards)))
            blocks_from_train.extend([t[0] for t in block_w_rewards])

        book_keeper.log("Eval at iteration {}".format(report_iteration))
        gen_blocks = eval_darts_cons_acc(report_iteration, actor, env, evaluator, top_acc_set,
                                         n_desired_blocks_eval + i * n_desired_blocks_inc,
                                         max_cons_score, cons_predictor, book_keeper,
                                         valid_block_keys=valid_block_keys,
                                         update_callback=acc_update_callback)

        if debug:
            top_blocks = top_acc_set.get_curr_blocks()
            truth_blocks = top_acc_set.get_curr_blocks()
            book_keeper.log("")
            book_keeper.log("====== Start of debug info ======")
            truth_keys = set(b.get_str_id() for b in truth_blocks)
            train_block_keys = set(b.get_str_id() for b in blocks_from_train.tolist())
            common_keys = train_block_keys.intersection(truth_keys)
            book_keeper.log("{} unique blocks from train".format(len(train_block_keys)))
            book_keeper.log("{}% blocks from train are from truth".format(100. * float(len(common_keys)) / len(train_block_keys)))
            truth_keys = set(b.get_str_id() for b in truth_blocks)
            gen_keys = set(b.get_str_id() for b in gen_blocks)
            common_keys = gen_keys.intersection(truth_keys)
            book_keeper.log("{} unique gen blocks".format(len(gen_keys)))
            book_keeper.log("{}% gen blocks are from truth".format(100. * float(len(common_keys)) / len(gen_keys)))
            book_keeper.log("------")
            pair_dis_pred_check(truth_blocks, top_blocks, reward_module.get_model(),
                                log_f=book_keeper.log, desc="truth")
            pair_dis_pred_check(gen_blocks, top_blocks, reward_module.get_model(),
                                log_f=book_keeper.log, desc="generated")
            book_keeper.log("------")
            report_darts_blocks_info(truth_blocks, log_f=book_keeper.log, desc="truth")
            report_darts_blocks_info(gen_blocks, log_f=book_keeper.log, desc="generated")
            book_keeper.log("====== End of debug info ======")
            book_keeper.log("")
        book_keeper.log("Num blocks evaluated so far: {}".format(interval_info.num_block_eval))

        # Hack to terminate search early
        curr_tops = top_acc_set.get_top_blocks_with_acc()
        top_acc = sorted(curr_tops, key=lambda t: t[1], reverse=True)[0][1]
        if early_stop and report_iteration > 5 and top_acc > 0.8671:
            book_keeper.log("Curr known best valid acc: {}".format(top_acc))
            book_keeper.log("Search terminated early at epoch: {}".format(report_iteration))
            break
    book_keeper.log("Search seconds: {}".format(time.time() - g_start))


def get_random_darts_blocks(init_state:DartsBlockState, n_internal_cell_nodes,
                            n_input_edges_per_node, n_ops, scaffold_labels, op2idx,
                            n_input_cell_nodes=2, output_node_label=DARTS_OUTPUT_NODE_LABEL, n_blocks=100,
                            watchdog_count=10000, max_cons_score=None, cons_predictor=None):
    n_attempts = 0
    bar = tqdm(total=n_blocks, desc="Random sampling DARTS cells", ascii=True)
    blocks = UniqueList(key_func=lambda b: b.get_str_id())
    with torch.no_grad():
        while len(blocks) < n_blocks and n_attempts < watchdog_count:
            curr_block = copy.deepcopy(init_state)
            for ni in range(n_internal_cell_nodes):
                normal_src_node_inds = list(range(n_input_cell_nodes + ni))
                reduce_src_node_inds = list(range(n_input_cell_nodes + ni))
                random.shuffle(normal_src_node_inds)
                random.shuffle(reduce_src_node_inds)
                for src_i in range(n_input_edges_per_node):
                    normal_alpha = torch.ones(n_ops).float()
                    reduce_alpha = torch.ones(n_ops).float()
                    normal_alpha /= torch.sum(normal_alpha)
                    reduce_alpha /= torch.sum(reduce_alpha)
                    normal_node_idx = get_tsr_value(Categorical(probs=normal_alpha).sample())
                    reduce_node_idx = get_tsr_value(Categorical(probs=reduce_alpha).sample())
                    normal_label = scaffold_labels[normal_node_idx]
                    reduce_label = scaffold_labels[reduce_node_idx]
                    normal_edge = [(normal_src_node_inds[src_i], n_input_cell_nodes + ni)]
                    reduce_edge = [(reduce_src_node_inds[src_i], n_input_cell_nodes + ni)]
                    curr_block = curr_block.gen_block_state_with_new_node(normal_label,
                                                                          op2idx[normal_label],
                                                                          normal_edge,
                                                                          reduce_label,
                                                                          op2idx[reduce_label],
                                                                          reduce_edge)
            input_cell_node_edges = get_output_cell_node_input_edges()
            curr_block = curr_block.gen_block_state_with_new_node(output_node_label,
                                                                  op2idx[output_node_label],
                                                                  input_cell_node_edges,
                                                                  output_node_label,
                                                                  op2idx[output_node_label],
                                                                  input_cell_node_edges)
            n_attempts += 1
            if cons_predictor is not None and cons_predictor(curr_block) > max_cons_score:
                continue
            assert is_darts_cell_valid(curr_block)
            if blocks.append(curr_block):
                n_attempts = 0
                bar.update(1)
    bar.close()
    return blocks.tolist()


def get_darts_init_block_state(input_node_label, op2idx, de2ni_map, complete_nb101_edge_list):
    state = DartsBlockState([input_node_label, input_node_label],
                            [op2idx[input_node_label], op2idx[input_node_label]],
                            [],
                            [input_node_label, input_node_label],
                            [op2idx[input_node_label], op2idx[input_node_label]],
                            [],
                            de2ni_map, complete_nb101_edge_list)
    return state


def init_darts_acc_set(init_blocks, evaluator, top_k, log_f=print):
    init_block2acc = {}
    for block in tqdm(init_blocks, desc="Getting acc for init blocks", ascii=True):
        acc, _ = evaluator.get_perf_values(block)
        # acc = np.random.normal(0.5, 0.1, 1).item()
        # acc = min(1.0, acc)
        # acc = max(0.0, acc)
        init_block2acc[block.get_str_id()] = (block, acc)
    log_f("Initial block2acc size: {}".format(len(init_block2acc)))
    top_acc_set = BlockTopKAccSet(init_block2acc, top_k, log_f=log_f)
    log_f("Initial top acc set size: {}".format(len(top_acc_set.top_block2acc())))
    return top_acc_set


def pair_dis_pred_check(gen_blocks, top_blocks, d_reward_module, log_f=print, desc="fake"):
    avg_fake_prob = AverageMeter()
    avg_truth_prob = AverageMeter()
    avg_pred_truth_prob = AverageMeter()
    pred_truth_blocks = []
    for block in tqdm(gen_blocks, desc="Discriminator pred check", ascii=True, leave=False):
        assert isinstance(block, DartsBlockState)
        max_true_prob = 0.
        for p_block in top_blocks:
            logits = d_reward_module(build_d_pair_input(p_block, block))
            prob = torch.softmax(logits, dim=-1).squeeze()
            prob = get_tsr_value(prob)
            max_true_prob = max(max_true_prob, prob[1])
        avg_fake_prob.update(1. - max_true_prob)
        avg_truth_prob.update(max_true_prob)
        if max_true_prob >= 0.5:
            pred_truth_blocks.append(block)
            avg_pred_truth_prob.update(max_true_prob)
    log_f("Discriminator pred check on {} {} blocks".format(len(gen_blocks), desc))
    log_f("Avg fake prob: {}".format(avg_fake_prob.avg))
    log_f("Avg truth prob: {}".format(avg_truth_prob.avg))
    log_f("Avg pred truth prob: {}".format(avg_pred_truth_prob.avg))
    log_f("Number of pred truth blocks: {}".format(len(pred_truth_blocks)))


def report_darts_blocks_info(blocks, log_f=print, desc=""):
    log_f("Reporting info on {} {} blocks".format(len(blocks), desc))
    normal_op_ratio = collections.defaultdict(float)
    reduce_op_ratio = collections.defaultdict(float)
    n_normal_ops = 0
    n_reduce_ops = 0
    normal_edges_meter = RunningStatMeter()
    reduce_edges_meter = RunningStatMeter()
    normal_d_alpha_meter = RunningStatMeter()
    reduce_d_alpha_meter = RunningStatMeter()
    for block in blocks:
        assert isinstance(block, DartsBlockState)
        for op in block.normal_op_labels[2:-1]:
            normal_op_ratio[op] += 1
            n_normal_ops += 1
        for op in block.reduce_op_labels[2:-1]:
            reduce_op_ratio[op] += 1
            n_reduce_ops += 1
        normal_edges_meter.update(len(block.normal_op_edges))
        reduce_edges_meter.update(len(block.reduce_op_edges))
        _normal_d_alphas = darts_cell_to_discrete_alphas(block.normal_op_labels, block.normal_op_edges,
                                                         block.de2ni_map, DARTS_OP2IDX, fill_alpha_val=0.)
        _reduce_d_alphas = darts_cell_to_discrete_alphas(block.reduce_op_labels, block.reduce_op_edges,
                                                         block.de2ni_map, DARTS_OP2IDX, fill_alpha_val=0.)
        normal_d_alpha_meter.update(get_tsr_value(_normal_d_alphas.sum()))
        reduce_d_alpha_meter.update(get_tsr_value(_reduce_d_alphas.sum()))
    normal_op_ratio = [(k, v / n_normal_ops) for k, v in normal_op_ratio.items()]
    reduce_op_ratio = [(k, v / n_reduce_ops) for k, v in reduce_op_ratio.items()]
    normal_op_ratio.sort(key=lambda t: t[0])
    reduce_op_ratio.sort(key=lambda t: t[0])
    log_f("Number of searchable normal ops: {}".format(n_normal_ops))
    for k, v in normal_op_ratio:
        log_f("- {}: {:.2f}%".format(k, v * 100.))
    log_f("Number of searchable reduce ops: {}".format(n_reduce_ops))
    for k, v in reduce_op_ratio:
        log_f("- {}: {:.2f}%".format(k, v * 100.))
    log_f("Max normal cell edges: {}".format(normal_edges_meter.max))
    log_f("Min normal cell edges: {}".format(normal_edges_meter.min))
    log_f("Avg normal cell edges: {}".format(normal_edges_meter.avg))
    log_f("Max reduce cell edges: {}".format(reduce_edges_meter.max))
    log_f("Min reduce cell edges: {}".format(reduce_edges_meter.min))
    log_f("Avg reduce cell edges: {}".format(reduce_edges_meter.avg))
    log_f("Max normal num d alphas: {}".format(normal_d_alpha_meter.max))
    log_f("Min normal num d alphas: {}".format(normal_d_alpha_meter.min))
    log_f("Avg normal num d alphas: {}".format(normal_d_alpha_meter.avg))
    log_f("Max reduce num d alphas:: {}".format(reduce_d_alpha_meter.max))
    log_f("Min reduce num d alphas:: {}".format(reduce_d_alpha_meter.min))
    log_f("Avg reduce num d alphas:: {}".format(reduce_d_alpha_meter.avg))


class DartsPTEvaluator(BlockPerfEvaluator):

    def __init__(self, n_epochs, ni2de_map, train_loader, dev_loader, test_loader, lat_predictor, n_classes,
                 enable_perf_memo=True):
        super(DartsPTEvaluator, self).__init__(enable_perf_memo=enable_perf_memo)
        self.n_epochs = n_epochs
        self.ni2de_map = ni2de_map
        self.n_classes = n_classes
        self.train_loader = train_loader
        self.dev_loader = dev_loader
        self.test_loader = test_loader
        self.lat_predictor = lat_predictor

    def get_perf_reward(self, state):
        raise NotImplementedError

    def _silent_log_f(self, *args):
        pass

    def get_perf_values(self, block:DartsBlockState):
        key = block.get_str_id()
        if self.enable_perf_memo and key in self.perf_memo:
            pred_acc, pred_lat = self.perf_memo[key]
        else:
            concat = range(2, 6)  # Concat the last 4 internal states
            pred_acc = full_train_darts_v2(block.normal_op_labels[2:-1], # Chop two input + 1 output nodes
                                           block.reduce_op_labels[2:-1], self.ni2de_map,
                                           concat, epochs=self.n_epochs, n_internal_nodes=4, n_classes=self.n_classes,
                                           train_loader=self.train_loader, dev_loader=self.dev_loader,
                                           test_loader=self.test_loader, log_f=self._silent_log_f)
            pred_acc /= 100.
            pred_lat = self.lat_predictor(block)
            self.perf_memo[key] = (pred_acc, pred_lat)
            interval_info.num_block_eval += 1
        return pred_acc, pred_lat


class NB301Evaluator(BlockPerfEvaluator):

    def __init__(self, api, op2idx, pred_runtime=True, default_runtime=1.0, default_clue=1.0):
        super(NB301Evaluator, self).__init__(default_clue=default_clue)
        self.api = api
        self.op2idx = op2idx
        self.pred_runtime = pred_runtime
        self.default_runtime = default_runtime

    def get_perf_reward(self, state):
        raise NotImplementedError

    def get_perf_values(self, block:DartsBlockState):
        key = block.get_str_id()
        if self.enable_perf_memo and key in self.perf_memo:
            acc, lat = self.perf_memo[key]
        else:
            normal_d_alphas = darts_cell_to_discrete_alphas(block.normal_op_labels, block.normal_op_edges,
                                                            block.de2ni_map, self.op2idx)
            reduce_d_alphas = darts_cell_to_discrete_alphas(block.reduce_op_labels, block.reduce_op_edges,
                                                            block.de2ni_map, self.op2idx)
            genotype = alphas_to_genotype(n_steps=4, alphas_normal=normal_d_alphas, alphas_reduce=reduce_d_alphas)
            acc = self.api.get_geno_acc(genotype) / 100.
            if self.pred_runtime:
                lat = self.api.get_geno_runtime(genotype)
            else:
                lat = self.default_runtime
            self.perf_memo[key] = (acc, lat)
            interval_info.num_block_eval += 1
        return acc, lat

