from constants import *
from tqdm import tqdm
from model_nas_base import compute_t_step_gae, NodeAggr, run_seq_exploration, run_parallel_exploration, \
    BlockPerfEvaluator
from model_ga_nas import BlockTopKAccSet
from torch.distributions import Categorical
from utils.model_utils import device
from torch_geometric.data import Data, Batch
from utils.nas_utils import *
from utils.debug_utils import IntervalInfo as interval_info
from model_helpers import get_nb201_key
from sklearn.metrics import f1_score
from utils.darts_utils import AverageMeter


class GraphEncoder(torch.nn.Module):

    def __init__(self, in_channels, hidden_size, out_channels, gnn_constructor, node_pre_embedding,
                 activ=torch.nn.ReLU(), num_layers=2, dropout_prob=0.0):
        super(GraphEncoder, self).__init__()
        self.node_pre_embedding = node_pre_embedding
        self.in_channels = in_channels
        self.hidden_size = hidden_size
        self.out_channels = out_channels
        self.gnn_layers = torch.nn.ModuleList()
        for i in range(num_layers):
            input_size, output_size = hidden_size, hidden_size
            if i == 0:
                input_size = in_channels
            if i == num_layers - 1:
                output_size = out_channels
            gnn_layer = gnn_constructor(input_size, output_size)
            self.gnn_layers.append(gnn_layer)
        self.activ = activ
        self.dropout = torch.nn.Dropout(dropout_prob)
        self.init_ff = torch.nn.Linear(2 * in_channels, in_channels)

    def init_node_embedding_prior(self, node_embedding):
        prior = np.random.normal(size=node_embedding.shape)
        prior = torch.from_numpy(prior).float().to(device())
        node_embedding = torch.cat([node_embedding, prior], dim=-1)
        return self.init_ff(node_embedding)

    def forward(self, batch_node_tsr, edge_tsr_list, batch_last_node_idx_list):
        node_embedding = self.node_pre_embedding(batch_node_tsr.to(device()))
        if node_embedding.shape[1] == 1:
            node_embedding = self.init_node_embedding_prior(node_embedding)
        data_list = [
            Data(x=node_embedding[i,:], edge_index=edge_tsr_list[i].to(device()), edge_attr=None)
                for i in range(node_embedding.shape[0])
        ]
        torch_geo_batch = Batch.from_data_list(data_list)
        edge_index_tsr = torch_geo_batch.edge_index
        curr_gnn_output = torch_geo_batch.x
        for li, gnn_layer in enumerate(self.gnn_layers):
            node_embedding_tsr = gnn_layer(curr_gnn_output, edge_index_tsr)
            curr_gnn_output = self.activ(node_embedding_tsr)
        curr_gnn_output = self.dropout(curr_gnn_output)
        batch_embedding_list = torch_geo_batch_to_data_list(curr_gnn_output, batch_last_node_idx_list)
        batch_embedding = torch.cat([t.unsqueeze(0) for t in batch_embedding_list], dim=0)
        return batch_embedding


def build_nb201_pair_input(b1, b2):
    tsr_list_0, edge_list_0, idx_list_0 = b1.get_gpn_compatible_input()
    tsr_list_1, edge_list_1, idx_list_1 = b2.get_gpn_compatible_input()
    block_0 = (tsr_list_0, edge_list_0, idx_list_0)
    block_1 = (tsr_list_1, edge_list_1, idx_list_1)
    return block_0, block_1


class NB201BlockGPN(torch.nn.Module):

    def __init__(self, graph_encoder, action_predictor):
        super(NB201BlockGPN, self).__init__()
        self.graph_encoder = graph_encoder
        self.action_predictor = action_predictor

    def forward(self, b_tsr, edge_tsr_list, b_last_node_idx_list):
        batch_embedding = self.graph_encoder(b_tsr, edge_tsr_list, b_last_node_idx_list)
        batch_action_probs = self.action_predictor(batch_embedding)
        return batch_embedding, batch_action_probs


class NB201ActionPredictor(torch.nn.Module):

    def __init__(self, node_embed_size, hidden_size, node_selector):
        super(NB201ActionPredictor, self).__init__()
        self.graph_ff = torch.nn.Linear(node_embed_size, hidden_size)
        self.node_selector = node_selector

    def forward(self, node_embedding):
        graph_embedding = self.graph_ff(node_embedding.sum(dim=1).to(device()))
        ctx = self.node_selector(graph_embedding)
        node_probs = torch.softmax(ctx, dim=-1)
        return node_probs


class NB201BlockDiscriminator(torch.nn.Module):

    def __init__(self, cell_encoder, predictor):
        super(NB201BlockDiscriminator, self).__init__()
        self.cell_encoder = cell_encoder
        self.predictor = predictor

    def forward(self, data_list):
        ctx = []
        for data in data_list:
            tsr, edge_list, last_node_idx_list = data
            node_embedding = self.cell_encoder(tsr, edge_list, last_node_idx_list)
            graph_embedding = node_embedding[:, -1, :]
            ctx.append(graph_embedding)
        ctx = torch.cat([ctx[0], ctx[1]], dim=-1)
        batch_logits = self.predictor(ctx)
        return batch_logits


class NB201BlockState:

    def __init__(self, node_labels, node_tsr, curr_edge_list, complete_edge_list):
        self.node_labels = node_labels
        self.node_tsr = node_tsr
        self.curr_edge_list = curr_edge_list
        self.complete_edge_list = complete_edge_list
        self.alphas = [] # Contains numpy arrays
        self.str_id = None

    @property
    def curr_node_label_list(self):
        return self.node_labels

    def de_tensorize(self):
        b = copy.deepcopy(self)
        if isinstance(b.node_tsr, torch.Tensor):
            b.node_tsr = b.node_tsr.tolist()
        return b

    def re_tensorize(self):
        if isinstance(self.node_tsr, list):
            self.node_tsr = torch.LongTensor(self.node_tsr)
        return self

    def __len__(self):
        return len(self.node_labels)

    def __str__(self):
        return "NB201BlockState[{}] Nodes: {} | Edges: {}".format(len(self.node_labels),
                ",".join(self.node_labels), str(edge_list_to_edge_pairs(self.curr_edge_list)))

    def __repr__(self):
        return str(self)

    def __deepcopy__(self, memodict={}):
        _state = NB201BlockState(
            node_labels=copy.deepcopy(self.node_labels),
            node_tsr=copy.deepcopy(self.node_tsr),
            curr_edge_list=copy.deepcopy(self.curr_edge_list),
            complete_edge_list=self.complete_edge_list
        )
        _state.alphas = copy.deepcopy(self.alphas)
        return _state

    def to(self, dev):
        self.node_tsr = self.node_tsr.to(dev)
        return self

    def get_gpn_compatible_input(self):
        # Returns node_feature_tsr_list, edge_tsr_list, partial_graph_idx_tsr_list, last_node_idx_list
        # But only for a single instance, i.e., batch dim is 1
        edge_tsr = torch.LongTensor(self.curr_edge_list).to(device())
        return self.node_tsr.unsqueeze(0), [edge_tsr], [self.node_tsr.shape[0] - 1]

    def gen_block_state_with_new_node(self, new_label, new_node_tsr, new_alpha_tsr):
        new_labels = [v for v in self.node_labels] + [new_label]
        new_tsr = torch.cat([self.node_tsr, new_node_tsr], dim=0)
        n_nodes = len(new_labels)
        complete_edge_pairs = edge_list_to_edge_pairs(self.complete_edge_list)
        new_edge_pairs = [p for p in complete_edge_pairs if p[1] < n_nodes]
        assert len(new_edge_pairs) > 0
        # Manually add self edges for input node
        new_edge_list = [[0] + [p[0] for p in new_edge_pairs], [0] + [p[1] for p in new_edge_pairs]]
        new_state = NB201BlockState(node_labels=new_labels,
                                    node_tsr=new_tsr,
                                    curr_edge_list=new_edge_list,
                                    complete_edge_list=self.complete_edge_list)
        new_alphas = copy.deepcopy(self.alphas)
        if new_alpha_tsr is not None:
            new_alphas.append(new_alpha_tsr.detach().cpu().unsqueeze(0).numpy())
        new_state.alphas = new_alphas
        return new_state

    def get_alpha_tensor(self):
        alphas_np = np.vstack(self.alphas)
        return torch.from_numpy(alphas_np).float().to(device())

    def get_str_id(self):
        if self.str_id is not None:
            return self.str_id
        from utils.nas_utils import edge_list_to_edge_matrix
        src_list, dst_list = prune_self_edge_from_edge_list(self.curr_edge_list)
        if isinstance(self.node_tsr, torch.Tensor):
            node_indices = self.node_tsr.tolist()
        else:
            node_indices = self.node_tsr
        matrix = edge_list_to_edge_matrix([src_list, dst_list], len(node_indices))
        fingerprint = hash_module(matrix, node_indices)
        self.str_id = fingerprint
        return self.str_id


class BatchNB201BlockStates:
    """Only used during k epoch PPO updates"""
    def __init__(self, states):
        self.states = states
        self.node_tsr = None
        self.edge_tsr_list = []
        self.last_idx_list = []
        self._build_gpn_compatible_input()

    def __len__(self):
        return len(self.states)

    def __str__(self):
        return "BatchNB201BlockStates of size {}".format(len(self))

    def __repr__(self):
        return str(self)

    def _build_gpn_compatible_input(self, remove_self_edges=False):
        tsr_list = []
        self.edge_tsr_list = []
        self.last_idx_list = []
        aggr_node_size = 0
        for state in self.states:
            node_tsr = state.node_tsr.unsqueeze(0)
            if remove_self_edges:
                edge_list = prune_self_edge_from_edge_list(state.curr_edge_list)
            else:
                edge_list = state.curr_edge_list
            edge_tsr = torch.LongTensor(edge_list)
            tsr_list.append(node_tsr)
            self.edge_tsr_list.append(edge_tsr)
            self.last_idx_list.append(aggr_node_size + node_tsr.shape[1] - 1)
            aggr_node_size += node_tsr.shape[1]
        self.node_tsr = torch.cat(tsr_list)

    def get_gpn_compatible_input(self):
        return self.node_tsr, self.edge_tsr_list, self.last_idx_list


class NB201Action:

    def __init__(self, actor_name, timestep, node_probs):
        self.actor_name = actor_name
        self.timestep = timestep
        self.node_probs = node_probs
        self.node_idx = Categorical(probs=self.node_probs).sample()
        self.entropy = Categorical(probs=self.node_probs).entropy()

    def __deepcopy__(self, memodict={}):
        a = NB201Action(self.actor_name, self.timestep, copy.deepcopy(self.node_probs))
        a.node_idx = copy.deepcopy(self.node_idx)
        a.entropy = copy.deepcopy(self.entropy)
        return a

    def de_tensorize(self):
        a = copy.deepcopy(self)
        if isinstance(a.node_probs, torch.Tensor):
            a.node_probs = a.node_probs.tolist()
        if isinstance(a.node_idx, torch.Tensor):
            a.node_idx = get_tsr_value(a.node_idx)
        if isinstance(a.entropy, torch.Tensor):
            a.entropy = get_tsr_value(a.entropy)
        return a

    def re_tensorize(self):
        if isinstance(self.node_probs, list):
            self.node_probs = torch.FloatTensor(self.node_probs)
        if isinstance(self.node_idx, int):
            self.node_idx = torch.LongTensor([self.node_idx]).squeeze()
        if isinstance(self.entropy, float):
            self.entropy = torch.FloatTensor([self.entropy]).squeeze()
        return self

    def to(self, dev):
        self.node_probs = self.node_probs.to(dev)
        self.node_idx = self.node_idx.to(dev)
        self.entropy = self.entropy.to(dev)
        return self

    def re_sample_action(self):
        self.node_idx = Categorical(probs=self.node_probs).sample()
        return self.get_discrete_action()

    def get_discrete_action(self, get_raw_value=True):
        node_idx = get_tsr_value(self.node_idx) if get_raw_value else self.node_idx
        return node_idx

    def get_entropy(self):
        return self.entropy

    def __str__(self):
        s = "NB201Action at time {} by actor {}".format(self.timestep, self.actor_name)
        s += ", Node idx: {}".format(str(get_tsr_value(self.node_idx)))
        s += ", Node probs: {}".format(str(self.node_probs.tolist()))
        return s

    def __repr__(self):
        return str(self)


class BatchNB201Actions:

    def __init__(self, actions):
        self.actions = actions
        batch_node_probs = []
        batch_node_indices = []
        batch_entropy = []
        for action in self.actions:
            assert isinstance(action, NB201Action)
            batch_node_probs.append(action.node_probs.unsqueeze(0))
            batch_node_indices.append(action.node_idx.unsqueeze(0))
            batch_entropy.append(action.get_entropy().unsqueeze(0))
        self.batch_node_probs = torch.cat(batch_node_probs, dim=0)
        self.batch_node_indices = torch.cat(batch_node_indices, dim=0)
        self.entropy = torch.cat(batch_entropy, dim=0).mean()
        self.batch_node_log_probs = Categorical(probs=self.batch_node_probs).log_prob(self.batch_node_indices)

    def __len__(self):
        return len(self.actions)

    def __str__(self):
        return "BatchNB201Actions of size {}".format(len(self))

    def __repr__(self):
        return str(self)

    def to(self, dev):
        for action in self.actions:
            action.to(dev)
        return self

    def get_entropy(self):
        return self.entropy

    def get_action_log_probs(self, batch_node_indices=None):
        if batch_node_indices is None:
            batch_node_log_probs = self.batch_node_log_probs
        else:
            batch_node_log_probs = Categorical(probs=self.batch_node_probs).log_prob(batch_node_indices)
        return batch_node_log_probs


class NB201Env:

    def __init__(self, actor, init_state, node_label2feat, reward_module, max_allowed_steps, output_node_label,
                 scaffold_node_labels, step_reward=0.):
        self._actor = actor
        self._output_node_label = output_node_label
        self._node_label2feat = node_label2feat
        self._reward_module = reward_module
        self._curr_state = init_state
        self._init_state = init_state
        self._max_allowed_steps = max_allowed_steps
        self._num_completed_steps = 0
        self._step_reward = step_reward
        self.scaffold_node_labels = scaffold_node_labels

    @property
    def final_reward_module(self):
        return self._reward_module

    def load_state_dict(self, actor_state_dict):
        if actor_state_dict is not None:
            self._actor.load_state_dict(actor_state_dict)

    def get_init_state(self):
        return self._init_state

    def set_actor(self, actor):
        self._actor = actor

    def get_actor(self):
        return self._actor

    def get_curr_state(self):
        return self._curr_state

    def set_curr_state(self, state):
        self._curr_state = state

    def get_num_steps(self):
        return self._num_completed_steps

    def eval(self):
        self._reward_module.eval()
        self._actor.eval()

    def train(self):
        self._reward_module.train()
        self._actor.train()

    def to(self, dev):
        self._reward_module = self._reward_module.to(dev)
        self._actor = self._actor.to(dev)
        return self

    def reset(self):
        self._curr_state = self._init_state
        self._num_completed_steps = 0

    def get_reward_for_state(self, state):
        return self._reward_module.get_reward(state)

    def step(self, action:NB201Action):
        input_state = self._curr_state
        new_state = self.gen_new_state_from_action(self._curr_state, action)
        self._num_completed_steps += 1
        final_state_reached = self._num_completed_steps >= self._max_allowed_steps
        if final_state_reached:
            # Manually add the output node
            new_state = new_state.gen_block_state_with_new_node(self._output_node_label,
                                                                self._node_label2feat[self._output_node_label],
                                                                None)
            final_reward = self.get_reward_for_state(new_state)
            reward = final_reward
        else:
            reward = self._step_reward
        self._curr_state = new_state
        return final_state_reached, input_state, reward, new_state

    def gen_new_state_from_action(self, input_state, action:NB201Action):
        new_node_idx = action.get_discrete_action()
        new_node_label = self.scaffold_node_labels[new_node_idx]
        new_state = input_state.gen_block_state_with_new_node(new_node_label,
                                                              self._node_label2feat[new_node_label],
                                                              action.node_probs)
        return new_state

    def __str__(self):
        return "NB201Env for actor {}, num_steps: {}".format(self._actor.get_name(), self._num_completed_steps)

    def __repr__(self):
        return str(self)


class NB201EnvManager:

    def __init__(self, parallel_actors, n_parallel_actors, init_state, node_label2feat, scaffold_node_labels,
                 reward_module, max_allowed_steps, output_node_label):
        self.init_state = init_state
        self.max_allowed_steps = max_allowed_steps
        self.node_label2feat = node_label2feat
        self.reward_module = reward_module
        self.output_node_label = output_node_label
        self.n_parallel_actors = n_parallel_actors
        self.parallel_actors = parallel_actors
        self.curr_env_idx = 0
        for i, actor in enumerate(self.parallel_actors):
            actor.name += "_{}".format(i)
        self.parallel_envs = [NB201Env(actor, copy.deepcopy(init_state), node_label2feat, reward_module,
                                       max_allowed_steps, output_node_label, scaffold_node_labels)
                              for actor in self.parallel_actors]

    def share_memory(self):
        self.reward_module.share_memory()

    def load_checkpoint(self, actor_state_dict):
        for env in self.parallel_envs:
            env.load_state_dict(actor_state_dict)

    def get_actor_rep(self):
        return self.parallel_envs[0].get_actor()

    def get_env_rep(self):
        return self.parallel_envs[0]

    def get_env_actor_rep(self):
        return self.get_env_rep(), self.get_actor_rep()

    def re_clone_actors(self, new_actor):
        for i, actor in enumerate(self.parallel_actors):
            if actor == new_actor: continue
            actor.load_state_dict(new_actor.state_dict())
            self.parallel_envs[i].set_actor(actor)

    def to(self, dev):
        for env in self.parallel_envs:
            env.to(dev)
        return self

    def reset(self):
        for env in self.parallel_envs:
            env.reset()

    def eval(self):
        for env in self.parallel_envs:
            env.eval()

    def train(self):
        for env in self.parallel_envs:
            env.train()

    def __len__(self):
        return len(self.parallel_envs)

    def __str__(self):
        return "NB201EnvManager with {} parallel environments".format(len(self.parallel_envs))

    def __repr__(self):
        return str(self)

    def __iter__(self):
        self.curr_env_idx = 0
        return self

    def __next__(self):
        if self.curr_env_idx < len(self.parallel_envs):
            env = self.parallel_envs[self.curr_env_idx]
            self.curr_env_idx += 1
            return env
        else:
            self.curr_env_idx = 0
            raise StopIteration


class NB201ActorCritic(torch.nn.Module):

    def __init__(self, name, policy_net, value_net):
        super(NB201ActorCritic, self).__init__()
        self.name = name
        self.policy_net = policy_net
        self.value_net = value_net

    def forward(self, *args):
        return self.policy_net(*args)

    def get_value_for_state(self, state):
        node_embedding, _ = self.policy_net(*state.get_gpn_compatible_input())
        value = self.value_net(node_embedding).squeeze()
        return value

    def get_batch_values_for_states(self, batch_states):
        node_embedding, _ = self.policy_net(*batch_states.get_gpn_compatible_input())
        values = self.value_net(node_embedding).squeeze()
        return values

    def get_action_for_state(self, state, timestep=0):
        _, b_node_probs = self.policy_net(*state.get_gpn_compatible_input())
        assert b_node_probs.shape[0] == 1
        inst_node_probs = b_node_probs.squeeze(0)
        action = NB201Action(self.name, timestep, inst_node_probs)
        return action

    def get_batch_actions_for_states(self, batch_states, timestep=0):
        _, b_node_probs = self.policy_net(*batch_states.get_gpn_compatible_input())
        actions = []
        for si, state in enumerate(batch_states.states):
            inst_node_probs = b_node_probs[si, :]
            action = NB201Action(self.name, timestep, inst_node_probs).to(device())
            actions.append(action)
        batch_actions = BatchNB201Actions(actions)
        return batch_actions

    def get_name(self):
        return self.name


class NB201PPOBatchDataLoader:

    def __init__(self, batch_size, scale_reward=False, reward_discount_factor=1.0, gae_lambda=0.97,
                 min_reward=0., max_reward=None):
        self.batch_size = batch_size
        self.curr_batch_idx = 0
        self.trajectories = []
        self.batches = []
        self.n_instances = 0
        self.scale_reward = scale_reward
        self.data_built = False
        self.reward_discount_factor = reward_discount_factor
        self.gae_lambda = gae_lambda
        self.min_reward = min_reward
        self.max_reward = max_reward

    def clear(self):
        self.curr_batch_idx = 0
        self.trajectories = []
        self.batches = []
        self.data_built = False

    def add_trajectory(self, state_pairs, actions, rewards, input_state_values, end_state_value):
        assert len(state_pairs) == len(actions) == len(rewards) == len(input_state_values)
        self.trajectories.append( (state_pairs, actions, rewards, input_state_values, end_state_value) )

    def get_next_batch(self):
        batch = self.batches[self.curr_batch_idx]
        self.curr_batch_idx += 1
        return batch

    def build_batches(self):
        min_reward, max_reward = None, None
        if self.scale_reward:
            max_reward = self.max_reward if self.max_reward is not None else global_max([t[2] for t in self.trajectories])
            min_reward = self.min_reward if self.min_reward is not None else global_min([t[2] for t in self.trajectories])
        instances = []
        for state_pairs, actions, rewards, input_state_values, end_state_value in tqdm(self.trajectories,
                                                                                       desc="Building instances from trajectories",
                                                                                       ascii=True, leave=False):
            if self.scale_reward:
                scaled_rewards = [
                    exp_scale(r, old_range_min=min_reward, old_range_max=max_reward) if min_reward <= r <= max_reward else r
                        for r in rewards
                ]
                advantages = compute_t_step_gae(scaled_rewards, input_state_values, end_state_value,
                                                self.reward_discount_factor, self.gae_lambda)
            else:
                advantages = compute_t_step_gae(rewards, input_state_values, end_state_value,
                                                self.reward_discount_factor, self.gae_lambda)
            for (input_state, output_state), action, reward, value, adv in zip(state_pairs, actions, rewards,
                                                                               input_state_values, advantages):
                instances.append( (input_state, action, reward, value, adv) )
        random.shuffle(instances)
        self.n_instances = len(instances)
        print("Building batches")
        # Group instances by state dimensions
        size2instances = collections.defaultdict(list)
        for inst in instances:
            partial_graph_size = len(inst[0])
            size2instances[partial_graph_size].append(inst)
        for size, instances in size2instances.items():
            inst_i = 0
            while inst_i < len(instances):
                batch_instances = instances[inst_i: inst_i + self.batch_size]
                batch_states = BatchNB201BlockStates([t[0] for t in batch_instances])
                batch_actions = BatchNB201Actions([t[1] for t in batch_instances])
                batch_rewards = torch.FloatTensor([t[2] for t in batch_instances])
                batch_values = torch.FloatTensor([t[3] for t in batch_instances])
                batch_advantages = torch.FloatTensor([t[4] for t in batch_instances])
                self.batches.append((batch_states, batch_actions, batch_rewards, batch_values, batch_advantages))
                inst_i += self.batch_size
        self.data_built = True
        self.shuffle()

    def shuffle(self):
        random.shuffle(self.batches)

    @property
    def n_trajectories(self):
        return len(self.trajectories)

    @property
    def n_batches(self):
        return len(self)

    def __len__(self):
        return len(self.batches)

    def __iter__(self):
        self.curr_batch_idx = 0
        return self

    def __next__(self):
        if not self.data_built:
            self.build_batches()
        if self.curr_batch_idx < self.n_batches:
            batch = self.get_next_batch()
            return batch
        else:
            self.curr_batch_idx = 0
            self.shuffle()
            raise StopIteration


class NB201DisRewardModule:

    def __init__(self, d_trainer, evaluator, output_node_label, n_random_eval,
                 node_label2feature, rm_cand_node_labels,
                 n_truth_per_src=100, n_fake_total=400, aggr_fake=True, rm_depth=2, n_mutations=100,
                 enable_mutation=False, pos_reward=0.0, neg_reward=-0.1,
                 lat_weight=0.5, max_d_reward=1.0, valid_block_keys=None):
        self.d_trainer = d_trainer
        self.model = self.d_trainer.get_model()
        self.max_d_reward = max_d_reward
        self.evaluator = evaluator
        self.output_node_label = output_node_label
        self.pos_reward = pos_reward
        self.neg_reward = neg_reward
        self.n_truth_per_src = n_truth_per_src
        self.n_fake_total = n_fake_total
        self.src_block2perfs = {}
        self.src_blocks = []
        self.truth_block2perfs = {}
        self.truth_blocks = []
        self.src2truth = {}
        self.n_random_eval = n_random_eval
        self.node_label2feature = node_label2feature
        self.rm_cand_node_labels = rm_cand_node_labels
        self.rm_depth = rm_depth
        self.n_mutations = n_mutations
        self.lat_weight = lat_weight
        self.valid_block_keys = valid_block_keys
        self.enable_mutation = enable_mutation
        self.aggr_fake = aggr_fake
        self.past_fake_blocks = []

    def train(self):
        self.model.train()

    def eval(self):
        self.model.eval()

    def to(self, dev):
        self.d_trainer.to(dev)
        return self

    def get_model(self):
        return self.model

    def share_memory(self):
        self.model.share_memory()

    def clear(self):
        self.src_block2perfs = {}
        self.src_blocks = []
        self.truth_block2perfs = {}
        self.truth_blocks = []
        self.src2truth = {}
        self.past_fake_blocks = []

    def get_fake_data_size(self):
        if self.enable_mutation:
            return self.n_fake_total
        else:
            return len(self.truth_blocks)

    def build_truth_blocks(self, src_block2perfs, log_f=print):
        self.clear()
        self.src_blocks = [v[0] for k, v in src_block2perfs.items()]
        self.src_block2perfs = src_block2perfs
        if self.enable_mutation:
            for src_b in self.src_blocks:
                mutated_blocks = random_mutate_nb201_blocks(self.rm_depth, [src_b], self.n_mutations,
                                                            self.node_label2feature, self.rm_cand_node_labels,
                                                            self.output_node_label, verbose=False)
                blocks = mutated_blocks[:self.n_truth_per_src]
                self.truth_blocks.extend(blocks)
                self.src2truth[src_b.get_str_id()] = blocks
        else:
            self.truth_blocks = [b for b in self.src_blocks]
            for src_b in self.src_blocks:
                self.src2truth[src_b.get_str_id()] = self.truth_blocks
        log_f("{} truth blocks built".format(len(self.truth_blocks)))

    def random_eval_blocks(self):
        random.shuffle(self.truth_blocks)
        cand_blocks = self.truth_blocks[:self.n_random_eval]
        for block in cand_blocks:
            key = block.get_str_id()
            if key not in self.truth_block2perfs:
                acc, lat = self.evaluator.get_perf_values(block)
                self.truth_block2perfs[key] = (block, (acc, lat))
        return self.truth_block2perfs

    def d_train(self, fake_blocks, log_f=None):
        log_f("Building data pairs")
        if self.aggr_fake:
            unique_fake_blocks = UniqueList(key_func=lambda _b: _b.get_str_id())
            unique_fake_blocks.extend(fake_blocks)
            unique_fake_blocks.extend(self.past_fake_blocks)
            unique_fake_blocks = unique_fake_blocks.tolist()
        else:
            unique_fake_blocks = fake_blocks
        # Build truth pairs
        truth_block_pairs = []
        truth_keys = set()
        for src_key, truth_blocks in self.src2truth.items():
            src_block = self.src_block2perfs[src_key][0]
            for b in truth_blocks:
                truth_block_pairs.append((src_block, b))
                truth_keys.add(b.get_str_id())
        # Build fake pairs
        fake_block_pairs = []
        fake_keys = set()
        for src_block in self.src_blocks:
            for b in unique_fake_blocks:
                k = b.get_str_id()
                fake_keys.add(k)
                # if k in truth_keys:
                #     # We do not let the discriminator to get confused, for the most accurate reward
                #     continue
                fake_block_pairs.append((src_block, b))
        common_keys = truth_keys.intersection(fake_keys)
        log_f("Truth/fake key overlap: {}%".format(100. * len(common_keys) / (len(truth_keys.union(fake_keys)))))
        self.d_trainer.train(truth_block_pairs, fake_block_pairs, log_f)
        # Update past fake blocks
        unique_fake_blocks = UniqueList(key_func=lambda _b: _b.get_str_id())
        unique_fake_blocks.extend(fake_blocks)
        unique_fake_blocks.extend(self.past_fake_blocks)
        self.past_fake_blocks = unique_fake_blocks.tolist()

    def _get_d_reward(self, block):
        d_model = self.model
        reward = 0.
        with torch.no_grad():
            d_model.eval()
            for p_block in self.src_blocks:
                p_block = copy.deepcopy(p_block)
                gen_block_score = d_model(build_nb201_pair_input(p_block, block))
                gen_block_score = torch.softmax(gen_block_score, dim=-1).squeeze()
                reward = max(reward, get_tsr_value(gen_block_score[1]))
        return reward

    def _get_validity_reward(self, state, encourage_length=False):
        n_violations = _get_n_darts_block_violations(state, self.output_node_label)
        if n_violations == 0:
            if encourage_length:
                reward = self.pos_reward * max(0, len(state.curr_node_label_list) - 2)
            else:
                reward = self.pos_reward
            return reward
        else:
            return n_violations * self.neg_reward

    def get_perf_diff(self, acc, lat):
        assert 0 <= acc <= 1
        assert 0 <= lat <= 1
        # Find the block that is closest to the input block in terms of L1-distance
        src_d = []
        for _, (s_block, perf) in self.src_block2perfs.items():
            s_acc, s_lat = perf
            src_d.append( (s_block, abs(s_acc - acc) + abs(s_lat - lat), s_acc, s_lat) )
        src_d.sort(key=lambda t: t[1])
        s_acc, s_lat = src_d[0][2], src_d[0][3]
        acc_diff = acc - s_acc
        lat_diff = (1. - lat) - (1. - s_lat)
        return acc_diff, lat_diff

    def get_reward(self, state):
        assert self.src_block2perfs is not None
        key = state.get_str_id()
        validity_reward = self._get_validity_reward(state)
        if validity_reward < 0.:  # No other reward if block is invalid
            rv = validity_reward
        elif key in self.src_block2perfs:
            rv = validity_reward  # No additional reward if generated the same src blocks!
        else:
            rv = self._get_d_reward(state)
            if key in self.truth_block2perfs:
                # Check if this state is in block2perfs
                # If yes, the reward depends on the dominance over the src perfs
                # If complete dominance, d_reward reward is max, then the perf improvements are added.
                # If partial/no dominance, reward = improvements + d_reward
                _, (acc, lat) = self.truth_block2perfs[key]
                acc_diff, lat_diff = self.get_perf_diff(acc, lat)
                lat_diff *= self.lat_weight  # Reduce the impact of easy-to-improve metric
                perf_diff = acc_diff + lat_diff
                if perf_diff > 0:
                    rv = self.max_d_reward + perf_diff
                else:
                    rv += perf_diff
        return rv


class NB201DisTrainer:

    def __init__(self, n_epochs, model, init_lr, batch_size, max_grad_norm=5.0, early_stop_train_f1=0.9):
        self.criterion = torch.nn.CrossEntropyLoss()
        self.init_lr = init_lr
        self.original_model = model
        self.curr_model = model
        self.n_epochs = n_epochs
        self.batch_size = batch_size
        self.max_grad_norm = max_grad_norm
        self.early_stop_train_f1 = early_stop_train_f1

    def to(self, dev):
        self.curr_model = self.curr_model.to(dev)
        self.criterion = self.criterion.to(dev)
        return self

    def _make_optimizer(self, model):
        return torch.optim.Adam(model.parameters(), lr=self.init_lr)

    def get_model(self):
        return self.curr_model

    def train(self, truth_block_pairs, fake_block_pairs, log_f=None):
        self.curr_model.train()
        optimizer = self._make_optimizer(self.curr_model)
        truth_data = [(p[0], p[1], torch.LongTensor([1]).to(device())) for p in truth_block_pairs]
        fake_data = [(p[0], p[1], torch.LongTensor([0]).to(device())) for p in fake_block_pairs]
        if log_f is not None:
            log_f("Training discriminator with {} truth, {} fake instances".format(len(truth_data), len(fake_data)))
        assert len(truth_data) > 0, "DisTrainer received 0 truth data"
        assert len(fake_data) > 0, "DisTrainer received 0 fake data"
        data = truth_data + fake_data
        for epoch in range(self.n_epochs):
            random.shuffle(data)
            total_loss, n_insts = 0, 0
            y_pred, y_true = [], []
            batch_loss = torch.FloatTensor([0.]).to(device())
            for b1, b2, target in tqdm(data, desc="Training on {} instances".format(len(data)),
                                      leave=False, ascii=True):
                target.requires_grad = False
                block_score = self.curr_model(build_nb201_pair_input(b1, b2))
                inst_loss = self.criterion(block_score, target)
                batch_loss += inst_loss
                total_loss += get_tsr_value(inst_loss)
                if (n_insts + 1) % self.batch_size == 0 or n_insts == len(data) - 1:
                    optimizer.zero_grad()
                    batch_loss.backward()
                    torch.nn.utils.clip_grad_norm_(self.curr_model.parameters(), self.max_grad_norm)
                    optimizer.step()
                    batch_loss = torch.FloatTensor([0.]).to(device())
                n_insts += 1
                _, pred_ind = block_score.max(1)
                pred_ind = get_tsr_value(pred_ind)
                truth_ind = get_tsr_value(target)
                y_pred.append(pred_ind)
                y_true.append(truth_ind)
            rv_loss = total_loss / n_insts
            if len(set(y_pred)) <= 1:
                train_f1 = 0.
            else:
                train_f1 = f1_score(y_true=y_true, y_pred=y_pred)
            if log_f is not None: log_f("D-train epoch {} of {} loss: {:.3f}, f1: {:.3f}".format(epoch+1, self.n_epochs,
                                                                                                 rv_loss, train_f1))
            if train_f1 >= self.early_stop_train_f1:
                if log_f is not None: log_f("D-train early stop at epoch {}".format(epoch + 1))
                break


def random_mutate_nb201_block(block:NB201BlockState, node_label2feature,
                              searchable_ops=NB201_SEARCHABLE_OPS):
    # For a DARTS cell, only swap node is supported
    new_block = copy.deepcopy(block)
    idx = random.randint(1, len(new_block) - 2)
    old_label = new_block.node_labels[idx]
    cand_labels = [lab for lab in searchable_ops if lab != old_label]
    new_label = random.choice(cand_labels)
    new_feature_tsr = node_label2feature[new_label]
    new_block.node_labels[idx] = new_label
    new_block.node_tsr[idx] = new_feature_tsr
    return new_block


def random_mutate_nb201_blocks(depth, blocks, num_mutations, node_label2feature,
                               searchable_ops=NB201_SEARCHABLE_OPS,
                               output_node_label=NB201_OUTPUT_NODE_LABEL, verbose=True):
    src_blocks = [b for b in blocks]
    new_blocks = []
    generated_block_keys = {b.get_str_id() for b in blocks}
    bar = None
    if verbose: bar = tqdm(total=len(src_blocks), desc="Mutating blocks", ascii=True)
    for src_block in src_blocks:
        curr_mutants = [src_block]
        curr_mutant_keys = set()
        for d in range(depth):
            new_mutants = []
            for block in curr_mutants:
                for mi in range(num_mutations):
                    new_block = random_mutate_nb201_block(block, node_label2feature, searchable_ops)
                    if not is_nb201_block_valid(new_block, output_node_label=output_node_label):
                        continue # ignore invalid mutants
                    new_block_id = new_block.get_str_id()
                    if new_block_id not in curr_mutant_keys:
                        curr_mutant_keys.add(new_block_id)
                        new_mutants.append(new_block)
                    if new_block_id not in generated_block_keys:
                        generated_block_keys.add(new_block_id)
                        new_blocks.append(new_block)
            curr_mutants.extend(new_mutants)
        if bar is not None: bar.update(1)
    if bar is not None: bar.close()
    return new_blocks


def is_nb201_block_valid(block:NB201BlockState, output_node_label):
    return _get_n_darts_block_violations(block, output_node_label) == 0


def _get_n_darts_block_violations(block:NB201BlockState, output_node_label="output"):
    n_violations = 0
    # Hack to make this compatible
    from model_components import SimpleBlock
    b = SimpleBlock(block.node_labels, block.node_tsr, block.curr_edge_list)
    if block_has_invalid_output(b, output_node_label):
        n_violations += 1
    if block_has_invalid_edges(b):
        n_violations += 1
    if len(get_output_node_indices(b.curr_edge_list)) > 1:
        n_violations += 1
    input_set = get_input_node_indices(b.curr_edge_list)
    if len(input_set) > 2:
        n_violations += 1
    if block_has_loner_nodes(b):
        n_violations += 1
    return n_violations


def make_nb201_block_discriminator(num_unique_nodes, dis_in_channels, dis_hidden_size, dis_out_channels,
                                   dis_gnn_constructor, dis_num_layers, dis_dropout_prob):
    pre_embedding = torch.nn.Embedding(num_unique_nodes, dis_in_channels)
    encoder = GraphEncoder(dis_in_channels, dis_hidden_size, dis_out_channels, dis_gnn_constructor,
                           node_pre_embedding=pre_embedding, activ=torch.nn.Tanh(),
                           num_layers=dis_num_layers, dropout_prob=dis_dropout_prob)
    discriminator = NB201BlockDiscriminator(encoder, torch.nn.Linear(2 * dis_out_channels, 2))
    return discriminator


def make_nb201_gpn(num_unique_nodes, gpn_in_channels, gpn_hidden_size, gpn_out_channels, gpn_gnn_constructor,
                   gpn_num_layers, gpn_dropout_prob):
    pre_embedding = torch.nn.Embedding(num_unique_nodes, gpn_in_channels)
    action_predictor = NB201ActionPredictor(gpn_out_channels, gpn_hidden_size,
                                            torch.nn.Linear(gpn_out_channels, num_unique_nodes - 2))
    encoder = GraphEncoder(gpn_in_channels, gpn_hidden_size, gpn_out_channels, gpn_gnn_constructor,
                           node_pre_embedding=pre_embedding, activ=torch.nn.Tanh(),
                           num_layers=gpn_num_layers, dropout_prob=gpn_dropout_prob)
    model = NB201BlockGPN(encoder, action_predictor)
    return model


def make_nb201_actor_critic(policy_net, gpn_out_channels):
    value_net = torch.nn.Sequential(NodeAggr(gpn_out_channels, aggr_dim=1, method="sum", enable_aggr_net=False),
                                    torch.nn.Linear(gpn_out_channels, 1))
    actor_critic = NB201ActorCritic("DartsActorCritic", policy_net, value_net)
    return actor_critic


def make_nb201_env_manager(actor_critic, n_parallel_actors, init_state, node_label2feat,
                           reward_module, max_allowed_steps, output_node_label, scaffold_node_labels):
    parallel_actors = [actor_critic] + [copy.deepcopy(actor_critic) for _ in range(n_parallel_actors - 1)]
    env_manager = NB201EnvManager(parallel_actors, n_parallel_actors, init_state, node_label2feat, scaffold_node_labels,
                                  reward_module, max_allowed_steps, output_node_label)
    return env_manager


def nb201_ppo_train(env_manager, optimizer, value_criterion, n_train_rounds, k_epochs, t_timesteps, batch_size,
                    reward_discount_factor, clip_epsilon, gae_lambda, c1, c2, book_keeper,
                    actor_update_episode_interval, parallel_explore=False):
    env_manager.train()
    ppo_data_loader = NB201PPOBatchDataLoader(batch_size, reward_discount_factor=reward_discount_factor,
                                              gae_lambda=gae_lambda)
    block_w_rewards = UniqueList(key_func=lambda t: t[0].get_str_id())
    for i in range(n_train_rounds):
        book_keeper.log("Training round {} of {}".format(i + 1, n_train_rounds))
        book_keeper.log("Running {} actor(s), each for {} steps".format(len(env_manager), t_timesteps))
        book_keeper.log("Episodes per actor: {}".format(actor_update_episode_interval))
        env_manager.train()
        # Exploration
        if parallel_explore:
            br_list = run_parallel_exploration(env_manager, ppo_data_loader, actor_update_episode_interval, t_timesteps,
                                               log_f=book_keeper.log)
        else:
            br_list = run_seq_exploration(env_manager, ppo_data_loader, actor_update_episode_interval, t_timesteps,
                                          log_f=book_keeper.log)
        block_w_rewards.extend(br_list)
        # Train generator(actor)
        updated_actor = env_manager.get_actor_rep()
        ppo_loss = run_nb201_k_ppo_batch_update_epoch(updated_actor, optimizer, ppo_data_loader, value_criterion,
                                                      k_epochs, clip_epsilon, c1, c2)
        book_keeper.log("Avg PPO loss: {}".format(ppo_loss))
        env_manager.re_clone_actors(updated_actor)
        ppo_data_loader.clear()
    return env_manager.get_actor_rep(), block_w_rewards


def run_nb201_k_ppo_batch_update_epoch(actor, optimizer, data_loader, value_criterion, k_epochs, clip_epsilon, c1, c2):
    actor.train()
    avg_loss, num_updates = 0., 0
    for e in tqdm(range(k_epochs), desc="PPO train", ascii=True, leave=False):
        for i, batch in enumerate(data_loader):
            b_input_states, b_old_actions, _, old_state_values, advantages = batch
            b_old_actions = b_old_actions.to(device())
            advantages = advantages.to(device())
            old_state_values = old_state_values.to(device())
            b_new_actions = actor.get_batch_actions_for_states(b_input_states)
            new_state_values = actor.get_batch_values_for_states(b_input_states)
            old_node_logp = b_old_actions.get_action_log_probs()
            new_node_logp = b_new_actions.get_action_log_probs(batch_node_indices=b_old_actions.batch_node_indices)
            b_ratios = torch.exp(new_node_logp - old_node_logp.detach())
            surr1 = b_ratios * advantages
            surr2 = torch.clamp(b_ratios, 1 - clip_epsilon, 1 + clip_epsilon) * advantages
            policy_loss = -1. * torch.min(surr1, surr2)
            policy_loss = policy_loss.mean()
            target_values = old_state_values + advantages
            value_loss = c1 * value_criterion(new_state_values, target_values).mean()
            entropy_loss = -1. * c2 * b_new_actions.get_entropy()
            batch_loss = policy_loss + value_loss + entropy_loss
            optimizer.zero_grad()
            batch_loss.backward()
            optimizer.step()
            avg_loss += batch_loss.item()
            num_updates += 1
        data_loader.shuffle()
    assert num_updates > 0, "data_loader len: {}".format(len(data_loader))
    return avg_loss / num_updates


def _generate_nb201_blocks(actor, env, n_desired_blocks, output_node_label,
                           max_n_attempts=2000, verbose=True, desc="Generating blocks",
                           valid_block_keys=None):
    blocks = UniqueList(key_func=lambda b: b.get_str_id())
    n_attempts = 0
    bar = None
    if verbose: bar = tqdm(total=n_desired_blocks, desc=desc, ascii=True, leave=False)
    while n_attempts < max_n_attempts and len(blocks) < n_desired_blocks:
        env.reset()
        with torch.no_grad():
            final_state_reached = False
            while not final_state_reached:
                action = actor.get_action_for_state(env.get_curr_state(), env.get_num_steps())
                final_state_reached, input_state, reward, new_state = env.step(action)
            valid = is_nb201_block_valid(new_state, output_node_label)
            if valid_block_keys is not None:
                if hasattr(valid_block_keys, "get_id"):
                    key = valid_block_keys.get_id(new_state)
                else:
                    key = new_state.get_str_id()
                valid = valid and key in valid_block_keys
            if valid and blocks.append(new_state):
                n_attempts = 0
                if bar is not None: bar.update(1)
        n_attempts += 1
    env.reset()
    if bar is not None: bar.close()
    blocks = blocks.tolist()
    return blocks


def eval_nb201_acc(iteration, actor, env, evaluator:BlockPerfEvaluator, acc_set,
                   n_desired_blocks, output_node_label, book_keeper,
                   update_callback=None, valid_block_keys=None):
    new_block2acc = {}
    new_blocks = _generate_nb201_blocks(actor, env, n_desired_blocks, output_node_label,
                                        desc="Eval", valid_block_keys=valid_block_keys)
    for block in tqdm(new_blocks, desc="Getting performance", ascii=True):
        key = block.get_str_id()
        acc, lat = evaluator.get_perf_values(block)
        new_block2acc[key] = (block, acc)

    if len(new_block2acc) < n_desired_blocks:
        book_keeper.log("Only generated {} unique blocks".format(len(new_block2acc)))

    if len(new_block2acc) == 0:
        book_keeper.log("No blocks with acc generated")
    else:
        node_lens = [len(b) for _, (b, v) in new_block2acc.items()]
        edge_counts = [len(b.curr_edge_list[0]) for _, (b, v) in new_block2acc.items()]
        book_keeper.log("Avg number of nodes in generated blocks: {}".format(mean(node_lens)))
        book_keeper.log("Max number of nodes in generated blocks: {}".format(max(node_lens)))
        book_keeper.log("Min number of nodes in generated blocks: {}".format(min(node_lens)))
        book_keeper.log("Avg number of edges in generated blocks: {}".format(mean(edge_counts)))
        book_keeper.log("Max number of edges in generated blocks: {}".format(max(edge_counts)))
        book_keeper.log("Min number of edges in generated blocks: {}".format(min(edge_counts)))
        gen_acc_values = [acc for _, (b, acc) in new_block2acc.items()]
        book_keeper.log("Max acc in generated blocks: {}".format(max(gen_acc_values)))
        book_keeper.log("Min acc in generated blocks: {}".format(min(gen_acc_values)))
        acc_set.update(new_block2acc, log_f=book_keeper.log)
        if update_callback is not None:
            update_callback(acc_set, iteration, book_keeper.model_name)
        book_keeper.checkpoint_state_dict(acc_set.state_dict(),
                                          "{}_top_acc_set.pkl".format(book_keeper.model_name))
        acc_vals = acc_set.get_top_acc_values()
        book_keeper.log("Best acc so far: {}".format(max(acc_vals)))
        book_keeper.log("New top acc set size: {}, new_block2acc len: {}".format(len(acc_set),
                                                                                 len(new_block2acc)))
    return [b for _, (b, _) in new_block2acc.items()]


class NB201Evaluator(BlockPerfEvaluator):

    def __init__(self, perf_dict, ni2de_map, default_clue=1.0):
        super(NB201Evaluator, self).__init__(default_clue=default_clue)
        self.perf_dict = perf_dict
        self.ni2de_map = ni2de_map

    def get_perf_reward(self, state):
        pred_acc, pred_lat = self.get_perf_values(state)
        assert 0 <= pred_acc <= 1, "Acc value should be between 0 and 1, got {}".format(pred_acc)
        assert 0 <= pred_lat <= 1, "Lat value should be between 0 and 1, got {}".format(pred_lat)
        acc_reward = self.clue * pred_acc
        lat_reward = (1. - self.clue) * (1 - pred_lat)
        aggr_reward = acc_reward + lat_reward
        return aggr_reward

    def get_perf_values(self, state:NB201BlockState):
        key = get_nb201_key(state.node_labels, self.ni2de_map)
        if self.enable_perf_memo and key in self.perf_memo:
            acc, lat = self.perf_memo[key]
        else:
            acc, lat = self.perf_dict[key]
            self.perf_memo[key] = acc, lat
            interval_info.num_block_eval += 1
        return acc, lat


def nb201_ga_nas_acc_search(env_manager, ppo_optimizer, value_criterion, top_acc_set,
                            evaluator, n_train_iterations, n_train_rounds,
                            k_epochs, t_timesteps, batch_size, n_desired_blocks_eval,
                            output_node_label, book_keeper, parallel_explore=False,
                            m_repeats=2, reward_discount_factor=1.0, debug=True,
                            gae_lambda=0.97, clip_epsilon=0.2, c1=0.5, c2=0.01,
                            n_desired_blocks_inc=0,
                            actor_update_episode_interval=500,
                            acc_update_callback=None, valid_block_keys=None):
    env_manager = env_manager.to(device())
    value_criterion = value_criterion.to(device())
    reward_module = env_manager.reward_module
    g_start = time.time()
    # PPO training iterations
    for i in range(n_train_iterations):
        report_iteration = i + 1
        book_keeper.log("Running iteration {} of {}".format(report_iteration, n_train_iterations))
        env, actor = env_manager.get_env_actor_rep()
        src_block2perfs = top_acc_set.top_block2acc()
        reward_module.build_truth_blocks(src_block2perfs, log_f=book_keeper.log)

        blocks_from_train = UniqueList(key_func=lambda b: b.get_str_id())
        for mi in range(m_repeats):
            book_keeper.log("Inner training loop {} of {}".format(mi + 1, m_repeats))
            # Train discriminator first
            n_desired_blocks_dis = reward_module.get_fake_data_size()
            cand_blocks = _generate_nb201_blocks(actor, env, n_desired_blocks_dis, output_node_label,
                                                 desc="Gen blocks for discriminator training")
            book_keeper.log("{} unique blocks for training discriminator".format(len(cand_blocks)))
            reward_module.d_train(cand_blocks, log_f=book_keeper.log)

            actor, block_w_rewards = nb201_ppo_train(env_manager, ppo_optimizer, value_criterion, n_train_rounds,
                                                     k_epochs, t_timesteps, batch_size, reward_discount_factor,
                                                     clip_epsilon, gae_lambda, c1, c2, book_keeper,
                                                     parallel_explore=parallel_explore,
                                                     actor_update_episode_interval=actor_update_episode_interval)
            book_keeper.log("{} unique valid blocks generated during training".format(len(block_w_rewards)))
            blocks_from_train.extend([t[0] for t in block_w_rewards])

        book_keeper.log("Eval at iteration {}".format(report_iteration))
        gen_blocks = eval_nb201_acc(report_iteration, actor, env, evaluator, top_acc_set,
                                    n_desired_blocks_eval + i * n_desired_blocks_inc,
                                    output_node_label, book_keeper,
                                    valid_block_keys=valid_block_keys,
                                    update_callback=acc_update_callback)

        if debug:
            top_blocks = top_acc_set.get_curr_blocks()
            truth_blocks = top_acc_set.get_curr_blocks()
            book_keeper.log("")
            book_keeper.log("====== Start of debug info ======")
            truth_keys = set(b.get_str_id() for b in truth_blocks)
            train_block_keys = set(b.get_str_id() for b in blocks_from_train.tolist())
            common_keys = train_block_keys.intersection(truth_keys)
            book_keeper.log("{} unique blocks from train".format(len(train_block_keys)))
            book_keeper.log("{}% blocks from train are from truth".format(100. * float(len(common_keys)) / len(train_block_keys)))
            truth_keys = set(b.get_str_id() for b in truth_blocks)
            gen_keys = set(b.get_str_id() for b in gen_blocks)
            common_keys = gen_keys.intersection(truth_keys)
            book_keeper.log("{} unique gen blocks".format(len(gen_keys)))
            book_keeper.log("{}% gen blocks are from truth".format(100. * float(len(common_keys)) / len(gen_keys)))
            book_keeper.log("------")
            pair_dis_pred_check(truth_blocks, top_blocks, reward_module.get_model(),
                                log_f=book_keeper.log, desc="truth")
            pair_dis_pred_check(gen_blocks, top_blocks, reward_module.get_model(),
                                log_f=book_keeper.log, desc="fake")
            book_keeper.log("------")
            report_nb201_blocks_info(truth_blocks, log_f=book_keeper.log, desc="truth")
            report_nb201_blocks_info(gen_blocks, log_f=book_keeper.log, desc="fake")
            book_keeper.log("====== End of debug info ======")
            book_keeper.log("")
        book_keeper.log("Num blocks evaluated so far: {}".format(interval_info.num_block_eval))
    book_keeper.log("Search seconds: {}".format(time.time() - g_start))


def nb201_ga_nas_pareto_search(env_manager, ppo_optimizer, value_criterion, pareto_front,
                               evaluator:BlockPerfEvaluator, n_train_iterations, n_train_rounds,
                               k_epochs, t_timesteps, batch_size, n_desired_blocks_eval,
                               output_node_label, book_keeper, parallel_explore=False,
                               m_repeats=2, reward_discount_factor=1.0, debug=True,
                               gae_lambda=0.97, clip_epsilon=0.2, c1=0.5, c2=0.1, c3=0.1,
                               actor_update_episode_interval=500,
                               pareto_update_callback=None, valid_block_keys=None):
    env_manager = env_manager.to(device())
    value_criterion = value_criterion.to(device())
    reward_module = env_manager.final_reward_module
    g_start = time.time()
    for i in range(n_train_iterations):
        report_iteration = i + 1
        book_keeper.log("Running iteration {} of {}".format(report_iteration, n_train_iterations))
        env, actor = env_manager.get_env_actor_rep()
        book_keeper.log("Current c2: {}".format(c2))
        book_keeper.log("Current c3: {}".format(c3))
        src_block2perfs = pareto_front.get_pareto_neighborhood_block2perfs(max_runs=0, log_f=book_keeper.log)
        reward_module.build_truth_blocks(src_block2perfs, log_f=book_keeper.log)
        reward_module.random_eval_blocks()

        blocks_from_train = UniqueList(key_func=lambda t: t[0].get_str_id())
        for mi in range(m_repeats):
            book_keeper.log("Inner training loop {} of {}".format(mi + 1, m_repeats))
            # Train discriminator first
            n_desired_blocks_dis = reward_module.get_fake_data_size()
            cand_blocks = _generate_nb201_blocks(actor, env, n_desired_blocks_dis, output_node_label,
                                                 desc="Gen blocks for discriminator training")
            book_keeper.log("{} unique blocks for training discriminator".format(len(cand_blocks)))
            reward_module.d_train(cand_blocks, log_f=book_keeper.log)

            # PPO train actor (generator)
            actor, block_w_rewards = nb201_ppo_train(env_manager, ppo_optimizer, value_criterion, n_train_rounds,
                                                     k_epochs, t_timesteps, batch_size, reward_discount_factor,
                                                     clip_epsilon, gae_lambda, c1, c2, book_keeper,
                                                     parallel_explore=parallel_explore,
                                                     actor_update_episode_interval=actor_update_episode_interval)
            book_keeper.log("{} unique valid blocks generated during training".format(len(block_w_rewards)))
            blocks_from_train.extend(block_w_rewards)
        blocks_from_train = blocks_from_train.tolist()
        blocks_from_train.sort(key=lambda t: t[1], reverse=True)

        book_keeper.log("Eval at iteration {}".format(report_iteration))
        pareto_blocks = pareto_front.get_curr_blocks()
        truth_blocks = [b for b in reward_module.truth_blocks]
        gen_blocks = eval_nb201_acc_lat(report_iteration, actor, env, evaluator, pareto_front,
                                        n_desired_blocks_eval, output_node_label, book_keeper,
                                        valid_block_keys=valid_block_keys,
                                        pareto_update_callback=pareto_update_callback)

        if debug:
            book_keeper.log("")
            book_keeper.log("====== Start of debug info ======")
            pair_dis_pred_check(truth_blocks, pareto_blocks, reward_module.get_model(),
                                log_f=book_keeper.log, desc="truth")
            pair_dis_pred_check(gen_blocks, pareto_blocks, reward_module.get_model(),
                                log_f=book_keeper.log, desc="fake")
            book_keeper.log("------")
            truth_keys = set(b.get_str_id() for b in truth_blocks)
            train_block_keys = set([b.get_str_id() for b, _ in blocks_from_train])
            common_keys = train_block_keys.intersection(truth_keys)
            book_keeper.log("{} unique blocks from train".format(len(train_block_keys)))
            book_keeper.log("{}% blocks from train are from truth".format(100. * float(len(common_keys)) / len(train_block_keys)))
            truth_keys = set(b.get_str_id() for b in truth_blocks)
            gen_keys = set(b.get_str_id() for b in gen_blocks)
            common_keys = gen_keys.intersection(truth_keys)
            book_keeper.log("{} unique gen blocks".format(len(gen_keys)))
            book_keeper.log("{}% gen blocks are from truth".format(100. * float(len(common_keys)) / len(gen_keys)))
            book_keeper.log("------")
            report_nb201_blocks_info(truth_blocks, log_f=book_keeper.log, desc="truth")
            report_nb201_blocks_info(gen_blocks, log_f=book_keeper.log, desc="fake")
            book_keeper.log("====== End of debug info ======")
            book_keeper.log("")
        book_keeper.log("Num blocks evaluated so far: {}".format(interval_info.num_block_eval))
    book_keeper.log("Search seconds: {}".format(time.time() - g_start))


def eval_nb201_acc_lat(iteration, actor, env, evaluator:BlockPerfEvaluator, pareto_front,
                       n_desired_blocks, output_node_label, book_keeper,
                       pareto_update_callback=None, valid_block_keys=None):
    new_block2perfs = {}
    new_block2count = collections.defaultdict(int)
    new_blocks = _generate_nb201_blocks(actor, env, n_desired_blocks, output_node_label,
                                        desc="Eval", valid_block_keys=valid_block_keys)
    assert len(new_blocks) > 0, "No new blocks sampled"
    for block in tqdm(new_blocks, desc="Getting performance", ascii=True):
        key = block.get_str_id()
        new_block2count[key] += 1
        acc, lat = evaluator.get_perf_values(block)
        new_block2perfs[key] = (block, (acc, lat))
    book_keeper.log("Sampled {} blocks, {} unique ones".format(len(new_blocks), len(new_block2perfs)))

    if len(new_block2perfs) == 0:
        book_keeper.log("No valid blocks with perf generated")
    else:
        node_lens = [len(b) for _, (b, v) in new_block2perfs.items()]
        edge_counts = [len(b.curr_edge_list[0]) for _, (b, v) in new_block2perfs.items()]
        book_keeper.log("Avg number of nodes in generated blocks: {}".format(mean(node_lens)))
        book_keeper.log("Max number of nodes in generated blocks: {}".format(max(node_lens)))
        book_keeper.log("Min number of nodes in generated blocks: {}".format(min(node_lens)))
        book_keeper.log("Avg number of edges in generated blocks: {}".format(mean(edge_counts)))
        book_keeper.log("Max number of edges in generated blocks: {}".format(max(edge_counts)))
        book_keeper.log("Min number of edges in generated blocks: {}".format(min(edge_counts)))
        gen_acc_values = [v[0] for _, (b, v) in new_block2perfs.items()]
        book_keeper.log("Max acc in generated blocks: {}".format(max(gen_acc_values)))
        book_keeper.log("Min acc in generated blocks: {}".format(min(gen_acc_values)))
        pareto_front.update_pareto_front(new_block2perfs, log_f=book_keeper.log)
        if pareto_update_callback is not None:
            pareto_update_callback(pareto_front, iteration, book_keeper.model_name)
        book_keeper.checkpoint_state_dict(pareto_front.state_dict(),
                                          "{}_pareto_front.pkl".format(book_keeper.model_name))
        pareto_perf_values = pareto_front.get_pareto_block_perf_values()
        p_acc_values = [t[0] for t in pareto_perf_values]
        p_lat_values = [t[1] for t in pareto_perf_values]
        book_keeper.log("Max acc in current Pareto: {}".format(max(p_acc_values)))
        book_keeper.log("Min acc in current Pareto: {}".format(min(p_acc_values)))
        book_keeper.log("Max lat in current Pareto: {}".format(max(p_lat_values)))
        book_keeper.log("Min lat in current Pareto: {}".format(min(p_lat_values)))
        book_keeper.save_acc_lat_pareto_values(p_acc_values, p_lat_values, "iter {}".format(iteration))
        book_keeper.save_block2perfs_values(new_block2perfs, iteration)
        book_keeper.log("New front size: {}, new_block2perfs len: {}".format(len(pareto_front),
                                                                             len(new_block2perfs)))
    return [b for _, (b, _) in new_block2perfs.items()]


def get_random_nb201_blocks(init_state:NB201BlockState, n_darts_edges, n_ops, scaffold_labels, node_label2feat,
                            output_node_label=NB201_OUTPUT_NODE_LABEL, n_blocks=100):
    blocks = []
    with torch.no_grad():
        for i in range(n_blocks):
            curr_block = copy.deepcopy(init_state)
            for j in range(n_darts_edges):
                # Adds a bit more randomness
                node_alpha = np.random.uniform(size=n_ops)
                node_alpha = torch.from_numpy(node_alpha).float()
                node_alpha = torch.softmax(node_alpha, dim=-1)
                # # Uniform
                # node_alpha = torch.FloatTensor([1./n_ops for _ in range(n_ops)])
                node_node_idx = get_tsr_value(Categorical(probs=node_alpha).sample())
                node_label = scaffold_labels[node_node_idx]
                curr_block = curr_block.gen_block_state_with_new_node(node_label,
                                                                      node_label2feat[node_label],
                                                                      node_alpha)
            curr_block = curr_block.gen_block_state_with_new_node(output_node_label,
                                                                  node_label2feat[output_node_label],
                                                                  None)
            blocks.append(curr_block)
    return blocks


def get_nb201_init_block_state(input_node_label, node_label2feat, complete_edge_list):
    state = NB201BlockState([input_node_label],
                            node_label2feat[input_node_label],
                            [[0], [0]],
                            complete_edge_list)
    return state


def init_nb201_acc_set(init_blocks, evaluator, top_k, log_f=print):
    init_block2acc = {}
    for block in tqdm(init_blocks, desc="Getting acc for init blocks", ascii=True):
        acc, _ = evaluator.get_perf_values(block)
        init_block2acc[block.get_str_id()] = (block, acc)
    log_f("Initial block2acc size: {}".format(len(init_block2acc)))
    top_acc_set = BlockTopKAccSet(init_block2acc, top_k, log_f=log_f)
    log_f("Initial top acc set size: {}".format(len(top_acc_set.top_block2acc())))
    return top_acc_set


def pair_dis_pred_check(gen_blocks, top_blocks, d_reward_module, log_f=print, desc="fake"):
    avg_fake_prob = AverageMeter()
    avg_truth_prob = AverageMeter()
    avg_pred_truth_prob = AverageMeter()
    pred_truth_blocks = []
    for block in tqdm(gen_blocks, desc="Discriminator pred check", ascii=True, leave=False):
        max_true_prob = 0.
        for p_block in top_blocks:
            logits = d_reward_module(build_nb201_pair_input(p_block, block))
            prob = torch.softmax(logits, dim=-1).squeeze()
            prob = get_tsr_value(prob)
            max_true_prob = max(max_true_prob, prob[1])
        avg_fake_prob.update(1. - max_true_prob)
        avg_truth_prob.update(max_true_prob)
        if max_true_prob >= 0.5:
            pred_truth_blocks.append(block)
            avg_pred_truth_prob.update(max_true_prob)
    log_f("Discriminator pred check on {} {} blocks".format(len(gen_blocks), desc))
    log_f("Avg fake prob: {}".format(avg_fake_prob.avg))
    log_f("Avg truth prob: {}".format(avg_truth_prob.avg))
    log_f("Avg pred truth prob: {}".format(avg_pred_truth_prob.avg))
    log_f("Number of pred truth blocks: {}".format(len(pred_truth_blocks)))


def report_nb201_blocks_info(blocks, log_f=print, desc=""):
    log_f("Reporting info on {} {} blocks".format(len(blocks), desc))
    op_ratio = collections.defaultdict(float)
    n_ops = 0
    for block in blocks:
        for op in block.node_labels[1:-1]:
            op_ratio[op] += 1
            n_ops += 1
    op_ratio = [(k, v / n_ops) for k, v in op_ratio.items()]
    op_ratio.sort(key=lambda t: t[0])
    log_f("Number of searchable ops: {}".format(n_ops))
    for k, v in op_ratio:
        log_f("- {}: {:.2f}%".format(k, v * 100.))


class NB201BlockStateSampler:

    def __init__(self, cand_op_labels, n_nodes, complete_edge_list,
                 node_label2feat,
                 input_node_label=NB201_INPUT_NODE_LABEL,
                 output_node_label=NB201_OUTPUT_NODE_LABEL):
        self.cand_op_labels = cand_op_labels
        self.complete_edge_list = complete_edge_list
        self.n_nodes = n_nodes
        self.node_label2feat = node_label2feat
        self.input_node_label = input_node_label
        self.output_node_label = output_node_label

    def sample(self):
        n_intermediate_nodes = self.n_nodes - 2
        assert n_intermediate_nodes >= 0
        labels = [random.choice(self.cand_op_labels) for _ in range(n_intermediate_nodes)]
        labels = [self.input_node_label] + labels + [self.output_node_label]
        node_tsr = torch.cat([self.node_label2feat[lab] for lab in labels])
        block = NB201BlockState(labels, node_tsr, self.complete_edge_list, self.complete_edge_list)
        return block


def nb201_ga_nas_ws_acc_search(env_manager, ppo_optimizer, value_criterion, top_acc_set,
                               evaluator, n_train_iterations, n_train_rounds,
                               k_epochs, t_timesteps, batch_size, n_desired_blocks_eval,
                               output_node_label, book_keeper, parallel_explore=False,
                               m_repeats=2, reward_discount_factor=1.0, debug=True,
                               gae_lambda=0.97, clip_epsilon=0.2, c1=0.5, c2=0.01,
                               n_desired_blocks_inc=0,
                               actor_update_episode_interval=500,
                               acc_update_callback=None, valid_block_keys=None):
    env_manager = env_manager.to(device())
    value_criterion = value_criterion.to(device())
    reward_module = env_manager.reward_module
    g_start = time.time()
    # PPO training iterations
    for i in range(n_train_iterations):
        report_iteration = i + 1
        book_keeper.log("Running iteration {} of {}".format(report_iteration, n_train_iterations))
        env, actor = env_manager.get_env_actor_rep()
        src_block2perfs = top_acc_set.top_block2acc()
        reward_module.build_truth_blocks(src_block2perfs, log_f=book_keeper.log)

        blocks_from_train = UniqueList(key_func=lambda b: b.get_str_id())
        for mi in range(m_repeats):
            book_keeper.log("Inner training loop {} of {}".format(mi + 1, m_repeats))
            # Train discriminator first
            n_desired_blocks_dis = reward_module.get_fake_data_size()
            cand_blocks = _generate_nb201_blocks(actor, env, n_desired_blocks_dis, output_node_label,
                                                 desc="Gen blocks for discriminator training")
            book_keeper.log("{} unique blocks for training discriminator".format(len(cand_blocks)))
            reward_module.d_train(cand_blocks, log_f=book_keeper.log)

            actor, block_w_rewards = nb201_ppo_train(env_manager, ppo_optimizer, value_criterion, n_train_rounds,
                                                     k_epochs, t_timesteps, batch_size, reward_discount_factor,
                                                     clip_epsilon, gae_lambda, c1, c2, book_keeper,
                                                     parallel_explore=parallel_explore,
                                                     actor_update_episode_interval=actor_update_episode_interval)
            book_keeper.log("{} unique valid blocks generated during training".format(len(block_w_rewards)))
            blocks_from_train.extend([t[0] for t in block_w_rewards])

        book_keeper.log("Eval at iteration {}".format(report_iteration))
        gen_blocks = eval_nb201_ws_acc(report_iteration, actor, env, evaluator, top_acc_set,
                                       n_desired_blocks_eval + i * n_desired_blocks_inc,
                                       output_node_label, book_keeper,
                                       valid_block_keys=valid_block_keys,
                                       update_callback=acc_update_callback)

        if debug:
            top_blocks = top_acc_set.get_curr_blocks()
            truth_blocks = top_acc_set.get_curr_blocks()
            book_keeper.log("")
            book_keeper.log("====== Start of debug info ======")
            truth_keys = set(b.get_str_id() for b in truth_blocks)
            train_block_keys = set(b.get_str_id() for b in blocks_from_train.tolist())
            common_keys = train_block_keys.intersection(truth_keys)
            book_keeper.log("{} unique blocks from train".format(len(train_block_keys)))
            book_keeper.log("{}% blocks from train are from truth".format(100. * float(len(common_keys)) / len(train_block_keys)))
            truth_keys = set(b.get_str_id() for b in truth_blocks)
            gen_keys = set(b.get_str_id() for b in gen_blocks)
            common_keys = gen_keys.intersection(truth_keys)
            book_keeper.log("{} unique gen blocks".format(len(gen_keys)))
            book_keeper.log("{}% gen blocks are from truth".format(100. * float(len(common_keys)) / len(gen_keys)))
            book_keeper.log("------")
            pair_dis_pred_check(truth_blocks, top_blocks, reward_module.get_model(),
                                log_f=book_keeper.log, desc="truth")
            pair_dis_pred_check(gen_blocks, top_blocks, reward_module.get_model(),
                                log_f=book_keeper.log, desc="fake")
            book_keeper.log("------")
            report_nb201_blocks_info(truth_blocks, log_f=book_keeper.log, desc="truth")
            report_nb201_blocks_info(gen_blocks, log_f=book_keeper.log, desc="fake")
            book_keeper.log("====== End of debug info ======")
            book_keeper.log("")
        book_keeper.log("Num blocks evaluated so far: {}".format(interval_info.num_block_eval))
    book_keeper.log("Search seconds: {}".format(time.time() - g_start))


def eval_nb201_ws_acc(iteration, actor, env, evaluator, acc_set,
                      n_desired_blocks, output_node_label, book_keeper,
                      update_callback=None, valid_block_keys=None):
    new_block2acc = {}
    new_blocks = _generate_nb201_blocks(actor, env, n_desired_blocks, output_node_label,
                                        desc="Eval", valid_block_keys=valid_block_keys)

    def _optim_maker(_p):
        return torch.optim.Adam(_p, lr=0.0001)
    evaluator.train_on_family(new_blocks, epochs=50, optim_maker=_optim_maker)

    for block in tqdm(new_blocks, desc="Getting performance", ascii=True):
        key = block.get_str_id()
        acc, lat = evaluator.get_perf_values(block)
        new_block2acc[key] = (block, acc)

    if len(new_block2acc) < n_desired_blocks:
        book_keeper.log("Only generated {} unique blocks".format(len(new_block2acc)))

    if len(new_block2acc) == 0:
        book_keeper.log("No blocks with acc generated")
    else:
        node_lens = [len(b) for _, (b, v) in new_block2acc.items()]
        edge_counts = [len(b.curr_edge_list[0]) for _, (b, v) in new_block2acc.items()]
        book_keeper.log("Avg number of nodes in generated blocks: {}".format(mean(node_lens)))
        book_keeper.log("Max number of nodes in generated blocks: {}".format(max(node_lens)))
        book_keeper.log("Min number of nodes in generated blocks: {}".format(min(node_lens)))
        book_keeper.log("Avg number of edges in generated blocks: {}".format(mean(edge_counts)))
        book_keeper.log("Max number of edges in generated blocks: {}".format(max(edge_counts)))
        book_keeper.log("Min number of edges in generated blocks: {}".format(min(edge_counts)))
        gen_acc_values = [acc for _, (b, acc) in new_block2acc.items()]
        book_keeper.log("Max acc in generated blocks: {}".format(max(gen_acc_values)))
        book_keeper.log("Min acc in generated blocks: {}".format(min(gen_acc_values)))
        acc_set.update(new_block2acc, log_f=book_keeper.log)
        if update_callback is not None:
            update_callback(acc_set, iteration, book_keeper.model_name)
        book_keeper.checkpoint_state_dict(acc_set.state_dict(),
                                          "{}_top_acc_set.pkl".format(book_keeper.model_name))
        acc_vals = acc_set.get_top_acc_values()
        book_keeper.log("Best acc so far: {}".format(max(acc_vals)))
        book_keeper.log("New top acc set size: {}, new_block2acc len: {}".format(len(acc_set),
                                                                                 len(new_block2acc)))
    return [b for _, (b, _) in new_block2acc.items()]

