import math
import pickle
from params import *
from constants import *
from utils.misc_utils import *
from utils.model_utils import model_save
from utils.nas_utils import gen_block_visuals, is_block_valid
from tqdm import tqdm


_work_dir = os.path.dirname(os.path.abspath(__file__))


class BookKeeper:

    def __init__(self, log_file_name, model_name,
                 saved_model_dir=P_SEP.join([_work_dir, "saved_models"]),
                 verbose=True, init_eval_perf=0., init_eval_iter=0,
                 eval_perf_comp_func=lambda old, new: new > old,  # how to judge which eval perf is better
                 saved_model_file=None, log_dir=P_SEP.join([_work_dir, "logs"]),
                 plot_dir=P_SEP.join([_work_dir, "logs", "plots"]),
                 block_visual_dir=P_SEP.join([_work_dir, "logs", "blocks"]),
                 vis_env=None):
        self.saved_model_dir = saved_model_dir
        self._log_dir = log_dir
        self._plot_dir = plot_dir
        self._block_visual_dir = block_visual_dir
        self.model_name = model_name
        self.log_file = P_SEP.join([self._log_dir, log_file_name])
        self.verbose = verbose
        self.curr_best_eval_perf = init_eval_perf
        self.curr_best_eval_iter = init_eval_iter
        self.eval_perf_comp_func = eval_perf_comp_func
        self._init_eval_perf = init_eval_perf
        self._init_eval_iter = init_eval_iter
        self.saved_model_file = saved_model_file
        self._vis_env = vis_env
        self._vis_plotter = None
        self._vis_index = None
        if self._vis_env is not None:
            try:
                import visdom
                self._vis_plotter = visdom.Visdom(env=self._vis_env)
                self._vis_index = {}
            except ModuleNotFoundError:
                pass
            except ConnectionRefusedError:
                pass
        self.log("Model name: {}".format(self.model_name))
        self.log("Saved model dir: {}".format(self.saved_model_dir))

    def vis_plot(self, name, y, x, **kwargs):
        if self._vis_plotter is None: return
        step = self._vis_index.get(name, 0)
        self._vis_plotter.line(Y=np.array([y]), X=np.array([x]),
                               win=str(name), opts=dict(title=name),
                               update=None if step == 0 else 'append',
                               **kwargs)
        self._vis_index[name] = step + 1

    def gen_block_visuals(self, blocks, group_dir_name,
                          prune_existing=True, block_perfs=None, x_axis_idx=1, name_prefix="block"):
        gen_block_visuals(blocks, group_dir_name, block_visual_dir=self._block_visual_dir,
                          model_name=self.model_name, block_perfs=block_perfs, x_axis_idx=x_axis_idx,
                          name_prefix=name_prefix, prune_existing=prune_existing)

    def save_acc_lat_pareto_values(self, acc_values, lat_values, tag):
        line = "{}|".format(tag)
        line += get_compact_str_for_seg_list(acc_values, value_delim=",")
        line += "|" + get_compact_str_for_seg_list(lat_values, value_delim=",")
        write_line_to_file(line, P_SEP.join([self._log_dir, "{}_pareto_front_values.txt".format(self.model_name)]))

    def save_block2perfs_values(self, block2perfs, i):
        lines = []
        for _, (b, (acc, lat)) in block2perfs.items():
            lines.append("{}|{}|{}".format(b, acc, lat))
        line = "iter {}".format(i) + "$" + ";".join(lines)
        write_line_to_file(line, P_SEP.join([self._log_dir, "{}_past_pareto_values.txt".format(self.model_name)]),
                           new_file=i==1)

    def save_acc_lat_pareto_plot(self, acc_values, lat_values, iteration):
        import matplotlib.pyplot as plt
        _v = list(zip(lat_values, acc_values))
        _v.sort(key=lambda t:t[0])
        lat_values = [t[0] for t in _v]
        acc_values = [t[1] for t in _v]
        plt.plot(lat_values, acc_values, 'bo-')
        title = "Acc vs. Lat Pareto front {} {}".format(iteration, self.model_name)
        plt.title(title)
        plt.ylabel("Acc")
        plt.xlabel("Lat")
        plt.savefig(P_SEP.join([self._plot_dir, title + ".png"]))
        plt.close()

    def reset_eval_perfs(self, eval_perf=None, eval_iter=None):
        self.curr_best_eval_perf = self._init_eval_perf if eval_perf is None else eval_perf
        self.curr_best_eval_iter = self._init_eval_iter if eval_iter is None else eval_iter

    def log(self, msg, verbose=None):
        if verbose is None:
            verbose = self.verbose
        if not isinstance(msg, str):
            msg = str(msg)
        write_line_to_file(msg, self.log_file, verbose=verbose)

    def report_curr_best(self):
        self.log("Model name: {}".format(self.model_name))
        self.log("curr_best_eval_perf: {}, curr_best_eval_iter: {}".format(self.curr_best_eval_perf,
                                                                           self.curr_best_eval_iter))

    def load_model_checkpoint_w_suffix(self, model, suffix, model_key=CHKPT_MODEL, skip_eval_perfs=False,
                                       allow_silent_fail=True):
        checkpoint_file = P_SEP.join([self.saved_model_dir, self.model_name + suffix])
        return self.load_model_checkpoint(model, model_key=model_key, checkpoint_file=checkpoint_file,
                                          skip_eval_perfs=skip_eval_perfs, allow_silent_fail=allow_silent_fail)

    def load_model_checkpoint(self, model, model_key=CHKPT_MODEL, checkpoint_file=None,
                              skip_eval_perfs=False, allow_silent_fail=True):
        from utils.model_utils import model_load
        if checkpoint_file is None:
            checkpoint_file = self.saved_model_file
        if os.path.isfile(checkpoint_file):
            self.log("Found checkpoint: {}, loading".format(checkpoint_file))
            sd = model_load(checkpoint_file)
            model.load_state_dict(sd[model_key])
            self.log("Found best_eval_perf: {}, best_eval_iter: {}".format(sd[CHKPT_BEST_EVAL_RESULT],
                                                                           sd[CHKPT_BEST_EVAL_ITERATION]))
            if not skip_eval_perfs:
                self.curr_best_eval_perf = sd[CHKPT_BEST_EVAL_RESULT]
                self.curr_best_eval_iter = sd[CHKPT_BEST_EVAL_ITERATION]
                self.log("Loaded curr_best_eval_perf: {}, curr_best_eval_iter: {}".format(self.curr_best_eval_perf,
                                                                                          self.curr_best_eval_iter))
            if CHKPT_ITERATION in sd:
                completed_iterations = sd[CHKPT_ITERATION]
                self.log("Completed iterations: {}".format(completed_iterations))
                return completed_iterations
            elif CHKPT_COMPLETED_EPOCHS in sd:
                completed_epochs = sd[CHKPT_COMPLETED_EPOCHS]
                self.log("Completed epochs: {}".format(completed_epochs))
                return completed_epochs
        elif not allow_silent_fail:
            raise FileNotFoundError("checkpoint_file: {} not found".format(checkpoint_file))
        return 0

    def load_model_optim_checkpoint(self, model, optimizer, model_key=CHKPT_MODEL, optimizer_key=CHKPT_OPTIMIZER,
                                    checkpoint_file=None, skip_eval_perfs=False, allow_silent_fail=True):
        from utils.model_utils import model_load
        if checkpoint_file is None:
            checkpoint_file = self.saved_model_file
        if os.path.isfile(checkpoint_file):
            self.log("Found checkpoint: {}, loading".format(checkpoint_file))
            sd = model_load(checkpoint_file)
            model.load_state_dict(sd[model_key])
            optimizer.load_state_dict(sd[optimizer_key])
            self.log("Found best_eval_perf: {}, best_eval_iter: {}".format(sd[CHKPT_BEST_EVAL_RESULT],
                                                                           sd[CHKPT_BEST_EVAL_ITERATION]))
            if not skip_eval_perfs:
                self.curr_best_eval_perf = sd[CHKPT_BEST_EVAL_RESULT]
                self.curr_best_eval_iter = sd[CHKPT_BEST_EVAL_ITERATION]
                self.log("Loaded curr_best_eval_perf: {}, curr_best_eval_iter: {}".format(self.curr_best_eval_perf,
                                                                                          self.curr_best_eval_iter))
            if CHKPT_ITERATION in sd:
                completed_iterations = sd[CHKPT_ITERATION]
                self.log("Completed iterations: {}".format(completed_iterations))
                return completed_iterations
            elif CHKPT_COMPLETED_EPOCHS in sd:
                completed_epochs = sd[CHKPT_COMPLETED_EPOCHS]
                self.log("Completed epochs: {}".format(completed_epochs))
                return completed_epochs
        elif not allow_silent_fail:
            raise FileNotFoundError("checkpoint_file: {} not found".format(checkpoint_file))
        return 0

    def load_state_dict_checkpoint(self, obj, checkpoint_file, allow_silent_fail=False):
        assert hasattr(obj, "load_state_dict")
        import pickle
        checkpoint_file = P_SEP.join([self.saved_model_dir, checkpoint_file])
        if os.path.isfile(checkpoint_file):
            self.log("Found state dict checkpoint: {}, loading".format(checkpoint_file))
            with open(checkpoint_file, "rb") as f:
                sd = pickle.load(f)
            obj.load_state_dict(sd)
        elif not allow_silent_fail:
            raise FileNotFoundError("checkpoint_file: {} not found".format(checkpoint_file))
        return None

    def load_object_checkpoint(self, checkpoint_file, allow_silent_fail=False):
        import pickle
        checkpoint_file = P_SEP.join([self.saved_model_dir, checkpoint_file])
        if os.path.isfile(checkpoint_file):
            self.log("Found object checkpoint: {}, loading".format(checkpoint_file))
            with open(checkpoint_file, "rb") as f:
                obj = pickle.load(f)
            return obj
        elif not allow_silent_fail:
            raise FileNotFoundError("checkpoint_file: {} not found".format(checkpoint_file))
        return None

    def checkpoint_model(self, suffix, iteration, model, optimizer, eval_perf=None,
                         model_key=CHKPT_MODEL, optim_key=CHKPT_OPTIMIZER):
        if eval_perf is not None:
            if self.eval_perf_comp_func(self.curr_best_eval_perf, eval_perf):
                self.curr_best_eval_perf = eval_perf
                self.curr_best_eval_iter = iteration
                self._checkpoint_model(suffix, iteration, model, optimizer,
                                       model_key, optim_key)
        else:
            self._checkpoint_model(suffix, iteration, model, optimizer,
                                   model_key, optim_key)

    def _checkpoint_model(self, suffix, iteration, model, optimizer,
                          model_key=CHKPT_MODEL, optim_key=CHKPT_OPTIMIZER):
        file_path = P_SEP.join([self.saved_model_dir, self.model_name + suffix])
        sv = {
            CHKPT_ITERATION: iteration,
            model_key: model.state_dict(),
            optim_key: optimizer.state_dict() if optimizer is not None else None,
            CHKPT_BEST_EVAL_RESULT: self.curr_best_eval_perf,
            CHKPT_BEST_EVAL_ITERATION: self.curr_best_eval_iter,
        }
        self.log("Saving model to {}, please do not terminate".format(file_path))
        model_save(file_path, sv)
        self.log("Checkpoint complete")

    def checkpoint_object(self, obj, checkpoint_file):
        import pickle
        checkpoint_file = P_SEP.join([self.saved_model_dir, checkpoint_file])
        self.log("Saving object to {}, please do not terminate".format(checkpoint_file))
        with open(checkpoint_file, "wb") as f:
            pickle.dump(obj, f, protocol=4)
        self.log("Checkpoint complete")

    def checkpoint_state_dict(self, sd, checkpoint_file):
        import pickle
        checkpoint_file = P_SEP.join([self.saved_model_dir, checkpoint_file])
        self.log("Saving state dict to {}, please do not terminate".format(checkpoint_file))
        with open(checkpoint_file, "wb") as f:
            pickle.dump(sd, f, protocol=4)
        self.log("Checkpoint complete")


class CosAnnealOptimizer:

    def __init__(self, optimizer, num_epochs, eta_min=0.0001, last_epoch=-1):
        self._optimizer = optimizer
        self._num_epochs = num_epochs
        self._eta_min = eta_min
        self._last_epoch = last_epoch
        self._scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(self._optimizer, self._num_epochs,
                                                                     eta_min=self._eta_min,
                                                                     last_epoch=self._last_epoch)

    def zero_grad(self):
        self._optimizer.zero_grad()

    def step(self):
        self._optimizer.step()

    def schedule_lr(self, epoch, log_f=None, plot_f=None):
        self._scheduler.step(epoch)
        self._last_epoch += 1
        if log_f is not None:
            log_f("CosAnnealOptimizer new lr: {}".format(self.get_curr_lr()))
        if plot_f is not None:
            plot_f("CosAnnealOptimizer lr", self.get_curr_lr(), epoch)

    def get_curr_lr(self):
        return self._scheduler.get_lr()[0]

    def state_dict(self):
        return {
            "_optimizer": self._optimizer.state_dict(),
            "_scheduler_num_epochs": self._num_epochs,
            "_scheduler_eta_min": self._eta_min,
            "_scheduler_last_epoch": self._last_epoch
        }

    def load_state_dict(self, state_dict):
        self._optimizer.load_state_dict(state_dict["_optimizer"])
        self._scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(self._optimizer,
                                                                     state_dict["_scheduler_num_epochs"],
                                                                     eta_min=state_dict["_scheduler_eta_min"],
                                                                     last_epoch=state_dict["_scheduler_last_epoch"])


class LRDecayOptimizer:

    def __init__(self, optimizer, init_lr=0.001, min_lr=1e-3, min_score_diff=0., decay_ratio=0.9,
                 decay_constant=1e-3, n_allowed_fails=0, verbose=True):
        assert min_lr > 0.
        self.verbose = verbose
        self._optimizer = optimizer
        self._curr_lr = init_lr
        self._min_lr = min_lr
        self._min_score_diff = min_score_diff
        self._best_score = None
        self._decay_ratio = decay_ratio
        self._decay_constant = decay_constant
        self._n_allowed_fails = n_allowed_fails
        self._curr_fail_count = 0
        self._set_lr(init_lr)
        if self.verbose and self._decay_ratio is not None and self._decay_constant is not None:
            print("LRDecayOptimizer: decay_constant will be ignored when decay_ratio is set")

    def _set_lr(self, new_lr, log_f=None):
        if new_lr < self._min_lr:
            if log_f is not None:
                log_f("LRDecayOptimizer new proposed lr: {} < min lr: {}".format(new_lr, self._min_lr))
            new_lr = self._min_lr
        if log_f is not None and self._curr_lr != new_lr:
            log_f("LRDecayOptimizer new lr:{}".format(new_lr))
        self._curr_lr = new_lr
        for g in self._optimizer.param_groups:
            g['lr'] = self._curr_lr

    def zero_grad(self):
        self._optimizer.zero_grad()

    def step(self):
        self._optimizer.step()

    def update_lr(self, new_score, comp_fail_f=lambda new, old: new < old, log_f=None):
        if self._best_score is None:
            self._best_score = new_score
            return
        if comp_fail_f(new_score, self._best_score) or \
                abs(new_score - self._best_score) < self._min_score_diff:
            if self._decay_ratio is not None:
                new_lr = self._curr_lr * self._decay_ratio
            else:
                new_lr = self._curr_lr - self._decay_constant
            if self._curr_fail_count >= self._n_allowed_fails:
                self._set_lr(new_lr, log_f)
                self._curr_fail_count = 0
                self._best_score = None
                self._n_allowed_fails += 1
            else:
                self._curr_fail_count += 1
                if log_f is not None and self._curr_lr != new_lr:
                    log_f("LRDecayOptimizer new fail count: {}".format(self._curr_fail_count))
                    log_f("LRDecayOptimizer allowed fail count: {}".format(self._n_allowed_fails))
        else:
            self._best_score = new_score
            self._curr_fail_count = 0

    def state_dict(self):
        return {
            "_optimizer": self._optimizer.state_dict(),
            "verbose": self.verbose,
            "_curr_lr": self._curr_lr,
            "_min_lr": self._min_lr,
            "_min_score_diff": self._min_score_diff,
            "_best_score": self._best_score,
            "_decay_ratio": self._decay_ratio,
            "_decay_constant": self._decay_constant,
            "_n_allowed_fails": self._n_allowed_fails,
            "_curr_fail_count": self._curr_fail_count,
        }

    def load_state_dict(self, state_dict):
        self._optimizer.load_state_dict(state_dict["_optimizer"])
        self.verbose = state_dict["verbose"]
        self._curr_lr = state_dict["_curr_lr"]
        self._min_lr = state_dict["_min_lr"]
        self._min_score_diff = state_dict["_min_score_diff"]
        self._best_score = state_dict["_best_score"]
        self._decay_ratio = state_dict["_decay_ratio"]
        self._decay_constant = state_dict["_decay_constant"]
        self._n_allowed_fails = state_dict["_n_allowed_fails"]
        self._curr_fail_count = state_dict["_curr_fail_count"]


class LinearAnnealScheduler:

    def __init__(self, init_lr, end_lr, n_updates, verbose=True):
        assert init_lr >= end_lr
        self.init_lr = init_lr
        self.end_lr = end_lr
        self.n_updates = n_updates
        self.curr_lr = self.init_lr
        self.decrement = float(self.init_lr - self.end_lr) / n_updates
        self.verbose = verbose
        if self.verbose:
            print("LinearAnnealScheduler.init_lr={}".format(self.init_lr))
            print("LinearAnnealScheduler.end_lr={}".format(self.end_lr))
            print("LinearAnnealScheduler.decrement={}".format(self.decrement))

    def get_lr(self, verbose=None):
        if verbose is None:
            verbose = self.verbose
        lr = self.curr_lr
        if verbose:
            print("LinearAnnealScheduler.curr_lr={}".format(lr))
        self.curr_lr -= self.decrement
        self.curr_lr = max(self.curr_lr, self.end_lr)
        return lr


class RatioDecayScheduler:

    def __init__(self, init_lr, min_lr, ratio, decay_threshold, verbose=True):
        assert init_lr >= min_lr
        self.init_lr = init_lr
        self.min_lr = min_lr
        self.ratio = ratio
        self.curr_lr = self.init_lr
        self.verbose = verbose
        self.decay_threshold = decay_threshold
        self.n_steps = 0
        if self.verbose:
            print("RatioDecayScheduler.init_lr={}".format(self.init_lr))
            print("RatioDecayScheduler.min_lr={}".format(self.min_lr))
            print("RatioDecayScheduler.ratio={}".format(self.ratio))
            print("RatioDecayScheduler.decay_threshold={}".format(self.decay_threshold))

    def _decay(self):
        self.curr_lr *= self.ratio
        self.curr_lr = max(self.curr_lr, self.min_lr)
        self.n_steps = 0

    def get_lr(self):
        lr = self.curr_lr
        self.n_steps += 1
        if self.n_steps == self.decay_threshold:
            self._decay()
        return lr


class CustomLROptimizer:

    def __init__(self, optimizer, init_lr, lr_and_thresholds, verbose=False):
        # Expected format for lr_and_thresholds: [(new_lr, update threshold), ...]
        assert len(lr_and_thresholds) > 0
        self.verbose = verbose
        self.lr_and_thresholds = sorted(lr_and_thresholds, key=lambda t:t[1])
        self._optimizer = optimizer
        self._curr_lr = init_lr
        self._n_updates = 0
        self._curr_threshold_idx = 0
        self._set_lr(init_lr)
        if self.verbose:
            print("CustomLROptimizer.init_lr={}".format(self._curr_lr))
            print("CustomLROptimizer.lr_and_thresholds={}".format(self.lr_and_thresholds))

    def _set_lr(self, new_lr):
        self._curr_lr = new_lr
        for g in self._optimizer.param_groups:
            g['lr'] = self._curr_lr

    def update(self):
        self._n_updates += 1
        if self._curr_threshold_idx < len(self.lr_and_thresholds) \
            and self._n_updates >= self.lr_and_thresholds[self._curr_threshold_idx][1]:
            new_lr = self.lr_and_thresholds[self._curr_threshold_idx][0]
            if self.verbose:
                print("CustomLROptimizer._curr_lr={}".format(new_lr))
            self._set_lr(new_lr)
            self._curr_threshold_idx += 1

    def zero_grad(self):
        self._optimizer.zero_grad()

    def step(self):
        self._optimizer.step()

    def state_dict(self):
        return {
            "_optimizer": self._optimizer.state_dict(),
            "verbose": self.verbose,
            "lr_and_thresholds": self.lr_and_thresholds,
            "_curr_lr": self._curr_lr,
            "_n_updates": self._n_updates,
            "_curr_threshold_idx": self._curr_threshold_idx,
        }

    def load_state_dict(self, state_dict):
        self._optimizer.load_state_dict(state_dict["_optimizer"])
        self.verbose = state_dict["verbose"]
        self._curr_lr = state_dict["_curr_lr"]
        self.lr_and_thresholds = state_dict["lr_and_thresholds"]
        self._n_updates = state_dict["_n_updates"]
        self._curr_threshold_idx = state_dict["_curr_threshold_idx"]


class RandomLROptimizer:

    def __init__(self, optimizer, lr_candidates, init_lr=0.001, min_score_diff=0., n_allowed_fails=0):
        assert len(lr_candidates) > 0
        assert all(v > 0 for v in lr_candidates)
        self.lr_candidates = lr_candidates
        self._optimizer = optimizer
        self._curr_lr = init_lr
        self._min_score_diff = min_score_diff
        self._best_score = None
        self._n_allowed_fails = n_allowed_fails
        self._curr_fail_count = 0
        self._set_lr(init_lr)

    def sample_new_lr(self, log_f=None):
        new_lr_list = [lr for lr in self.lr_candidates if abs(self._curr_lr - lr) > 1e-5]
        if len(new_lr_list) == 0: return self._curr_lr
        new_lr = random.choice(new_lr_list)
        if log_f is not None:
            log_f("RandomLROptimizer new lr: {}".format(new_lr))
        return new_lr

    def _set_lr(self, new_lr):
        self._curr_lr = new_lr
        for g in self._optimizer.param_groups:
            g['lr'] = self._curr_lr

    def zero_grad(self):
        self._optimizer.zero_grad()

    def step(self):
        self._optimizer.step()

    def update_lr(self, new_score, comp_fail_f=lambda new, old: new < old, log_f=None):
        if self._best_score is None:
            self._best_score = new_score
            return
        if comp_fail_f(new_score, self._best_score) or \
                abs(new_score - self._best_score) < self._min_score_diff:
            if self._curr_fail_count >= self._n_allowed_fails:
                self._set_lr(self.sample_new_lr(log_f=log_f))
                self._curr_fail_count = 0
                self._best_score = None
                self._n_allowed_fails += 1
            else:
                self._curr_fail_count += 1
                if log_f is not None:
                    log_f("RandomLROptimizer new fail count: {}".format(self._curr_fail_count))
                    log_f("RandomLROptimizer allowed fail count: {}".format(self._n_allowed_fails))
        else:
            self._best_score = new_score
            self._curr_fail_count = 0

    def state_dict(self):
        return {
            "_optimizer": self._optimizer.state_dict(),
            "_curr_lr": self._curr_lr,
            "lr_candidates": self.lr_candidates,
            "_min_score_diff": self._min_score_diff,
            "_best_score": self._best_score,
            "_n_allowed_fails": self._n_allowed_fails,
            "_curr_fail_count": self._curr_fail_count,
        }

    def load_state_dict(self, state_dict):
        self._optimizer.load_state_dict(state_dict["_optimizer"])
        self.lr_candidates = state_dict["lr_candidates"]
        self._curr_lr = state_dict["_curr_lr"]
        self._min_score_diff = state_dict["_min_score_diff"]
        self._best_score = state_dict["_best_score"]
        self._n_allowed_fails = state_dict["_n_allowed_fails"]
        self._curr_fail_count = state_dict["_curr_fail_count"]


class ExponentialDecayOptimizer:

    def __init__(self, optimizer, init_lr, decay_steps, decay_ratio,
                 staircase=True, log_f=print):
        self._staircase = staircase
        self._decay_steps = decay_steps
        self._decay_ratio = decay_ratio
        self._optimizer = optimizer
        self._init_lr = init_lr
        self._n_updates = 0
        self._set_lr(init_lr)
        log_f("ExponentialDecayOptimizer.init_lr={}".format(self._init_lr))
        log_f("ExponentialDecayOptimizer.staircase={}".format(self._staircase))
        log_f("ExponentialDecayOptimizer.decay_steps={}".format(self._decay_steps))
        log_f("ExponentialDecayOptimizer.decay_ratio={}".format(self._decay_ratio))

    def get_n_updates(self):
        return self._n_updates

    def get_lr(self):
        lr = None
        for g in self._optimizer.param_groups:
            lr = g['lr'] if lr is None else min(lr, g['lr'])
        return lr

    def _set_lr(self, new_lr):
        for g in self._optimizer.param_groups:
            g['lr'] = new_lr

    def update(self):
        self._n_updates += 1
        p = float(self._n_updates) / self._decay_steps
        if self._staircase:
            p = math.floor(p)
        new_lr = self._init_lr * math.pow(self._decay_ratio, p)
        self._set_lr(new_lr)

    def zero_grad(self):
        self._optimizer.zero_grad()

    def step(self):
        self._optimizer.step()

    def state_dict(self):
        return {
            "_optimizer": self._optimizer.state_dict(),
            "_decay_steps": self._decay_steps,
            "_staircase": self._staircase,
            "_init_lr": self._init_lr,
            "_n_updates": self._n_updates,
            "_decay_ratio": self._decay_ratio,
        }

    def load_state_dict(self, state_dict):
        self._optimizer.load_state_dict(state_dict["_optimizer"])
        self._init_lr = state_dict["_init_lr"]
        self._decay_steps= state_dict["_decay_steps"]
        self._staircase= state_dict["_staircase"]
        self._n_updates = state_dict["_n_updates"]
        self._decay_ratio = state_dict["_decay_ratio"]


def linear_combination(x, y, epsilon):
    return epsilon*x + (1-epsilon)*y


def reduce_loss(loss, reduction='mean'):
    return loss.mean() if reduction == 'mean' else loss.sum() if reduction == 'sum' else loss


class LabelSmoothingCrossEntropy(torch.nn.Module):
    def __init__(self, epsilon=0.1, reduction='mean'):
        super().__init__()
        self.epsilon = epsilon
        self.reduction = reduction

    def forward(self, preds, target):
        n = preds.size()[-1]
        log_preds = torch.nn.functional.log_softmax(preds, dim=-1)
        loss = reduce_loss(-log_preds.sum(dim=-1), self.reduction)
        nll = torch.nn.functional.nll_loss(log_preds, target, reduction=self.reduction)
        return linear_combination(loss / n, nll, self.epsilon)


def report_blocks_info(blocks, log_f=print, desc=""):
    from utils.darts_utils import AverageMeter
    log_f("Reporting info on {} {} blocks".format(len(blocks), desc))
    avg_n_nodes = AverageMeter()
    avg_n_edges = AverageMeter()
    max_n_nodes, min_n_nodes = 0, float("inf")
    max_n_edges, min_n_edges = 0, float("inf")
    node_size_counts = collections.defaultdict(int)
    edge_size_counts = collections.defaultdict(int)
    op_ratio = collections.defaultdict(float)
    n_ops = 0
    for block in blocks:
        avg_n_nodes.update(len(block))
        max_n_nodes = max(max_n_nodes, len(block))
        min_n_nodes = min(min_n_nodes, len(block))
        node_size_counts[len(block)] += 1
        avg_n_edges.update(len(block.curr_edge_list[0]))
        max_n_edges = max(max_n_edges, len(block.curr_edge_list[0]))
        min_n_edges = min(min_n_edges, len(block.curr_edge_list[0]))
        edge_size_counts[len(block.curr_edge_list[0])] += 1
        for op in block.curr_node_label_list[1:-1]:
            op_ratio[op] += 1
            n_ops += 1
    op_ratio = [(k, v / n_ops) for k, v in op_ratio.items()]
    op_ratio.sort(key=lambda t: t[0])
    log_f("Avg number of nodes: {:.2f}".format(avg_n_nodes.avg))
    log_f("Max number of nodes: {:.2f}".format(max_n_nodes))
    log_f("Min number of nodes: {:.2f}".format(min_n_nodes))
    log_f("Node size counts: {}".format(node_size_counts))
    log_f("Avg number of edges: {:.2f}".format(avg_n_edges.avg))
    log_f("Max number of edges: {:.2f}".format(max_n_edges))
    log_f("Min number of edges: {:.2f}".format(min_n_edges))
    log_f("Edge size counts: {}".format(edge_size_counts))
    log_f("Number of searchable ops: {}".format(n_ops))
    for k, v in op_ratio:
        log_f("- {}: {:.2f}%".format(k, v * 100.))


def report_blocks_info(blocks, log_f=print, desc=""):
    from utils.darts_utils import AverageMeter
    log_f("Reporting info on {} {} blocks".format(len(blocks), desc))
    avg_n_nodes = AverageMeter()
    avg_n_edges = AverageMeter()
    max_n_nodes, min_n_nodes = 0, float("inf")
    max_n_edges, min_n_edges = 0, float("inf")
    node_size_counts = collections.defaultdict(int)
    edge_size_counts = collections.defaultdict(int)
    op_ratio = collections.defaultdict(float)
    n_ops = 0
    for block in blocks:
        avg_n_nodes.update(len(block))
        max_n_nodes = max(max_n_nodes, len(block))
        min_n_nodes = min(min_n_nodes, len(block))
        node_size_counts[len(block)] += 1
        avg_n_edges.update(len(block.curr_edge_list[0]))
        max_n_edges = max(max_n_edges, len(block.curr_edge_list[0]))
        min_n_edges = min(min_n_edges, len(block.curr_edge_list[0]))
        edge_size_counts[len(block.curr_edge_list[0])] += 1
        for op in block.curr_node_label_list[1:-1]:
            op_ratio[op] += 1
            n_ops += 1
    op_ratio = [(k, v / n_ops) for k, v in op_ratio.items()]
    op_ratio.sort(key=lambda t: t[0])
    log_f("Avg number of nodes: {:.2f}".format(avg_n_nodes.avg))
    log_f("Max number of nodes: {:.2f}".format(max_n_nodes))
    log_f("Min number of nodes: {:.2f}".format(min_n_nodes))
    log_f("Node size counts: {}".format(node_size_counts))
    log_f("Avg number of edges: {:.2f}".format(avg_n_edges.avg))
    log_f("Max number of edges: {:.2f}".format(max_n_edges))
    log_f("Min number of edges: {:.2f}".format(min_n_edges))
    log_f("Edge size counts: {}".format(edge_size_counts))
    log_f("Number of searchable ops: {}".format(n_ops))
    for k, v in op_ratio:
        log_f("- {}: {:.2f}%".format(k, v * 100.))


def step_wise_check(truth_block, actor, env, log_f=print):
    with torch.no_grad():
        actor.eval()
        log_f("Step-wise check for truth block: {}".format(truth_block))
        states = truth_block.get_evolving_states(include_self=True)
        for si, state in enumerate(states[:-1]):
            log_f("Input state {}: {}".format(si, str(state)))
            action = actor.get_action_for_state(state)
            log_f("Action for this state: {}".format(str(action)))
            pred_state, _ = env.gen_new_state_from_action(state, action)
            log_f("Generated next state: {}".format(pred_state))
            log_f("Expected next state: {}".format(states[si + 1]))
            log_f("")


def linear_schedule(init_val, end_val, curr_iter, total_iter):
    delta = float(end_val - init_val) / total_iter
    return init_val + delta * curr_iter


def blocks_diff(blocks1, blocks2):
    from utils.nas_utils import is_isomorphic
    extra_blocks = []
    for b1 in blocks1:
        found_in_other = False
        label1 = get_tsr_value(b1.curr_node_feature_tsr)
        mat1 = edge_list_to_adj_mat(b1.curr_edge_list[0], b1.curr_edge_list[1])
        for b2 in blocks2:
            label2 = get_tsr_value(b2.curr_node_feature_tsr)
            mat2 = edge_list_to_adj_mat(b2.curr_edge_list[0], b2.curr_edge_list[1])
            if is_isomorphic([mat1, label1], [mat2, label2]):
                found_in_other = True
                break
        if not found_in_other:
            extra_blocks.append(b1)
    return extra_blocks


def get_common_blocks(blocks1, blocks2):
    from utils.nas_utils import is_isomorphic
    common_blocks = []
    for b1 in blocks1:
        found_in_other = False
        label1 = get_tsr_value(b1.curr_node_feature_tsr)
        mat1 = edge_list_to_adj_mat(b1.curr_edge_list[0], b1.curr_edge_list[1])
        b1_key = b1.get_str_id()
        for b2 in blocks2:
            label2 = get_tsr_value(b2.curr_node_feature_tsr)
            mat2 = edge_list_to_adj_mat(b2.curr_edge_list[0], b2.curr_edge_list[1])
            b2_key = b2.get_str_id()
            if b1_key == b2_key or is_isomorphic([mat1, label1], [mat2, label2]):
                found_in_other = True
                break
        if found_in_other:
            common_blocks.append(b1)
    return common_blocks


def random_mutate_blocks(depth, blocks, num_mutations, node_label2feature, max_num_nodes=7,
                         available_node_labels=NB101_SEARCHABLE_OPS,
                         output_node_label="output", valid_block_keys=None, verbose=True,
                         valid_mutations=("swap node", "add node",
                                          "remove node","add edge", "remove edge"),
                         add_node_depth_limit=None):
    valid_node_set = set(available_node_labels)
    src_blocks = [b for b in blocks]
    new_blocks = []
    generated_block_keys = {b.get_str_id() for b in blocks}
    sampled_mutation_counts = collections.defaultdict(int)
    success_mutation_counts = collections.defaultdict(float)
    bar = None
    if verbose: bar = tqdm(total=len(src_blocks), desc="Mutating blocks", ascii=True)
    for src_block in src_blocks:
        curr_mutants = [src_block]
        curr_mutant_keys = set()
        for d in range(depth):
            new_mutants = []
            for block in curr_mutants:
                for mi in range(num_mutations):
                    mutations = set(valid_mutations)
                    # Exclude impossible mutations beforehand
                    if len(set(block.curr_node_label_list).intersection(valid_node_set)) == 0:
                        # if only contains input and output, cannot swap a node, remove a node/edge, add edge
                        mutations = {"add node"}
                    elif len(block.curr_node_label_list) >= max_num_nodes:
                        # if already contains the max number of nodes, cannot add a new node
                        mutations.remove("add node")
                    elif add_node_depth_limit is not None and d > add_node_depth_limit:
                        mutations.remove("add node")

                    if is_complete_dag(len(block.curr_node_label_list), block.curr_edge_list):
                        # cannot add an edge if there is no place to add
                        if "add edge" in mutations: mutations.remove("add edge")

                    mutation = random.choice(list(mutations))
                    sampled_mutation_counts[mutation] += 1
                    new_block = apply_mutation_to_block(block, mutation, node_label2feature, available_node_labels,
                                                        valid_node_set)
                    if not is_block_valid(new_block, output_node_label=output_node_label):
                        continue # ignore invalid mutants
                    success_mutation_counts[mutation] += 1.
                    new_block_id = new_block.get_str_id()
                    if hasattr(valid_block_keys, "get_id"):
                        key = valid_block_keys.get_id(new_block)
                    else:
                        key = new_block_id
                    if valid_block_keys is not None and key not in valid_block_keys:
                        continue # If we have a finite set of valid blocks
                    if new_block_id not in curr_mutant_keys:
                        curr_mutant_keys.add(new_block_id)
                        new_mutants.append(new_block)
                    if new_block_id not in generated_block_keys:
                        generated_block_keys.add(new_block_id)
                        new_blocks.append(new_block)
            curr_mutants.extend(new_mutants)
        if bar is not None: bar.update(1)
    if bar is not None: bar.close()
    success_rates = {k: v / max(sampled_mutation_counts[k], 1) for k, v in success_mutation_counts.items()}
    return new_blocks, success_rates, sampled_mutation_counts


def apply_mutation_to_block(block, mutation, node_label2feature, available_node_labels, valid_node_set):
    if mutation == "swap node":
        candidates = [(ni, label) for ni, label in enumerate(block.curr_node_label_list) if label in valid_node_set]
        assert len(candidates) > 0
        ni, old_node_label = random.choice(candidates)
        new_node_labels = [n for n in list(available_node_labels) if n != old_node_label]
        assert len(new_node_labels) > 0
        new_node_label = random.choice(new_node_labels)
        new_block = copy.deepcopy(block)
        new_feature_tsr = node_label2feature[new_node_label]
        new_block.curr_node_label_list[ni] = new_node_label
        new_block.curr_node_feature_tsr[ni] = new_feature_tsr
    elif mutation == "add node":
        insert_idx = random.randint(0, len(block.curr_node_label_list) - 2) + 1
        n_ancestors = max(insert_idx, 1)
        n_descendants = max(len(block.curr_node_label_list) - insert_idx, 1)
        new_node_label = random.choice(available_node_labels)
        new_feature_tsr = node_label2feature[new_node_label]
        n_incoming_edges = random.randint(1, n_ancestors)
        n_outgoing_edges = random.randint(1, n_descendants)
        parent_nodes = random.sample(range(n_ancestors), k=n_incoming_edges)
        child_nodes = random.sample(range(n_descendants), k=n_outgoing_edges)
        child_nodes = [ci + insert_idx for ci in child_nodes]
        new_block = copy.deepcopy(block)
        new_block.curr_node_feature_tsr = torch.cat([
            new_block.curr_node_feature_tsr[:insert_idx], new_feature_tsr, new_block.curr_node_feature_tsr[insert_idx:]
        ])
        new_block.curr_node_label_list = new_block.curr_node_label_list[:insert_idx] + [new_node_label] + \
                                         new_block.curr_node_label_list[insert_idx:]
        new_src_list, new_dst_list = [], []
        for src, dst in zip(block.curr_edge_list[0], block.curr_edge_list[1]):
            if src < insert_idx:
                new_src_list.append(src)
            else:
                new_src_list.append(src + 1)
            if dst < insert_idx:
                new_dst_list.append(dst)
            else:
                new_dst_list.append(dst + 1)
        for pi in parent_nodes:
            new_src_list.append(pi)
            new_dst_list.append(insert_idx)
        for ci in child_nodes:
            new_src_list.append(insert_idx)
            new_dst_list.append(ci + 1)
        new_block.curr_edge_list = [new_src_list, new_dst_list]
    elif mutation == "remove node":
        candidates = [(ni, label) for ni, label in enumerate(block.curr_node_label_list) if label in valid_node_set]
        assert len(candidates) > 0
        ni, _ = random.choice(candidates)
        new_block = copy.deepcopy(block)
        del new_block.curr_node_label_list[ni]
        new_block.curr_node_feature_tsr = torch.cat([
            new_block.curr_node_feature_tsr[:ni], new_block.curr_node_feature_tsr[ni+1:]
        ])
        new_src_list, new_dst_list = [], []
        for src, dst in zip(block.curr_edge_list[0], block.curr_edge_list[1]):
            if src == ni or dst == ni: continue
            if src < ni:
                new_src_list.append(src)
            else:
                new_src_list.append(src - 1)
            if dst < ni:
                new_dst_list.append(dst)
            else:
                new_dst_list.append(dst - 1)
        new_block.curr_edge_list = [new_src_list, new_dst_list]
    elif mutation == "add edge":
        curr_edge_list = prune_self_edge_from_edge_list(block.curr_edge_list)
        existing_edge_pairs = set(zip(curr_edge_list[0], curr_edge_list[1]))
        complete_edge_pairs = get_complete_dag_edges(len(block.curr_node_label_list))
        candidates = list(complete_edge_pairs - existing_edge_pairs)
        assert len(candidates) > 0
        src, dst = random.choice(candidates)
        assert (src, dst) not in existing_edge_pairs
        new_block = copy.deepcopy(block)
        new_block.curr_edge_list[0].append(src)
        new_block.curr_edge_list[1].append(dst)
    elif mutation == "remove edge":
        candidates = []
        for src, dst in zip(block.curr_edge_list[0], block.curr_edge_list[1]):
            # # Maybe we do not need to exclude edges that are connected to input/output nodes?
            # if block.curr_node_label_list[src] not in valid_node_set: continue
            # if block.curr_node_label_list[dst] not in valid_node_set: continue
            candidates.append( (src, dst) )
        assert len(candidates) > 0
        tgt_src, tgt_dst = random.choice(candidates)
        new_src_list, new_dst_list = [], []
        for src, dst in zip(block.curr_edge_list[0], block.curr_edge_list[1]):
            if src == tgt_src and dst == tgt_dst: continue
            new_src_list.append(src)
            new_dst_list.append(dst)
        new_block = copy.deepcopy(block)
        new_block.curr_edge_list = [new_src_list, new_dst_list]
    else:
        raise ValueError("Unknown mutation: {}".format(mutation))
    return new_block


def load_complete_nb201(data_file=P_SEP.join([DATA_DIR, "NB201-v1_1-096897.pth"]), log_f=print,
                        d_set="cifar10", eval_method="ori-test", show_example=False,
                        simple_data_dict_cache=P_SEP.join([DATA_DIR, "nb201_acc_flops_dict.pkl"])):
    from nas_201_api import NASBench201API as API
    api = API(data_file)
    log_f("{} archs loaded from NasBench201".format(len(api)))
    keys = set()
    max_test_acc, min_test_acc = 0., 1.
    max_flops, min_flops = 0., float("inf")
    avg_test_acc = 0.
    avg_flops = 0
    simple_data_dict = {}
    best_block_key = None
    for i in tqdm(range(len(api)), desc="Processing", ascii=True):
        info = api.query_meta_info_by_index(i)
        keys.add(info.arch_str)
        res_metrics = info.get_metrics(d_set, eval_method)
        cost_metrics = info.get_compute_costs(d_set)  # flops, params, latency
        test_acc = res_metrics["accuracy"] / 100.
        flops = cost_metrics["flops"]
        simple_data_dict[info.arch_str] = [test_acc, flops]
        if test_acc > max_test_acc:
            best_block_key = info.arch_str
        max_test_acc = max(max_test_acc, test_acc)
        min_test_acc = min(min_test_acc, test_acc)
        max_flops = max(flops, max_flops)
        min_flops = min(flops, min_flops)
        avg_test_acc += test_acc
        avg_flops += flops
    avg_test_acc /= len(api)
    avg_flops /= len(api)
    log_f("Best block key: {}".format(best_block_key))
    log_f("Max test acc: {}, min test acc: {}, avg test acc: {}".format(max_test_acc, min_test_acc, avg_test_acc))
    log_f("Max flops: {}, min flops: {}, avg flops: {}".format(max_flops, min_flops, avg_flops))
    log_f("{} unique keys recorded".format(len(keys)))
    log_f("Normalizing flops to between 0 and 1")
    for k, v in simple_data_dict.items():
        flops = v[1]
        norm_flops = map_to_range(flops, min_flops, max_flops, 0., 1.)
        v[1] = norm_flops
    for v in simple_data_dict.values():
        assert 0. <= v[0] <= 1.
        assert 0. <= v[1] <= 1.
    if simple_data_dict_cache is not None:
        with open(simple_data_dict_cache, "wb") as f:
            pickle.dump(simple_data_dict, f, protocol=4)
    if show_example:
        # Get the detailed information
        results = api.query_by_index(1, 'cifar10') # A dict of all trials for 1st net on cifar100, where the key is the seed
        print('There are {:} trials for this architecture [{:}] on cifar10'.format(len(results), api[1]))
        print("results: {}".format(str(results)))

        index = api.query_index_by_arch('|nor_conv_3x3~0|+|nor_conv_3x3~0|avg_pool_3x3~1|+|skip_connect~0|nor_conv_3x3~1|skip_connect~2|')
        info = api.query_meta_info_by_index(index)  # This is an instance of `ArchResults`
        res_metrics = info.get_metrics('cifar10', 'ori-test')  # This is a dict with metric names as keys
        print("ori-test res_metrics: {}".format(res_metrics))

        res_metrics = info.get_metrics('cifar10', 'train')  # This is a dict with metric names as keys
        print("train res_metrics: {}".format(res_metrics))

        cost_metrics = info.get_compute_costs('cifar10')  # This is a dict with metric names as keys, e.g., flops, params, latency
        print("cost_metrics: {}".format(cost_metrics))

        # for the metric after a specific epoch
        index = api.query_index_by_arch('|nor_conv_3x3~0|+|nor_conv_3x3~0|avg_pool_3x3~1|+|skip_connect~0|nor_conv_3x3~1|skip_connect~2|')
        api.show(index)
    return api


def get_nb201_key(node_labels, ni2de_map):
    """
    :param node_labels: Must include input/output labels
    :param ni2de_map: Mapping between node idx to darts edge pair. E.g. node idx=3: (darts_node=0, darts_node=4)
    :return: str key
    """
    darts_dst2ni = collections.defaultdict(list)
    for ni, (src, dst) in ni2de_map.items():
        darts_dst2ni[dst].append( (ni, src) )
        darts_dst2ni[dst].sort(key=lambda t: t[0])
    darts_dst_ni_pairs = [(k, v) for k,v in darts_dst2ni.items()]
    darts_dst_ni_pairs.sort(key=lambda t: t[0])
    key = ""
    for dart_dst, adj_list in darts_dst_ni_pairs:
        labels = [node_labels[ni]+"~"+str(dart_src) for ni, dart_src in adj_list]
        partial_key = "|" + "|".join(labels) + "|"
        key += partial_key  + "+"
    return key[:-1]


def nb201_key2labels(key):
    """
    Reverse process of get_nb201_key()
    NOTE: does not include input and output nodes
    """
    rv = []
    node_strs = key.split("+")
    for node_str in node_strs:
        op_keys = node_str.split("|")
        op_keys = [v for v in op_keys if len(v) > 0]
        curr_src_idx = -1
        for op_key in op_keys:
            op, src_idx = op_key.split("~")
            assert int(src_idx) > curr_src_idx
            curr_src_idx = int(src_idx) # Strictly increasing
            rv.append(op)
    return rv


if __name__ == "__main__":
    # print("Loading cifar10")
    # load_complete_nb201(d_set="cifar10", eval_method="ori-test",
    #                     simple_data_dict_cache=P_SEP.join([CACHE_DIR, "nb201_acc_flops_dict_c10.pkl"])
    #                     )
    # print("Loading cifar100")
    # load_complete_nb201(d_set="cifar100", eval_method="x-test",
    #                     simple_data_dict_cache=P_SEP.join([CACHE_DIR, "nb201_acc_flops_dict_c100.pkl"])
    #                     )
    print("Loading ImageNet16-120")
    load_complete_nb201(d_set="ImageNet16-120", eval_method="x-test",
                        #simple_data_dict_cache=P_SEP.join([CACHE_DIR, "nb201_acc_flops_dict_img_net.pkl"])
                        )

    # _map = {1: (0, 1), 2: (0, 2), 3: (1, 2), 4: (0, 3), 5: (1, 3), 6: (2, 3)}
    # _k = get_nb201_key(["input", "nor_conv_3x3", "none", "nor_conv_1x1", "none", "nor_conv_1x1", "avg_pool_3x3", "output"],
    #                    _map)
    # print(_k)
    # _labels = nb201_key2labels(_k)
    # print(_labels)

    print("done")
