from tqdm import tqdm
from constants import *
from model_components import SimpleRNN
from torch.distributions import Categorical
from utils.model_utils import device
from torch_geometric.data import Data, Batch
from utils.nas_utils import *
from utils.debug_utils import IntervalInfo as interval_info


"""
Helper functions
"""
def get_gpn_compatible_batch(blocks):
    batch_node_tsr_list = []
    batch_edge_lists = []
    batch_last_idx_list = []
    for block in blocks:
        node_tsr_list, edge_list, _, last_idx_list = block.get_gpn_compatible_input(include_scaffold=False)
        batch_node_tsr_list.append(node_tsr_list[0])
        batch_edge_lists.append(edge_list[0])
        if len(batch_last_idx_list) == 0:
            batch_last_idx_list.append(last_idx_list[0])
        else:
            batch_last_idx_list.append(batch_last_idx_list[-1] + last_idx_list[0] + 1)
    return batch_node_tsr_list, batch_edge_lists, batch_last_idx_list


def get_aggr_perf(perf_values, smaller_is_better, weights=None):
    """
    Returns a single value based on a multi-value perf array
    """
    assert len(perf_values) == len(smaller_is_better)
    if weights is not None:
        assert len(weights) == len(smaller_is_better)
    else:
        weights = [1.0] * len(smaller_is_better)
    weighted_vals = []
    for i, perf in enumerate(perf_values):
        assert 0 <= perf<= 1, "Performance value should be between 0 and 1"
        if smaller_is_better[i]:
            perf = 1.0 - perf
        weighted_vals.append(weights[i] * perf)
    return sum(weighted_vals)


def get_validity_metrics(state, output_node_label):
    rv = {
        "Contains output node": output_node_label in state.curr_node_label_list,
        "No loner nodes": not block_has_loner_nodes(state),
        "No more than 1 input": len(get_input_node_indices(state.curr_edge_list)) == 1,
        "No more than 1 output": len(get_output_node_indices(state.curr_edge_list)) == 1
    }
    return rv


def data_dict_to_state(data_dict, scaffold_node_labels, scaffold_node_feature_tsr,
                       allow_self_edge_for_non_input=False):
    # convert 1-hot feature to label indices
    one_hot_lists = data_dict[ARCH_TOPO_ENCODING_LISTS_NAS_BENCH]
    idx_list = one_hot_lists_to_idx_list(one_hot_lists)
    node_feature_tsr = torch.LongTensor(idx_list)
    edge_list = [
        [t[0] for t in data_dict[ARCH_EDGE_LIST_NAS_BENCH]],
        [t[1] for t in data_dict[ARCH_EDGE_LIST_NAS_BENCH]],
    ]
    if not allow_self_edge_for_non_input:
        edge_list = prune_self_edge_from_edge_list(edge_list)
        edge_list[0] = [0] + edge_list[0]
        edge_list[1] = [0] + edge_list[1]
    label_list = data_dict[ARCH_OP_LIST_NAS_BENCH]
    state = BlockState(label_list, node_feature_tsr, scaffold_node_labels, scaffold_node_feature_tsr, edge_list)
    return state


"""
Neural models
"""
class BlockGPN(torch.nn.Module):

    def __init__(self, in_channels, hidden_size, out_channels, gnn_constructor, action_predictor,
                 node_pre_embedding, activ=torch.nn.ReLU(), num_layers=4, dropout_prob=0.1):
        super(BlockGPN, self).__init__()
        self.node_pre_embedding = node_pre_embedding
        self.in_channels = in_channels
        self.hidden_size = hidden_size
        self.out_channels = out_channels
        self.gnn_layers = torch.nn.ModuleList()
        for i in range(num_layers):
            input_size, output_size = hidden_size, hidden_size
            if i == 0:
                input_size = in_channels
            if i == num_layers - 1:
                output_size = out_channels
            gnn_layer = gnn_constructor(input_size, output_size)
            self.gnn_layers.append(gnn_layer)
        self.activ = activ
        self.action_predictor = action_predictor
        self.dropout = torch.nn.Dropout(dropout_prob)
        self.init_ff = torch.nn.Linear(2 * in_channels, in_channels)

    def init_node_embedding_prior(self, node_embedding):
        prior = np.random.normal(size=node_embedding.shape)
        prior = torch.from_numpy(prior).float().to(device())
        node_embedding = torch.cat([node_embedding, prior], dim=-1)
        return self.init_ff(node_embedding)

    def forward(self, batch_node_tsr, edge_tsr_list, scaffold_node_tsr, batch_last_node_idx_list):
        node_embedding = self.node_pre_embedding(batch_node_tsr.to(device()))
        if node_embedding.shape[1] == 1:
            node_embedding = self.init_node_embedding_prior(node_embedding)
        data_list = [
            Data(x=node_embedding[i,:], edge_index=edge_tsr_list[i].to(device()), edge_attr=None)
                for i in range(node_embedding.shape[0])
        ]
        torch_geo_batch = Batch.from_data_list(data_list)
        edge_index_tsr = torch_geo_batch.edge_index
        curr_gnn_output = torch_geo_batch.x
        for li, gnn_layer in enumerate(self.gnn_layers):
            node_embedding_tsr = gnn_layer(curr_gnn_output, edge_index_tsr)
            curr_gnn_output = self.activ(node_embedding_tsr)
        curr_gnn_output = self.dropout(curr_gnn_output)
        batch_embedding_list = torch_geo_batch_to_data_list(curr_gnn_output, batch_last_node_idx_list)
        batch_embedding = torch.cat([t.unsqueeze(0) for t in batch_embedding_list], dim=0)
        batch_action_probs = self.action_predictor(batch_embedding)
        return batch_embedding, batch_action_probs


class RNNActionPredictor(torch.nn.Module):

    def __init__(self, node_embed_size, hidden_size, n_cand_nodes):
        super(RNNActionPredictor, self).__init__()
        self.graph_ff = torch.nn.Linear(node_embed_size, hidden_size)
        # self.scaffold_ff = torch.nn.Linear(scaffold_embed_size, hidden_size)
        self.node_selector = torch.nn.Linear(hidden_size, n_cand_nodes)
        self.edge_selector = torch.nn.Linear(hidden_size, 2)
        self.hidden_init = torch.nn.Linear(n_cand_nodes, hidden_size)
        self.edge_embed_layer = torch.nn.Embedding(2, hidden_size)
        self.rnn = SimpleRNN(2 * hidden_size, hidden_size, rnn_dir=1)

    def forward(self, node_embedding):
        # Node probs
        node_embedding = node_embedding.to(device())
        graph_embedding = self.graph_ff(node_embedding.sum(dim=1))
        new_node_ctx = self.node_selector(graph_embedding)
        new_node_probs = torch.softmax(new_node_ctx, dim=-1)
        # # Edge probs
        # scaffold_embedding = self.scaffold_ff(scaffold_embedding.to(device()))
        # scaffold_embedding = scaffold_embedding.repeat(new_node_probs.shape[0], 1, 1)
        # weighted_scaffold_ctx = torch.bmm(new_node_probs.unsqueeze(1), scaffold_embedding)
        # curr_hidden = weighted_scaffold_ctx.transpose(0, 1)
        # edge_ctx_list = []
        # for i in range(node_embedding.shape[1] - 1, -1, -1):
        #     input_state = node_embedding[:, i, :].unsqueeze(1)
        #     output_state, curr_hidden = self.rnn(input_state, curr_hidden)
        #     edge_ctx_list.append(output_state)
        # edge_ctx_list = edge_ctx_list[::-1]
        # new_edge_ctx = torch.cat(edge_ctx_list, dim=1)
        # new_edge_ctx = self.edge_selector(new_edge_ctx)
        # new_edge_probs = torch.softmax(new_edge_ctx, dim=-1)

        curr_hidden = self.hidden_init(new_node_ctx.detach()).unsqueeze(0)
        edge_probs_list = []
        input_tsr = torch.zeros(node_embedding.shape[0]).long().reshape(-1, 1).to(device())
        for i in range(node_embedding.shape[1] - 1, -1, -1):
            input_tsr = self.edge_embed_layer(input_tsr)
            input_tsr = torch.cat([input_tsr, node_embedding[:, i, :].unsqueeze(1)], dim=-1)
            output_state, curr_hidden = self.rnn(input_tsr, curr_hidden)
            edge_logits = self.edge_selector(output_state)
            edge_probs = torch.softmax(edge_logits, dim=-1)
            edge_probs_list = [edge_probs] + edge_probs_list
            _, idx = edge_probs.max(dim=-1)
            input_tsr = idx.reshape(-1, 1)
        new_edge_probs = torch.cat(edge_probs_list, dim=1)
        return new_node_probs, new_edge_probs


class NodeAggr(torch.nn.Module):

    def __init__(self, hidden_size, aggr_dim=1, method="sum",
                 enable_aggr_net=False, enable_node_gating=False):
        super(NodeAggr, self).__init__()
        self.hidden_size = hidden_size
        self.method = method.rstrip().lower()
        self.enable_node_gating = enable_node_gating
        self.enable_aggr_net = enable_aggr_net
        self.aggr_dim = aggr_dim
        self.gating_net = torch.nn.Linear(self.hidden_size, 1)
        self.aggr_net = torch.nn.Linear(self.hidden_size, 1)

    def forward(self, node_embedding_tsr, aggr_dim=None):
        if aggr_dim is None: aggr_dim = self.aggr_dim
        if self.enable_node_gating:
            gate_probs = torch.sigmoid(self.gating_net(node_embedding_tsr).squeeze())
            node_embedding_tsr = gate_probs.repeat(1, 1, node_embedding_tsr.shape[-1]) * node_embedding_tsr
        if self.method == "sum":
            aggr_tsr = torch.sum(node_embedding_tsr, dim=aggr_dim, keepdim=False)
        elif self.method == "mean":
            aggr_tsr = torch.mean(node_embedding_tsr, dim=aggr_dim, keepdim=False)
        elif self.method == "max":
            _, aggr_tsr = torch.max(node_embedding_tsr, dim=aggr_dim, keepdim=False)
        elif self.method == "last":
            aggr_tsr = node_embedding_tsr[-1, :]
        else:
            raise ValueError("Unknown aggregation method: {}".format(self.method))
        return self.aggr_net(aggr_tsr) if self.enable_aggr_net else aggr_tsr


"""
RL-related components
"""
class BlockState:

    def __init__(self, curr_node_label_list, curr_node_feature_tsr,
                 scaffold_node_label_list, scaffold_node_feature_tsr,
                 curr_edge_list):
        """
        :param curr_node_label_list: identifier for each node that are already in the graph, i.e., no scaffold nodes
        :param curr_node_feature_tsr: current nodes in the graph, 1 x num_nodes x feature_size
        :param scaffold_node_label_list: this won't change between different states
        :param scaffold_node_feature_tsr: potential future nodes to add 1 x num_nodes x feature_size
        :param curr_edge_list: edge list like [ [0,1,2,3],[1,2,3,4] ]
        """
        self.curr_node_label_list = curr_node_label_list
        self.curr_node_feature_tsr = curr_node_feature_tsr
        self.scaffold_node_label_list = scaffold_node_label_list
        self.scaffold_node_feature_tsr = scaffold_node_feature_tsr
        self.curr_edge_list = copy.deepcopy(curr_edge_list)

    def de_tensorize(self):
        b = copy.deepcopy(self)
        if isinstance(b.curr_node_feature_tsr, torch.Tensor):
            b.curr_node_feature_tsr = b.curr_node_feature_tsr.tolist()
        if isinstance(b.scaffold_node_feature_tsr, torch.Tensor):
            b.scaffold_node_feature_tsr = b.scaffold_node_feature_tsr.tolist()
        return b

    def re_tensorize(self):
        if isinstance(self.curr_node_feature_tsr, list):
            self.curr_node_feature_tsr = torch.LongTensor(self.curr_node_feature_tsr)
        if isinstance(self.scaffold_node_feature_tsr, list):
            self.scaffold_node_feature_tsr = torch.LongTensor(self.scaffold_node_feature_tsr)
        return self

    def __len__(self):
        return len(self.curr_node_label_list)

    def __str__(self):
        return "BlockState[{}]:{}".format(len(self.curr_node_label_list),
                                          ",".join(self.curr_node_label_list)) + \
               "|Edges: {}".format(str(edge_list_to_edge_pairs(self.curr_edge_list)))

    def __repr__(self):
        return str(self)

    def __deepcopy__(self, memodict={}):
        _state = BlockState(
            curr_node_label_list=copy.deepcopy(self.curr_node_label_list),
            curr_node_feature_tsr=copy.deepcopy(self.curr_node_feature_tsr),
            scaffold_node_label_list=copy.deepcopy(self.scaffold_node_label_list),
            scaffold_node_feature_tsr=copy.deepcopy(self.scaffold_node_feature_tsr),
            curr_edge_list=copy.deepcopy(self.curr_edge_list),
        )
        return _state

    def to(self, dev):
        self.curr_node_feature_tsr = self.curr_node_feature_tsr.to(dev)
        self.scaffold_node_feature_tsr = self.scaffold_node_feature_tsr.to(dev)
        return self

    def get_node_out_degrees(self):
        deg = [0] * len(self)
        for src, dst in zip(self.curr_edge_list[0], self.curr_edge_list[1]):
            if src == dst: continue
            deg[src] += 1
        return deg

    def fix_dangling_nodes(self, output_node_label, output_node_feature_tsr, add_self_edge=False):
        dangling_input_nodes = get_input_node_indices(self.curr_edge_list)
        for node_idx in dangling_input_nodes:
            if node_idx == 0: continue
            self.curr_edge_list[0].append(0)
            self.curr_edge_list[1].append(node_idx)

        # fix for output
        dangling_output_nodes = get_output_node_indices(self.curr_edge_list)
        if len(dangling_output_nodes) <= 1: return
        if output_node_label not in self.curr_node_label_list:
            output_node_idx = len(self.curr_node_label_list)
            self.curr_node_label_list.append(output_node_label)
            self.curr_node_feature_tsr = torch.cat([self.curr_node_feature_tsr, output_node_feature_tsr], dim=0)
            if add_self_edge:
                self.curr_edge_list[0].append(output_node_idx)
                self.curr_edge_list[1].append(output_node_idx)
            for node_idx in dangling_output_nodes:
                self.curr_edge_list[0].append(node_idx)
                self.curr_edge_list[1].append(output_node_idx)
        else:
            output_node_idx = None
            for i, n in enumerate(self.curr_node_label_list):
                if n == output_node_label:
                    output_node_idx = i
                    break
            assert output_node_idx is not None
            for node_idx in dangling_output_nodes:
                if node_idx == output_node_idx: continue
                self.curr_edge_list[0].append(node_idx)
                self.curr_edge_list[1].append(output_node_idx)

    def get_gpn_compatible_input(self, include_scaffold=False, remove_self_edges=False):
        # Returns node_feature_tsr_list, edge_tsr_list, partial_graph_idx_tsr_list, last_node_idx_list
        # But only for a single instance, i.e., batch dim is 1
        if include_scaffold:
            node_feature_tsr = torch.cat([self.curr_node_feature_tsr, self.scaffold_node_feature_tsr], dim=0)
        else:
            node_feature_tsr = self.curr_node_feature_tsr
        if remove_self_edges:
            edge_list = prune_self_edge_from_edge_list(self.curr_edge_list)
        else:
            edge_list = self.curr_edge_list
        edge_tsr = torch.LongTensor(edge_list).to(device())
        return node_feature_tsr.unsqueeze(0), [edge_tsr], self.scaffold_node_feature_tsr, [node_feature_tsr.shape[0] - 1]

    def gen_block_state_with_new_edge(self, src_idx, dst_idx):
        """
        Only add one edge between two existing nodes, returns a new state object
        """
        new_state = BlockState(
            curr_node_label_list=[v for v in self.curr_node_label_list],
            curr_node_feature_tsr=self.curr_node_feature_tsr,
            scaffold_node_label_list=self.scaffold_node_label_list,
            scaffold_node_feature_tsr=self.scaffold_node_feature_tsr,
            curr_edge_list=self.curr_edge_list,
        )
        edge_set = {tuple(t) for t in zip(self.curr_edge_list[0], self.curr_edge_list[1])}
        if (src_idx, dst_idx) not in edge_set:
            new_state.curr_edge_list[0].append(src_idx)
            new_state.curr_edge_list[1].append(dst_idx)
        return new_state

    def gen_block_state_with_new_node(self, new_node_label, new_node_feature_tsr, edge_connections,
                                      add_self_edge=False):
        """
        Adds a new node and a new edge to the graph
        :param new_node_label: identifier for the new node
        :param new_node_feature_tsr:
        :param edge_connections:
        :param add_self_edge: whether to add a self edge
        """
        assert len(edge_connections) == len(self.curr_node_label_list)
        new_node_idx = self.curr_node_feature_tsr.shape[0]
        new_state = BlockState(
            curr_node_label_list=[v for v in self.curr_node_label_list] + [new_node_label],
            curr_node_feature_tsr=torch.cat([self.curr_node_feature_tsr, new_node_feature_tsr], dim=0),
            scaffold_node_label_list=self.scaffold_node_label_list,
            scaffold_node_feature_tsr=self.scaffold_node_feature_tsr,
            curr_edge_list=self.curr_edge_list,
        )
        if add_self_edge:
            new_state.curr_edge_list[0].append(new_node_idx)
            new_state.curr_edge_list[1].append(new_node_idx)
        for src_idx, connection in enumerate(edge_connections):
            if get_tsr_value(connection) > 0:
                new_state.curr_edge_list[0].append(src_idx)
                new_state.curr_edge_list[1].append(new_node_idx)
        return new_state

    def get_str_id(self):
        from utils.nas_utils import edge_list_to_edge_matrix
        if isinstance(self.curr_node_feature_tsr, torch.Tensor):
            node_indices = self.curr_node_feature_tsr.tolist()
        else:
            node_indices = self.curr_node_feature_tsr
        src_list, dst_list = prune_self_edge_from_edge_list(self.curr_edge_list)
        matrix = edge_list_to_edge_matrix([src_list, dst_list], len(node_indices))
        fingerprint = hash_module(matrix, node_indices)
        return fingerprint

    def get_evolving_states(self, include_self=False):
        """
        There could be more than 1 valid sequences, this just returns 1
        :return: A list of states that builds towards the current state
        """
        rv = []
        for bi in range(1, len(self)):
            state = copy.deepcopy(self)
            state.curr_node_label_list = state.curr_node_label_list[:bi]
            state.curr_node_feature_tsr = state.curr_node_feature_tsr[:bi]
            new_src_list, new_dst_list = [], []
            for src, dst in zip(state.curr_edge_list[0], state.curr_edge_list[1]):
                if src >= bi or dst >= bi: continue
                new_src_list.append(src)
                new_dst_list.append(dst)
            state.curr_edge_list = [new_src_list, new_dst_list]
            rv.append(state)
        if include_self:
            rv.append(self)
        return rv


class BatchBlockStates:
    """Only used during k epoch PPO updates"""
    def __init__(self, states):
        self.states = states
        self.node_feature_tsr = None
        self.edge_tsr_list = []
        self.scaffold_node_feature_tsr = None
        self.last_idx_list = []
        self._build_gpn_compatible_input()

    def __len__(self):
        return len(self.states)

    def __str__(self):
        return "BatchBlockStates of size {}".format(len(self))

    def __repr__(self):
        return str(self)

    def _build_gpn_compatible_input(self, remove_self_edges=False):
        node_feature_tsr_list = []
        self.edge_tsr_list = []
        self.last_idx_list = []
        aggr_node_size = 0
        self.scaffold_node_feature_tsr = self.states[0].scaffold_node_feature_tsr
        for state in self.states:
            node_feature_tsr = state.curr_node_feature_tsr.unsqueeze(0)
            if remove_self_edges:
                edge_list = prune_self_edge_from_edge_list(state.curr_edge_list)
            else:
                edge_list = state.curr_edge_list
            edge_tsr = torch.LongTensor(edge_list)
            node_feature_tsr_list.append(node_feature_tsr)
            self.edge_tsr_list.append(edge_tsr)
            self.last_idx_list.append(aggr_node_size + node_feature_tsr.shape[1] - 1)
            aggr_node_size += node_feature_tsr.shape[1]
        self.node_feature_tsr = torch.cat(node_feature_tsr_list)

    def get_gpn_compatible_input(self):
        return self.node_feature_tsr, self.edge_tsr_list, self.scaffold_node_feature_tsr, self.last_idx_list


class GPNAction:

    def __init__(self, actor_name, timestep, node_probs, edge_connection_probs):
        assert len(edge_connection_probs.shape) == 2
        self.actor_name = actor_name
        self.timestep = timestep
        self.node_probs = node_probs
        self.edge_connection_probs = edge_connection_probs
        self.node_idx = Categorical(probs=self.node_probs).sample()
        self.edge_connections = Categorical(probs=self.edge_connection_probs).sample()
        self.node_entropy = Categorical(probs=self.node_probs).entropy()
        self.edge_entropy = Categorical(probs=self.edge_connection_probs).entropy().mean()
        self.aggr_entropy = torch.sum(
            torch.cat([
                self.get_node_entropy().unsqueeze(0),
                self.get_edge_entropy().unsqueeze(0)
            ], dim=0)
        )

    def __deepcopy__(self, memodict={}):
        a = GPNAction(self.actor_name, self.timestep,
                      copy.deepcopy(self.node_probs),
                      copy.deepcopy(self.edge_connection_probs))
        a.node_idx = copy.deepcopy(self.node_idx)
        a.edge_connections = copy.deepcopy(self.edge_connections)
        a.node_entropy = copy.deepcopy(self.node_entropy)
        a.edge_entropy = copy.deepcopy(self.edge_entropy)
        a.aggr_entropy = copy.deepcopy(self.aggr_entropy)
        return a

    def de_tensorize(self):
        a = copy.deepcopy(self)
        if isinstance(a.node_probs, torch.Tensor):
            a.node_probs = a.node_probs.tolist()
        if isinstance(a.edge_connection_probs, torch.Tensor):
            a.edge_connection_probs = a.edge_connection_probs.tolist()
        if isinstance(a.node_idx, torch.Tensor):
            a.node_idx = get_tsr_value(a.node_idx)
        if isinstance(a.edge_connections, torch.Tensor):
            a.edge_connections = a.edge_connections.tolist()
        if isinstance(a.node_entropy, torch.Tensor):
            a.node_entropy = get_tsr_value(a.node_entropy)
        if isinstance(a.edge_entropy, torch.Tensor):
            a.edge_entropy = get_tsr_value(a.edge_entropy)
        if isinstance(a.aggr_entropy, torch.Tensor):
            a.aggr_entropy = get_tsr_value(a.aggr_entropy)
        return a

    def re_tensorize(self):
        if isinstance(self.node_probs, list):
            self.node_probs = torch.FloatTensor(self.node_probs)
        if isinstance(self.edge_connection_probs, list):
            self.edge_connection_probs = torch.FloatTensor(self.edge_connection_probs)
        if isinstance(self.node_idx, int):
            self.node_idx = torch.LongTensor([self.node_idx]).squeeze()
        if isinstance(self.edge_connections, list):
            self.edge_connections = torch.LongTensor(self.edge_connections)
        if isinstance(self.node_entropy, float):
            self.node_entropy = torch.FloatTensor([self.node_entropy]).squeeze()
        if isinstance(self.edge_entropy, float):
            self.edge_entropy = torch.FloatTensor([self.edge_entropy]).squeeze()
        if isinstance(self.aggr_entropy, float):
            self.aggr_entropy = torch.FloatTensor([self.aggr_entropy]).squeeze()
        return self

    def to(self, dev):
        self.node_probs = self.node_probs.to(dev)
        self.edge_connection_probs = self.edge_connection_probs.to(dev)
        self.node_idx = self.node_idx.to(dev)
        self.edge_connections = self.edge_connections.to(dev)
        return self

    def re_sample_action(self):
        self.node_idx = Categorical(probs=self.node_probs).sample()
        self.edge_connections = Categorical(probs=self.edge_connection_probs).sample()
        return self.get_discrete_action()

    def get_discrete_action(self, get_raw_value=True):
        node_idx = self.node_idx.squeeze().item() if get_raw_value else self.node_idx
        edge_connections = []
        for ei in range(self.edge_connections.shape[0]):
            conn = self.edge_connections[ei]
            edge_connections.append(get_tsr_value(conn) if get_raw_value else conn)
        return node_idx, edge_connections

    def get_node_entropy(self):
        return self.node_entropy

    def get_edge_entropy(self):
        return self.edge_entropy

    def get_entropy(self):
        return self.aggr_entropy

    def get_discrete_action_str_id(self):
        s = "Node idx: {}".format(str(self.node_idx.squeeze().item()))
        s += ", Edge connections: {}".format(str(self.edge_connections))
        return s

    def __str__(self):
        s = "GPN Action at time {} by actor {}".format(self.timestep, self.actor_name)
        s += ", Node idx: {}".format(str(self.node_idx.squeeze().item()))
        s += ", Edge connections: {}".format(str(self.edge_connections))
        s += ", Node probs: {}".format(str(self.node_probs.tolist()))
        s += ", Edge probs: {}".format(str(self.edge_connection_probs))
        return s

    def __repr__(self):
        return str(self)


class BatchGPNActions:
    """Only used during k epoch PPO updates"""
    def __init__(self, actions):
        self.actions = actions
        batch_node_probs = []
        batch_edge_probs = []
        batch_node_indices = []
        batch_edge_connections = []
        batch_node_entropy = []
        batch_edge_entropy = []
        batch_aggr_entropy = []
        for action in self.actions:
            batch_node_probs.append(action.node_probs.unsqueeze(0))
            batch_edge_probs.append(action.edge_connection_probs.unsqueeze(0))
            batch_node_indices.append(action.node_idx.unsqueeze(0))
            batch_edge_connections.append(action.edge_connections.unsqueeze(0))
            batch_node_entropy.append(action.get_node_entropy().unsqueeze(0))
            batch_edge_entropy.append(action.get_edge_entropy().unsqueeze(0))
            batch_aggr_entropy.append(action.get_entropy().unsqueeze(0))
        self.batch_node_probs = torch.cat(batch_node_probs, dim=0)
        self.batch_edge_probs = torch.cat(batch_edge_probs, dim=0)
        self.batch_node_indices = torch.cat(batch_node_indices, dim=0)
        self.batch_edge_connections = torch.cat(batch_edge_connections, dim=0)
        self.node_entropy = torch.cat(batch_node_entropy, dim=0).mean()
        self.edge_entropy = torch.cat(batch_edge_entropy, dim=0).mean()
        self.aggr_entropy = torch.cat(batch_aggr_entropy, dim=0).mean()
        self.batch_node_log_probs = Categorical(probs=self.batch_node_probs).log_prob(self.batch_node_indices)
        self.batch_edge_log_probs = Categorical(probs=self.batch_edge_probs).log_prob(self.batch_edge_connections)

    def __len__(self):
        return len(self.actions)

    def __str__(self):
        return "BatchGPNActions of size {}".format(len(self))

    def __repr__(self):
        return str(self)

    def to(self, dev):
        for action in self.actions:
            action.to(dev)
        return self

    def get_node_entropy(self):
        return self.node_entropy

    def get_edge_entropy(self):
        return self.edge_entropy

    def get_entropy(self):
        return self.aggr_entropy

    def get_action_log_probs(self, batch_node_indices=None, batch_edge_connections=None):
        if batch_node_indices is None:
            batch_node_log_probs = self.batch_node_log_probs
        else:
            batch_node_log_probs = Categorical(probs=self.batch_node_probs).log_prob(batch_node_indices)
        if batch_edge_connections is None:
            batch_edge_log_probs = self.batch_edge_log_probs
        else:
            batch_edge_log_probs = Categorical(probs=self.batch_edge_probs).log_prob(batch_edge_connections)
        return batch_node_log_probs, batch_edge_log_probs


class NASEnv:

    def __init__(self, actor, init_state, node_label2feat, step_reward_module, final_reward_module,
                 output_node_label, max_allowed_steps, add_self_edge=False):
        self._actor = actor
        self._node_label2feat = node_label2feat
        self._step_reward_module = step_reward_module
        self._final_reward_module = final_reward_module
        self._curr_state = init_state
        self._init_state = init_state
        self._add_self_edge = add_self_edge
        self._output_node_label = output_node_label
        self._max_allowed_steps = max_allowed_steps
        self._num_steps = 0

    @property
    def final_reward_module(self):
        return self._final_reward_module

    def load_state_dict(self, actor_state_dict):
        if actor_state_dict is not None:
            self._actor.load_state_dict(actor_state_dict)

    def get_init_state(self):
        return self._init_state

    def set_actor(self, actor):
        self._actor = actor

    def get_actor(self):
        return self._actor

    def get_curr_state(self):
        return self._curr_state

    def set_curr_state(self, state):
        self._curr_state = state

    def get_num_steps(self):
        return self._num_steps

    def one_step_left(self):
        return self._num_steps == self._max_allowed_steps - 1

    def eval(self):
        self._final_reward_module.eval()

    def train(self):
        self._final_reward_module.train()

    def to(self, dev):
        self._final_reward_module = self._final_reward_module.to(dev)
        self._actor = self._actor.to(dev)
        return self

    def render(self, mode='human'):
        pass

    def reset(self):
        self._curr_state = self._init_state
        self._num_steps = 0

    def get_step_reward_for_action(self, action):
        return self._step_reward_module.get_reward(action)

    def get_final_reward_for_state(self, state):
        return self._final_reward_module.get_reward(state)

    def step(self, action:GPNAction):
        input_state = self._curr_state
        new_state, stop = self.gen_new_state_from_action(self._curr_state, action)
        final_state_reached = stop or self._num_steps + 1 >= self._max_allowed_steps
        step_reward = self.get_step_reward_for_action(action)
        if final_state_reached:
            # if stop is predicted, need to evaluate the old state + predicted action and get a final reward
            final_reward = self.get_final_reward_for_state(new_state)
            reward = step_reward + final_reward
        elif step_reward < 0.:
            # Force stop if invalid action is generated, also return final reward
            final_state_reached = True
            final_reward = self.get_final_reward_for_state(new_state)
            reward = step_reward + final_reward
        else:
            reward = step_reward
        self._curr_state = new_state
        self._num_steps += 1
        return final_state_reached, input_state, reward, new_state

    def gen_new_state_from_action(self, input_state, action):
        new_node_idx, edge_connections = action.get_discrete_action()
        new_node_label = input_state.scaffold_node_label_list[new_node_idx]
        new_state = input_state.gen_block_state_with_new_node(new_node_label,
                                                              self._node_label2feat[new_node_label],
                                                              edge_connections,
                                                              add_self_edge=self._add_self_edge)
        if new_node_label == self._output_node_label:
            # Whether to fix output node manually?
            # new_state.fix_dangling_nodes(new_node_label, self.node_label2feat[new_node_label], add_self_edge=self.add_self_edge)
            return new_state, True
        return new_state, False

    def __str__(self):
        return "NASEnv for actor {}, num_steps: {}".format(self._actor.get_name(), self._num_steps)

    def __repr__(self):
        return str(self)


class NASEnvManager:

    def __init__(self, parallel_actors, num_parallel_actors, init_state, node_label2feat,
                 step_reward_module, final_reward_module,
                 max_allowed_steps, output_node_label, add_self_edge=False):
        self.init_state = init_state
        self.max_allowed_steps = max_allowed_steps
        self.node_label2feat = node_label2feat
        self.step_reward_module = step_reward_module
        self.final_reward_module = final_reward_module
        self.add_self_edge = add_self_edge
        self.output_node_label = output_node_label
        self.num_parallel_actors = num_parallel_actors
        self.parallel_actors = parallel_actors
        self.curr_env_idx = 0
        for i, actor in enumerate(self.parallel_actors):
            actor.name += "_{}".format(i)
        self.parallel_envs = [
            NASEnv(actor, copy.deepcopy(init_state), node_label2feat, step_reward_module,
                   final_reward_module, output_node_label, max_allowed_steps, add_self_edge)
                for actor in self.parallel_actors]

    def share_memory(self):
        self.final_reward_module.share_memory()

    def load_checkpoint(self, actor_state_dict):
        for env in self.parallel_envs:
            env.load_state_dict(actor_state_dict)

    def get_actor_rep(self):
        return self.parallel_envs[0].get_actor()

    def get_env_rep(self):
        return self.parallel_envs[0]

    def get_env_actor_rep(self):
        return self.get_env_rep(), self.get_actor_rep()

    def re_clone_actors(self, new_actor):
        for i, actor in enumerate(self.parallel_actors):
            if actor == new_actor: continue
            actor.load_state_dict(new_actor.state_dict())
            self.parallel_envs[i].set_actor(actor)

    def to(self, dev):
        for env in self.parallel_envs:
            env.to(dev)
        return self

    def reset(self):
        for env in self.parallel_envs:
            env.reset()

    def eval(self):
        for env in self.parallel_envs:
            env.eval()

    def train(self):
        for env in self.parallel_envs:
            env.train()

    def __len__(self):
        return len(self.parallel_envs)

    def __str__(self):
        return "NASEnvManager with {} parallel environments".format(len(self.parallel_envs))

    def __repr__(self):
        return str(self)

    def __iter__(self):
        self.curr_env_idx = 0
        return self

    def __next__(self):
        if self.curr_env_idx < len(self.parallel_envs):
            env = self.parallel_envs[self.curr_env_idx]
            self.curr_env_idx += 1
            return env
        else:
            self.curr_env_idx = 0
            raise StopIteration


class GPNActorCritic(torch.nn.Module):

    def __init__(self, name, policy_net, value_net):
        super(GPNActorCritic, self).__init__()
        self.name = name
        self.policy_net = policy_net
        self.value_net = value_net

    def forward(self, *args):
        return self.policy_net(*args)  # fall-through to policy network

    def get_value_for_state(self, state):
        node_embedding, _ = self.policy_net(*state.get_gpn_compatible_input())
        value = self.value_net(node_embedding).squeeze()
        return value

    def get_batch_values_for_states(self, batch_states):
        node_embedding, _ = self.policy_net(*batch_states.get_gpn_compatible_input())
        values = self.value_net(node_embedding).squeeze(1)
        return values

    def get_action_for_state(self, state, timestep=0):
        _, batch_action_probs = self.policy_net(*state.get_gpn_compatible_input())
        node_probs, multi_edge_probs = batch_action_probs
        node_probs = node_probs.squeeze(0)
        multi_edge_probs = multi_edge_probs.squeeze(0)
        action = GPNAction(self.name, timestep, node_probs, multi_edge_probs)
        return action

    def get_batch_actions_for_states(self, batch_states, timestep=0):
        _, (batch_node_probs, batch_edge_probs) = self.policy_net(*batch_states.get_gpn_compatible_input())
        # TODO: get rid of this loop
        actions = []
        for si, state in enumerate(batch_states.states):
            node_probs, multi_edge_probs = batch_node_probs[si, :], batch_edge_probs[si, :]
            action = GPNAction(self.name, timestep, node_probs, multi_edge_probs).to(device())
            actions.append(action)
        batch_actions = BatchGPNActions(actions)
        return batch_actions

    def get_name(self):
        return self.name


class StepRewardModule:

    def __init__(self, input_node_label, output_node_label, pos_reward=0.1, neg_reward=-0.1):
        self.pos_reward = pos_reward
        self.neg_reward = neg_reward
        self.input_node_label = input_node_label
        self.output_node_label = output_node_label

    def is_action_valid(self, action):
        node_idx, edge_connections = action.get_discrete_action()
        # should not predict no connections
        if all(c == 0 for c in edge_connections):
            return False
        return True

    def get_reward(self, action, pos_reward=None, neg_reward=None):
        cum_neg_reward = 0.
        action_valid = True
        if not self.is_action_valid(action):
            action_valid = False
            cum_neg_reward += self.neg_reward
        if action_valid:
            step_reward = self.pos_reward
        else:
            step_reward = cum_neg_reward
        return step_reward


class FinalRewardModule(torch.nn.Module):

    def __init__(self, perf_reward_module, pareto_reward_module, output_node_label,
                 enable_perf_reward=True, enable_pareto_reward=True,
                 input_node_idx=0, fixed_pos_reward=0.1, fixed_neg_reward=-0.1):
        super(FinalRewardModule, self).__init__()
        self.perf_reward_module = perf_reward_module
        self.pareto_reward_module = pareto_reward_module
        self.output_node_label = output_node_label
        self.enable_pareto_reward = enable_pareto_reward
        self.enable_perf_reward = enable_perf_reward
        self.fixed_pos_reward = fixed_pos_reward
        self.fixed_neg_reward = fixed_neg_reward
        self.input_node_idx = input_node_idx

    def forward(self):
        raise NotImplementedError("Should call get_reward() instead")

    def _get_validity_reward(self, state, encourage_length=False):
        n_violations = get_num_block_violations(state, self.output_node_label, self.input_node_idx)
        if n_violations == 0:
            if encourage_length:
                reward = self.fixed_pos_reward * max(0, len(state.curr_node_label_list) - 2)
            else:
                reward = self.fixed_pos_reward
            return reward
        else:
            return n_violations * self.fixed_neg_reward

    def get_reward(self, state, validity_reward_only=False):
        perf_reward, pareto_reward = 0., 0.
        validity_reward = self._get_validity_reward(state)
        if validity_reward_only or validity_reward < 0.: # no other reward if block is invalid
            rv = validity_reward
        else:
            if self.enable_perf_reward and self.perf_reward_module is not None:
                perf_reward = self.perf_reward_module.get_perf_reward(state)
            if self.enable_pareto_reward and self.pareto_reward_module is not None:
                pareto_reward = self.pareto_reward_module.get_pareto_reward(state)
            rv = pareto_reward + perf_reward + validity_reward
        return rv


class ParetoRewardModule(torch.nn.Module):

    def __init__(self, pareto_front):
        super(ParetoRewardModule, self).__init__()
        self.pareto_front = pareto_front

    def get_pareto_reward(self, block):
        raise NotImplementedError("Should override in subclasses")

    def forward(self, *args):
        raise NotImplementedError("Should override in subclasses")


class ParetoL2RewardModule:

    def __init__(self, evaluator, pareto_front, output_node_label, output_node_tsr,
                 fixed_pos_reward=0., fixed_neg_reward=-0.1):
        super(ParetoL2RewardModule, self).__init__()
        self.evaluator = evaluator
        self.pareto_front = pareto_front
        self.output_node_label = output_node_label
        self.output_node_tsr = output_node_tsr
        self.fixed_pos_reward = fixed_pos_reward
        self.fixed_neg_reward = fixed_neg_reward

    def forward(self):
        raise NotImplementedError("Should call get_reward() instead")

    def get_validity_reward(self, state: BlockState, encourage_length=False):
        n_violations = get_num_block_violations(state, self.output_node_label)
        if n_violations == 0:
            if encourage_length:
                reward = self.fixed_pos_reward * max(0, len(state.curr_node_label_list) - 2)
            else:
                reward = self.fixed_pos_reward
            return reward
        else:
            return n_violations * self.fixed_neg_reward

    def get_reward(self, state, validity_reward_only=False, perfs_dict=None):
        validity_reward = self.get_validity_reward(state)
        if validity_reward_only or validity_reward < 0.: # no other reward if block is invalid
            rv = validity_reward
        else:
            perfs = self.evaluator.get_perf_values(state)
            if perfs_dict is not None:
                perfs_dict[state.get_str_id()] = (state, perfs)
            l2_p_dists = get_pareto_l2_distances(perfs, self.pareto_front)
            scores = [t[1] for t in l2_p_dists]
            if self.pareto_front.is_dominated(perfs):
                pareto_reward = max([-s for s in scores])
            else:
                pareto_reward = min(scores)
            rv = validity_reward + pareto_reward
        return rv


class PPOBatchDataLoader:

    def __init__(self, batch_size, scale_reward=False,
                 reward_discount_factor=0.99, gae_lambda=0.95,
                 min_reward=0., max_reward=None):
        self.batch_size = batch_size
        self.curr_batch_idx = 0
        self.trajectories = []
        self.batches = []
        self.n_instances = 0
        self.scale_reward = scale_reward
        self.seen_state_action_pairs = set()
        self.data_built = False
        self.reward_discount_factor = reward_discount_factor
        self.gae_lambda = gae_lambda
        self.min_reward = min_reward
        self.max_reward = max_reward

    def clear(self):
        self.curr_batch_idx = 0
        self.trajectories = []
        self.batches = []
        self.seen_state_action_pairs = set()
        self.data_built = False

    def add_trajectory(self, state_pairs, actions, rewards, input_state_values, end_state_value):
        assert len(state_pairs) == len(actions) == len(rewards) == len(input_state_values)
        self.trajectories.append( (state_pairs, actions, rewards, input_state_values, end_state_value) )

    def get_next_batch(self):
        batch = self.batches[self.curr_batch_idx]
        self.curr_batch_idx += 1
        return batch

    def build_batches(self):
        min_reward, max_reward = None, None
        if self.scale_reward:
            max_reward = self.max_reward if self.max_reward is not None else global_max([t[2] for t in self.trajectories])
            min_reward = self.min_reward if self.min_reward is not None else global_min([t[2] for t in self.trajectories])
        instances = []
        for state_pairs, actions, rewards, input_state_values, end_state_value in tqdm(self.trajectories,
                                                                                       desc="Building instances from trajectories",
                                                                                       ascii=True, leave=False):
            if self.scale_reward:
                scaled_rewards = [
                    exp_scale(r, old_range_min=min_reward, old_range_max=max_reward) if min_reward <= r <= max_reward else r
                        for r in rewards
                ]
                advantages = compute_t_step_gae(scaled_rewards, input_state_values, end_state_value,
                                                self.reward_discount_factor, self.gae_lambda)
            else:
                advantages = compute_t_step_gae(rewards, input_state_values, end_state_value,
                                                self.reward_discount_factor, self.gae_lambda)
            for (input_state, output_state), action, reward, value, adv in zip(state_pairs, actions, rewards,
                                                                               input_state_values, advantages):
                instances.append( (input_state, action, reward, value, adv) )
        random.shuffle(instances)
        self.n_instances = len(instances)
        print("Building batches")
        # group instances by state dimensions
        size2instances = collections.defaultdict(list)
        for inst in instances:
            partial_graph_size = len(inst[0])
            size2instances[partial_graph_size].append(inst)
        for size, instances in size2instances.items():
            inst_i = 0
            while inst_i < len(instances):
                batch_instances = instances[inst_i: inst_i + self.batch_size]
                batch_states = BatchBlockStates([t[0] for t in batch_instances])
                batch_actions = BatchGPNActions([t[1] for t in batch_instances])
                batch_rewards = torch.FloatTensor([t[2] for t in batch_instances])
                batch_values = torch.FloatTensor([t[3] for t in batch_instances])
                batch_advantages = torch.FloatTensor([t[4] for t in batch_instances])
                self.batches.append((batch_states, batch_actions, batch_rewards, batch_values, batch_advantages))
                inst_i += self.batch_size
        self.data_built = True
        self.shuffle()

    def shuffle(self):
        random.shuffle(self.batches)

    @property
    def n_trajectories(self):
        return len(self.trajectories)

    @property
    def n_batches(self):
        return len(self)

    def __len__(self):
        return len(self.batches)

    def __iter__(self):
        self.curr_batch_idx = 0
        return self

    def __next__(self):
        if not self.data_built:
            self.build_batches()
        if self.curr_batch_idx < self.n_batches:
            batch = self.get_next_batch()
            return batch
        else:
            self.curr_batch_idx = 0
            self.shuffle()
            raise StopIteration


class BlockParetoFront:

    def __init__(self, init_block2perfs, perf_smaller_better_list,
                 num_samples_per_batch=1024, ignore_init_update=False,
                 ignore_init_perf_record=True, log_f=print):
        # whether smaller is better, must match the length and the order of the input block2perfs dict
        self.perf_smaller_is_better = tuple(perf_smaller_better_list)
        self.num_samples_per_batch = num_samples_per_batch  # how many truth data to sample during training
        if ignore_init_update:
            log_f("ignore_init_update=True!")
            self._curr_block2perfs = init_block2perfs
            self._curr_blocks = [b for _, (b, _) in init_block2perfs.items()]
            self._past_block2perfs = copy.deepcopy(init_block2perfs)
        else:
            self._curr_block2perfs = {}
            self._curr_blocks = []
            self._past_block2perfs = {}
            self.update_pareto_front(init_block2perfs, log_f=log_f)
            if ignore_init_perf_record:
                self._past_block2perfs = copy.deepcopy(self._curr_block2perfs)

    def state_dict(self):
        return {
            "perf_smaller_is_better": self.perf_smaller_is_better,
            "num_samples_per_batch": self.num_samples_per_batch,
            "_curr_block2perfs": self._curr_block2perfs,
            "_curr_blocks": self._curr_blocks,
            "_past_block2perfs": self._past_block2perfs,
        }

    def load_state_dict(self, sd):
        self.perf_smaller_is_better = sd["perf_smaller_is_better"]
        self.num_samples_per_batch = sd["num_samples_per_batch"]
        self._curr_block2perfs = sd["_curr_block2perfs"]
        self._curr_blocks = sd["_curr_blocks"]
        self._past_block2perfs = sd["_past_block2perfs"]

    @property
    def curr_block2perfs(self):
        return self._curr_block2perfs

    def __contains__(self, item):
        key = item.get_str_id()
        if key in self._curr_block2perfs:
            return True
        return False

    def is_dominated(self, perfs):
        for p_perfs in self.get_pareto_block_perf_values():
            dominated = True
            pi = 0
            for p_perf, perf in zip(p_perfs, perfs):
                if self.perf_smaller_is_better[pi]:
                    p_perf, perf = -p_perf, -perf
                if p_perf < perf:
                    dominated = False
                    break
                pi += 1
            if dominated: return True
        return False

    @staticmethod
    def get_pareto_for_blocks(block2perfs, perf_smaller_is_better, past_block_perfs):
        # expected block2perfs format: {block_key: (block_object, perf_values)}
        idx2key, costs = [], []
        for block_key, v in block2perfs.items():
            idx2key.append(block_key)
            _, perf_vals = v
            past_block_perfs[block_key] = v
            block_cost = []
            for i, perf in enumerate(perf_vals):
                # Convert every performance to smaller is better
                cost = perf
                if not perf_smaller_is_better[i]:
                    cost = - perf
                block_cost.append(cost)
            costs.append(block_cost)
        costs = np.asarray(costs)
        pareto_masks = BlockParetoFront.is_pareto_efficient(costs, return_mask=True)
        pareto_indices = np.argwhere(pareto_masks).squeeze(1).tolist()
        kept_block_keys = [idx2key[idx] for idx in pareto_indices]
        pareto_block2perfs = {key: block2perfs[key] for key in kept_block_keys}
        return pareto_block2perfs

    def get_pareto_neighborhood_blocks(self, max_runs=2, log_f=print):
        rv = [[b for b in self._curr_blocks]]
        dup_keys = {b.get_str_id() for b in self._curr_blocks}
        cand_block2perfs = {k: v for k, v in self._past_block2perfs.items() if k not in dup_keys}
        completed_runs = 0
        n_total_blocks = len(rv[0])
        while len(cand_block2perfs) > 0 and completed_runs < max_runs:
            new_blocks = []
            new_block2perfs = self.get_pareto_for_blocks(cand_block2perfs, self.perf_smaller_is_better, {})
            for key, (b, perf) in new_block2perfs.items():
                if key not in dup_keys:
                    dup_keys.add(key)
                    new_blocks.append(b)
                    del cand_block2perfs[key]
            rv.append(new_blocks)
            n_total_blocks += len(new_blocks)
            completed_runs += 1
        log_f("{} pareto neighborhood blocks collected after {} run(s)".format(n_total_blocks, completed_runs))
        return rv

    def get_pareto_neighborhood_block2perfs(self, max_runs=2, log_f=print):
        rv = {k: v for k, v in self._curr_block2perfs.items()}
        cand_block2perfs = {k: v for k, v in self._past_block2perfs.items() if k not in rv}
        completed_runs = 0
        while len(cand_block2perfs) > 0 and completed_runs < max_runs:
            new_block2perfs = self.get_pareto_for_blocks(cand_block2perfs, self.perf_smaller_is_better, {})
            for k, v in new_block2perfs.items():
                rv[k] = v
                del cand_block2perfs[k]
            completed_runs += 1
        log_f("{} pareto neighborhood block2perfs collected after {} run(s)".format(len(rv), completed_runs))
        return rv

    @property
    def past_block2perfs(self):
        return self._past_block2perfs

    def __sub__(self, other):
        assert isinstance(other, BlockParetoFront)
        from utils.misc_utils import edge_list_to_adj_mat
        from utils.nas_utils import is_isomorphic
        from model_helpers import blocks_diff
        extra_blocks = blocks_diff(self._curr_blocks, other.get_curr_blocks())
        extra_block2perfs = {}
        for block in extra_blocks:
            key = block.get_str_id()
            extra_block2perfs[key] = self._curr_block2perfs[key]
        # Old way of checking
        self_keys = set()
        for block in self._curr_blocks:
            self_keys.add(block.get_str_id())
        other_keys = set()
        for block in other.get_curr_blocks():
            other_keys.add(block.get_str_id())
        extra_keys = self_keys - other_keys
        for key, v in self._curr_block2perfs.items():
            if key in extra_keys:
                extra_block2perfs[key] = v
        return extra_block2perfs

    def get_pareto_blocks_with_perf(self):
        rv = []
        for key, (block, perfs) in self._curr_block2perfs.items():
            rv.append((block, perfs))
        rv.sort(key=lambda t: t[1], reverse=True)
        return rv

    def get_curr_blocks(self):
        return [b for b in self._curr_blocks]

    def get_pareto_block_perf_values(self):
        rv = []
        for key, (block, perfs) in self._curr_block2perfs.items():
            rv.append(perfs)
        rv.sort(reverse=True)
        return rv

    @staticmethod
    def is_pareto_efficient(costs, return_mask=True):
        """
        Find the pareto-efficient points
        :param costs: An (n_points, n_costs) array
        :param return_mask: True to return a mask
        :return: An array of indices of pareto-efficient points.
            If return_mask is True, this will be an (n_points, ) boolean array
            Otherwise it will be a (n_efficient_points, ) integer array of indices.
        """
        is_efficient = np.arange(costs.shape[0])
        n_points = costs.shape[0]
        next_point_index = 0  # Next index in the is_efficient array to search for
        while next_point_index < len(costs):
            nondominated_point_mask = np.any(costs < costs[next_point_index], axis=1)
            nondominated_point_mask[next_point_index] = True
            is_efficient = is_efficient[nondominated_point_mask]  # Remove dominated points
            costs = costs[nondominated_point_mask]
            next_point_index = np.sum(nondominated_point_mask[:next_point_index]) + 1
        if return_mask:
            is_efficient_mask = np.zeros(n_points, dtype=bool)
            is_efficient_mask[is_efficient] = True
            return is_efficient_mask
        else:
            return is_efficient

    def update_pareto_front(self, new_block2perfs, log_f=print):
        # new_block2perfs expected format: {block_key: (block_object, perf_values)}
        log_f("Updating Pareto front")
        old_block_keys = set(self._curr_block2perfs.keys())
        new_block_keys = set(new_block2perfs.keys())
        # merge new blocks with current front, overwrite existing block's perf if duplicated
        block2perfs = {**self._curr_block2perfs, **new_block2perfs}
        self._curr_block2perfs = self.get_pareto_for_blocks(block2perfs, self.perf_smaller_is_better,
                                                            self._past_block2perfs)
        self._curr_blocks = [v[0] for _, v in self._curr_block2perfs.items()]
        selected_block_keys = set(self._curr_block2perfs.keys())
        log_f("New Pareto front size: {}".format(len(self._curr_blocks)))
        log_f("{} blocks from previous front are kept".format(len(selected_block_keys.intersection(old_block_keys))))
        log_f("{} new blocks added".format(len(selected_block_keys.intersection(new_block_keys - old_block_keys))))
        log_f("{} past unique blocks recorded in Pareto front".format(len(self._past_block2perfs)))
        aggr_perfs = []
        for k, t in self._curr_block2perfs.items():
            _, perfs = t
            for i, p in enumerate(perfs):
                if i >= len(aggr_perfs):
                    aggr_perfs.append(p)
                else:
                    aggr_perfs[i] += p
        rv = [p / len(self._curr_block2perfs) for p in aggr_perfs]
        log_f("New pareto performance: {} ".format(rv))
        return rv

    def re_map_performance(self, perf_dict, log_f=print):
        log_f("Re-mapping Pareto")
        missing_blocks = set()
        new_block2perfs = {}
        for key, (block, perf_values) in self._curr_block2perfs.items():
            if key not in perf_dict:
                missing_blocks.add(key)
            else:
                new_block2perfs[key] = (block, perf_dict[key])
        self._curr_block2perfs = {}
        self._curr_blocks = []
        self.update_pareto_front(new_block2perfs, log_f=log_f)
        log_f("Number of missing blocks: {}".format(len(missing_blocks)))

    def __str__(self):
        return "BlockParetoFront[size: {}]".format(len(self._curr_block2perfs))

    def __repr__(self):
        return str(self)

    def __len__(self):
        return len(self._curr_blocks)


class BlockPerfEvaluator:

    def __init__(self, enable_perf_memo=True, default_clue=1.0):
        self.enable_perf_memo = enable_perf_memo
        self.clue = default_clue
        self.perf_memo = {}

    def set_clue(self, clue):
        self.clue = clue

    def get_perf_reward(self, state):
        raise NotImplementedError("Should override in child classes")

    def get_perf_values(self, state):
        # Assuming this is the expensive part
        raise NotImplementedError("Should override in child classes")


class NasBenchEvaluator(BlockPerfEvaluator):

    def __init__(self, perf_dict, default_acc, default_lat, default_clue=1.0):
        super(NasBenchEvaluator, self).__init__(default_clue=default_clue)
        self.perf_dict = perf_dict
        self.default_acc = default_acc
        self.default_lat = default_lat

    def get_perf_reward(self, state):
        pred_acc, pred_lat = self.get_perf_values(state)
        assert 0 <= pred_acc <= 1, "Acc value should be between 0 and 1, got {}".format(pred_acc)
        assert 0 <= pred_lat <= 1, "Lat value should be between 0 and 1, got {}".format(pred_lat)
        acc_reward = self.clue * pred_acc
        lat_reward = (1. - self.clue) * (1 - pred_lat)
        aggr_reward = acc_reward + lat_reward
        return aggr_reward

    def get_perf_values(self, state):
        key = state.get_str_id()
        if self.enable_perf_memo and key in self.perf_memo:
            acc, lat = self.perf_memo[key]
        else:
            if key in self.perf_dict:
                acc, lat = self.perf_dict[key]
            else:
                acc, lat = self.default_acc, self.default_lat
            self.perf_memo[key] = acc, lat
            interval_info.num_block_eval += 1
            interval_info.n_evals_to_find_block[key] = interval_info.num_block_eval
        return acc, lat


class CombinedEvaluator(BlockPerfEvaluator):

    def __init__(self, perf_predictor, perf_dict, default_clue=1.0):
        super(CombinedEvaluator, self).__init__(default_clue=default_clue)
        self.perf_predictor = perf_predictor.to(device())
        self.perf_dict = perf_dict

    def get_perf_reward(self, state):
        pred_acc, pred_lat = self.get_perf_values(state)
        assert 0 <= pred_acc <= 1, "Acc value should be between 0 and 1, got {}".format(pred_acc)
        assert 0 <= pred_lat <= 1, "Lat value should be between 0 and 1, got {}".format(pred_lat)
        acc_reward = self.clue * pred_acc
        lat_reward = (1. - self.clue) * (1 - pred_lat)
        aggr_reward = acc_reward + lat_reward
        return aggr_reward

    def get_perf_values(self, state):
        key = state.get_str_id()
        if self.enable_perf_memo and key in self.perf_memo:
            acc, lat = self.perf_memo[key]
        else:
            if key in self.perf_dict:
                acc, lat = self.perf_dict[key]
            else:
                interval_info.predictor_n_used_update()
                self.perf_predictor.eval()
                node_tsr, edge_list, _, last_idx_list = state.get_gpn_compatible_input(remove_self_edges=True)
                with torch.no_grad():
                    pred_acc, pred_lat = self.perf_predictor(node_tsr, edge_list, last_idx_list)
                pred_acc = get_tsr_value(pred_acc)
                pred_lat = get_tsr_value(pred_lat)
                acc, lat = pred_acc, pred_lat
                self.perf_memo[key] = (acc, lat)
                interval_info.num_block_eval += 1
        return acc, lat


class RandomEvaluator(BlockPerfEvaluator):

    def __init__(self, default_clue=1.0):
        super(RandomEvaluator, self).__init__(default_clue=default_clue)

    def get_perf_reward(self, state):
        pred_acc, pred_lat = self.get_perf_values(state)
        assert 0 <= pred_acc <= 1, "Acc value should be between 0 and 1, got {}".format(pred_acc)
        assert 0 <= pred_lat <= 1, "Lat value should be between 0 and 1, got {}".format(pred_lat)
        acc_reward = self.clue * pred_acc
        lat_reward = (1. - self.clue) * (1 - pred_lat)
        aggr_reward = acc_reward + lat_reward
        return aggr_reward

    def get_perf_values(self, state):
        key = state.get_str_id()
        if self.enable_perf_memo and key in self.perf_memo:
            acc, lat = self.perf_memo[key]
        else:
            acc = np.random.normal(0.5, 0.1, 1).item()
            acc = min(1.0, acc)
            acc = max(0.0, acc)
            lat = np.random.normal(0.5, 0.1, 1).item()
            lat = min(1.0, lat)
            lat = max(0.0, lat)
            self.perf_memo[key] = acc, lat
            interval_info.num_block_eval += 1
        return acc, lat


class ParetoClues:

    def __init__(self, enable, num_points, start=0., end=1.):
        self.enable = enable
        self.points = []
        self.num_points = num_points
        self.start = start
        self.end = end
        self.curr_idx = 0
        self.curr_point = 0
        self._gen_uniform_points()

    def _gen_uniform_points(self):
        if not self.enable:
            self.points = [self.end]
        else:
            self.points = np.linspace(self.start, self.end, self.num_points).tolist()[::-1]

    def get_curr_clue(self):
        return self.curr_point

    def __len__(self):
        return len(self.points)

    def __iter__(self):
        self.curr_idx = 0
        return self

    def __next__(self):
        if self.curr_idx < len(self.points):
            self.curr_point = self.points[self.curr_idx]
            self.curr_idx += 1
            return self.curr_point
        else:
            self.curr_idx = 0
            raise StopIteration


"""
Main training/evaluation routines and related functions
"""
def init_nas_bench_pareto_front(data_list, perf_key_list, perf_smaller_better_list,
                                scaffold_node_labels, scaffold_node_feature_tsr, book_keeper,
                                init_perf_rank=0.5, num_samples_per_batch=10,
                                ignore_init_perf_record=True,
                                allow_self_edge_for_non_input=False, perf_func=None):
    """
    :param data_list: data
    :param perf_key_list: list of str keys used to get performance metrics for each instance
    :param perf_smaller_better_list: list of bool indicating whether target at index i is smaller the better or not
    :param scaffold_node_labels: needed for building states
    :param scaffold_node_feature_tsr: needed for building states
    :param book_keeper:
    :param init_perf_rank: randomly take ratio of blocks that are bottom as init pareto front
    :param num_samples_per_batch: Pareto front optional settings
    :param ignore_init_perf_record:
    :param allow_self_edge_for_non_input:
    :param perf_func:
    :return: ParetoBlockSet
    """
    block2perfs = get_nas_bench_lower_rank_block2perfs(data_list, perf_key_list, perf_smaller_better_list,
                                                       scaffold_node_labels, scaffold_node_feature_tsr,
                                                       init_perf_rank=init_perf_rank, log_f=book_keeper.log,
                                                       perf_func=perf_func,
                                                       allow_self_edge_for_non_input=allow_self_edge_for_non_input)
    book_keeper.log("Initial block2perfs size: {}".format(len(block2perfs)))
    pareto_front = BlockParetoFront(block2perfs, perf_smaller_better_list,
                                    ignore_init_perf_record=ignore_init_perf_record,
                                    num_samples_per_batch=num_samples_per_batch, log_f=book_keeper.log)
    return pareto_front


def get_nas_bench_lower_rank_block2perfs(data_list, perf_key_list, perf_smaller_better_list,
                                         scaffold_node_labels, scaffold_node_feature_tsr,
                                         init_perf_rank=0.5, log_f=print, perf_func=None,
                                         allow_self_edge_for_non_input=False):
    # Finds the worst init_perf_rank% of blocks in terms of the performance specified in perf_key_list.
    assert len(perf_key_list) == len(perf_smaller_better_list)
    log_f("Candidate data size: {}".format(len(data_list)))
    log_f("init_perf_rank: {}".format(init_perf_rank))
    if perf_func is None:
        log_f("Initialize perfs using truth values!")
    else:
        log_f("Initialize perfs using provided function!")
    keep_length = int(len(data_list) * init_perf_rank)
    idx2data = {idx: data_dict for idx, data_dict in enumerate(data_list)}
    data_list = [(i, d) for i, d in enumerate(data_list)]
    key2sorted_indices = {}
    for ki, key in enumerate(perf_key_list):
        assert key not in key2sorted_indices
        key2sorted_indices[key] = sorted(data_list, key=lambda t: t[1][key], reverse=not perf_smaller_better_list[ki])
        key2sorted_indices[key] = key2sorted_indices[key][-keep_length:]
        key2sorted_indices[key] = [t[0] for t in key2sorted_indices[key]]
    common_set = None
    for key, sorted_indices in key2sorted_indices.items():
        indices = set(sorted_indices)
        if common_set is None:
            common_set = indices
        else:
            common_set = common_set.intersection(indices)
    selected_indices = list(common_set)
    pareto_data = [idx2data[i] for i in selected_indices]
    # need to build state from data dict, then convert to block2perfs format
    block2perfs = {}
    for data_dict in pareto_data:
        state = data_dict_to_state(data_dict, scaffold_node_labels, scaffold_node_feature_tsr,
                                   allow_self_edge_for_non_input)
        state_key = state.get_str_id()
        if perf_func is None:
            perfs = [data_dict[perf_key] for perf_key in perf_key_list]
        else:
            perfs = perf_func(state)
        block2perfs[state_key] = (state, perfs)
    return block2perfs


"""
Component makers
"""
def make_block_gpn(num_unique_nodes, gpn_in_channels, gpn_hidden_size, gpn_out_channels, gpn_gnn_constructor,
                   gpn_num_layers, gpn_dropout_prob):
    pre_embedding = torch.nn.Embedding(num_unique_nodes, gpn_in_channels)
    action_predictor = RNNActionPredictor(
        gpn_out_channels,
        gpn_out_channels,
        num_unique_nodes - 1, # Exclude input node
    )
    model = BlockGPN(gpn_in_channels, gpn_hidden_size, gpn_out_channels, gpn_gnn_constructor, action_predictor,
                     node_pre_embedding=pre_embedding, activ=torch.nn.Tanh(),
                     num_layers=gpn_num_layers, dropout_prob=gpn_dropout_prob)
    return model


def make_nas_bench_evaluator(perf_dict, default_acc, default_lat, default_clue=1.0):
    evaluator = NasBenchEvaluator(perf_dict, default_acc, default_lat, default_clue)
    return evaluator


def make_actor_critic(policy_net, gpn_out_channels):
    value_net = torch.nn.Sequential(
        NodeAggr(gpn_out_channels, aggr_dim=1, method="sum", enable_aggr_net=False),
        torch.nn.Linear(gpn_out_channels, 1)
    )
    actor_critic = GPNActorCritic("GPNActorCritic", policy_net, value_net)
    return actor_critic


def make_reward_modules(input_node_label, output_node_label, pos_step_reward, neg_step_reward,
                        perf_reward_module, pareto_reward_module, pos_validity_reward, neg_validity_reward,
                        enable_pareto_reward, enable_perf_reward):
    step_reward_module = StepRewardModule(input_node_label, output_node_label, pos_step_reward, neg_step_reward)
    final_reward_module = FinalRewardModule(perf_reward_module, pareto_reward_module, output_node_label,
                                            fixed_pos_reward=pos_validity_reward, fixed_neg_reward=neg_validity_reward,
                                            enable_pareto_reward=enable_pareto_reward,
                                            enable_perf_reward=enable_perf_reward)
    return step_reward_module, final_reward_module


def make_nas_env_manager(actor_critic, num_parallel_actors, init_state, node_label2feat,
                         step_reward_module, final_reward_module, max_allowed_steps, output_node_label,
                         add_self_edge=False):
    parallel_actors = [actor_critic] + [copy.deepcopy(actor_critic) for _ in range(num_parallel_actors - 1)]
    env_manager = NASEnvManager(parallel_actors, num_parallel_actors, init_state, node_label2feat,
                                step_reward_module, final_reward_module, max_allowed_steps, output_node_label,
                                add_self_edge=add_self_edge)
    return env_manager


"""
Train/Eval routines
"""
def ppo_train(env_manager, optimizer, value_criterion, n_train_rounds, k_epochs, t_timesteps, batch_size,
              reward_discount_factor, clip_epsilon, gae_lambda, c1, c2, c3, book_keeper,
              actor_update_episode_interval, parallel_explore=True):
    env_manager.train()
    ppo_data_loader = PPOBatchDataLoader(batch_size, reward_discount_factor=reward_discount_factor,
                                         gae_lambda=gae_lambda)
    block_w_rewards = UniqueList(key_func=lambda t: t[0].get_str_id())
    for i in range(n_train_rounds):
        book_keeper.log("Training round {} of {}".format(i + 1, n_train_rounds))
        book_keeper.log("Running {} actor(s), each for {} steps".format(len(env_manager), t_timesteps))
        book_keeper.log("Episodes per actor: {}".format(actor_update_episode_interval))
        # Exploration
        if parallel_explore:
            br_list = run_parallel_exploration(env_manager, ppo_data_loader, actor_update_episode_interval, t_timesteps,
                                               log_f=book_keeper.log)
        else:
            br_list = run_seq_exploration(env_manager, ppo_data_loader, actor_update_episode_interval, t_timesteps,
                                          log_f=book_keeper.log)
        block_w_rewards.extend(br_list)
        # Train generator(actor)
        updated_actor = env_manager.get_actor_rep()
        ppo_loss = run_k_ppo_batch_update_epoch(updated_actor, optimizer, ppo_data_loader, value_criterion,
                                                k_epochs, batch_size, clip_epsilon, c1, c2, c3)
        book_keeper.log("Avg PPO loss: {}".format(ppo_loss))
        env_manager.re_clone_actors(updated_actor)
        ppo_data_loader.clear()
    return env_manager.get_actor_rep(), sorted(block_w_rewards.tolist(), key=lambda t: t[1], reverse=True)


def run_seq_exploration(env_manager, data_loader, actor_update_episode_interval, t_timesteps, log_f=print):
    max_reward, avg_reward, min_reward, max_size = -float("inf"), 0., float("inf"), 0
    num_non_neg_reward_traj = 0
    num_pos_reward_traj = 0
    num_high_reward_traj = 0
    block_w_rewards = UniqueList(key_func=lambda t: t[0].get_str_id())
    start_time = time.time()
    for i in tqdm(range(actor_update_episode_interval), desc="Sequential exploration", ascii=True, leave=False):
        env_manager.reset()
        completed_actors = set()
        while len(completed_actors) < len(env_manager):
            for ei, env in enumerate(env_manager):
                actor = env.get_actor()
                if actor in completed_actors: continue
                end_reached, states, actions, rewards, end_state = run_actor_t_steps(env, actor, t_timesteps)
                if end_reached:
                    completed_actors.add(actor)
                    if rewards[-1] >= 0.:
                        num_non_neg_reward_traj += 1
                        num_pos_reward_traj += 1 if rewards[-1] > 0. else 0
                        num_high_reward_traj += 1 if rewards[-1] >= 0.5 else 0
                        block_w_rewards.append([end_state, rewards[-1]])
                with torch.no_grad():
                    input_state_values = [get_tsr_value(actor.get_value_for_state(s)) for s, _ in states]
                    if end_reached:
                        end_state_value = 0.
                    else:
                        end_state_value = get_tsr_value(actor.get_value_for_state(end_state))
                max_size = max(max_size, len(end_state))
                max_reward = max(max_reward, max(rewards))
                min_reward = min(min_reward, min(rewards))
                avg_reward += sum([get_tsr_value(r) for r in rewards])
                data_loader.add_trajectory(states, actions, rewards, input_state_values, end_state_value)

    log_f("Exploration elapsed seconds: {:.2f}".format(time.time() - start_time))
    data_loader.build_batches()
    avg_reward = avg_reward / data_loader.n_trajectories
    log_f("Dataloader n_trajectories: {}".format(data_loader.n_trajectories))
    log_f("Dataloader n_instances: {}".format(data_loader.n_instances))
    log_f("Dataloader n_batches: {}, max batch_size: {}".format(data_loader.n_batches, data_loader.batch_size))
    log_f("Max generated block size: {}".format(max_size))
    log_f("Max reward: {:.3f}".format(max_reward))
    log_f("Min reward: {:.3f}".format(min_reward))
    log_f("Avg reward: {:.3f}".format(avg_reward))
    log_f("Num traj with non-neg reward: {}".format(num_non_neg_reward_traj))
    log_f("Num traj with pos reward: {}".format(num_pos_reward_traj))
    log_f("Num traj with high reward: {}".format(num_high_reward_traj))
    log_f("Num unique valid end states: {}".format(len(block_w_rewards)))
    return block_w_rewards.tolist()


def run_parallel_exploration(env_manager, data_loader, actor_update_episode_interval, t_timesteps, log_f=print):
    import torch.multiprocessing as mp
    try:
        mp.set_start_method('spawn', force=True)
        mp.set_sharing_strategy('file_system')
    except RuntimeError:
        pass
    log_f("Parallel exploration...")
    env_manager.share_memory()
    manager = mp.Manager()
    traj_q = manager.Queue()
    start_time = time.time()
    processes = []
    for env in env_manager:
        arg_tuple = (traj_q, env, actor_update_episode_interval, t_timesteps)
        p = mp.Process(target=_parallel_explore, args=(arg_tuple,))
        p.start()
        processes.append(p)
    for p in processes:
        p.join()
    trajectories = []
    while not traj_q.empty():
        packed_states, packed_actions, rewards, input_state_values, end_state_value, end_state = traj_q.get()
        states, actions = [], []
        for si, so in packed_states:
            states.append((si.re_tensorize(), so.re_tensorize()))
        for action in packed_actions:
            actions.append(action.re_tensorize().to(device()))
        trajectories.append((states, actions, rewards, input_state_values, end_state_value, end_state.re_tensorize()))
    assert len(trajectories) > 0
    log_f("Number of trajectories: {}".format(len(trajectories)))
    block_w_rewards = UniqueList(key_func=lambda t: t[0].get_str_id())
    max_reward, avg_reward, min_reward, max_size = -float("inf"), 0., float("inf"), 0
    num_non_neg_reward_traj = 0
    num_pos_reward_traj = 0
    num_high_reward_traj = 0
    for states, actions, rewards, input_state_values, end_state_value, end_state in trajectories:
        final_reward = rewards[-1]
        block = end_state
        num_non_neg_reward_traj += 1 if final_reward >= 0. else 0
        num_pos_reward_traj += 1 if final_reward > 0. else 0
        num_high_reward_traj += 1 if final_reward >= 0.5 else 0
        if final_reward >= 0.:
            max_size = max(max_size, len(block))
            block_w_rewards.append([block, final_reward])
        max_reward = max(max_reward, max(rewards))
        min_reward = min(min_reward, min(rewards))
        avg_reward += sum([get_tsr_value(r) for r in rewards])
        data_loader.add_trajectory(states, actions, rewards, input_state_values, end_state_value)
    data_loader.build_batches()
    avg_reward = avg_reward / data_loader.n_trajectories
    log_f("Exploration elapsed seconds: {:.2f}".format(time.time() - start_time))
    log_f("Dataloader n_trajectories: {}".format(data_loader.n_trajectories))
    log_f("Dataloader n_instances: {}".format(data_loader.n_instances))
    log_f("Dataloader n_batches: {}, max batch_size: {}".format(data_loader.n_batches, data_loader.batch_size))
    log_f("Max generated block size: {}".format(max_size))
    log_f("Max reward: {:.3f}".format(max_reward))
    log_f("Min reward: {:.3f}".format(min_reward))
    log_f("Avg reward: {:.3f}".format(avg_reward))
    log_f("Num traj with non-neg reward: {}".format(num_non_neg_reward_traj))
    log_f("Num traj with pos reward: {}".format(num_pos_reward_traj))
    log_f("Num traj with high reward: {}".format(num_high_reward_traj))
    log_f("Num unique valid end states: {}".format(len(block_w_rewards)))
    return block_w_rewards.tolist()


def _parallel_explore(arg_tuple):
    traj_q, env, actor_update_episode_interval, t_timesteps = arg_tuple
    actor = env.get_actor()
    for i in range(actor_update_episode_interval):
        env.reset()
        end_reached, states, actions, rewards, end_state = run_actor_t_steps(env, actor, t_timesteps)
        assert end_reached
        input_state_values = [get_tsr_value(actor.get_value_for_state(s)) for s, _ in states]
        if end_reached:
            end_state_value = 0.
        else:
            end_state_value = get_tsr_value(actor.get_value_for_state(end_state))
        packed_states = []
        for si, so in states:
            packed_states.append((si.de_tensorize(), so.de_tensorize()))
        packed_actions = []
        for action in actions:
            packed_actions.append(action.de_tensorize())
        traj_q.put((packed_states, packed_actions, rewards, input_state_values,
                    end_state_value, end_state.de_tensorize()))


def gpn_generate_blocks(actor, env, max_n_steps, n_desired_blocks, output_node_label,
                        max_n_samples=None, max_n_attempts=2000, verbose=True, desc="Generating blocks",
                        valid_block_keys=None):
    n_unique_blocks_rv = n_desired_blocks
    if max_n_samples is not None:
        # If max_n_samples is set, n_desired_blocks will only be used to limit the return set
        n_desired_blocks = max_n_samples
        blocks = []
    else:
        blocks = UniqueList(key_func=lambda b: b.get_str_id())
    n_attempts = 0
    bar = None
    if verbose: bar = tqdm(total=n_desired_blocks, desc=desc, ascii=True)
    while n_attempts < max_n_attempts and len(blocks) < n_desired_blocks:
        env.reset()
        with torch.no_grad():
            for i in range(max_n_steps):
                action = actor.get_action_for_state(env.get_curr_state(), env.get_num_steps())
                final_state_reached, input_state, reward, new_state = env.step(action)
                if final_state_reached:
                    valid = is_block_valid(new_state, output_node_label)
                    if valid_block_keys is not None:
                        if hasattr(valid_block_keys, "get_id"):
                            key = valid_block_keys.get_id(new_state)
                        else:
                            key = new_state.get_str_id()
                        valid = valid and key in valid_block_keys
                    if max_n_samples is not None:
                        if valid:
                            blocks.append(new_state)
                            n_attempts = 0
                            if bar is not None: bar.update(1)
                    else:
                        if valid and blocks.append(new_state):
                            n_attempts = 0
                            if bar is not None: bar.update(1)
                    break
        n_attempts += 1
    env.reset()
    if bar is not None: bar.close()
    if isinstance(blocks, UniqueList):
        return blocks.tolist()
    else:
        block2count = collections.defaultdict(int)
        for block in blocks:
            block2count[block.get_str_id()] += 1
        unique_blocks = UniqueList(key_func=lambda b: b.get_str_id())
        unique_blocks.extend(blocks)
        block_w_counts = []
        for block in unique_blocks:
            block_w_counts.append((block, block2count[block.get_str_id()]))
        block_w_counts.sort(key=lambda t: t[1], reverse=True)
        block_w_counts = block_w_counts[:n_unique_blocks_rv]
        return [t[0] for t in block_w_counts]


def run_actor_t_steps(env, actor, t):
    final_state_reached = False
    states, actions, rewards = [], [], []
    for i in range(t):
        with torch.no_grad():
            actor.train()
            action = actor.get_action_for_state(env.get_curr_state(), env.get_num_steps())
        final_state_reached, input_state, reward, new_state = env.step(action)
        states.append( (input_state, new_state) )
        actions.append(action)
        rewards.append(reward)
        if final_state_reached:
            break
    return final_state_reached, states, actions, rewards, env.get_curr_state()


def compute_t_step_gae(rewards, input_state_values, end_state_value, discount_factor, lambda_):
    assert len(input_state_values) == len(rewards)
    assert len(rewards) > 0
    state_values = [v for v in input_state_values]
    state_values.append(end_state_value)

    def compute_delta(t):
        return rewards[t] + discount_factor * state_values[t + 1] - state_values[t]

    def compute_advantage(t):
        adv = 0.
        for i in range(t, len(rewards)):
            discounter = i - t
            multiplier = (discount_factor * lambda_) ** discounter
            delta_i = compute_delta(i)
            adv += delta_i * multiplier
        return adv

    advantages = [compute_advantage(t) for t in range(len(rewards))]
    assert not any(abs(t) == float("inf") for t in advantages), "Inf detected in advantages"
    return advantages


def run_k_ppo_batch_update_epoch(actor, optimizer, data_loader, value_criterion,
                                 k_epochs, batch_size, clip_epsilon, c1, c2, c3):
    actor.train()
    avg_loss, num_updates = 0., 0
    for e in tqdm(range(k_epochs), desc="PPO train", ascii=True, leave=False):
        for i, batch in enumerate(data_loader):
            batch_input_states, batch_old_actions, _, old_state_values, advantages = batch
            batch_old_actions = batch_old_actions.to(device())
            advantages = advantages.to(device())
            old_state_values = old_state_values.to(device())
            batch_new_actions = actor.get_batch_actions_for_states(batch_input_states)
            new_state_values = actor.get_batch_values_for_states(batch_input_states)
            old_node_logp, old_connection_logp = batch_old_actions.get_action_log_probs()
            new_node_logp, new_connection_logp = \
                batch_new_actions.get_action_log_probs(batch_node_indices=batch_old_actions.batch_node_indices,
                                                       batch_edge_connections=batch_old_actions.batch_edge_connections)
            node_ratio = torch.exp(new_node_logp - old_node_logp.detach())
            edge_ratio = torch.exp(new_connection_logp - old_connection_logp.detach()).mean(dim=-1)

            batch_ratios = torch.cat([node_ratio.unsqueeze(1), edge_ratio.unsqueeze(1)], dim=1).mean(dim=1)
            surr1 = batch_ratios * advantages
            surr2 = torch.clamp(batch_ratios, 1 - clip_epsilon, 1 + clip_epsilon) * advantages
            policy_loss = -1. * torch.min(surr1, surr2)
            policy_loss = policy_loss.mean()
            target_values = old_state_values + advantages
            value_loss = c1 * value_criterion(new_state_values, target_values).mean()
            node_entropy_loss = -1. * c2 * batch_new_actions.get_node_entropy()
            edge_entropy_loss = -1. * c3 * batch_new_actions.get_edge_entropy()
            batch_loss = policy_loss + value_loss + node_entropy_loss + edge_entropy_loss

            assert not torch.isnan(batch_loss).any(), "Nan detected in batch_loss"
            optimizer.zero_grad()
            batch_loss.backward()
            optimizer.step()
            avg_loss += batch_loss.item()
            num_updates += 1
        data_loader.shuffle()
    assert num_updates > 0, "data_loader len: {}, batch_size: {}".format(len(data_loader), batch_size)
    return avg_loss / num_updates


def evaluate_block_w_rewards(iteration, blocks_w_rewards, evaluator, pareto_front, output_node_label,
                             book_keeper, num_examples_to_print, report_validity_only=False):
    new_block2perfs, new_block2validity = {}, {}
    avg_reward = 0
    for block, reward in blocks_w_rewards:
        avg_reward += reward
        key = block.get_str_id()
        if key in new_block2validity: continue
        validity = get_validity_metrics(block, output_node_label)
        new_block2validity[key] = (block, validity)
        if not report_validity_only and is_block_valid(block):
            # NOTE: perf values should already be in memo else this could be slow!
            perf_values = evaluator.get_perf_values(block)
            new_block2perfs[key] = (block, perf_values)
    if len(blocks_w_rewards) > 0:
        avg_reward /= len(blocks_w_rewards)

    candidates = [(k, v) for k, v in new_block2validity.items()]
    if len(candidates) > num_examples_to_print:
        samples = random.sample(candidates, k=num_examples_to_print)
    else:
        samples = candidates
    for _, tup in samples:
        state, _ = tup
        book_keeper.log("Generated examples at iteration {}:".format(iteration))
        book_keeper.log(str(state))

    valid_counts = collections.defaultdict(float)
    for _, tup in new_block2validity.items():
        state, validity = tup
        for k, v in validity.items():
            valid_counts[k] += 1. if v else 0.
    valid_acc = {k: v / len(new_block2validity) for k, v in valid_counts.items()}
    book_keeper.log("Validity report at iteration {}: {}, num unique blocks: {}".format(iteration, valid_acc,
                                                                                        len(new_block2validity)))
    if report_validity_only:
        return avg_reward
    elif len(new_block2perfs) == 0:
        book_keeper.log("No valid blocks generated")
        return avg_reward
    else:
        book_keeper.log("Avg number of nodes in generated blocks: {}".format(
            mean([len(b) for _, (b, v) in new_block2perfs.items()])))
        book_keeper.log("Avg number of edges in generated blocks: {}".format(
            mean([len(b.curr_edge_list[0]) for _, (b, v) in new_block2perfs.items()])))
        gen_acc_values = [v[0] for _, (b, v) in new_block2perfs.items()]
        gen_lat_values = [v[1] for _, (b, v) in new_block2perfs.items()]
        book_keeper.log("Max acc in generated blocks: {}".format(max(gen_acc_values)))
        book_keeper.log("Min acc in generated blocks: {}".format(min(gen_acc_values)))
        book_keeper.log("Max lat in generated blocks: {}".format(max(gen_lat_values)))
        book_keeper.log("Min lat in generated blocks: {}".format(min(gen_lat_values)))
        best_perfs = pareto_front.update_pareto_front(new_block2perfs, log_f=book_keeper.log)
        book_keeper.checkpoint_state_dict(pareto_front.state_dict(),
                                          "{}_pareto_front.pkl".format(book_keeper.model_name))
        pareto_perf_values = pareto_front.get_pareto_block_perf_values()
        p_acc_values = [t[0] for t in pareto_perf_values]
        p_lat_values = [t[1] for t in pareto_perf_values]
        book_keeper.log("Max acc in current Pareto: {}".format(max(p_acc_values)))
        book_keeper.log("Min acc in current Pareto: {}".format(min(p_acc_values)))
        book_keeper.log("Max lat in current Pareto: {}".format(max(p_lat_values)))
        book_keeper.log("Min lat in current Pareto: {}".format(min(p_lat_values)))
        book_keeper.save_acc_lat_pareto_values(p_acc_values, p_lat_values, "iter {}".format(iteration))
        book_keeper.save_block2perfs_values(new_block2perfs, iteration)
        book_keeper.log(
            "New pareto front size: {}, new_block2perfs len: {}".format(len(pareto_front), len(new_block2perfs))
        )
        book_keeper.log("Avg pareto performance at iteration {}: {} ".format(iteration, best_perfs))
        return avg_reward


def get_pareto_l2_distances(perfs, pareto_front: BlockParetoFront):
    import math
    p_block_dists = []
    for p_block, p_perfs in pareto_front.get_pareto_blocks_with_perf():
        d = 0.
        for p_perf , perf in zip(p_perfs, perfs):
            d += (p_perf - perf) ** 2
        p_block_dists.append( (p_block, math.sqrt(d)) )
    p_block_dists.sort(key=lambda t: t[1])
    return p_block_dists

