import time
import copy
import random
import pickle
from tqdm import tqdm
from operations import *
from constants import *
from params import *
from model_nas_base import BlockPerfEvaluator
from torch.distributions import Categorical
from utils.darts_utils import AverageMeter, accuracy
from utils.misc_utils import label_tsr_to_one_hot_tsr, get_tsr_value, map_to_range
from utils.model_utils import device, can_parallel, config_bn
from utils.debug_utils import IntervalInfo as interval_info


NB101_PRIMITIVES = [
    'maxpool3x3',
    'conv1x1-bn-relu',
    'conv3x3-bn-relu'
]


NB201_PRIMITIVES = list(NB201_SEARCHABLE_OPS)



def build_input_inds(src_inds, dst_inds):
    assert len(src_inds) == len(dst_inds) > 0
    input_inds = [set() for _ in range(max(dst_inds) + 1)]
    for src, dst in zip(src_inds, dst_inds):
        if src == dst: continue # Ignore self-edges
        assert src < dst
        input_inds[dst].add(src)
    assert len(input_inds[0]) == 0
    assert all(len(v) > 0 for v in input_inds[1:])
    input_inds = [list(s) for s in input_inds][1:] # First node has no input to it, remove it
    return input_inds


def input_inds_to_edge_list(input_inds):
    src_inds, dst_inds = [], []
    for dst, src_list in enumerate(input_inds):
        for src in src_list:
            assert src < dst + 1
            src_inds.append(src)
            dst_inds.append(dst + 1) # Assuming input_inds is created with build_input_inds()
    return [src_inds, dst_inds]


def build_ext_samples_from_blocks(blocks, op_label2idx, output_node_idx=NB101_SUPER_BLOCK_OUTPUT_NODE_IDX):
    ext_samples = []
    for block in blocks:
        b_alphas = []
        for node_label in block.curr_node_label_list:
            if node_label not in op_label2idx:
                continue
            idx = op_label2idx[node_label]
            label_tsr = torch.LongTensor([[idx]])
            alphas = label_tsr_to_one_hot_tsr(label_tsr, len(op_label2idx)).squeeze()
            b_alphas.append(alphas)
        b_alphas.append([])
        node_inds = list(range(len(block.curr_node_label_list[1:-1])))
        node_inds.append(output_node_idx)
        input_inds = build_input_inds(block.curr_edge_list[0], block.curr_edge_list[1])
        assert len(b_alphas) == len(node_inds)
        ext_samples.append((node_inds, b_alphas, input_inds))
    return ext_samples


def fix_dangling_output(old_b_input_inds):
    # Manually connect any dangling ops to output node
    input_set = set()
    b_input_inds = copy.deepcopy(old_b_input_inds)
    for input_inds in b_input_inds:
        input_set.update(set(input_inds))
    dangling_nodes = [n+1 for n, _ in enumerate(b_input_inds) if n+1 not in input_set]
    assert len(dangling_nodes) >= 1
    for n in dangling_nodes:
        # By default, last node is the output node
        if n < len(old_b_input_inds):
            b_input_inds[-1].append(n)
            b_input_inds[-1].sort()
    return b_input_inds


def sample_discrete_alphas(num_ops, prob_tsr=None):
    if prob_tsr is not None:
        prob_tsr = prob_tsr.squeeze()
        assert len(prob_tsr.shape) == 1 and prob_tsr.shape[0] == num_ops
    else:
        prob_tsr = torch.FloatTensor([1. / num_ops] * num_ops)
    i = Categorical(probs=prob_tsr).sample()
    alphas = torch.zeros_like(prob_tsr)
    alphas[i] = 1.
    return alphas


def get_block_flops(model, block_config, C_in=3, W=32, H=32):
    from thop import profile
    model = model.to(device())
    rand_input = torch.randn(1, C_in, W, H).to(device())
    node_inds, b_alphas, input_inds = block_config
    macs, params = profile(model, inputs=(rand_input, node_inds, b_alphas, input_inds))
    return macs


class MixedOp(nn.Module):
    """A searchable node"""
    def __init__(self, C_op, stride_op, op_primitives):
        super(MixedOp, self).__init__()
        self.op_label2idx = {}
        self.ops = nn.ModuleList()
        self.num_modifiable_ops = len(op_primitives)
        self.primitives = op_primitives
        for pi, primitive in enumerate(self.primitives):
            self.op_label2idx[primitive] = pi
            op = OPS[primitive](C_op, stride_op, False)
            self.ops.append(op)

    def __len__(self):
        return len(self.ops)

    def sample_discrete_alphas(self, prob_tsr=None):
        return sample_discrete_alphas(len(self.ops), prob_tsr)

    def forward(self, x, alphas=None):
        if alphas is None:
            alphas = [1] * len(self.ops)
        # return sum(a * op(x) for a, op in zip(alphas, self._ops))
        alphas = get_tsr_value(alphas)
        op_inds = [i for i, a in enumerate(alphas) if a > 0]
        return sum(alphas[i] * self.ops[i](x) for i in op_inds)


class MixedOutput(nn.Module):
    """Output node"""
    def __init__(self):
        super(MixedOutput, self).__init__()

    def sample_discrete_alphas(self, *args):
        return []

    def forward(self, x, *args):
        return x


class SuperBlock(nn.Module):

    def __init__(self, C_in, C_b, num_searchable_nodes, op_primitives,
                 preprocess=Identity(), b_stride=1, add_C_proj=False):
        super(SuperBlock, self).__init__()
        self.C_in = C_in
        self.C_b = C_b
        self.b_stride = b_stride
        self.add_C_proj = add_C_proj
        self.num_searchable_nodes = num_searchable_nodes
        self.steps = self.num_searchable_nodes + 1
        self.preprocess = preprocess
        self.mixed_ops = nn.ModuleList()
        self.projections = nn.ModuleList()
        self.op_label2idx = None
        for i in range(self.num_searchable_nodes):
            mixed_op = MixedOp(self.C_b, stride_op=1, op_primitives=op_primitives)
            self.mixed_ops.append(mixed_op)
            self.op_label2idx = mixed_op.op_label2idx
        self.mixed_ops.append(MixedOutput())
        # Add projections for reduction
        if self.add_C_proj:
            for _ in range(len(self.mixed_ops)):
                proj = torch.nn.Sequential(
                    nn.Conv2d(self.C_in, self.C_b, kernel_size=1, stride=self.b_stride, padding=0),
                    nn.BatchNorm2d(self.C_b, affine=True))
                self.projections.append(proj)

    @property
    def num_modifiable_ops(self):
        if len(self.mixed_ops) <= 1: return 0
        return self.mixed_ops[0].num_modifiable_ops

    def __len__(self):
        return len(self.mixed_ops)

    @property
    def channel_size(self):
        return self.C_b

    def forward(self, x, node_inds, b_alphas, input_inds):
        states = [self.preprocess(x)]
        ops = [self.mixed_ops[i] for i in node_inds]
        assert isinstance(ops[-1], MixedOutput)
        for i, op in enumerate(ops):
            inputs = []
            for j in input_inds[i]:
                if self.add_C_proj and j == 0:
                    # apply reduction first if needed
                    inputs.append(self.projections[i](states[j]))
                else:
                    inputs.append(states[j])
            alphas = b_alphas[i]
            input_s = sum(inputs)
            s = op(input_s, alphas)
            states.append(s)
        return states[-1]


class NB101SuperNet(nn.Module):

    def __init__(self, C_input, C_init=128, n_classes=10, n_stacks=3, n_blocks_per_stack=3):
        super(NB101SuperNet, self).__init__()
        self.stem_layer = nn.Conv2d(C_input, C_init, kernel_size=3, padding=1, stride=1)
        self.blocks = nn.ModuleList()
        self.op_label2idx = None
        C_block = C_init
        for stack_i in range(n_stacks):
            for block_i in range(n_blocks_per_stack):
                block = SuperBlock(C_block, C_block, op_primitives=NB101_PRIMITIVES,
                                   num_searchable_nodes=NB101_NUM_INTERNAL_NODES_PER_BLOCK,
                                   preprocess=Identity())
                self.blocks.append(block)
                self.op_label2idx = block.op_label2idx
            if stack_i < n_stacks - 1:
                down_sample = nn.Sequential(nn.MaxPool2d(2, stride=2, padding=0),
                                            nn.Conv2d(C_block, C_block * 2, 1, stride=1))
                self.blocks.append(down_sample)
                C_block *= 2
        self.global_pooling = nn.AdaptiveAvgPool2d(1)
        self.classifier = nn.Linear(C_block, n_classes)

    def sample(self):
        assert len(self.blocks) > 0
        return self.blocks[0].sample()

    def forward(self, x, node_inds, b_alphas, input_inds):
        x = self.stem_layer(x)
        for bi, block in enumerate(self.blocks):
            if not isinstance(block, SuperBlock):
                x = block(x)
            else:
                x = block(x, node_inds, b_alphas, input_inds)
        out = self.global_pooling(x)
        logits = self.classifier(out.view(out.size(0), -1))
        return logits

    @property
    def summary(self):
        msg = "NB101SuperNet with {} block(s)\n".format(len(self.blocks))
        return msg


class SuperBlockSampler:

    def __init__(self, num_searchable_ops, output_node_idx,
                 topo_file=P_SEP.join([DATA_DIR, "random_max_n7_topos.pkl"]),
                 ext_samples=None, n_nodes_set=None, max_unique_samples=None):
        self.num_searchable_ops = num_searchable_ops
        self.output_node_idx = output_node_idx
        with open(topo_file, "rb") as f:
            self.topos = pickle.load(f)
        if n_nodes_set is not None:
            print("BlockSampler.n_nodes_set={}".format(n_nodes_set))
            filtered_topos = []
            for n_nodes, adj_list in self.topos:
                if n_nodes in n_nodes_set:
                    filtered_topos.append((n_nodes, adj_list))
            self.topos = filtered_topos
        self.ext_samples = ext_samples
        if max_unique_samples is not None:
            print("BlockSampler.max_unique_samples={}".format(max_unique_samples))
        self.max_unique_samples = max_unique_samples
        self.past_sample_keys = set()
        self.past_samples = []

    def get_largest_random_block(self):
        sorted_topos = sorted(self.topos, key=lambda t: (t[0], len(t[1][0])), reverse=True)
        n_nodes, largest_topo = sorted_topos[0]
        print("largest_topo contains {} edges".format(len(largest_topo[0])))
        return self._sample_supernet_block(n_nodes, largest_topo)

    def get_smallest_block(self):
        return [self.output_node_idx], [[]], [[0]]

    def _sample_supernet_block(self, n_nodes, adj_list):
        n_internal_nodes = n_nodes - 2
        assert n_internal_nodes >= 0
        node_inds = list(range(n_internal_nodes))
        b_alphas = []
        for i in range(n_internal_nodes):
            alphas = sample_discrete_alphas(self.num_searchable_ops)
            b_alphas.append(alphas)
        input_inds = build_input_inds(adj_list[0], adj_list[1])
        node_inds.append(self.output_node_idx)
        b_alphas.append([])
        return node_inds, b_alphas, input_inds

    def sample(self):
        if self.ext_samples is not None:
            return random.choice(self.ext_samples)

        if self.max_unique_samples is not None and len(self.past_samples) >= self.max_unique_samples:
            return random.choice(self.past_samples)

        n_nodes, adj_list = random.choice(self.topos)
        node_inds, b_alphas, input_inds = self._sample_supernet_block(n_nodes, adj_list)

        if self.max_unique_samples is not None:
            key = "#".join([str(node_inds), str(b_alphas), str(input_inds)])
            if key not in self.past_sample_keys:
                self.past_sample_keys.add(key)
                self.past_samples.append((node_inds, b_alphas, input_inds))
        return node_inds, b_alphas, input_inds


class SupernetEvaluator(BlockPerfEvaluator):

    def __init__(self, model, train_loader, test_loader, lat_predictor,
                 output_node_idx=NB101_SUPER_BLOCK_OUTPUT_NODE_IDX,
                 criterion=torch.nn.CrossEntropyLoss(), enable_perf_memo=True):
        super(SupernetEvaluator, self).__init__(enable_perf_memo=enable_perf_memo)
        self.model = model.to(device())
        self.train_loader = train_loader
        self.test_loader = test_loader
        self.criterion = criterion.to(device())
        self.lat_predictor = lat_predictor
        self.output_node_idx = output_node_idx

    def get_perf_reward(self, state):
        pred_acc, pred_lat = self.get_perf_values(state)
        assert 0 <= pred_acc <= 1, "Acc value should be between 0 and 1, got {}".format(pred_acc)
        assert 0 <= pred_lat <= 1, "Lat value should be between 0 and 1, got {}".format(pred_lat)
        acc_reward = self.clue * pred_acc
        lat_reward = (1. - self.clue) * (1 - pred_lat)
        aggr_reward = acc_reward + lat_reward
        return aggr_reward

    def get_perf_values(self, block):
        key = block.get_str_id()
        if self.enable_perf_memo and key in self.perf_memo:
            pred_acc, pred_lat = self.perf_memo[key]
        else:
            with torch.no_grad():
                self.model.eval()
                pred_acc, _ = run_supernet_inference(self.model, self.test_loader,
                                                     output_node_idx=self.output_node_idx,
                                                     op_labels=block.curr_node_label_list[1:],
                                                     src_inds=block.curr_edge_list[0],
                                                     dst_inds=block.curr_edge_list[1])
            pred_acc /= 100.
            pred_lat = self.lat_predictor(block)
            self.perf_memo[key] = (pred_acc, pred_lat)
            interval_info.num_block_eval += 1
        return pred_acc, pred_lat

    def train_on_family(self, blocks, epochs, optim_maker, verbose=True):
        self.model = self.model.to(device())
        optimizer = optim_maker(self.model.parameters())
        self.model.train()
        bar = None
        if verbose:
            bar = tqdm(range(epochs), desc="Training with {} blocks, train acc=0.".format(len(blocks)),
                       ascii=True)
        for i in range(epochs):
            avg_top_1_acc = AverageMeter()
            for si, (feature, target) in enumerate(self.train_loader):
                block = random.choice(blocks)
                node_inds, b_alphas, input_inds = block_to_supernet_configs(block.curr_node_label_list[1:],
                                                                            self.model.op_label2idx,
                                                                            block.curr_edge_list[0],
                                                                            block.curr_edge_list[1],
                                                                            self.output_node_idx)
                feature = feature.to(device())
                target = target.to(device())
                logits = supernet_inst_fwd(feature, target, self.model, node_inds, b_alphas, input_inds,
                                           criterion=self.criterion, optimizer=optimizer)
                p1 = accuracy(logits, target, topk=[1])
                avg_top_1_acc.update(p1[0].item(), feature.shape[0])
            train_acc = avg_top_1_acc.avg
            if verbose:
                bar.desc = "Training with {} blocks, train acc={:.2f}".format(len(blocks), train_acc)
                bar.update(1)
        if bar is not None:
            bar.close()


class SupernetFlopsCounter:
    """
    flops_counter = SupernetFlopsCounter(model, sampler.get_largest_random_block(),
                                         sampler.get_smallest_block(),
                                         HIAML_SUPER_BLOCK_OUTPUT_NODE_IDX)
    """
    def __init__(self, model, large_block, smallest_block, output_node_idx,
                 C_in=3, W=32, H=32, max_flop_op="conv3x3"):
        # Remember to run benchmark again if ops set ever changes!
        self.model = model
        self.output_node_idx = output_node_idx
        self.C_in = C_in
        self.W = W
        self.H = H
        self.max_flop_op = max_flop_op
        op_idx = self.model.op_label2idx[self.max_flop_op]
        alphas = label_tsr_to_one_hot_tsr(torch.LongTensor([[op_idx]]),
                                          vocab_size=len(self.model.op_label2idx)).squeeze()
        node_inds, _, input_inds = large_block
        b_alphas = [alphas] * len(node_inds)
        self.max_flops = get_block_flops(model, (node_inds, b_alphas, input_inds), C_in=C_in, W=W, H=H)
        self.min_flops = get_block_flops(model, smallest_block, C_in=C_in, W=W, H=H)
        print("Max flops: {}".format(self.max_flops))
        print("Min flops: {}".format(self.min_flops))

    def __call__(self, block):
        op_labels = block.curr_node_label_list[1:]
        op_label2idx = self.model.op_label2idx
        src_list, dst_list = block.curr_edge_list
        node_inds, b_alphas, input_inds = block_to_supernet_configs(op_labels, op_label2idx,
                                                                    src_list, dst_list, self.output_node_idx)
        flops = get_block_flops(self.model, (node_inds, b_alphas, input_inds),
                                C_in=self.C_in, W=self.W, H=self.H)
        flops = min(self.max_flops, flops)
        flops = max(self.min_flops, flops)
        flops = map_to_range(flops, self.min_flops, self.max_flops, 0., 1.)
        assert 0. <= flops <= 1., "Invalid flops: {}".format(flops)
        return flops


def make_nb101_supernet(C_input=3, C_block=16, n_classes=10):
    model = NB101SuperNet(C_input, C_init=C_block, n_classes=n_classes)
    config_bn(model)
    return model


def run_supernet_train_epoch(i, model, loader, sampler, criterion, optimizer,
                             log_f=print, plot_f=None, desc="Train",
                             max_grad_norm=5.0, global_unique_archs=None):
    avg_loss = AverageMeter()
    avg_top_1_acc = AverageMeter()
    avg_top_5_acc = AverageMeter()
    unique_archs = set()
    for feature, target in tqdm(loader, desc=desc, ascii=True, leave=False):
        feature = feature.to(device())
        target = target.to(device())
        node_inds, b_alphas, input_inds = sampler.sample()
        unique_archs.add("#".join([str(node_inds), str(b_alphas), str(input_inds)]))
        logits = model(feature, node_inds, b_alphas, input_inds)
        loss = criterion(logits, target)
        optimizer.zero_grad()
        loss.backward()
        nn.utils.clip_grad_norm_(model.parameters(), max_grad_norm)
        optimizer.step()
        p1, p5 = accuracy(logits, target, topk=(1, 5))
        avg_loss.update(loss.item(), feature.shape[0])
        avg_top_1_acc.update(p1.item(), feature.shape[0])
        avg_top_5_acc.update(p5.item(), feature.shape[0])
    msg = "{} epoch {} loss: {}, top-1 acc: {}, top-5 acc: {}".format(desc, i, avg_loss.avg,
                                                                      avg_top_1_acc.avg, avg_top_5_acc.avg)
    if global_unique_archs is not None:
        global_unique_archs.update(unique_archs)
    if log_f is not None:
        log_f(msg)
        log_f("Num unique arch sampled: {}".format(len(unique_archs)))
    if plot_f is not None:
        plot_f("{}/avg epoch loss".format(desc), avg_loss.avg, i)
        plot_f("{}/avg epoch top-1 acc".format(desc), avg_top_1_acc.avg, i)
        plot_f("{}/avg epoch top-5 acc".format(desc), avg_top_5_acc.avg, i)
    return avg_loss.avg, avg_top_1_acc.avg


def run_supernet_dev_epoch(i, model, dev_loader, sampler, criterion, warmup_loader=None, warmup_steps=10,
                           log_f=print, plot_f=None, desc="Dev", global_unique_archs=None):
    avg_loss = AverageMeter()
    avg_top_1_acc = AverageMeter()
    avg_top_5_acc = AverageMeter()
    unique_archs = set()
    for feature, target in tqdm(dev_loader, desc=desc, ascii=True, leave=False):
        feature = feature.to(device())
        target = target.to(device())
        node_inds, b_alphas, input_inds = sampler.sample()
        unique_archs.add("#".join([str(node_inds), str(b_alphas), str(input_inds)]))
        if warmup_loader is not None:
            warmup_model = warmup_model_for_block(model, warmup_loader, criterion,
                                                  node_inds, b_alphas, input_inds,
                                                  steps=warmup_steps)
        else:
            warmup_model = model
        with torch.no_grad():
            warmup_model.eval()
            logits = warmup_model(feature, node_inds, b_alphas, input_inds)
        loss = criterion(logits, target)
        p1, p5 = accuracy(logits, target, topk=(1, 5))
        avg_loss.update(loss.item(), feature.shape[0])
        avg_top_1_acc.update(p1.item(), feature.shape[0])
        avg_top_5_acc.update(p5.item(), feature.shape[0])
    msg = "{} epoch {} loss: {}, top-1 acc: {}, top-5 acc: {}".format(desc, i, avg_loss.avg,
                                                                      avg_top_1_acc.avg, avg_top_5_acc.avg)
    if global_unique_archs is not None:
        global_unique_archs.update(unique_archs)
    if log_f is not None:
        log_f(msg)
        log_f("Num unique arch sampled: {}".format(len(unique_archs)))
    if plot_f is not None:
        plot_f("{}/avg epoch loss".format(desc), avg_loss.avg, i)
        plot_f("{}/avg epoch top-1 acc".format(desc), avg_top_1_acc.avg, i)
        plot_f("{}/avg epoch top-5 acc".format(desc), avg_top_5_acc.avg, i)
    return avg_loss.avg, avg_top_1_acc.avg


def train_supernet(num_epochs, model, train_loader, sampler, criterion, optimizer, book_keeper,
                   max_grad_norm=5.0, completed_epochs=0, dev_loader=None, allow_parallel=False,
                   checkpoint=True, dev_warmup_steps=0, ext_dev_func=None):
    model = model.to(device())
    criterion = criterion.to(device())
    unique_archs = set()
    for epoch in range(num_epochs):
        report_epoch = epoch + completed_epochs + 1
        model.train()
        if allow_parallel and can_parallel():
            model = torch.nn.DataParallel(model)
        train_loss, train_acc = run_supernet_train_epoch(report_epoch, model, train_loader, sampler, criterion,
                                                         optimizer=optimizer, log_f=book_keeper.log,
                                                         desc="Train", max_grad_norm=max_grad_norm,
                                                         global_unique_archs=unique_archs)
        if isinstance(model, torch.nn.DataParallel):
            model = model.module
        book_keeper.log("Trained on {} unique archs so far".format(len(unique_archs)))
        if hasattr(optimizer, "schedule_lr"):
            optimizer.schedule_lr(epoch, log_f=book_keeper.log)
        elif ext_dev_func is None and dev_loader is None and hasattr(optimizer, "update_lr"):
            optimizer.update_lr(train_acc, log_f=book_keeper.log)
        if checkpoint:
            book_keeper.checkpoint_model("_latest.pt", report_epoch, model, optimizer)

        if ext_dev_func is not None:
            dev_score = ext_dev_func(model, report_epoch, train_loss, train_acc)
            if dev_score is not None:
                if hasattr(optimizer, "update_lr"):
                    optimizer.update_lr(dev_score, log_f=book_keeper.log)
                if checkpoint:
                    book_keeper.checkpoint_model("_best.pt", report_epoch, model, optimizer, eval_perf=dev_score)
                    book_keeper.report_curr_best()
        elif dev_loader is not None:
            _, dev_acc = run_supernet_dev_epoch(report_epoch, model, dev_loader, sampler, criterion,
                                                warmup_loader=train_loader, log_f=book_keeper.log,
                                                desc="Dev", warmup_steps=dev_warmup_steps)
            if hasattr(optimizer, "update_lr"):
                optimizer.update_lr(dev_acc, log_f=book_keeper.log)
            if checkpoint:
                book_keeper.checkpoint_model("_best.pt", report_epoch, model, optimizer, eval_perf=dev_acc)
                book_keeper.report_curr_best()
        book_keeper.log("")


def block_to_supernet_configs(op_labels, op_label2idx, src_list, dst_list,
                              output_node_idx=NB101_SUPER_BLOCK_OUTPUT_NODE_IDX):
    node_inds = list(range(len(op_labels) - 1))
    node_inds.append(output_node_idx)  # Must manually add output node
    b_alphas = []
    for label in op_labels:
        if label == "output":
            b_alphas.append([])
        else:
            op_idx = op_label2idx[label]
            alphas = label_tsr_to_one_hot_tsr(torch.LongTensor([[op_idx]]),
                                              vocab_size=len(op_label2idx))
            b_alphas.append(alphas.squeeze().to(device()))
    input_inds = build_input_inds(src_list, dst_list)
    return node_inds, b_alphas, input_inds


def run_supernet_inference(model, loader, op_labels=None, src_inds=None, dst_inds=None,
                           output_node_idx=NB101_SUPER_BLOCK_OUTPUT_NODE_IDX,
                           criterion=None, optimizer=None, max_grad_norm=5.0,
                           node_inds=None, b_alphas=None, input_inds=None,
                           verbose=False, step_limit=None, allow_parallel=False,
                           desc="Running supernet inference"):
    if node_inds is None or b_alphas is None or input_inds is None:
        if op_labels is None or src_inds is None or dst_inds is None:
            node_inds = list(range(output_node_idx + 1)) # fully connected
            b_alphas = [None for _ in node_inds]
            input_inds = [list(range(i + 1)) for i in range(len(node_inds))]
        else:
            node_inds, b_alphas, input_inds = block_to_supernet_configs(op_labels, model.op_label2idx,
                                                                        src_inds, dst_inds, output_node_idx)
    model = model.to(device())
    avg_top_1_acc = AverageMeter()
    bar = None
    if verbose:
        bar = tqdm(total=len(loader), desc=desc, ascii=True, leave=False)
    if allow_parallel and can_parallel():
        model = torch.nn.DataParallel(model)
    start_time = time.time()
    for si, (feature, target) in enumerate(loader):
        if step_limit is not None and si == step_limit:
            break
        feature = feature.to(device())
        target = target.to(device())
        logits = supernet_inst_fwd(feature, target, model, node_inds, b_alphas, input_inds,
                                   criterion=criterion, optimizer=optimizer, max_grad_norm=max_grad_norm)
        p1 = accuracy(logits, target, topk=[1])
        avg_top_1_acc.update(p1[0].item(), feature.shape[0])
        if bar is not None: bar.update(1)
    lat = time.time() - start_time
    if bar is not None: bar.close()
    return avg_top_1_acc.avg, lat


def warmup_model_for_block(model, loader, criterion, node_inds, b_alphas, input_inds,
                           steps=20, init_lr=0.0001):
    if steps == 0: return model
    model = copy.deepcopy(model).to(device())
    criterion = criterion.to(device())
    optimizer = torch.optim.Adam(model.parameters(), lr=init_lr)
    n_steps = 0
    model.train()
    for feature, target in loader:
        if n_steps == steps:
            return model
        feature = feature.to(device())
        target = target.to(device())
        supernet_inst_fwd(feature, target, model, node_inds, b_alphas, input_inds,
                          criterion=criterion, optimizer=optimizer)
        n_steps += 1
    return model


def supernet_inst_fwd(feature, target, model, node_inds, b_alphas, input_inds,
                      criterion=None, optimizer=None, max_grad_norm=5.0):
    logits = model(feature, node_inds, b_alphas, input_inds)
    if criterion is not None:
        loss = criterion(logits, target)
        if optimizer is not None:
            optimizer.zero_grad()
            loss.backward()
            nn.utils.clip_grad_norm_(model.parameters(), max_grad_norm)
            optimizer.step()
    return logits


class NB201SupernetBlockSampler:

    def __init__(self, num_searchable_ops, output_node_idx, n_nodes, edge_list):
        self.num_searchable_ops = num_searchable_ops
        self.output_node_idx = output_node_idx
        self.past_sample_keys = set()
        self.past_samples = []
        self.edge_list = edge_list
        self.n_nodes = n_nodes

    def _sample_supernet_block(self):
        n_internal_nodes = self.n_nodes - 2
        assert n_internal_nodes >= 0
        node_inds = list(range(n_internal_nodes))
        b_alphas = []
        for i in range(n_internal_nodes):
            alphas = sample_discrete_alphas(self.num_searchable_ops)
            b_alphas.append(alphas)
        input_inds = build_input_inds(self.edge_list[0], self.edge_list[1])
        node_inds.append(self.output_node_idx)
        b_alphas.append([])
        return node_inds, b_alphas, input_inds

    def sample(self):
        node_inds, b_alphas, input_inds = self._sample_supernet_block()
        return node_inds, b_alphas, input_inds


class NB201SuperNet(nn.Module):

    def __init__(self, C_input, C_init=16, n_classes=10, n_stacks=3, n_blocks_per_stack=5):
        super(NB201SuperNet, self).__init__()
        self.stem_layer = ConvBN(C_input, C_init, kernel_size=3, padding=1, stride=1, affine=False)
        self.blocks = nn.ModuleList()
        self.op_label2idx = None
        C_block = C_init
        for stack_i in range(n_stacks):
            for block_i in range(n_blocks_per_stack):
                block = SuperBlock(C_block, C_block, op_primitives=NB201_PRIMITIVES,
                                   num_searchable_nodes=NB201_SUPER_BLOCK_OUTPUT_NODE_IDX,
                                   preprocess=Identity())
                self.blocks.append(block)
                self.op_label2idx = block.op_label2idx
            if stack_i < n_stacks - 1:
                res_block = NB201ResidualBlock(C_block, 2 * C_block)
                self.blocks.append(res_block)
                C_block *= 2
        self.global_pooling = nn.AdaptiveAvgPool2d(1)
        self.classifier = nn.Linear(C_block, n_classes)

    def sample(self):
        assert len(self.blocks) > 0
        return self.blocks[0].sample()

    def forward(self, x, node_inds, b_alphas, input_inds):
        x = self.stem_layer(x)
        for bi, block in enumerate(self.blocks):
            if not isinstance(block, SuperBlock):
                x = block(x)
            else:
                x = block(x, node_inds, b_alphas, input_inds)
        out = self.global_pooling(x)
        logits = self.classifier(out.view(out.size(0), -1))
        return logits

    @property
    def summary(self):
        msg = "NB201SuperNet with {} block(s)\n".format(len(self.blocks))
        return msg


class NB201ResidualBlock(nn.Module):

    def __init__(self, C_in, C_out):
        super(NB201ResidualBlock, self).__init__()
        self.conv1 = nn.Conv2d(C_in, C_out, kernel_size=3, stride=2, padding=1)
        self.conv2 = nn.Conv2d(C_out, C_out, kernel_size=3, stride=1, padding=1)
        self.res = nn.Sequential(nn.AvgPool2d(2, stride=2, padding=0),
                                 nn.Conv2d(C_in, C_out, kernel_size=1, stride=1))

    def forward(self, x, *args):
        res_x = self.res(x)
        conv_x = self.conv2(self.conv1(x))
        return res_x + conv_x


def make_nb201_supernet(C_input=3, C_block=16, n_classes=10):
    model = NB201SuperNet(C_input, C_init=C_block, n_classes=n_classes)
    config_bn(model)
    return model
