from constants import *
import utils.model_utils as m_util
from model_helpers import BookKeeper, ExponentialDecayOptimizer, LabelSmoothingCrossEntropy
from model_efficientnet import *
from params import *
from utils.data_utils import *
from utils.model_utils import set_random_seed, add_weight_decay
import utils.debug_utils as d_util


def prepare_local_params(parser):
    parser.add_argument("-model_name", required=False, type=str,
                        default="eb0_supernet_imagenet_120")
    parser.add_argument("-imagenet_data_dir", required=False, type=str,
                        default=P_SEP.join([DATA_DIR, "ImageNet-224-120"]))
    parser.add_argument("-train_batch_size", required=False, type=int,
                        default=128)
    parser.add_argument("-test_batch_size", required=False, type=int,
                        default=128)
    parser.add_argument("-epochs", required=False, type=int,
                        default=500)
    parser.add_argument("-n_ops", required=False, type=int,
                        default=7)
    parser.add_argument("-dev_ratio", required=False, type=float,
                        default=0.0)
    parser.add_argument("-n_train_images", required=False, type=int,
                        default=NUM_TRAIN_IMAGES_IMAGENET_120)
    parser.add_argument("-lr_decay_epoch_interval", required=False, type=float,
                        default=2.4)
    parser.add_argument("-base_lr", required=False, type=float,
                        default=0.016)
    parser.add_argument("-rmsprop_eps", required=False, type=float,
                        default=0.1)
    parser.add_argument("-rmsprop_exp_decay_ratio", required=False, type=float,
                        default=0.99)
    return parser.parse_args()


def get_unique_labels(loader, desc="Recording unique class labels"):
    labels = set()
    for feature, target in tqdm(loader, desc=desc, ascii=True):
        target = target.tolist()
        for label in target:
            labels.add(label)
    return labels


def main(params):
    d_util.IntervalInfo.env = params.model_name
    book_keeper = BookKeeper(log_file_name=params.model_name + ".txt",
                             model_name=params.model_name, saved_model_dir=params.saved_models_dir,
                             init_eval_perf=float("-inf"), eval_perf_comp_func=lambda old, new: new > old,
                             saved_model_file=params.saved_model_file,
                             log_dir=params.logs_dir, plot_dir=params.plots_dir,
                             block_visual_dir=params.blocks_dir)
    book_keeper.log("Params: {}".format(params), verbose=False)
    set_random_seed(params.seed, log_f=book_keeper.log)
    train_trans, test_trans = get_efficientnet_train_test_transforms(224)
    # Dataloaders
    classes, train_loader, dev_loader, _ = get_imagenet_train_dev_test_set(name="ImageNet-224-120",
                                                                           data_folder=params.imagenet_data_dir,
                                                                           batch_size=params.train_batch_size,
                                                                           batch_size_test=params.test_batch_size,
                                                                           dev_ratio=params.dev_ratio,
                                                                           train_transform_override=train_trans,
                                                                           test_transform_override=test_trans)
    book_keeper.log("Number of classes: {}".format(classes))
    book_keeper.log("Train loader size: {}".format(len(train_loader)))
    book_keeper.log("Dev loader size: {}".format(len(dev_loader)))

    # Search space
    k_e_list = sorted(build_k_e_pairs(EB0_SEARCHABLE_OPS))
    book_keeper.log("Cand ops: {}".format(EB0_SEARCHABLE_OPS))
    book_keeper.log("Cand k_e_pairs: {}".format(k_e_list))

    # Make model
    model = make_eb0_sp_supernet(k_e_list, classes).to(device())

    book_keeper.log("Model name: {}".format(params.model_name))
    n_params = sum(p.numel() for p in model.parameters() if p.requires_grad)
    book_keeper.log("Number of trainable parameters: {}".format(n_params))
    book_keeper.log("Number of total parameters: {}".format(sum(p.numel() for p in model.parameters())))
    model_params = add_weight_decay(model, weight_decay=1e-5)
    # Optimizer
    init_lr = params.base_lr * (params.train_batch_size / 256.0)
    book_keeper.log("RMSProp eps: {}".format(params.rmsprop_eps))
    optimizer = torch.optim.RMSprop(model_params, lr=init_lr,
                                    alpha=0.9, eps=params.rmsprop_eps, weight_decay=0, momentum=0.9)
    lr_decay_interval = params.n_train_images / params.train_batch_size * params.lr_decay_epoch_interval
    optimizer = ExponentialDecayOptimizer(optimizer, init_lr, lr_decay_interval,
                                          decay_ratio=params.rmsprop_exp_decay_ratio,
                                          log_f=book_keeper.log)
    criterion = LabelSmoothingCrossEntropy(epsilon=0.1)
    ema = ExponentialMovingAverage(model.parameters(), decay=0.9999)
    if os.path.isfile(params.saved_model_file):
        book_keeper.load_model_optim_checkpoint(model, optimizer, checkpoint_file=params.saved_model_file,
                                                allow_silent_fail=False)
        ema.num_updates = optimizer.get_n_updates()
        params.train_steps -= ema.num_updates
    # Train
    book_keeper.log("Training")
    book_keeper.log("Num epochs: {}".format(params.epochs))
    train_eb0_sp_supernet(params.epochs, model, train_loader, criterion, optimizer,
                          params.n_ops, k_e_list, book_keeper, ema=ema, dev_loader=dev_loader)


if __name__ == "__main__":
    _parser = prepare_global_params()
    _args = prepare_local_params(_parser)
    m_util.DEVICE_STR_OVERRIDE = _args.device_str
    main(_args)
    print("done")

