import pickle
import utils.model_utils as m_util
import utils.debug_utils as d_util
import torch_geometric
from model_ga_nas import *
from params import *
from utils.misc_utils import one_hot_list_to_idx
from utils.debug_utils import DatasetInfo as di
from model_helpers import BookKeeper
from utils.model_utils import set_random_seed
from model_supernet import SupernetEvaluator, make_nb101_supernet
from utils.data_utils import get_cifar_train_dev_test_set


def prepare_local_params(parser):
    parser.add_argument("-model_name", required=False, type=str,
                        default="nb101_ga_nas_acc_ws")
    parser.add_argument("-supernet_checkpoint", required=False, type=str,
                        default=P_SEP.join([SAVED_MODELS_DIR, "nb101_supernet_c64_latest.pt"]))
    parser.add_argument("-supernet_C_block", required=False, type=int,
                        default=64)
    parser.add_argument("-data_seed", required=False, type=int,
                        default=12345)
    parser.add_argument("-search_seed", required=False, type=int,
                        default=12345)
    parser.add_argument("-initial_lr", required=False, type=float,
                        default=0.0001)
    parser.add_argument("-top_k", required=False, type=int,
                        default=50)
    parser.add_argument("-ppo_random_init_set_size", required=False, type=int,
                        default=100)
    parser.add_argument("-ppo_num_iterations", required=False, type=int,
                        default=5)
    parser.add_argument("-actor_update_episode_interval", required=False, type=int,
                        default=200)
    parser.add_argument("-ppo_n_train_rounds", required=False, type=int,
                        default=10)
    parser.add_argument("-ppo_k_epochs", required=False, type=int,
                        default=10)
    parser.add_argument("-ppo_num_desired_blocks_eval", required=False, type=int,
                        default=100)
    parser.add_argument("-ppo_num_desired_blocks_inc", required=False, type=int,
                        default=100)
    parser.add_argument("-ppo_num_block_samples_eval", required=False, type=int,
                        default=None)
    parser.add_argument("-ppo_num_block_samples_dis", required=False, type=int,
                        default=None)
    parser.add_argument("-ppo_num_parallel_actors", required=False, type=int,
                        default=4)
    parser.add_argument("-ppo_parallel_explore", action="store_true", required=False,
                        default=False)
    parser.add_argument("-ppo_gan_m_repeats", required=False, type=int,
                        default=1)
    parser.add_argument("-ppo_t_timesteps", required=False, type=int,
                        default=6)
    parser.add_argument("-ppo_max_generate_steps", required=False, type=int,
                        default=6)
    parser.add_argument("-ppo_batch_size", required=False, type=int,
                        default=4096)
    parser.add_argument("-n_random_eval", help="", type=int,
                        default=0, required=False)
    parser.add_argument("-random_mutated_truth", action="store_true", required=False,
                        default=False)
    parser.add_argument("-n_truth_per_src", help="", type=int,
                        default=100, required=False)
    parser.add_argument("-n_fake_total", help="", type=int,
                        default=400, required=False)
    parser.add_argument("-aggr_fake", action="store_true", required=False,
                        default=False)
    parser.add_argument("-ppo_reward_discount_factor", required=False, type=float,
                        default=1.0)
    parser.add_argument("-ppo_gae_lambda", required=False, type=float,
                        default=0.97)
    parser.add_argument("-ppo_clip_epsilon", required=False, type=float,
                        default=0.2)
    parser.add_argument("-ppo_c1", help="value loss", required=False, type=float,
                        default=0.5)
    parser.add_argument("-ppo_c2", help="node entropy", required=False, type=float,
                        default=0.1)
    parser.add_argument("-ppo_c3", help="edge entropy", required=False, type=float,
                        default=0.1)
    parser.add_argument("-ppo_pos_step_reward", required=False, type=float,
                        default=0.0)
    parser.add_argument("-ppo_neg_step_reward", required=False, type=float,
                        default=-0.1)
    parser.add_argument("-ppo_pos_validity_reward", required=False, type=float,
                        default=0.0)
    parser.add_argument("-ppo_neg_validity_reward",required=False,  type=float,
                        default=-0.1)
    parser.add_argument("-dis_in_channels", required=False, type=int,
                        default=128)
    parser.add_argument("-dis_hidden_size", required=False, type=int,
                        default=256)
    parser.add_argument("-dis_out_channels", required=False, type=int,
                        default=256)
    parser.add_argument("-dis_num_layers", required=False, type=int,
                        default=2)
    parser.add_argument("-dis_dropout_prob", required=False, type=float,
                        default=0.1)
    parser.add_argument("-dis_train_batch_size", required=False, type=int,
                        default=1024)
    parser.add_argument("-dis_train_epochs_per_iteration", required=False, type=int,
                        default=50)
    parser.add_argument("-dis_train_init_lr", required=False, type=float,
                        default=0.001)
    parser.add_argument("-gpn_in_channels", help="", type=int,
                        default=128, required=False)
    parser.add_argument("-gpn_hidden_size", help="", type=int,
                        default=256, required=False)
    parser.add_argument("-gpn_out_channels", help="", type=int,
                        default=256, required=False)
    parser.add_argument("-gpn_num_layers", help="", type=int,
                        default=2, required=False)
    parser.add_argument("-gpn_dropout_prob", help="", type=float,
                        default=0.0, required=False)
    return parser.parse_args()


def main(params):
    d_util.IntervalInfo.env = params.model_name
    book_keeper = BookKeeper(log_file_name=params.model_name + ".txt",
                             model_name=params.model_name,
                             saved_model_dir=params.saved_models_dir,
                             init_eval_perf=float("-inf"), eval_perf_comp_func=lambda old, new: new > old,
                             saved_model_file=params.saved_model_file,
                             log_dir=params.logs_dir, plot_dir=params.plots_dir,
                             block_visual_dir=params.blocks_dir)
    book_keeper.log("Params: {}".format(params), verbose=False)
    set_random_seed(params.data_seed, log_f=book_keeper.log)
    _, train_loader, dev_loader, _ = get_cifar_train_dev_test_set(name="CIFAR10", data_folder=DATA_DIR, dev_ratio=0.2)
    set_random_seed(params.search_seed, log_f=book_keeper.log)

    book_keeper.log("Loading NAS Bench acc lat dict")
    with open(P_SEP.join([CACHE_DIR, "nas_bench_acc_lat_dict.pkl"]), "rb") as f:
        perf_dict = pickle.load(f)
    book_keeper.log("{} block performance values loaded".format(len(perf_dict)))

    book_keeper.log("Reading NAS Bench PPO data")
    with open(P_SEP.join([CACHE_DIR, "nas_bench_ppo_data.pkl"]), "rb") as f:
        ppo_data = pickle.load(f)
    ppo_data_list = ppo_data[DATA]
    unique_node2feature = ppo_data[UNIQUE_NODE2FEATURE]
    unique_feature_list = ppo_data[UNIQUE_FEATURE_LIST]
    # Convert 1-hot to compact indices
    unique_feature_list = [(t[0], torch.LongTensor([one_hot_list_to_idx(t[1])])) for t in unique_feature_list]
    unique_node2feature = {k: torch.LongTensor([one_hot_list_to_idx(v)]) for k, v in unique_node2feature.items()}
    perf_key_list = [TEST_ACC_NAS_BENCH]
    perf_smaller_better_list = [False]
    scaffold_node_labels = [t[0] for t in unique_feature_list[1:]] # Drop the input node
    scaffold_node_tsr = torch.LongTensor([t[1] for t in unique_feature_list[1:]]).squeeze()
    book_keeper.log("Node features: {}".format(unique_node2feature))
    book_keeper.log("Performance metrics: {}".format(perf_key_list))
    book_keeper.log("Scaffold nodes: {}".format(scaffold_node_labels))

    # log the best/worst/avg perf for debugging
    for inst in ppo_data_list:
        di.update_acc_lat(inst[TEST_ACC_NAS_BENCH], inst[NORM_TRAIN_TIME_NAS_BENCH])
    di.self_check()
    book_keeper.log("Number of recorded data instances: {}".format(di.n_instances))
    book_keeper.log("Max acc: {}, avg_acc: {}, min acc in data: {}".format(di.max_acc, di.avg_acc, di.min_acc))
    book_keeper.log("Max lat: {}, avg_lat: {}, min lat in data: {}".format(di.max_lat, di.avg_lat, di.min_lat))

    # For debugging only
    book_keeper.log("Initializing truth acc set for comparison")
    truth_acc_set = init_nas_bench_top_k_acc_set(ppo_data_list, [TEST_ACC_NAS_BENCH], [False], params.top_k,
                                                 scaffold_node_labels, scaffold_node_tsr, book_keeper,
                                                 init_perf_rank=1.0)

    book_keeper.log("Initializing components for {}".format(params.model_name))

    # Make block generator
    gpn = make_block_gpn(num_unique_nodes=2 + len(NB101_SEARCHABLE_OPS),
                         gpn_in_channels=params.gpn_in_channels, gpn_hidden_size=params.gpn_hidden_size,
                         gpn_out_channels=params.gpn_out_channels, gpn_gnn_constructor=torch_geometric.nn.GraphConv,
                         gpn_num_layers=params.gpn_num_layers, gpn_dropout_prob=params.gpn_dropout_prob)

    # Wrap gpn into actor_critic
    actor_critic = make_actor_critic(policy_net=gpn, gpn_out_channels=params.gpn_out_channels)

    # Make discriminator
    discriminator = make_block_discriminator(num_unique_nodes=2 + len(NB101_SEARCHABLE_OPS),
                                             dis_in_channels=params.dis_in_channels,
                                             dis_hidden_size=params.dis_hidden_size,
                                             dis_out_channels=params.dis_out_channels,
                                             dis_gnn_constructor=torch_geometric.nn.GraphConv,
                                             dis_num_layers=params.dis_num_layers,
                                             dis_dropout_prob=params.dis_dropout_prob)

    supernet = make_nb101_supernet(C_block=params.supernet_C_block)
    book_keeper.load_model_checkpoint(supernet, checkpoint_file=params.supernet_checkpoint, allow_silent_fail=False,
                                      skip_eval_perfs=True)
    # Latency does not matter here, just use truth training time
    def _dummy_lat(_b):
        k = _b.get_str_id()
        return perf_dict[k][1] if k in perf_dict else di.avg_lat
    evaluator = SupernetEvaluator(supernet, train_loader, dev_loader, _dummy_lat,
                                  output_node_idx=NB101_SUPER_BLOCK_OUTPUT_NODE_IDX,
                                  criterion=torch.nn.CrossEntropyLoss(), enable_perf_memo=False)

    book_keeper.log("Initializing acc set for {}".format(params.model_name))
    top_acc_set = init_nas_bench_top_k_acc_set(ppo_data_list, [TEST_ACC_NAS_BENCH], [False], params.top_k,
                                               scaffold_node_labels, scaffold_node_tsr, book_keeper,
                                               random_init_set_size=params.ppo_random_init_set_size,
                                               init_perf_rank=params.nas_bench_pareto_init_perf_rank,
                                               perf_func=evaluator.get_perf_values)

    # Make step and final reward modules
    step_reward_module = StepRewardModule(NB101_INPUT_NODE_LABEL, NB101_OUTPUT_NODE_LABEL,
                                          params.ppo_pos_step_reward, params.ppo_neg_step_reward)
    dis_trainer = DisTrainer(params.dis_train_epochs_per_iteration, discriminator,
                             params.dis_train_init_lr, params.dis_train_batch_size)
    final_reward_module = DisRewardModule(dis_trainer, evaluator, NB101_OUTPUT_NODE_LABEL,
                                          params.n_random_eval, unique_node2feature, NB101_SEARCHABLE_OPS,
                                          n_truth_per_src=params.n_truth_per_src, n_fake_total=params.n_fake_total,
                                          enable_mutation=params.random_mutated_truth, aggr_fake=params.aggr_fake,
                                          max_n_nodes=1 + params.ppo_max_generate_steps,
                                          pos_reward=params.ppo_pos_validity_reward,
                                          neg_reward=params.ppo_neg_validity_reward)

    # Define initial state
    init_state = BlockState(curr_node_label_list=[NB101_INPUT_NODE_LABEL],
                            curr_node_feature_tsr=unique_node2feature[NB101_INPUT_NODE_LABEL],
                            scaffold_node_label_list=scaffold_node_labels,
                            scaffold_node_feature_tsr=scaffold_node_tsr,
                            curr_edge_list=[[0], [0]])

    # Make environments
    env_manager = make_nas_env_manager(actor_critic=actor_critic, num_parallel_actors=params.ppo_num_parallel_actors,
                                       init_state=init_state, node_label2feat=unique_node2feature,
                                       step_reward_module=step_reward_module, final_reward_module=final_reward_module,
                                       max_allowed_steps=params.ppo_max_generate_steps,
                                       output_node_label=NB101_OUTPUT_NODE_LABEL)

    actor_critic = env_manager.get_actor_rep()
    value_criterion = torch.nn.MSELoss()
    actor_optimizer = torch.optim.Adam(actor_critic.parameters(), lr=params.initial_lr)
    book_keeper.log(actor_critic)
    book_keeper.log("Actor critic for {}".format(params.model_name))
    n_params = sum(p.numel() for p in actor_critic.parameters() if p.requires_grad)
    book_keeper.log("Number of trainable parameters: {}".format(n_params))
    book_keeper.log(discriminator)
    book_keeper.log("Discriminator for {}".format(params.model_name))
    n_params = sum(p.numel() for p in discriminator.parameters() if p.requires_grad)
    book_keeper.log("Number of trainable parameters: {}".format(n_params))

    book_keeper.log("Training")
    try:
        ga_nas_ws_acc_search(env_manager=env_manager, ppo_optimizer=actor_optimizer,
                             value_criterion=value_criterion, top_acc_set=top_acc_set,
                             evaluator=evaluator, n_train_iterations=params.ppo_num_iterations,
                             n_train_rounds=params.ppo_n_train_rounds,
                             k_epochs=params.ppo_k_epochs, t_timesteps=params.ppo_t_timesteps,
                             batch_size=params.ppo_batch_size, m_repeats=params.ppo_gan_m_repeats,
                             max_actor_steps=params.ppo_max_generate_steps,
                             n_desired_blocks_eval=params.ppo_num_desired_blocks_eval,
                             n_block_samples_eval=params.ppo_num_block_samples_eval,
                             n_block_samples_dis=params.ppo_num_block_samples_dis,
                             n_desired_blocks_eval_inc=params.ppo_num_desired_blocks_inc,
                             output_node_label=NB101_OUTPUT_NODE_LABEL,
                             book_keeper=book_keeper,
                             parallel_explore=params.ppo_parallel_explore,
                             reward_discount_factor=params.ppo_reward_discount_factor,
                             gae_lambda=params.ppo_gae_lambda, clip_epsilon=params.ppo_clip_epsilon,
                             c1=params.ppo_c1, c2=params.ppo_c2, c3=params.ppo_c3,
                             actor_update_episode_interval=params.actor_update_episode_interval,
                             valid_block_keys=perf_dict)

        block_acc_pairs = top_acc_set.get_top_blocks_with_acc()
        block_acc_pairs.sort(key=lambda t: t[1], reverse=True)
        truth_block_info = {}
        truth_block_acc_pairs = truth_acc_set.get_complete_block_with_acc()
        truth_block_acc_pairs.sort(key=lambda t: t[1], reverse=True)
        book_keeper.log("Note: this ranking does not account for blocks with the same acc!")
        for i, (block, acc) in enumerate(truth_block_acc_pairs):
            truth_block_info[block.get_str_id()] = (i + 1, acc)
        for i, (block, acc) in enumerate(block_acc_pairs[:10]):
            book_keeper.log("Top searched block {}: {}".format(i + 1, str(block)))
            book_keeper.log("Searched acc: {}".format(acc))
            bk = block.get_str_id()
            if bk in truth_block_info:
                book_keeper.log("Truth rank: {}".format(truth_block_info[bk][0]))
                book_keeper.log("Truth acc: {}".format(truth_block_info[bk][1]))
            else:
                book_keeper.log("Searched block not in truth")
            book_keeper.log("")

    except KeyboardInterrupt:
        book_keeper.log("Training interrupted")


if __name__ == "__main__":
    _parser = prepare_global_params()
    _args = prepare_local_params(_parser)
    m_util.DEVICE_STR_OVERRIDE = _args.device_str
    main(_args)
    print("done")
