import utils.model_utils as m_util
from model_helpers import BookKeeper, LRDecayOptimizer
from model_supernet import *
from params import *
from utils.data_utils import get_cifar_train_dev_test_set
from utils.model_utils import set_random_seed


def prepare_local_params(parser):
    parser.add_argument("-model_name", required=False, type=str,
                        default="nb101_supernet")
    parser.add_argument("-batch_size", required=False, type=int,
                        default=128)
    parser.add_argument("-epochs", required=False, type=int,
                        default=500)
    parser.add_argument("-C_block", required=False, type=int,
                        default=64)
    parser.add_argument("-dev_ratio", required=False, type=float,
                        default=0.2)
    parser.add_argument("-lrd_initial_lr", required=False, type=float,
                        default=0.001)
    parser.add_argument("-lrd_min_lr", required=False, type=float,
                        default=0.001 * 0.1)
    parser.add_argument("-lrd_decay_ratio", required=False, type=float,
                        default=0.1)
    parser.add_argument("-lrd_decay_constant", required=False, type=float,
                        default=None)
    parser.add_argument("-lrd_n_allowed_fails", required=False, type=int,
                        default=5)
    parser.add_argument("-lrd_min_score_diff", required=False, type=float,
                        default=0.)
    parser.add_argument("-num_random_ext_blocks", required=False, type=int,
                        default=None)
    parser.add_argument("-ext_block_file", required=False, type=str,
                        default=None)
    parser.add_argument("-min_num_nodes", required=False, type=int,
                        default=None)
    parser.add_argument("-max_num_nodes", required=False, type=int,
                        default=None)
    parser.add_argument("-dev_warmup_steps", required=False, type=int,
                        default=0)
    parser.add_argument("-max_unique_samples", required=False, type=int,
                        default=None)
    parser.add_argument("-nas_bench_sample_space", action="store_true", required=False,
                        default=False)
    parser.add_argument("-enable_ext_dev", action="store_true", required=False,
                        default=False)
    return parser.parse_args()


def main(params):
    params.model_name += "_c{}".format(params.C_block)
    book_keeper = BookKeeper(log_file_name=params.model_name + ".txt",
                             model_name=params.model_name, saved_model_dir=params.saved_models_dir,
                             init_eval_perf=-float("inf"), eval_perf_comp_func=lambda old, new: new > old,
                             saved_model_file=params.saved_model_file, log_dir=params.logs_dir,
                             plot_dir=params.plots_dir, block_visual_dir=params.blocks_dir,
                             vis_env=params.model_name if params.enable_vis_plot else None)
    book_keeper.log("Params: {}".format(params), verbose=False)
    set_random_seed(params.seed, log_f=book_keeper.log)
    classes, train_loader, dev_loader, _ = get_cifar_train_dev_test_set(name="CIFAR10", data_folder=DATA_DIR,
                                                                        batch_size=params.batch_size,
                                                                        dev_ratio=params.dev_ratio)
    book_keeper.log("Train loader size: {}".format(len(train_loader)))

    model = make_nb101_supernet(C_block=params.C_block, n_classes=classes)

    criterion = torch.nn.CrossEntropyLoss()
    optimizer = torch.optim.Adam(model.parameters())
    optimizer = LRDecayOptimizer(optimizer, params.lrd_initial_lr, params.lrd_min_lr, params.lrd_min_score_diff,
                                 params.lrd_decay_ratio, params.lrd_decay_constant, params.lrd_n_allowed_fails)

    sampler = SuperBlockSampler(num_searchable_ops=len(NB101_SEARCHABLE_OPS),
                           output_node_idx=NB101_SUPER_BLOCK_OUTPUT_NODE_IDX,
                           max_unique_samples=params.max_unique_samples)

    book_keeper.log(model)
    book_keeper.log("Model name: {}".format(params.model_name))
    n_params = sum(p.numel() for p in model.parameters() if p.requires_grad)
    book_keeper.log("Number of trainable parameters: {}".format(n_params))

    book_keeper.log("Training")
    book_keeper.log("Num epochs: {}".format(params.epochs))
    try:
        train_supernet(params.epochs, model, train_loader, sampler, criterion, optimizer, book_keeper,
                       dev_loader=None, allow_parallel=params.allow_parallel,
                       dev_warmup_steps=params.dev_warmup_steps)

    except KeyboardInterrupt:
        book_keeper.log("Training interrupted")


if __name__ == "__main__":
    _parser = prepare_global_params()
    _args = prepare_local_params(_parser)
    m_util.DEVICE_STR_OVERRIDE = _args.device_str
    main(_args)
    print("done")