import pickle
import utils.model_utils as m_util
import utils.debug_utils as d_util
import torch_geometric
from model_ga_nas_nb201 import *
from params import *
from model_helpers import BookKeeper
from utils.model_utils import set_random_seed
from utils.darts_utils import nb201_topo_to_node_based_topo


def prepare_local_params(parser):
    parser.add_argument("-model_name", required=False, type=str,
                        default="nb201_ga_nas_acc_c10")
    parser.add_argument("-perf_dict_cache", required=False, type=str,
                        default=P_SEP.join([CACHE_DIR, "nb201_acc_flops_dict_c10.pkl"]))
    parser.add_argument("-initial_lr", required=False, type=float,
                        default=0.0001)
    parser.add_argument("-top_k", required=False, type=int,
                        default=40)
    parser.add_argument("-ppo_random_init_set_size", required=False, type=int,
                        default=60)
    parser.add_argument("-ppo_num_iterations", required=False, type=int,
                        default=5)
    parser.add_argument("-actor_update_episode_interval", required=False, type=int,
                        default=200)
    parser.add_argument("-ppo_n_train_rounds", required=False, type=int,
                        default=10)
    parser.add_argument("-ppo_k_epochs", required=False, type=int,
                        default=10)
    parser.add_argument("-ppo_num_desired_blocks_eval", required=False, type=int,
                        default=60)
    parser.add_argument("-ppo_num_desired_blocks_inc", required=False, type=int,
                        default=10)
    parser.add_argument("-ppo_num_parallel_actors", required=False, type=int,
                        default=4)
    parser.add_argument("-ppo_parallel_explore", action="store_true", required=False,
                        default=False)
    parser.add_argument("-ppo_gan_m_repeats", required=False, type=int,
                        default=1)
    parser.add_argument("-ppo_t_timesteps", required=False, type=int,
                        default=6)
    parser.add_argument("-ppo_max_generate_steps", required=False, type=int,
                        default=6)
    parser.add_argument("-ppo_batch_size", required=False, type=int,
                        default=4096)
    parser.add_argument("-n_random_eval", help="", type=int,
                        default=0, required=False)
    parser.add_argument("-random_mutated_truth", action="store_true", required=False,
                        default=False)
    parser.add_argument("-n_truth_per_src", help="", type=int,
                        default=100, required=False)
    parser.add_argument("-n_fake_total", help="", type=int,
                        default=400, required=False)
    parser.add_argument("-aggr_fake", action="store_true", required=False,
                        default=False)
    parser.add_argument("-ppo_reward_discount_factor", required=False, type=float,
                        default=1.0)
    parser.add_argument("-ppo_gae_lambda", required=False, type=float,
                        default=0.97)
    parser.add_argument("-ppo_clip_epsilon", required=False, type=float,
                        default=0.2)
    parser.add_argument("-ppo_c1", help="value loss", required=False, type=float,
                        default=0.5)
    parser.add_argument("-ppo_c2", help="node entropy", required=False, type=float,
                        default=0.1)
    parser.add_argument("-ppo_pos_step_reward", required=False, type=float,
                        default=0.0)
    parser.add_argument("-ppo_neg_step_reward", required=False, type=float,
                        default=-0.1)
    parser.add_argument("-ppo_pos_validity_reward", required=False, type=float,
                        default=0.0)
    parser.add_argument("-ppo_neg_validity_reward",required=False,  type=float,
                        default=-0.1)
    parser.add_argument("-default_pareto_clue", required=False, type=float,
                        default=1.0)
    parser.add_argument("-dis_in_channels", required=False, type=int,
                        default=128)
    parser.add_argument("-dis_hidden_size", required=False, type=int,
                        default=256)
    parser.add_argument("-dis_out_channels", required=False, type=int,
                        default=256)
    parser.add_argument("-dis_num_layers", required=False, type=int,
                        default=2)
    parser.add_argument("-dis_dropout_prob", required=False, type=float,
                        default=0.1)
    parser.add_argument("-dis_train_batch_size", required=False, type=int,
                        default=1024)
    parser.add_argument("-dis_train_epochs_per_iteration", required=False, type=int,
                        default=50)
    parser.add_argument("-dis_train_init_lr", required=False, type=float,
                        default=0.001)
    parser.add_argument("-gpn_in_channels", help="", type=int,
                        default=128, required=False)
    parser.add_argument("-gpn_hidden_size", help="", type=int,
                        default=256, required=False)
    parser.add_argument("-gpn_out_channels", help="", type=int,
                        default=256, required=False)
    parser.add_argument("-gpn_num_layers", help="", type=int,
                        default=2, required=False)
    parser.add_argument("-gpn_dropout_prob", help="", type=float,
                        default=0.0, required=False)
    return parser.parse_args()


def main(params):
    d_util.IntervalInfo.env = params.model_name
    book_keeper = BookKeeper(log_file_name=params.model_name + ".txt",
                             model_name=params.model_name,
                             saved_model_dir=params.saved_models_dir,
                             init_eval_perf=float("-inf"), eval_perf_comp_func=lambda old, new: new > old,
                             saved_model_file=params.saved_model_file,
                             log_dir=params.logs_dir, plot_dir=params.plots_dir,
                             block_visual_dir=params.blocks_dir)
    book_keeper.log("Params: {}".format(params), verbose=False)
    set_random_seed(params.seed, log_f=book_keeper.log)

    book_keeper.log("Loading NB201 acc flops dict")
    with open(params.perf_dict_cache, "rb") as f:
        perf_dict = pickle.load(f)

    node_label2feat = {k: torch.LongTensor([v]) for k, v in NB201_OP2IDX.items()}
    book_keeper.log("Node features: {}".format(node_label2feat))
    ni2dep_map = {}
    complete_edge_list = nb201_topo_to_node_based_topo(map_rv=ni2dep_map)
    init_state = get_nb201_init_block_state(NB201_INPUT_NODE_LABEL, node_label2feat, complete_edge_list)
    init_blocks = get_random_nb201_blocks(init_state, n_darts_edges=6, n_ops=len(NB201_SCAFFOLD_LABELS),
                                          scaffold_labels=NB201_SCAFFOLD_LABELS, node_label2feat=node_label2feat,
                                          n_blocks=params.ppo_random_init_set_size)

    evaluator = NB201Evaluator(perf_dict, ni2dep_map)
    top_acc_set = init_nb201_acc_set(init_blocks, evaluator, params.top_k, log_f=book_keeper.log)

    block_truth_acc_pairs = []
    for key, (acc, flops) in perf_dict.items():
        block_truth_acc_pairs.append((key, acc))
    block_truth_acc_pairs.sort(key=lambda t: t[1], reverse=True)
    block2truth_rank = {t[0]: i+1 for i, t in enumerate(block_truth_acc_pairs)}
    book_keeper.log("Truth top-3 acc: {}".format([t[1] for t in block_truth_acc_pairs[:3]]))

    # Make block generator
    gpn = make_nb201_gpn(num_unique_nodes=len(NB201_OP2IDX),
                         gpn_in_channels=params.gpn_in_channels, gpn_hidden_size=params.gpn_hidden_size,
                         gpn_out_channels=params.gpn_out_channels, gpn_gnn_constructor=torch_geometric.nn.GraphConv,
                         gpn_num_layers=params.gpn_num_layers, gpn_dropout_prob=params.gpn_dropout_prob)

    # Wrap gpn into actor_critic
    actor_critic = make_nb201_actor_critic(policy_net=gpn, gpn_out_channels=params.gpn_out_channels)

    # Make discriminator
    discriminator = make_nb201_block_discriminator(num_unique_nodes=len(NB201_OP2IDX),
                                                   dis_in_channels=params.dis_in_channels,
                                                   dis_hidden_size=params.dis_hidden_size,
                                                   dis_out_channels=params.dis_out_channels,
                                                   dis_gnn_constructor=torch_geometric.nn.GraphConv,
                                                   dis_num_layers=params.dis_num_layers,
                                                   dis_dropout_prob=params.dis_dropout_prob)

    dis_trainer = NB201DisTrainer(params.dis_train_epochs_per_iteration, discriminator,
                                  params.dis_train_init_lr, params.dis_train_batch_size)
    reward_module = NB201DisRewardModule(dis_trainer, evaluator, NB201_OUTPUT_NODE_LABEL,
                                         params.n_random_eval, node_label2feat, NB201_SEARCHABLE_OPS,
                                         n_truth_per_src=params.n_truth_per_src, n_fake_total=params.n_fake_total,
                                         enable_mutation=params.random_mutated_truth, aggr_fake=params.aggr_fake,
                                         pos_reward=params.ppo_pos_validity_reward,
                                         neg_reward=params.ppo_neg_validity_reward)

    # Make environments
    env_manager = make_nb201_env_manager(actor_critic=actor_critic, n_parallel_actors=params.ppo_num_parallel_actors,
                                         init_state=init_state, node_label2feat=node_label2feat,
                                         reward_module=reward_module,
                                         max_allowed_steps=params.ppo_max_generate_steps,
                                         output_node_label=NB201_OUTPUT_NODE_LABEL,
                                         scaffold_node_labels=NB201_SCAFFOLD_LABELS)

    actor_critic = env_manager.get_actor_rep()
    value_criterion = torch.nn.MSELoss()
    actor_optimizer = torch.optim.Adam(actor_critic.parameters(), lr=params.initial_lr)
    book_keeper.log(actor_critic)
    book_keeper.log("Actor critic for {}".format(params.model_name))
    n_params = sum(p.numel() for p in actor_critic.parameters() if p.requires_grad)
    book_keeper.log("Number of trainable parameters: {}".format(n_params))
    book_keeper.log(discriminator)
    book_keeper.log("Discriminator for {}".format(params.model_name))
    n_params = sum(p.numel() for p in discriminator.parameters() if p.requires_grad)
    book_keeper.log("Number of trainable parameters: {}".format(n_params))
    book_keeper.log("Training")
    try:
        nb201_ga_nas_acc_search(env_manager=env_manager, ppo_optimizer=actor_optimizer,
                                value_criterion=value_criterion, top_acc_set=top_acc_set,
                                evaluator=evaluator, n_train_iterations=params.ppo_num_iterations,
                                n_train_rounds=params.ppo_n_train_rounds,
                                k_epochs=params.ppo_k_epochs, t_timesteps=params.ppo_t_timesteps,
                                batch_size=params.ppo_batch_size, m_repeats=params.ppo_gan_m_repeats,
                                n_desired_blocks_eval=params.ppo_num_desired_blocks_eval,
                                n_desired_blocks_inc=params.ppo_num_desired_blocks_inc,
                                output_node_label=NB201_OUTPUT_NODE_LABEL,
                                book_keeper=book_keeper,
                                parallel_explore=params.ppo_parallel_explore,
                                reward_discount_factor=params.ppo_reward_discount_factor,
                                gae_lambda=params.ppo_gae_lambda, clip_epsilon=params.ppo_clip_epsilon,
                                c1=params.ppo_c1, c2=params.ppo_c2,
                                actor_update_episode_interval=params.actor_update_episode_interval)

        block_acc_pairs = top_acc_set.get_top_blocks_with_acc()
        block_acc_pairs.sort(key=lambda t: t[1], reverse=True)
        book_keeper.log("Note: this ranking does not account for blocks with the same acc!")
        for i, (block, acc) in enumerate(block_acc_pairs[:10]):
            book_keeper.log("Top searched block {}: {}".format(i + 1, str(block)))
            book_keeper.log("Searched acc: {}".format(acc))
            nb201_key = get_nb201_key(block.node_labels, ni2dep_map)
            book_keeper.log("Truth rank: {}".format(block2truth_rank[nb201_key]))
            book_keeper.log("Truth acc: {}".format(perf_dict[nb201_key][0]))
            book_keeper.log("")

    except KeyboardInterrupt:
        book_keeper.log("Search interrupted")


if __name__ == "__main__":
    _parser = prepare_global_params()
    _args = prepare_local_params(_parser)
    m_util.DEVICE_STR_OVERRIDE = _args.device_str
    main(_args)
    print("done")
