import utils.model_utils as m_util
from model_helpers import BookKeeper, nb201_key2labels
from model_supernet import *
from params import *
from utils.data_utils import get_cifar_train_dev_test_set
from utils.model_utils import set_random_seed
from utils.darts_utils import nb201_topo_to_node_based_topo
from scipy.stats import spearmanr


def prepare_local_params(parser):
    parser.add_argument("-model_name", required=False, type=str,
                        default="nb201_supernet_test")
    parser.add_argument("-supernet_checkpoint", required=False, type=str,
                        default=P_SEP.join([SAVED_MODELS_DIR, "nb201_cifar10_supernet_c16_latest.pt"]))
    parser.add_argument("-supernet_C_block", required=False, type=int,
                        default=16)
    parser.add_argument("-perf_dict_cache", required=False, type=str,
                        default=P_SEP.join([CACHE_DIR, "nb201_acc_flops_dict_c10.pkl"]))
    return parser.parse_args()


def main(params):
    book_keeper = BookKeeper(log_file_name=params.model_name + ".txt",
                             model_name=params.model_name, saved_model_dir=params.saved_models_dir,
                             init_eval_perf=0., eval_perf_comp_func=lambda old, new: new > old,
                             saved_model_file=params.saved_model_file, log_dir=params.logs_dir,
                             plot_dir=params.plots_dir, block_visual_dir=params.blocks_dir,
                             vis_env=params.model_name if params.enable_vis_plot else None)
    book_keeper.log("Params: {}".format(params), verbose=False)
    set_random_seed(params.seed, log_f=book_keeper.log)
    _, _, dev_loader, _ = get_cifar_train_dev_test_set(name="CIFAR10", data_folder=DATA_DIR, dev_ratio=0.2,
                                                       disable_train_transform=True)

    supernet = make_nb201_supernet(C_block=params.supernet_C_block)
    book_keeper.load_model_checkpoint(supernet, checkpoint_file=params.supernet_checkpoint, allow_silent_fail=False,
                                      skip_eval_perfs=True)

    book_keeper.log("Loading NB201 acc flops dict")
    with open(params.perf_dict_cache, "rb") as f:
        perf_dict = pickle.load(f)
    complete_edge_list = nb201_topo_to_node_based_topo()
    block_acc_list = [(b, acc) for b, (acc, flops) in perf_dict.items()]
    random.shuffle(block_acc_list)
    sorted_block_acc_list = sorted(block_acc_list, key=lambda t: t[1], reverse=True)
    block_perfs_seq = sorted_block_acc_list[:500]
    block_perfs_rand = block_acc_list[:2000]
    truth_acc_seq = [t[1] for t in block_perfs_seq]
    truth_acc_rand = [t[1] for t in block_perfs_rand]
    pred_acc_seq = []
    pred_acc_rand = []
    with torch.no_grad():
        supernet.eval()
        for key, _ in tqdm(block_perfs_rand, desc="Running supernet inferences", ascii=True):
            node_labels = nb201_key2labels(key)
            node_labels.append(NB201_OUTPUT_NODE_LABEL)
            src_inds, dst_inds = complete_edge_list
            acc, lat = run_supernet_inference(supernet, dev_loader, node_labels, src_inds, dst_inds,
                                              output_node_idx=NB201_SUPER_BLOCK_OUTPUT_NODE_IDX)
            pred_acc_rand.append(acc)
        for key, _ in tqdm(block_perfs_seq, desc="Running supernet inferences", ascii=True):
            node_labels = nb201_key2labels(key)
            node_labels.append(NB201_OUTPUT_NODE_LABEL)
            src_inds, dst_inds = complete_edge_list
            acc, lat = run_supernet_inference(supernet, dev_loader, node_labels, src_inds, dst_inds,
                                              output_node_idx=NB201_SUPER_BLOCK_OUTPUT_NODE_IDX)
            pred_acc_seq.append(acc)
    assert len(truth_acc_seq) == len(pred_acc_seq)
    assert len(truth_acc_rand) == len(pred_acc_rand)
    seq_acc_rho, seq_acc_p = spearmanr(truth_acc_seq, pred_acc_seq)
    rand_acc_rho, rand_acc_p = spearmanr(truth_acc_rand, pred_acc_rand)
    book_keeper.log("Seq acc rho: {}, p: {}".format(seq_acc_rho, seq_acc_p))
    book_keeper.log("Rand acc rho: {}, p: {}".format(rand_acc_rho, rand_acc_p))
    book_keeper.log("First 10 pred acc seq: {}".format(pred_acc_seq[:10]))
    book_keeper.log("First 10 pred acc rand: {}".format(pred_acc_rand[:10]))
    book_keeper.log("First 10 truth seq acc: {}".format(truth_acc_seq[:10]))
    book_keeper.log("First 10 truth rand acc: {}".format(truth_acc_rand[:10]))


if __name__ == "__main__":
    _parser = prepare_global_params()
    _args = prepare_local_params(_parser)
    m_util.DEVICE_STR_OVERRIDE = _args.device_str
    main(_args)
    print("done")
