import utils.model_utils as m_util
from model_helpers import BookKeeper, LRDecayOptimizer
from model_supernet import *
from params import *
from utils.darts_utils import nb201_topo_to_node_based_topo
from utils.data_utils import get_cifar_train_dev_test_set
from utils.model_utils import set_random_seed


def prepare_local_params(parser):
    parser.add_argument("-model_name", required=False, type=str,
                        default="nb201_cifar10_supernet")
    parser.add_argument("-batch_size", required=False, type=int,
                        default=128)
    parser.add_argument("-epochs", required=False, type=int,
                        default=400)
    parser.add_argument("-C_block", required=False, type=int,
                        default=16)
    parser.add_argument("-dev_ratio", required=False, type=float,
                        default=0.2)
    parser.add_argument("-lrd_initial_lr", required=False, type=float,
                        default=0.001)
    parser.add_argument("-lrd_min_lr", required=False, type=float,
                        default=0.001 * 0.01)
    parser.add_argument("-lrd_decay_ratio", required=False, type=float,
                        default=0.5)
    parser.add_argument("-lrd_decay_constant", required=False, type=float,
                        default=None)
    parser.add_argument("-lrd_n_allowed_fails", required=False, type=int,
                        default=20)
    parser.add_argument("-lrd_min_score_diff", required=False, type=float,
                        default=0.)
    return parser.parse_args()


def main(params):
    params.model_name += "_c{}".format(params.C_block)
    book_keeper = BookKeeper(log_file_name=params.model_name + ".txt",
                             model_name=params.model_name, saved_model_dir=params.saved_models_dir,
                             init_eval_perf=-float("inf"), eval_perf_comp_func=lambda old, new: new > old,
                             saved_model_file=params.saved_model_file, log_dir=params.logs_dir,
                             plot_dir=params.plots_dir, block_visual_dir=params.blocks_dir,
                             vis_env=params.model_name if params.enable_vis_plot else None)
    book_keeper.log("Params: {}".format(params), verbose=False)
    set_random_seed(params.seed, log_f=book_keeper.log)
    classes, train_loader, dev_loader, _ = get_cifar_train_dev_test_set(name="CIFAR10", data_folder=DATA_DIR,
                                                                        batch_size=params.batch_size,
                                                                        dev_ratio=params.dev_ratio)
    book_keeper.log("Train loader size: {}".format(len(train_loader)))

    model = make_nb201_supernet(C_block=params.C_block, n_classes=classes)

    criterion = torch.nn.CrossEntropyLoss()
    optimizer = torch.optim.Adam(model.parameters())
    optimizer = LRDecayOptimizer(optimizer, params.lrd_initial_lr, params.lrd_min_lr, params.lrd_min_score_diff,
                                 params.lrd_decay_ratio, params.lrd_decay_constant, params.lrd_n_allowed_fails)

    complete_edge_list = nb201_topo_to_node_based_topo()
    sampler = NB201SupernetBlockSampler(num_searchable_ops=len(NB201_SEARCHABLE_OPS),
                                        output_node_idx=NB201_SUPER_BLOCK_OUTPUT_NODE_IDX,
                                        n_nodes=8, edge_list=complete_edge_list)

    book_keeper.log(model)
    book_keeper.log("Model name: {}".format(params.model_name))
    n_params = sum(p.numel() for p in model.parameters() if p.requires_grad)
    book_keeper.log("Number of trainable parameters: {}".format(n_params))
    book_keeper.log(model.summary)
    book_keeper.log("Training")
    book_keeper.log("Num epochs: {}".format(params.epochs))
    g_start = time.time()
    try:
        train_supernet(params.epochs, model, train_loader, sampler, criterion, optimizer, book_keeper)

    except KeyboardInterrupt:
        book_keeper.log("Training interrupted")
    book_keeper.log("Train seconds: {}".format(time.time() - g_start))


if __name__ == "__main__":
    _parser = prepare_global_params()
    _args = prepare_local_params(_parser)
    m_util.DEVICE_STR_OVERRIDE = _args.device_str
    main(_args)
    print("done")
