import utils.model_utils as m_util
import utils.debug_utils as d_util
import torch_geometric
from model_ga_nas_darts import *
from libs.nb301_api import NB301
from model_darts import alphas_to_genotype
from params import *
from model_helpers import BookKeeper
from utils.model_utils import set_random_seed
from utils.darts_utils import darts_topo_to_node_based_topo


def prepare_local_params(parser):
    parser.add_argument("-model_name", required=False, type=str,
                        default="nb301_ga_nas_cons_acc")
    parser.add_argument("-max_runtime", required=False, type=int,
                        default=0)
    parser.add_argument("-initial_lr", required=False, type=float,
                        default=0.0001)
    parser.add_argument("-top_k", required=False, type=int,
                        default=50)
    parser.add_argument("-ppo_random_init_set_size", required=False, type=int,
                        default=100)
    parser.add_argument("-ppo_num_iterations", required=False, type=int,
                        default=10)
    parser.add_argument("-actor_update_episode_interval", required=False, type=int,
                        default=500)
    parser.add_argument("-ppo_n_train_rounds", required=False, type=int,
                        default=3)
    parser.add_argument("-ppo_k_epochs", required=False, type=int,
                        default=5)
    parser.add_argument("-ppo_num_desired_blocks_eval", required=False, type=int,
                        default=50)
    parser.add_argument("-ppo_num_desired_blocks_inc", required=False, type=int,
                        default=50)
    parser.add_argument("-ppo_num_parallel_actors", required=False, type=int,
                        default=4)
    parser.add_argument("-ga_nas_m_repeats", required=False, type=int,
                        default=1)
    parser.add_argument("-ppo_t_timesteps", required=False, type=int,
                        default=4)
    parser.add_argument("-ppo_max_generate_steps", required=False, type=int,
                        default=4)
    parser.add_argument("-ppo_batch_size", required=False, type=int,
                        default=128)
    parser.add_argument("-n_random_eval", help="", type=int,
                        default=0, required=False)
    parser.add_argument("-random_mutated_truth", action="store_true", required=False,
                        default=False)
    parser.add_argument("-n_truth_per_src", help="", type=int,
                        default=100, required=False)
    parser.add_argument("-n_fake_total", help="", type=int,
                        default=400, required=False)
    parser.add_argument("-aggr_fake", action="store_true", required=False,
                        default=False)
    parser.add_argument("-ppo_reward_discount_factor", required=False, type=float,
                        default=1.0)
    parser.add_argument("-ppo_gae_lambda", required=False, type=float,
                        default=1.0)
    parser.add_argument("-ppo_clip_epsilon", required=False, type=float,
                        default=0.2)
    parser.add_argument("-ppo_c1", help="value loss", required=False, type=float,
                        default=0.5)
    parser.add_argument("-ppo_c2", help="op entropy", required=False, type=float,
                        default=0.1)
    parser.add_argument("-ppo_c3", help="edge entropy", required=False, type=float,
                        default=0.1)
    parser.add_argument("-ppo_pos_step_reward", required=False, type=float,
                        default=0.0)
    parser.add_argument("-ppo_neg_step_reward", required=False, type=float,
                        default=-0.1)
    parser.add_argument("-ppo_pos_validity_reward", required=False, type=float,
                        default=0.0)
    parser.add_argument("-ppo_neg_validity_reward",required=False,  type=float,
                        default=-0.1)
    parser.add_argument("-dis_in_channels", required=False, type=int,
                        default=128)
    parser.add_argument("-dis_hidden_size", required=False, type=int,
                        default=256)
    parser.add_argument("-dis_out_channels", required=False, type=int,
                        default=256)
    parser.add_argument("-dis_num_layers", required=False, type=int,
                        default=2)
    parser.add_argument("-dis_dropout_prob", required=False, type=float,
                        default=0.1)
    parser.add_argument("-dis_train_batch_size", required=False, type=int,
                        default=256)
    parser.add_argument("-dis_train_epochs_per_iteration", required=False, type=int,
                        default=50)
    parser.add_argument("-dis_train_init_lr", required=False, type=float,
                        default=0.001)
    parser.add_argument("-gpn_in_channels", help="", type=int,
                        default=128, required=False)
    parser.add_argument("-gpn_hidden_size", help="", type=int,
                        default=256, required=False)
    parser.add_argument("-gpn_out_channels", help="", type=int,
                        default=256, required=False)
    parser.add_argument("-gpn_num_layers", help="", type=int,
                        default=2, required=False)
    parser.add_argument("-gpn_dropout_prob", help="", type=float,
                        default=0.0, required=False)
    parser.add_argument("-parallel_explore", action="store_true", required=False,
                        default=False)
    return parser.parse_args()


def main(params):
    params.model_name += "_seed{}".format(params.seed)
    d_util.IntervalInfo.env = params.model_name
    book_keeper = BookKeeper(log_file_name=params.model_name + ".txt",
                             model_name=params.model_name,
                             saved_model_dir=params.saved_models_dir,
                             init_eval_perf=float("-inf"), eval_perf_comp_func=lambda old, new: new > old,
                             saved_model_file=params.saved_model_file,
                             log_dir=params.logs_dir, plot_dir=params.plots_dir,
                             block_visual_dir=params.blocks_dir)
    book_keeper.log("Params: {}".format(params), verbose=False)
    set_random_seed(params.seed, log_f=book_keeper.log)
    if params.max_runtime > 0:
        book_keeper.log("Specified max runtime: {}".format(params.max_runtime))
    else:
        book_keeper.log("Performing unconstrained acc search!")

    node_label2feat = {k: torch.LongTensor([v]) for k, v in DARTS_OP2IDX.items()}
    book_keeper.log("Node features: {}".format(node_label2feat))
    ni2de_map = {} # NB101 node index in graph to darts edge
    complete_nb101_edge_list = darts_topo_to_node_based_topo(map_rv=ni2de_map)
    de2ni_map = {"{}->{}".format(src, dst): k for k, (src, dst) in ni2de_map.items()}
    assert len(ni2de_map) == len(de2ni_map)

    api = NB301()

    def get_runtime(_block):
        _normal_d_alphas = darts_cell_to_discrete_alphas(_block.normal_op_labels, _block.normal_op_edges,
                                                         _block.de2ni_map, DARTS_OP2IDX)
        _reduce_d_alphas = darts_cell_to_discrete_alphas(_block.reduce_op_labels, _block.reduce_op_edges,
                                                         _block.de2ni_map, DARTS_OP2IDX)
        _genotype = alphas_to_genotype(n_steps=4, alphas_normal=_normal_d_alphas, alphas_reduce=_reduce_d_alphas)
        return api.get_geno_runtime(_genotype)

    evaluator = NB301Evaluator(api, DARTS_OP2IDX, pred_runtime=False)

    init_state = get_darts_init_block_state(DARTS_INPUT_NODE_LABEL, DARTS_OP2IDX, de2ni_map, complete_nb101_edge_list)

    init_blocks = get_random_darts_blocks(init_state, n_internal_cell_nodes=4, n_input_edges_per_node=2,
                                          n_ops=len(DARTS_SCAFFOLD_LABELS),
                                          scaffold_labels=DARTS_SCAFFOLD_LABELS, op2idx=DARTS_OP2IDX,
                                          n_blocks=params.ppo_random_init_set_size,
                                          max_cons_score=params.max_runtime,
                                          cons_predictor=get_runtime if params.max_runtime > 0 else None)

    top_acc_set = init_darts_acc_set(init_blocks, evaluator, params.top_k, log_f=book_keeper.log)

    # Make block generator
    gpn = make_darts_gpn(num_unique_nodes=len(DARTS_OP2IDX),
                         gpn_in_channels=params.gpn_in_channels, gpn_hidden_size=params.gpn_hidden_size,
                         gpn_out_channels=params.gpn_out_channels, gpn_gnn_constructor=torch_geometric.nn.GraphConv,
                         gpn_num_layers=params.gpn_num_layers, gpn_dropout_prob=params.gpn_dropout_prob)

    # Wrap gpn into actor_critic
    actor_critic = make_darts_actor_critic(policy_net=gpn, gpn_out_channels=params.gpn_out_channels)

    # Make discriminator
    discriminator = make_darts_block_discriminator(num_unique_nodes=len(DARTS_OP2IDX),
                                                   dis_in_channels=params.dis_in_channels,
                                                   dis_hidden_size=params.dis_hidden_size,
                                                   dis_out_channels=params.dis_out_channels,
                                                   dis_gnn_constructor=torch_geometric.nn.GraphConv,
                                                   dis_num_layers=params.dis_num_layers,
                                                   dis_dropout_prob=params.dis_dropout_prob)

    dis_trainer = DartsDisTrainer(params.dis_train_epochs_per_iteration, discriminator,
                                  params.dis_train_init_lr, params.dis_train_batch_size)
    reward_module = DartsDisRewardModule(dis_trainer, DARTS_OUTPUT_NODE_LABEL,
                                         params.n_random_eval, node_label2feat, DARTS_PRIMITIVES,
                                         n_truth_per_src=params.n_truth_per_src, n_fake_total=params.n_fake_total,
                                         enable_mutation=params.random_mutated_truth, aggr_fake=params.aggr_fake,
                                         max_n_nodes=1 + max(complete_nb101_edge_list[1]),
                                         pos_reward=params.ppo_pos_validity_reward,
                                         neg_reward=params.ppo_neg_validity_reward)

    # Make environments
    env_manager = make_darts_env_manager(actor_critic=actor_critic, n_parallel_actors=params.ppo_num_parallel_actors,
                                         init_state=init_state, op2idx=DARTS_OP2IDX,
                                         reward_module=reward_module,
                                         max_allowed_steps=params.ppo_max_generate_steps,
                                         output_node_label=DARTS_OUTPUT_NODE_LABEL,
                                         scaffold_node_labels=DARTS_SCAFFOLD_LABELS)

    actor_critic = env_manager.get_actor_rep()
    value_criterion = torch.nn.MSELoss()
    actor_optimizer = torch.optim.Adam(actor_critic.parameters(), lr=params.initial_lr)
    book_keeper.log(actor_critic)
    book_keeper.log("Actor critic for {}".format(params.model_name))
    n_params = sum(p.numel() for p in actor_critic.parameters() if p.requires_grad)
    book_keeper.log("Number of trainable parameters: {}".format(n_params))
    book_keeper.log(discriminator)
    book_keeper.log("Discriminator for {}".format(params.model_name))
    n_params = sum(p.numel() for p in discriminator.parameters() if p.requires_grad)
    book_keeper.log("Number of trainable parameters: {}".format(n_params))
    book_keeper.log("Training")
    ga_nas_darts_cons_acc_search(env_manager=env_manager, ppo_optimizer=actor_optimizer,
                                  value_criterion=value_criterion, top_acc_set=top_acc_set,
                                  evaluator=evaluator, n_train_iterations=params.ppo_num_iterations,
                                  n_train_rounds=params.ppo_n_train_rounds,
                                  k_epochs=params.ppo_k_epochs, t_timesteps=params.ppo_t_timesteps,
                                  batch_size=params.ppo_batch_size, m_repeats=params.ga_nas_m_repeats,
                                  n_desired_blocks_eval=params.ppo_num_desired_blocks_eval,
                                  n_desired_blocks_inc=params.ppo_num_desired_blocks_inc,
                                  max_cons_score=params.max_runtime,
                                  cons_predictor=get_runtime if params.max_runtime > 0 else None,
                                  book_keeper=book_keeper,
                                  parallel_explore=params.parallel_explore,
                                  reward_discount_factor=params.ppo_reward_discount_factor,
                                  gae_lambda=params.ppo_gae_lambda, clip_epsilon=params.ppo_clip_epsilon,
                                  c1=params.ppo_c1, c2=params.ppo_c2, c3=params.ppo_c3,
                                  actor_update_episode_interval=params.actor_update_episode_interval)

    best_blocks = top_acc_set.get_top_blocks_with_acc()
    best_blocks.sort(key=lambda t: t[1], reverse=True)
    for bi, (block, pred_acc) in enumerate(best_blocks[:10]):
        normal_d_alphas = darts_cell_to_discrete_alphas(block.normal_op_labels, block.normal_op_edges,
                                                        block.de2ni_map, DARTS_OP2IDX)
        reduce_d_alphas = darts_cell_to_discrete_alphas(block.reduce_op_labels, block.reduce_op_edges,
                                                        block.de2ni_map, DARTS_OP2IDX)
        genotype = alphas_to_genotype(n_steps=4, alphas_normal=normal_d_alphas, alphas_reduce=reduce_d_alphas)
        book_keeper.log("Best block {}: {}".format(bi + 1, genotype))
        book_keeper.log("Pred acc: {}".format(pred_acc))
        book_keeper.log("")


if __name__ == "__main__":
    _parser = prepare_global_params()
    _args = prepare_local_params(_parser)
    m_util.DEVICE_STR_OVERRIDE = _args.device_str
    main(_args)
    print("done")
