import os
import numpy as np
import torch
import shutil
from pathlib import Path


class AverageMeter(object):

    def __init__(self):
        self.avg = 0
        self.sum = 0
        self.cnt = 0

    def reset(self):
        self.avg = 0
        self.sum = 0
        self.cnt = 0
        return self

    def update(self, val, n=1):
        self.sum += val * n
        self.cnt += n
        self.avg = self.sum / self.cnt


def accuracy(output, target, topk=(1,)):
    maxk = max(topk)
    batch_size = target.size(0)
    _, pred = output.topk(maxk, 1, True, True)
    pred = pred.t()
    correct = pred.eq(target.view(1, -1).expand_as(pred))
    res = []
    for k in topk:
        correct_k = correct[:k].view(-1).float().sum(0)
        res.append(correct_k.mul_(100.0 / batch_size))
    return res


class Cutout(object):
    def __init__(self, length):
        self.length = length

    def __call__(self, img):
        h, w = img.size(1), img.size(2)
        mask = np.ones((h, w), np.float32)
        y = np.random.randint(h)
        x = np.random.randint(w)

        y1 = np.clip(y - self.length // 2, 0, h)
        y2 = np.clip(y + self.length // 2, 0, h)
        x1 = np.clip(x - self.length // 2, 0, w)
        x2 = np.clip(x + self.length // 2, 0, w)

        mask[y1: y2, x1: x2] = 0.
        mask = torch.from_numpy(mask)
        mask = mask.expand_as(img)
        img *= mask
        return img


# Note: This came with the original repo - I actually don't believe its accurate
def count_parameters_in_MB(model):
    return np.sum(np.prod(v.size()) for name, v in model.named_parameters() if "auxiliary" not in name) / 1e6


# https://stackoverflow.com/questions/48393608/pytorch-network-parameter-calculation
# Modified to only print the final number of parameters.
def count_parameters(model):
    total_param = 0
    for name, param in model.named_parameters():
        if param.requires_grad:
            num_param = np.prod(param.size())
            total_param += num_param
    return total_param


def save_checkpoint(state, is_best, save):
    filename = os.path.join(save, 'checkpoint.pth.tar')
    torch.save(state, filename)
    if is_best:
        best_filename = os.path.join(save, 'model_best.pth.tar')
        shutil.copyfile(filename, best_filename)


def save(model, model_path):
    torch.save(model.state_dict(), model_path)


def load(model, model_path):
    # state_dict = torch.load(model_path)
    # # create new ordererDict that does not contain 'module'
    # from collections import OrderedDict
    # new_state_dict = OrderedDict()
    # for k, v in state_dict.items():
    #   namekey = k[7:] # remove 'module'
    #   new_state_dict[namekey] = v
    # # load params
    model.load_state_dict(torch.load(model_path), strict=False)


def drop_path(x, drop_prob):
    if drop_prob > 0.:
        keep_prob = 1. - drop_prob
        mask = torch.cuda.FloatTensor(x.size(0), 1, 1, 1).bernoulli_(keep_prob)
        x.div_(keep_prob)
        x.mul_(mask)
    return x


def create_exp_dir(path, scripts_to_save=None):
    if not os.path.exists(path):
        Path(path).mkdir(parents=True, exist_ok=True)
    print('Experiment dir : {}'.format(path))

    if scripts_to_save is not None:
        os.mkdir(os.path.join(path, 'scripts'))
        for script in scripts_to_save:
            dst_file = os.path.join(path, 'scripts', os.path.basename(script))
            shutil.copyfile(script, dst_file)


def darts_topo_to_node_based_topo(n=7, n_input_nodes=2, verbose=True, log_f=print,
                                  map_rv=None):
    """
    Converts DARTS cell into a topology where each node is an operator
    :param n:
    :param n_input_nodes:
    :param verbose:
    :param log_f:
    :param map_rv: mapping between node indices to its input/output DARTS node
    :return: [src_inds, dst_inds]
    """
    from collections import namedtuple
    Operator = namedtuple("Operator", ["edge_based_label", "darts_src", "darts_dst", "connected_to_input"])
    if verbose:
        log_f("Converting the DARTS cell with N={} to the node-as-operator topology".format(n))
    n_intermediate_nodes = n - n_input_nodes - 1 # Exclude input and output nodes
    assert n_intermediate_nodes > 0
    # First get the edges between input nodes and intermediate nodes
    operators = []
    for ni in range(n_input_nodes):
        for nj in range(n_intermediate_nodes):
            src_ind = ni
            dst_ind = n_input_nodes + nj
            operators.append(
                Operator(edge_based_label="{}->{}".format(src_ind, dst_ind),
                         darts_src=src_ind, darts_dst=dst_ind,
                         connected_to_input=True)
            )
    # Then collect the edge between the intermediate nodes
    for ni in range(n_intermediate_nodes):
        for nj in range(n_intermediate_nodes):
            if ni >= nj: continue
            src_ind = n_input_nodes + ni
            dst_ind = n_input_nodes + nj
            operators.append(
                Operator(edge_based_label="{}->{}".format(src_ind, dst_ind),
                         darts_src=src_ind, darts_dst=dst_ind,
                         connected_to_input=False)
            )
    if verbose:
        log_f("Collected {} edge-based operators: ".format(len(operators)))
        for op in operators: log_f("{}".format(op))
    # We need to create more nodes for inputs and output in the new topology
    # Gives an id to each operator
    operators_w_id = _assign_id_to_darts_ops(operators, n_input_nodes, n_intermediate_nodes)
    if verbose:
        log_f("New operator node id map: ")
        for op_id, op in operators_w_id:
            log_f("{}, new node id: {}".format(op, op_id))
    if map_rv is not None:
        assert isinstance(map_rv, dict)
        for curr_op_id, curr_op in operators_w_id:
            assert curr_op_id not in map_rv
            map_rv[curr_op_id] = (curr_op.darts_src, curr_op.darts_dst)
    # Now, treat each operator as a node, figure out the new edge connections
    # For each op, figure out which previous ops connect to it by looking at the old darts feature map index
    src_inds, dst_inds = [], []
    for curr_op_id, curr_op in operators_w_id:
        if curr_op.darts_src < n_input_nodes:
            # If originally this node connects to one of the input, just connect it to the new input
            src_inds.append(curr_op.darts_src)
            dst_inds.append(curr_op_id)
        else:
            # We need to find all operators that provides the input for the current op
            input_feat_id = curr_op.darts_src
            connected_ops = [(op_id, op) for op_id, op in operators_w_id if op.darts_dst == input_feat_id]
            assert len(connected_ops) > 0
            # Connect the nodes
            for neighbor_op_id, _ in connected_ops:
                assert neighbor_op_id < curr_op_id
                src_inds.append(neighbor_op_id)
                dst_inds.append(curr_op_id)
    # Finally, we need to connect all intermediate nodes in darts to the output
    output_node_id = max(dst_inds) + 1
    if verbose: log_f("New output node id: {}".format(output_node_id))
    # Connect all ops that should be connected to the output
    for ni in range(n_intermediate_nodes):
        feat_node_id = ni + n_input_nodes
        connected_ops = [(op_id, op) for op_id, op in operators_w_id if op.darts_dst == feat_node_id]
        for neighbor_op_id, _ in connected_ops:
            assert neighbor_op_id < output_node_id
            src_inds.append(neighbor_op_id)
            dst_inds.append(output_node_id)
    # Check no duplicated edges
    assert len(set("{}|{}".format(s, d) for s, d in zip(src_inds, dst_inds))) == len(src_inds)
    if verbose:
        log_f("Converted edge pairs: {}".format(sorted(list(zip(src_inds, dst_inds)))))
    return src_inds, dst_inds


def _assign_id_to_darts_ops(operators, n_input_nodes, n_intermediate_nodes):
    """
    Assign an id to each operator, should match the assigning/calling order in DARTS supernet
    :return: operators_w_id
    """
    operators_w_id = []
    curr_id = n_input_nodes
    for i in range(n_intermediate_nodes):  # Loop on nodes
        tgt_darts_dst = i + n_input_nodes
        related_ops = [op for op in operators if op.darts_dst == tgt_darts_dst]
        related_ops.sort(key=lambda _op: (_op.darts_src, _op.darts_dst))
        for op in related_ops:
            operators_w_id.append((curr_id, op))
            curr_id += 1
    assert len(operators_w_id) == len(operators)
    return operators_w_id


def nb201_topo_to_node_based_topo(n=4, verbose=True, log_f=print, map_rv=None):
    """
    Converts NB201 cell into a topology where each node is an operator
    :param n:
    :param verbose:
    :param log_f:
    :param map_rv: mapping between node indices to its input/output DARTS node
    :return: [src_inds, dst_inds]
    """
    from collections import namedtuple
    Operator = namedtuple("Operator", ["edge_based_label", "darts_src", "darts_dst", "connected_to_input"])
    if verbose:
        log_f("Converting the NB201 cell with N={} to the node-as-operator topology".format(n))
    # Get the edges
    operators = []
    for dst_ind in range(n):
        for src_ind in range(n):
            if src_ind >= dst_ind: continue
            operators.append(
                Operator(edge_based_label="{}->{}".format(src_ind, dst_ind),
                         darts_src=src_ind, darts_dst=dst_ind,
                         connected_to_input=True)
            )
    if verbose:
        log_f("Collected {} edge-based operators: ".format(len(operators)))
        for op in operators: log_f("{}".format(op))
    # We need to create more nodes for inputs and output in the new topology
    # Gives an id to each operator
    operators_w_id = [(i + 1, op) for i, op in enumerate(operators)]
    if verbose:
        log_f("New operator node id map: ")
        for op_id, op in operators_w_id:
            log_f("{}, new node id: {}".format(op, op_id))
    if map_rv is not None:
        assert isinstance(map_rv, dict)
        for curr_op_id, curr_op in operators_w_id:
            assert curr_op_id not in map_rv
            map_rv[curr_op_id] = (curr_op.darts_src, curr_op.darts_dst)
    # Now, treat each operator as a node, figure out the new edge connections
    # For each op, figure out which previous ops connect to it by looking at the old darts feature map index
    src_inds, dst_inds = [], []
    for curr_op_id, curr_op in operators_w_id:
        if curr_op.darts_src == 0:
            # If originally this node connects to one of the input, just connect it to the new input
            src_inds.append(curr_op.darts_src)
            dst_inds.append(curr_op_id)
        else:
            # We need to find all operators that provides the input for the current op
            input_feat_id = curr_op.darts_src
            connected_ops = [(op_id, op) for op_id, op in operators_w_id if op.darts_dst == input_feat_id]
            assert len(connected_ops) > 0
            # Connect the nodes
            for neighbor_op_id, _ in connected_ops:
                assert neighbor_op_id < curr_op_id
                src_inds.append(neighbor_op_id)
                dst_inds.append(curr_op_id)
    # Finally, we need to connect all intermediate nodes in darts to the output
    output_node_id = max(dst_inds) + 1
    if verbose: log_f("New output node id: {}".format(output_node_id))
    # Connect all ops that should be connected to the output
    connected_ops = [(op_id, op) for op_id, op in operators_w_id if op.darts_dst == (n-1)]
    for neighbor_op_id, _ in connected_ops:
        assert neighbor_op_id < output_node_id
        src_inds.append(neighbor_op_id)
        dst_inds.append(output_node_id)
    # Check no duplicated edges
    assert len(set("{}|{}".format(s, d) for s, d in zip(src_inds, dst_inds))) == len(src_inds)
    if verbose:
        log_f("Converted edge pairs: {}".format(sorted(list(zip(src_inds, dst_inds)))))
    return src_inds, dst_inds


def get_nb101_edge_list_from_darts_cell(darts_ops, darts_op_edges, de2ni_map, complete_nb101_edge_list,
                                        n_input_nodes=2, add_self_edge_for_input=True, output_node_label="output"):
    """
    :param darts_ops: list of operation strings
    :param darts_op_edges: list of tuples: (darts_src_node, darts_dst_node), ...
    :param de2ni_map: maps darts edge to NB101 node index, note this is different from the op index
    :param complete_nb101_edge_list: edge list from a complete darts cell
    :param n_input_nodes:
    :param add_self_edge_for_input:
    :param output_node_label:
    :return:
    """
    complete_nb101_edge_pairs = edge_list_to_edge_pairs(complete_nb101_edge_list)
    output_node_idx = max(complete_nb101_edge_list[1])
    # First determine which NB101 node indices are selected
    nb101_node_indices = []
    if output_node_label in darts_ops:
        nb101_node_indices.append(output_node_idx)
    for src, dst in darts_op_edges:
        key = "{}->{}".format(src, dst)
        if key in de2ni_map:
            # If dst is the output node then key will not be in de2ni_map
            ind = de2ni_map[key]
            nb101_node_indices.append(ind)
    nb101_node_indices.sort()
    # Next fetch the edges that are associated with these NB101 nodes from the complete edge list
    nb101_edges = []
    for src, dst in complete_nb101_edge_pairs:
        if src in nb101_node_indices and dst in nb101_node_indices:
            nb101_edges.append([src, dst])
        elif src < n_input_nodes and dst in nb101_node_indices:
            nb101_edges.append([src, dst])
    # Perform node index reduction for the selected NB101 edges
    for ni, sel_ni in enumerate(nb101_node_indices):
        target_ni = ni + n_input_nodes
        if target_ni == sel_ni:
            # No need to perform node index reduction
            continue
        for pair in nb101_edges:
            if pair[0] == sel_ni:
                pair[0] = target_ni
            if pair[1] == sel_ni:
                pair[1] = target_ni
    if len(nb101_edges) > 0:
        assert global_max(nb101_edges) == len(nb101_node_indices) + 2 - 1
    self_edges = []
    if add_self_edge_for_input:
        for i in range(n_input_nodes):
            self_edges.append(i)
    return [self_edges + [t[0] for t in nb101_edges],
            self_edges + [t[1] for t in nb101_edges]]


def darts_cell_to_discrete_alphas(darts_ops, darts_op_edges, de2ni_map, op2idx,
                                  n_rows=14, n_input_nodes=2, fill_op_idx=0, fill_alpha_val=0.1):
    """
    Creates discrete complete n_rows x n_candidate_ops alpha matrix from a discrete cell
    :param darts_ops: list of operations
    :param darts_op_edges: list of tuples: (darts_src_node, darts_dst_node), ...
    :param de2ni_map: maps darts edge to NB101 node index, note this is different from the op index
    :param op2idx:
    :param n_rows: number of rows in the complete alpha matrix
    :param n_input_nodes:
    :param fill_op_idx:
    :param fill_alpha_val:
    :return:
    """
    n_cols = max(v for v in op2idx.values()) - n_input_nodes + 1
    darts_ops = darts_ops[n_input_nodes:]
    alphas = torch.zeros(n_rows, n_cols).float()
    nb101_node_indices = []
    for src, dst in darts_op_edges:
        key = "{}->{}".format(src, dst)
        if key in de2ni_map:
            # If dst is the output node then key will not be in de2ni_map
            ind = de2ni_map[key]
            nb101_node_indices.append(ind)
    for i, ni in enumerate(nb101_node_indices):
        if ni < n_input_nodes: continue # Ignore input nodes
        alpha_row_i = ni - n_input_nodes
        op = darts_ops[i]
        op_idx = op2idx[op] - n_input_nodes
        alphas[alpha_row_i, op_idx] = 1.
    # Fill in unused alpha rows
    if fill_alpha_val > 0.:
        for ri in range(n_rows):
            if ri + 2 not in nb101_node_indices:
                alphas[ri, fill_op_idx] = fill_alpha_val
    # Added in extra column for the None op
    extra_col = torch.zeros(n_rows, 1).float()
    alphas = torch.cat([extra_col, alphas], dim=1)
    return alphas


if __name__ == "__main__":
    # darts_topo_to_node_based_topo(n=7)
    _map_rv = {}
    nb201_topo_to_node_based_topo(n=4, map_rv=_map_rv)
    print(_map_rv)
    print("done")

