import random
from tqdm import tqdm
from utils.darts_utils import *


CIFAR10_MEAN_RGB = [0.49139968, 0.48215827, 0.44653124]
CIFAR10_STD_RGB = [0.24703233, 0.24348505, 0.26158768]
CIFAR100_MEAN_RGB = [0.5071, 0.4867, 0.4408]
CIFAR100_STD_RGB = [0.2675, 0.2565, 0.2761]
IMAGENET_MEAN_RGB = [0.485, 0.456, 0.406]
IMAGENET_STD_RGB = [0.229, 0.224, 0.225]
NUM_TRAIN_IMAGES_IMAGENET = 1281167
NUM_VAL_IMAGES_IMAGENET = 50000
NUM_TRAIN_IMAGES_IMAGENET_120 = 155367 # Class labels 1 to 120
NUM_VAL_IMAGES_IMAGENET_120 = 6000
NUM_TRAIN_IMAGES_IMAGENET_200 = 254640 # Class labels 1 and interval 5
NUM_VAL_IMAGES_IMAGENET_200 = 10000
NUM_TRAIN_IMAGES_IMAGENET_100 = 126720 # Class labels 1 and interval 10
NUM_VAL_IMAGES_IMAGENET_100 = 5000


def get_cifar_train_dev_test_set(name="CIFAR10", batch_size=128, dev_ratio=0.0, cutout=False,
                                 data_folder="./data", batch_size_test=None, verbose=True,
                                 disable_train_transform=False, num_workers=16, image_size=None,
                                 train_transform_override=None, test_transform_override=None):
    import torch
    classes, train_transform, test_transform = _get_transforms(name, cutout, image_size=image_size, verbose=verbose)
    if train_transform_override is not None:
        print("Overriding default train transforms with provided ones")
        train_transform = train_transform_override
    if test_transform_override is not None:
        print("Overriding default test transforms with provided ones")
        test_transform = test_transform_override
    if disable_train_transform:
        print("Warning: using test transforms for the train and dev sets")
        train_transform = test_transform
    train_set, test_set = _get_dataset(name, data_folder, train_transform,
                                       get_test=True, test_transform=test_transform,
                                       verbose=verbose)
    if batch_size_test is None:
        batch_size_test = batch_size
    test_loader = torch.utils.data.DataLoader(
        test_set, batch_size=batch_size_test,
        shuffle=False, pin_memory=True, num_workers=num_workers,
        sampler=None)
    train_size = len(train_set)
    indices = list(range(train_size))
    dev_size = int(np.floor(dev_ratio * train_size))
    if dev_ratio > 0.:
        train_loader = torch.utils.data.DataLoader(
            train_set, batch_size=batch_size,
            sampler=torch.utils.data.sampler.SubsetRandomSampler(indices[dev_size:]),
            pin_memory=True, num_workers=num_workers)
        dev_loader = torch.utils.data.DataLoader(
            train_set, batch_size=batch_size_test,
            sampler=torch.utils.data.sampler.SubsetRandomSampler(indices[:dev_size]),
            pin_memory=True, num_workers=num_workers)
    else:
        train_loader = torch.utils.data.DataLoader(
            train_set, batch_size=batch_size, shuffle=True,
            pin_memory=True, num_workers=num_workers)
        if verbose: print("Warning: dev_ratio is set to 0, using test_loader as dev_loader!")
        dev_loader = test_loader
    return classes, train_loader, dev_loader, test_loader


def data_transforms_cifar(name, cutout, image_size=None):
    import torchvision.transforms as transforms
    if '100' in name:
        # print("Transform for CIFAR100")
        CIFAR_MEAN = [0.5071, 0.4867, 0.4408]
        CIFAR_STD = [0.2675, 0.2565, 0.2761]
        CIFAR_CO = 8
        CIFAR_CLASS = 100
    else:
        # print("Transform for CIFAR10")
        CIFAR_MEAN = [0.49139968, 0.48215827, 0.44653124]
        CIFAR_STD = [0.24703233, 0.24348505, 0.26158768]
        CIFAR_CO = 16
        CIFAR_CLASS = 10
    if image_size is not None:
        trans_list = [
            transforms.RandomCrop(32, padding=4),
            transforms.RandomHorizontalFlip(),
            transforms.Resize(image_size),
            transforms.ToTensor(),
            transforms.Normalize(CIFAR_MEAN, CIFAR_STD),
        ]
    else:
        trans_list = [
            transforms.RandomCrop(32, padding=4),
            transforms.RandomHorizontalFlip(),
            transforms.ToTensor(),
            transforms.Normalize(CIFAR_MEAN, CIFAR_STD),
        ]
    train_transform = transforms.Compose(trans_list)
    if cutout:
        train_transform.transforms.append(Cutout(CIFAR_CO))
    if image_size is not None:
        test_trans_list = [
            transforms.Resize(image_size),
            transforms.ToTensor(),
            transforms.Normalize(CIFAR_MEAN, CIFAR_STD),
        ]
    else:
        test_trans_list = [
            transforms.ToTensor(),
            transforms.Normalize(CIFAR_MEAN, CIFAR_STD),
        ]
    test_transform = transforms.Compose(test_trans_list)
    return CIFAR_CLASS, train_transform, test_transform


def data_transforms_imagenet(name, cutout):
    import torchvision.transforms as transforms
    MEAN = [0.485, 0.456, 0.406]
    STD = [0.229, 0.224, 0.225]
    CO = 8
    if "-" in name:
        strs = [s.strip() for s in name.split("-") if len(s.strip()) > 0]
        CLASS = int(strs[-1])
    else:
        CLASS = 1001
    train_transform = transforms.Compose([
        transforms.RandomResizedCrop(224),
        transforms.RandomHorizontalFlip(),
        # transforms.ColorJitter(
        #     brightness=0.4,
        #     contrast=0.4,
        #     saturation=0.4,
        #     hue=0.2),
        transforms.ToTensor(),
        transforms.Normalize(MEAN, STD),
    ])
    if cutout:
        train_transform.transforms.append(Cutout(CO))
    test_transform = transforms.Compose([
        # transforms.Resize(256),
        transforms.CenterCrop(224),
        transforms.ToTensor(),
        transforms.Normalize(MEAN, STD),
    ])
    return CLASS, train_transform, test_transform


def _get_transforms(name, cutout, image_size=None, verbose=True):
    if "cifar" in name.lower():
        if verbose: print("Data is CIFAR")
        return data_transforms_cifar(name, cutout, image_size=image_size)
    elif "imagenet" in name.lower():
        if verbose: print("Data is ImageNet")
        return data_transforms_imagenet(name, cutout)
    else:
        raise ValueError("Only support CIFAR and ImageNet transforms")


def _get_dataset(name, data_folder, train_transform, get_test=False, test_transform=None, verbose=True):
    import torchvision.datasets as dset
    if name.lower() == "cifar100":
        if verbose: print("Fetching CIFAR100")
        train_set = dset.CIFAR100(root=data_folder, train=True, download=True, transform=train_transform)
        if get_test:
            test_set = dset.CIFAR100(root=data_folder, train=False, download=True, transform=test_transform)
            return train_set, test_set
        return train_set
    elif name.lower() == "cifar10":
        if verbose: print("Fetching CIFAR10")
        train_set = dset.CIFAR10(root=data_folder, train=True, download=True, transform=train_transform)
        if get_test:
            test_set = dset.CIFAR10(root=data_folder, train=False, download=True, transform=test_transform)
            return train_set, test_set
        return train_set
    elif name.lower().startswith("imagenet"):
        if verbose: print("Fetching {}".format(name))
        train_dir = os.path.sep.join([data_folder, "train"])
        valid_dir = os.path.sep.join([data_folder, "val"])
        assert os.path.isdir(train_dir), "Cannot find dir: {}".format(train_dir)
        assert os.path.isdir(valid_dir), "Cannot find dir: {}".format(valid_dir)
        train_data = dset.ImageFolder(train_dir, train_transform)
        valid_data = dset.ImageFolder(valid_dir, test_transform)
        return train_data, valid_data
    else:
        raise ValueError("Only support CIFAR10, CIFAR100, and ImageNet dataset")


def get_stem_opt(dataset_name):
    stem_opt = {}
    if 'cifar' in dataset_name.lower():
        stem_opt['channels'] = 3
        stem_opt['kernel_size'] = 3
        stem_opt['padding'] = 1
        return stem_opt
    return NotImplementedError


def read_col_word_delim_data(in_fn, ignore_header=False, col_delim="|", word_delim=" ",
                             shuffle=False, read_lines_limit=None, randomly_take_n_lines=None,
                             exclude_indices=set()):
    with open(in_fn, "r", encoding="utf-8") as f:
        lines = f.readlines()
    lines = lines[1:] if ignore_header else lines
    rv = []
    for line in tqdm(lines, desc="Processing lines from {}".format(in_fn), ascii=True):
        line = line.strip()
        if len(line) == 0: continue
        segs = [[w for w in (d.split(word_delim) if word_delim != "" else d) if len(w) > 0]
                for i, d in enumerate(line.split(col_delim)) if i not in exclude_indices]
        rv.append(segs)
        if read_lines_limit is not None and len(rv) >= read_lines_limit:
            break
    if randomly_take_n_lines is not None:
        random.shuffle(rv)
        rv = rv[:randomly_take_n_lines]
    elif shuffle:
        random.shuffle(rv)
    return rv


def get_imagenet_train_dev_test_set(name="ImageNet", batch_size=256, dev_ratio=0.0, cutout=False,
                                    data_folder="./data", batch_size_test=None,
                                    train_transform_override=None,
                                    test_transform_override=None,
                                    include_background_label=False,
                                    disable_train_transform=False, num_workers=32):
    classes, train_transform, test_transform = _get_transforms(name, cutout)
    if classes == 1001 and not include_background_label:
        classes -= 1
    if train_transform_override is not None:
        print("Overriding default train transforms with provided ones")
        train_transform = train_transform_override
    if test_transform_override is not None:
        print("Overriding default test transforms with provided ones")
        test_transform = test_transform_override
    if disable_train_transform:
        print("Warning: using test transforms for the train and dev sets")
        train_transform = test_transform
    train_set, test_set = _get_dataset(name, data_folder, train_transform,
                                       get_test=True, test_transform=test_transform)
    if batch_size_test is None:
        batch_size_test = batch_size
    test_loader = torch.utils.data.DataLoader(
        test_set, batch_size=batch_size_test,
        shuffle=False, pin_memory=True, num_workers=num_workers,
        sampler=None)
    train_size = len(train_set)
    indices = list(range(train_size))
    dev_size = int(np.floor(dev_ratio * train_size))
    if dev_ratio > 0.:
        train_loader = torch.utils.data.DataLoader(
            train_set, batch_size=batch_size,
            sampler=torch.utils.data.sampler.SubsetRandomSampler(indices[dev_size:]),
            pin_memory=True, num_workers=num_workers)
        dev_loader = torch.utils.data.DataLoader(
            train_set, batch_size=batch_size_test,
            sampler=torch.utils.data.sampler.SubsetRandomSampler(indices[:dev_size]),
            pin_memory=True, num_workers=num_workers)
    else:
        train_loader = torch.utils.data.DataLoader(
            train_set, batch_size=batch_size, shuffle=True,
            pin_memory=True, num_workers=num_workers)
        print("Warning: dev_ratio is set to 0, using test_loader as dev_loader!")
        dev_loader = test_loader
    return classes, train_loader, dev_loader, test_loader

