import nltk
import math
import random
import collections
from tqdm import tqdm
from sklearn.metrics import *
from nltk.translate.bleu_score import SmoothingFunction
from constants import OOV_TOKEN
from utils.misc_utils import mean, variance, flatten_nested_lists, list_element_wise_op


def get_pareto_for_perfs(id2perfs, perf_smaller_is_better):
    # expected block2perfs format: {id_str: perf_values}
    idx2key, costs = [], []
    for key, perf_vals in id2perfs.items():
        idx2key.append(key)
        block_cost = []
        for i, perf in enumerate(perf_vals):
            # Convert every performance to smaller is better
            cost = perf
            if not perf_smaller_is_better[i]:
                cost = - perf
            block_cost.append(cost)
        costs.append(block_cost)
    costs = np.asarray(costs)
    pareto_masks = _is_pareto_efficient(costs, return_mask=True)
    pareto_indices = np.argwhere(pareto_masks).squeeze(1).tolist()
    kept_keys = [idx2key[idx] for idx in pareto_indices]
    pareto_id2perfs = {key: id2perfs[key] for key in kept_keys}
    return pareto_id2perfs


def _is_pareto_efficient(costs, return_mask=True):
    is_efficient = np.arange(costs.shape[0])
    n_points = costs.shape[0]
    next_point_index = 0  # Next index in the is_efficient array to search for
    while next_point_index < len(costs):
        nondominated_point_mask = np.any(costs < costs[next_point_index], axis=1)
        nondominated_point_mask[next_point_index] = True
        is_efficient = is_efficient[nondominated_point_mask]  # Remove dominated points
        costs = costs[nondominated_point_mask]
        next_point_index = np.sum(nondominated_point_mask[:next_point_index]) + 1
    if return_mask:
        is_efficient_mask = np.zeros(n_points, dtype=bool)
        is_efficient_mask[is_efficient] = True
        return is_efficient_mask
    else:
        return is_efficient


def list_regression_report(pred_list, target_list):
    assert len(pred_list) == len(target_list)
    n = len(pred_list)
    exp_variance_score = explained_variance_score(target_list, pred_list)
    mean_sq_error = mean_squared_error(target_list, pred_list)
    mean_abs_error = mean_absolute_error(target_list, pred_list)
    r2 = r2_score(target_list, pred_list)
    max_err = max([abs(target_list[i] - pred_list[i]) for i in range(n)])
    if any(v == 0 for v in target_list):
        mape_pred_list, mape_target_list = [], []
        for i, v in enumerate(target_list):
            if v == 0: continue
            mape_pred_list.append(v)
            mape_target_list.append(v)
    else:
        mape_pred_list, mape_target_list = pred_list, target_list
    mape = mean([100 * abs(mape_pred_list[i] - truth) / truth for i, truth in enumerate(mape_target_list)], fallback_val=0)
    rv = {
        "mean_square_error": mean_sq_error,
        "mean_absolute_error": mean_abs_error,
        "explained_variance_score": exp_variance_score,
        "max_error": max_err,
        "r2_score": r2,
        "mean_absolute_percent_error": mape,
        "mape_effective_sample_size_diff": abs(len(mape_pred_list) - len(pred_list)),
    }
    return rv


def list_classification_report(pred_list, target_list, target_names):
    y_pred = np.asarray(pred_list)
    y_true = np.asarray(target_list)
    classify_report = classification_report(y_true, y_pred, target_names=target_names)
    return classify_report


def post_evaluate_results_file(f_name="test_results.txt", col_delim="|", word_delim=" ",
                               ignore_header=True, oov_token=OOV_TOKEN):
    with open(f_name, "r", encoding="utf-8") as f:
        lines = f.readlines()
    lines = lines[1:] if ignore_header else lines
    lsplit = [line.rstrip().strip().split(col_delim) for line in lines]
    gen_count = 0
    n_oov_tokens= 0
    n_tokens = 0
    rv = {}
    corpus_preds = []
    corpus_truth = []
    for val_list in lsplit:
        if len(val_list) != 3: continue
        gen = val_list[1]
        truth = val_list[2]
        gen_count += 1
        truth_seg = [w for w in truth.split(word_delim) if len(w) > 0]
        gen_seg = [w for w in gen.split(word_delim) if len(w) > 0]
        # key = " ".join(truth_seg) + "|" + " ".join(gen_seg)
        # if key in counted_results: continue
        # counted_results.add(key)
        n_oov_tokens += sum([1 for w in gen.split(word_delim) if w == oov_token])
        n_tokens += sum([1 for w in gen.split(word_delim) if len(w) > 0])
        corpus_preds.append(gen_seg)
        corpus_truth.append([truth_seg])
        multi_eval(gen_seg, truth_seg, result_dict=rv)
    if gen_count == 0: return rv
    for k, v in rv.items():
        rv[k] = v / gen_count
    rv["corpus_bleu_4"] = bleu_c(corpus_preds, corpus_truth)
    rv["percent_oov_in_output"] = n_oov_tokens/n_tokens if n_tokens > 0 else -1.0
    return rv


def corpus_eval(pred_lists, truth_lists):
    """
    :param pred_lists: shape like [ [pred_word1, pred_word2...], [pred_word1, pred_word2...] ]
    :param truth_lists: shape like [ [[truth_word1, truth_word2...]], [[truth_word1, truth_word2...]] ]
    :param result_dict:
    """
    assert len(pred_lists) > 0, "pred_lists cannot be empty"
    assert len(pred_lists) == len(truth_lists), "One prediction must correspond to one truth"
    rv = {
        "bleu_1":0,
        "bleu_2":0,
        "bleu_3":0,
        "bleu_4":0,
        "rouge_1":0,
        "rouge_2":0,
        "rouge_L":0,
        "em":0,
        "acc":0,
    }
    for i, pred_seg in enumerate(pred_lists):
        truth_seg = truth_lists[i][0] # TODO: support multiple truth?
        rv["bleu_1"] += bleu_1(pred_seg, truth_seg)
        rv["bleu_2"] += bleu_2(pred_seg, truth_seg)
        rv["bleu_3"] += bleu_3(pred_seg, truth_seg)
        rv["bleu_4"] += bleu_4(pred_seg, truth_seg)
        rv["rouge_1"] += rouge_1(pred_seg, truth_seg)
        rv["rouge_2"] += rouge_2(pred_seg, truth_seg)
        rv["rouge_L"] += rouge_L(pred_seg, truth_seg)
        rv["em"] += em(pred_seg, truth_seg)
        rv["acc"] += em(pred_seg, truth_seg)
    for k, v in rv.items():
        rv[k] = v / len(pred_lists) # macro average
    rv["corpus_bleu_4"] = bleu_c(pred_lists, truth_lists)
    rv["default_eval_metric"] = rv["bleu_1"]
    return rv


def multi_truth_recall_eval(ordered_multi_pred_lists, multi_truth_lists):
    assert len(ordered_multi_pred_lists) > 0, "pred_lists cannot be empty"
    assert len(ordered_multi_pred_lists) == len(multi_truth_lists), "One prediction must correspond to one truth"
    rv = {
        "recall_top_1": 0,
        "recall_top_5": 0,
        "recall_top_10": 0,
        "recall_top_20": 0,
        "recall_top_50": 0,
        # "recall_top_100": 0,
        # "recall_top_500": 0,
        "bleu_1":0,
        "bleu_2":0,
        "bleu_3":0,
        "bleu_4":0,
    }
    top_1_preds = []
    top_1_truths = []
    for i, pred_seg_lists in enumerate(ordered_multi_pred_lists):
        if len(pred_seg_lists) == 0: continue
        truth_seg_lists = multi_truth_lists[i]
        top_1_truths.append(truth_seg_lists)
        truth_sets = set([" ".join(val) for val in truth_seg_lists])
        pred_sets_top_1 = set([" ".join(val) for val in pred_seg_lists[:1]])
        pred_sets_top_5 = set([" ".join(val) for val in pred_seg_lists[:5]])
        pred_sets_top_10 = set([" ".join(val) for val in pred_seg_lists[:10]])
        pred_sets_top_20 = set([" ".join(val) for val in pred_seg_lists[:20]])
        pred_sets_top_50 = set([" ".join(val) for val in pred_seg_lists[:50]])
        # pred_sets_top_100 = set(["".join(val) for val in pred_seg_lists[:100]])
        # pred_sets_top_500 = set(["".join(val) for val in pred_seg_lists[:500]])
        rv["recall_top_1"] += min(len(pred_sets_top_1.intersection(truth_sets)), 1)
        rv["recall_top_5"] += min(len(pred_sets_top_5.intersection(truth_sets)), 1)
        rv["recall_top_10"] += min(len(pred_sets_top_10.intersection(truth_sets)), 1)
        rv["recall_top_20"] += min(len(pred_sets_top_20.intersection(truth_sets)), 1)
        rv["recall_top_50"] += min(len(pred_sets_top_50.intersection(truth_sets)), 1)
        top_1_pred_seg_list = pred_seg_lists[0]
        top_1_preds.append(top_1_pred_seg_list)
        rv["bleu_1"] += multi_truth_bleu_1(top_1_pred_seg_list, truth_seg_lists)
        rv["bleu_2"] += multi_truth_bleu_2(top_1_pred_seg_list, truth_seg_lists)
        rv["bleu_3"] += multi_truth_bleu_3(top_1_pred_seg_list, truth_seg_lists)
        rv["bleu_4"] += multi_truth_bleu_4(top_1_pred_seg_list, truth_seg_lists)
    for k, v in rv.items():
        rv[k] = v / len(ordered_multi_pred_lists) # macro average
    rv["corpus_bleu_4"] = bleu_c(top_1_preds, top_1_truths)
    rv["default_eval_metric"] = rv["recall_top_10"]
    return rv


def multi_eval(gen_seg, truth_seg, result_dict={}):
    if "em" not in result_dict: result_dict["em"] = 0
    result_dict["em"] += em(gen_seg, truth_seg)
    if "acc" not in result_dict: result_dict["acc"] = 0
    result_dict["acc"] += em(gen_seg, truth_seg)
    if "bleu_1" not in result_dict: result_dict["bleu_1"] = 0
    result_dict["bleu_1"] += bleu_1(gen_seg, truth_seg)
    if "bleu_2" not in result_dict: result_dict["bleu_2"] = 0
    result_dict["bleu_2"] += bleu_2(gen_seg, truth_seg)
    if "bleu_3" not in result_dict: result_dict["bleu_3"] = 0
    result_dict["bleu_3"] += bleu_3(gen_seg, truth_seg)
    if "bleu_4" not in result_dict: result_dict["bleu_4"] = 0
    result_dict["bleu_4"] += bleu_4(gen_seg, truth_seg)
    if "rouge_1" not in result_dict: result_dict["rouge_1"] = 0
    result_dict["rouge_1"] += rouge_1(gen_seg, truth_seg)
    if "rouge_2" not in result_dict: result_dict["rouge_2"] = 0
    result_dict["rouge_2"] += rouge_2(gen_seg, truth_seg)
    if "rouge_L" not in result_dict: result_dict["rouge_L"] = 0
    result_dict["rouge_L"] += rouge_L(gen_seg, truth_seg)
    return result_dict


def perplexity(probs, truth_idx):
    """
    Not working yet
    """
    n = truth_idx.shape[0]
    val = 0
    for i in range(n):
        wi = truth_idx[i].item()
        prob = probs[i, wi].item()
        val += math.log(prob, 2)
    return math.log(2, -1 / n * val)


def bleu_c(predict, truth, n=4):
    smoother = SmoothingFunction()
    if n == 1:
        return 100 * nltk.translate.bleu_score.corpus_bleu(truth, predict, weights=(1, 0, 0, 0), smoothing_function=smoother.method1)
    if n == 2:
        return 100 * nltk.translate.bleu_score.corpus_bleu(truth, predict, weights=(0.5, 0.5, 0, 0), smoothing_function=smoother.method1)
    if n == 3:
        return 100 * nltk.translate.bleu_score.corpus_bleu(truth, predict, weights=(0.33, 0.33, 0.33, 0), smoothing_function=smoother.method1)
    if n == 4:
        return 100 * nltk.translate.bleu_score.corpus_bleu(truth, predict, weights=(0.25, 0.25, 0.25, 0.25), smoothing_function=smoother.method1)
    print("n must be 1, 2, 3 or 4!")
    return None


def em(pred_seg_list, truth_seg_list):
    if " ".join(pred_seg_list) == " ".join(truth_seg_list):
        return 1
    else:
        return 0


def bleu_1(pred_seg_list, truth_seg_list):
    return 100 * _bleu_n(pred_seg_list, truth_seg_list, weights=(1,0,0,0))


def multi_truth_bleu_1(pred_seg_list, truth_seg_lists):
    return 100 * _multi_truth_bleu_n(pred_seg_list, truth_seg_lists, weights=(1,0,0,0))


def bleu_2(pred_seg_list, truth_seg_list):
    return 100 * _bleu_n(pred_seg_list, truth_seg_list, weights=(0.5,0.5,0,0)) # cumulative, not individual


def multi_truth_bleu_2(pred_seg_list, truth_seg_lists):
    return 100 * _multi_truth_bleu_n(pred_seg_list, truth_seg_lists, weights=(0.5,0.5,0,0))


def bleu_3(pred_seg_list, truth_seg_list):
    return 100 * _bleu_n(pred_seg_list, truth_seg_list, weights=(0.33,0.33,0.33,0))


def multi_truth_bleu_3(pred_seg_list, truth_seg_lists):
    return 100 * _multi_truth_bleu_n(pred_seg_list, truth_seg_lists, weights=(0.33,0.33,0.33,0))


def bleu_4(pred_seg_list, truth_seg_list):
    return 100 * _bleu_n(pred_seg_list, truth_seg_list, weights=(0.25,0.25,0.25,0.25))


def multi_truth_bleu_4(pred_seg_list, truth_seg_lists):
    return 100 * _multi_truth_bleu_n(pred_seg_list, truth_seg_lists, weights=(0.25,0.25,0.25,0.25))


def _bleu_n(pred_seg_list, truth_seg_list, weights):
    smoother = SmoothingFunction()
    score = nltk.translate.bleu_score.sentence_bleu([truth_seg_list], pred_seg_list,
                                                    weights=weights, smoothing_function=smoother.method1)
    return score


def _multi_truth_bleu_n(pred_seg_list, truth_seg_lists, weights):
    smoother = SmoothingFunction()
    score = nltk.translate.bleu_score.sentence_bleu(truth_seg_lists, pred_seg_list,
                                                    weights=weights, smoothing_function=smoother.method1)
    return score


def rouge_L(pred_seg_list, truth_seg_list):
    rouge_score_map = rouge([" ".join(pred_seg_list)], [" ".join(truth_seg_list)])
    return 100 * rouge_score_map["rouge_l/f_score"]


def rouge_1(pred_seg_list, truth_seg_list):
    rouge_score_map = rouge([" ".join(pred_seg_list)], [" ".join(truth_seg_list)])
    return 100 * rouge_score_map["rouge_1/f_score"]


def rouge_2(pred_seg_list, truth_seg_list):
    rouge_score_map = rouge([" ".join(pred_seg_list)], [" ".join(truth_seg_list)])
    return 100 * rouge_score_map["rouge_2/f_score"]


"""
ROUGE metric implementation.

Copy from tf_seq2seq/seq2seq/metrics/rouge.py.
This is a modified and slightly extended verison of
https://github.com/miso-belica/sumy/blob/squad_dev/sumy/evaluation/rouge.py.
"""

import itertools
import numpy as np

#pylint: disable=C0103


def _get_ngrams(n, text):
  """Calcualtes n-grams.

  Args:
    n: which n-grams to calculate
    text: An array of tokens

  Returns:
    A set of n-grams
  """
  ngram_set = set()
  text_length = len(text)
  max_index_ngram_start = text_length - n
  for i in range(max_index_ngram_start + 1):
    ngram_set.add(tuple(text[i:i + n]))
  return ngram_set


def _split_into_words(sentences):
  """Splits multiple sentences into words and flattens the result"""
  return list(itertools.chain(*[_.split(" ") for _ in sentences]))


def _get_word_ngrams(n, sentences):
  """Calculates word n-grams for multiple sentences.
  """
  assert len(sentences) > 0
  assert n > 0

  words = _split_into_words(sentences)
  return _get_ngrams(n, words)


def _len_lcs(x, y):
  """
  Returns the length of the Longest Common Subsequence between sequences x
  and y.
  Source: http://www.algorithmist.com/index.php/Longest_Common_Subsequence

  Args:
    x: sequence of words
    y: sequence of words

  Returns
    integer: Length of LCS between x and y
  """
  table = _lcs(x, y)
  n, m = len(x), len(y)
  return table[n, m]


def _lcs(x, y):
  """
  Computes the length of the longest common subsequence (lcs) between two
  strings. The implementation below uses a DP programming algorithm and runs
  in O(nm) time where n = len(x) and m = len(y).
  Source: http://www.algorithmist.com/index.php/Longest_Common_Subsequence

  Args:
    x: collection of words
    y: collection of words

  Returns:
    Table of dictionary of coord and len lcs
  """
  n, m = len(x), len(y)
  table = dict()
  for i in range(n + 1):
    for j in range(m + 1):
      if i == 0 or j == 0:
        table[i, j] = 0
      elif x[i - 1] == y[j - 1]:
        table[i, j] = table[i - 1, j - 1] + 1
      else:
        table[i, j] = max(table[i - 1, j], table[i, j - 1])
  return table


def _recon_lcs(x, y):
  """
  Returns the Longest Subsequence between x and y.
  Source: http://www.algorithmist.com/index.php/Longest_Common_Subsequence

  Args:
    x: sequence of words
    y: sequence of words

  Returns:
    sequence: LCS of x and y
  """
  i, j = len(x), len(y)
  table = _lcs(x, y)

  def _recon(i, j):
    """private recon calculation"""
    if i == 0 or j == 0:
      return []
    elif x[i - 1] == y[j - 1]:
      return _recon(i - 1, j - 1) + [(x[i - 1], i)]
    elif table[i - 1, j] > table[i, j - 1]:
      return _recon(i - 1, j)
    else:
      return _recon(i, j - 1)

  recon_tuple = tuple(map(lambda x: x[0], _recon(i, j)))
  return recon_tuple


def rouge_n(evaluated_sentences, reference_sentences, n=2):
  """
  Computes ROUGE-N of two text collections of sentences.
  Sourece: http://research.microsoft.com/en-us/um/people/cyl/download/
  papers/rouge-working-note-v1.3.1.pdf

  Args:
    evaluated_sentences: The sentences that have been picked by the summarizer
    reference_sentences: The sentences from the referene set
    n: Size of ngram.  Defaults to 2.

  Returns:
    A tuple (f1, precision, recall) for ROUGE-N

  Raises:
    ValueError: raises exception if a param has len <= 0
  """
  if len(evaluated_sentences) <= 0 or len(reference_sentences) <= 0:
    raise ValueError("Collections must contain at least 1 sentence.")

  evaluated_ngrams = _get_word_ngrams(n, evaluated_sentences)
  reference_ngrams = _get_word_ngrams(n, reference_sentences)
  reference_count = len(reference_ngrams)
  evaluated_count = len(evaluated_ngrams)

  # Gets the overlapping ngrams between evaluated and reference
  overlapping_ngrams = evaluated_ngrams.intersection(reference_ngrams)
  overlapping_count = len(overlapping_ngrams)

  # Handle edge case. This isn't mathematically correct, but it's good enough
  if evaluated_count == 0:
    precision = 0.0
  else:
    precision = overlapping_count / evaluated_count

  if reference_count == 0:
    recall = 0.0
  else:
    recall = overlapping_count / reference_count

  f1_score = 2.0 * ((precision * recall) / (precision + recall + 1e-8))

  # return overlapping_count / reference_count
  return f1_score, precision, recall


def _f_p_r_lcs(llcs, m, n):
  """
  Computes the LCS-based F-measure score
  Source: http://research.microsoft.com/en-us/um/people/cyl/download/papers/
  rouge-working-note-v1.3.1.pdf

  Args:
    llcs: Length of LCS
    m: number of words in reference summary
    n: number of words in candidate summary

  Returns:
    Float. LCS-based F-measure score
  """
  r_lcs = llcs / m
  p_lcs = llcs / n
  beta = p_lcs / (r_lcs + 1e-12)
  num = (1 + (beta**2)) * r_lcs * p_lcs
  denom = r_lcs + ((beta**2) * p_lcs)
  f_lcs = num / (denom + 1e-12)
  return f_lcs, p_lcs, r_lcs


def rouge_l_sentence_level(evaluated_sentences, reference_sentences):
  """
  Computes ROUGE-L (sentence level) of two text collections of sentences.
  http://research.microsoft.com/en-us/um/people/cyl/download/papers/
  rouge-working-note-v1.3.1.pdf

  Calculated according to:
  R_lcs = LCS(X,Y)/m
  P_lcs = LCS(X,Y)/n
  F_lcs = ((1 + beta^2)*R_lcs*P_lcs) / (R_lcs + (beta^2) * P_lcs)

  where:
  X = reference summary
  Y = Candidate summary
  m = length of reference summary
  n = length of candidate summary

  Args:
    evaluated_sentences: The sentences that have been picked by the summarizer
    reference_sentences: The sentences from the referene set

  Returns:
    A float: F_lcs

  Raises:
    ValueError: raises exception if a param has len <= 0
  """
  if len(evaluated_sentences) <= 0 or len(reference_sentences) <= 0:
    raise ValueError("Collections must contain at least 1 sentence.")
  reference_words = _split_into_words(reference_sentences)
  evaluated_words = _split_into_words(evaluated_sentences)
  m = len(reference_words)
  n = len(evaluated_words)
  lcs = _len_lcs(evaluated_words, reference_words)
  return _f_p_r_lcs(lcs, m, n)


def _union_lcs(evaluated_sentences, reference_sentence):
  """
  Returns LCS_u(r_i, C) which is the LCS score of the union longest common
  subsequence between reference sentence ri and candidate summary C. For example
  if r_i= w1 w2 w3 w4 w5, and C contains two sentences: c1 = w1 w2 w6 w7 w8 and
  c2 = w1 w3 w8 w9 w5, then the longest common subsequence of r_i and c1 is
  "w1 w2" and the longest common subsequence of r_i and c2 is "w1 w3 w5". The
  union longest common subsequence of r_i, c1, and c2 is "w1 w2 w3 w5" and
  LCS_u(r_i, C) = 4/5.

  Args:
    evaluated_sentences: The sentences that have been picked by the summarizer
    reference_sentence: One of the sentences in the reference summaries

  Returns:
    float: LCS_u(r_i, C)

  ValueError:
    Raises exception if a param has len <= 0
  """
  if len(evaluated_sentences) <= 0:
    raise ValueError("Collections must contain at least 1 sentence.")

  lcs_union = set()
  reference_words = _split_into_words([reference_sentence])
  combined_lcs_length = 0
  for eval_s in evaluated_sentences:
    evaluated_words = _split_into_words([eval_s])
    lcs = set(_recon_lcs(reference_words, evaluated_words))
    combined_lcs_length += len(lcs)
    lcs_union = lcs_union.union(lcs)

  union_lcs_count = len(lcs_union)
  union_lcs_value = union_lcs_count / combined_lcs_length
  return union_lcs_value


def rouge_l_summary_level(evaluated_sentences, reference_sentences):
  """
  Computes ROUGE-L (summary level) of two text collections of sentences.
  http://research.microsoft.com/en-us/um/people/cyl/download/papers/
  rouge-working-note-v1.3.1.pdf

  Calculated according to:
  R_lcs = SUM(1, u)[LCS<union>(r_i,C)]/m
  P_lcs = SUM(1, u)[LCS<union>(r_i,C)]/n
  F_lcs = ((1 + beta^2)*R_lcs*P_lcs) / (R_lcs + (beta^2) * P_lcs)

  where:
  SUM(i,u) = SUM from i through u
  u = number of sentences in reference summary
  C = Candidate summary made up of v sentences
  m = number of words in reference summary
  n = number of words in candidate summary

  Args:
    evaluated_sentences: The sentences that have been picked by the summarizer
    reference_sentence: One of the sentences in the reference summaries

  Returns:
    A float: F_lcs

  Raises:
    ValueError: raises exception if a param has len <= 0
  """
  if len(evaluated_sentences) <= 0 or len(reference_sentences) <= 0:
    raise ValueError("Collections must contain at least 1 sentence.")

  # total number of words in reference sentences
  m = len(_split_into_words(reference_sentences))

  # total number of words in evaluated sentences
  n = len(_split_into_words(evaluated_sentences))

  union_lcs_sum_across_all_references = 0
  for ref_s in reference_sentences:
    union_lcs_sum_across_all_references += _union_lcs(evaluated_sentences,
                                                      ref_s)
  return _f_p_r_lcs(union_lcs_sum_across_all_references, m, n)


def rouge(hypotheses, references):
  """Calculates average rouge scores for a list of hypotheses and
  references"""

  # Filter out hyps that are of 0 length
  # hyps_and_refs = zip(hypotheses, references)
  # hyps_and_refs = [_ for _ in hyps_and_refs if len(_[0]) > 0]
  # hypotheses, references = zip(*hyps_and_refs)

  # Calculate ROUGE-1 F1, precision, recall scores
  rouge_1 = [
      rouge_n([hyp], [ref], 1) for hyp, ref in zip(hypotheses, references)
  ]
  rouge_1_f, rouge_1_p, rouge_1_r = map(np.mean, zip(*rouge_1))

  # Calculate ROUGE-2 F1, precision, recall scores
  rouge_2 = [
      rouge_n([hyp], [ref], 2) for hyp, ref in zip(hypotheses, references)
  ]
  rouge_2_f, rouge_2_p, rouge_2_r = map(np.mean, zip(*rouge_2))

  # Calculate ROUGE-L F1, precision, recall scores
  rouge_l = [
      rouge_l_sentence_level([hyp], [ref])
      for hyp, ref in zip(hypotheses, references)
  ]
  rouge_l_f, rouge_l_p, rouge_l_r = map(np.mean, zip(*rouge_l))

  return {
      "rouge_1/f_score": rouge_1_f,
      "rouge_1/r_score": rouge_1_r,
      "rouge_1/p_score": rouge_1_p,
      "rouge_2/f_score": rouge_2_f,
      "rouge_2/r_score": rouge_2_r,
      "rouge_2/p_score": rouge_2_p,
      "rouge_l/f_score": rouge_l_f,
      "rouge_l/r_score": rouge_l_r,
      "rouge_l/p_score": rouge_l_p,
  }


"""
Information Retrieval metrics

Useful Resources:
http://www.cs.utexas.edu/~mooney/ir-course/slides/Evaluation.ppt
http://www.nii.ac.jp/TechReports/05-014E.pdf
http://www.stanford.edu/class/cs276/handouts/EvaluationNew-handout-6-per.pdf
http://hal.archives-ouvertes.fr/docs/00/72/67/60/PDF/07-busa-fekete.pdf
Learning to Rank for Information Retrieval (Tie-Yan Liu)
"""
def mean_reciprocal_rank(rs):
    """Score is reciprocal of the rank of the first relevant item

    First element is 'rank 1'.  Relevance is binary (nonzero is relevant).

    Example from http://en.wikipedia.org/wiki/Mean_reciprocal_rank
    >>> rs = [[0, 0, 1], [0, 1, 0], [1, 0, 0]]
    >>> mean_reciprocal_rank(rs)
    0.61111111111111105
    >>> rs = np.array([[0, 0, 0], [0, 1, 0], [1, 0, 0]])
    >>> mean_reciprocal_rank(rs)
    0.5
    >>> rs = [[0, 0, 0, 1], [1, 0, 0], [1, 0, 0]]
    >>> mean_reciprocal_rank(rs)
    0.75

    Args:
        rs: Iterator of relevance scores (list or numpy) in rank order
            (first element is the first item)

    Returns:
        Mean reciprocal rank
    """
    rs = (np.asarray(r).nonzero()[0] for r in rs)
    return np.mean([1. / (r[0] + 1) if r.size else 0. for r in rs])


def r_precision(r):
    """Score is precision after all relevant documents have been retrieved

    Relevance is binary (nonzero is relevant).

    >>> r = [0, 0, 1]
    >>> r_precision(r)
    0.33333333333333331
    >>> r = [0, 1, 0]
    >>> r_precision(r)
    0.5
    >>> r = [1, 0, 0]
    >>> r_precision(r)
    1.0

    Args:
        r: Relevance scores (list or numpy) in rank order
            (first element is the first item)

    Returns:
        R Precision
    """
    r = np.asarray(r) != 0
    z = r.nonzero()[0]
    if not z.size:
        return 0.
    return np.mean(r[:z[-1] + 1])


def precision_at_k(r, k):
    """Score is precision @ k

    Relevance is binary (nonzero is relevant).

    >>> r = [0, 0, 1]
    >>> precision_at_k(r, 1)
    0.0
    >>> precision_at_k(r, 2)
    0.0
    >>> precision_at_k(r, 3)
    0.33333333333333331
    >>> precision_at_k(r, 4)
    Traceback (most recent call last):
        File "<stdin>", line 1, in ?
    ValueError: Relevance score length < k


    Args:
        r: Relevance scores (list or numpy) in rank order
            (first element is the first item)

    Returns:
        Precision @ k

    Raises:
        ValueError: len(r) must be >= k
    """
    assert k >= 1
    r = np.asarray(r)[:k] != 0
    if r.size != k:
        raise ValueError('Relevance score length < k')
    return np.mean(r)


def average_precision(r):
    """Score is average precision (area under PR curve)

    Relevance is binary (nonzero is relevant).

    >>> r = [1, 1, 0, 1, 0, 1, 0, 0, 0, 1]
    >>> delta_r = 1. / sum(r)
    >>> sum([sum(r[:x + 1]) / (x + 1.) * delta_r for x, y in enumerate(r) if y])
    0.7833333333333333
    >>> average_precision(r)
    0.78333333333333333

    Args:
        r: Relevance scores (list or numpy) in rank order
            (first element is the first item)

    Returns:
        Average precision
    """
    r = np.asarray(r) != 0
    out = [precision_at_k(r, k + 1) for k in range(r.size) if r[k]]
    if not out:
        return 0.
    return np.mean(out)


def mean_average_precision(rs):
    """Score is mean average precision

    Relevance is binary (nonzero is relevant).

    >>> rs = [[1, 1, 0, 1, 0, 1, 0, 0, 0, 1]]
    >>> mean_average_precision(rs)
    0.78333333333333333
    >>> rs = [[1, 1, 0, 1, 0, 1, 0, 0, 0, 1], [0]]
    >>> mean_average_precision(rs)
    0.39166666666666666

    Args:
        rs: Iterator of relevance scores (list or numpy) in rank order
            (first element is the first item)

    Returns:
        Mean average precision
    """
    return np.mean([average_precision(r) for r in rs])


def dcg_at_k(r, k, method=0):
    """Score is discounted cumulative gain (dcg)

    Relevance is positive real values.  Can use binary
    as the previous methods.

    Example from
    http://www.stanford.edu/class/cs276/handouts/EvaluationNew-handout-6-per.pdf
    >>> r = [3, 2, 3, 0, 0, 1, 2, 2, 3, 0]
    >>> dcg_at_k(r, 1)
    3.0
    >>> dcg_at_k(r, 1, method=1)
    3.0
    >>> dcg_at_k(r, 2)
    5.0
    >>> dcg_at_k(r, 2, method=1)
    4.2618595071429155
    >>> dcg_at_k(r, 10)
    9.6051177391888114
    >>> dcg_at_k(r, 11)
    9.6051177391888114

    Args:
        r: Relevance scores (list or numpy) in rank order
            (first element is the first item)
        k: Number of results to consider
        method: If 0 then weights are [1.0, 1.0, 0.6309, 0.5, 0.4307, ...]
                If 1 then weights are [1.0, 0.6309, 0.5, 0.4307, ...]

    Returns:
        Discounted cumulative gain
    """
    r = np.asfarray(r)[:k]
    if r.size:
        if method == 0:
            return r[0] + np.sum(r[1:] / np.log2(np.arange(2, r.size + 1)))
        elif method == 1:
            return np.sum(r / np.log2(np.arange(2, r.size + 2)))
        else:
            raise ValueError('method must be 0 or 1.')
    return 0.


def ndcg_at_k(r, k, method=0):
    """Score is normalized discounted cumulative gain (ndcg)

    Relevance is positive real values.  Can use binary
    as the previous methods.

    Example from
    http://www.stanford.edu/class/cs276/handouts/EvaluationNew-handout-6-per.pdf
    >>> r = [3, 2, 3, 0, 0, 1, 2, 2, 3, 0]
    >>> ndcg_at_k(r, 1)
    1.0
    >>> r = [2, 1, 2, 0]
    >>> ndcg_at_k(r, 4)
    0.9203032077642922
    >>> ndcg_at_k(r, 4, method=1)
    0.96519546960144276
    >>> ndcg_at_k([0], 1)
    0.0
    >>> ndcg_at_k([1], 2)
    1.0

    Args:
        r: Relevance scores (list or numpy) in rank order
            (first element is the first item)
        k: Number of results to consider
        method: If 0 then weights are [1.0, 1.0, 0.6309, 0.5, 0.4307, ...]
                If 1 then weights are [1.0, 0.6309, 0.5, 0.4307, ...]

    Returns:
        Normalized discounted cumulative gain
    """
    dcg_max = dcg_at_k(sorted(r, reverse=True), k, method)
    if not dcg_max:
        return 0.
    return dcg_at_k(r, k, method) / dcg_max


def regression_results_to_ranking_results(pred_target_list, truth_target_list, n_candidates_list=(5, 10)):

    def _compute_global_ranking_metrics():
        from scipy.stats import spearmanr
        spearman_rho, spearman_p = spearmanr(truth_target_list, pred_target_list)
        metrics["spearman_rho"] = spearman_rho
        metrics["spearman_p"] = spearman_p

    def _compute_multi_cand_ranking_metrics(curr_i, key_prefix, _cand_indices, k_cands):
        ranked_pred_vals = [(ci, pred_target_list[ci]) for ci in _cand_indices]
        ranked_pred_vals.append((curr_i, pred_val))
        ranked_pred_vals.sort(key=lambda t: t[1], reverse=True)
        ranked_truth_vals = [(ci, truth_target_list[ci]) for ci in _cand_indices]
        ranked_truth_vals.append((curr_i, pred_val))
        ranked_truth_vals.sort(key=lambda t: t[1], reverse=True)
        truth_global_i2local_i = {
            t[0]: li for li, t in enumerate(ranked_truth_vals)
        }
        ranked_list = [truth_global_i2local_i[t[0]] for t in ranked_pred_vals][::-1]
        inst_ndcg = ndcg_at_k(ranked_list, k_cands)
        inst_dcg = dcg_at_k(ranked_list, k_cands)
        metrics["avg {} ndcg-{}".format(key_prefix, k_cands)] += inst_ndcg
        # metrics["avg {} dcg-{}".format(key_prefix, k_cands)] += inst_dcg
        n_inst_dict["avg {} ndcg-{}".format(key_prefix, k_cands)] += 1
        # n_inst_dict["avg {} dcg-{}".format(key_prefix, k_cands)] += 1

    def _add_knn_candidates(curr_i, k_cands):
        _knn_cand_indices = []
        front_idx, back_idx = curr_i - 1, curr_i + 1
        take_front = True if curr_i == len(pred_target_list)-1 else False
        while len(_knn_cand_indices) < k_cands:
            if take_front and 0 <= front_idx < len(pred_target_list):
                _knn_cand_indices.append(front_idx)
                front_idx -= 1
            elif 0 <= back_idx < len(pred_target_list):
                _knn_cand_indices.append(back_idx)
                back_idx += 1
            else:
                break
            take_front = not take_front
        assert len(_knn_cand_indices) > 0
        return _knn_cand_indices

    assert len(pred_target_list) == len(truth_target_list) > 0
    print("pred mean: {}, variance: {}".format(mean(pred_target_list), variance(pred_target_list)))
    print("truth mean: {}, variance: {}".format(mean(truth_target_list), variance((truth_target_list))))
    joint_list = list(zip(pred_target_list, truth_target_list))
    joint_list.sort(key=lambda t: t[1], reverse=True) # sort to make finding knn easier
    pred_target_list, truth_target_list = [t[0] for t in joint_list], [t[1] for t in joint_list]
    metrics = collections.defaultdict(float)
    n_inst_dict = collections.defaultdict(int)
    for k in n_candidates_list:
        bar = tqdm(total=len(pred_target_list), desc="Computing top-{} ranking results".format(k), ascii=True)
        for i, pred_val in enumerate(pred_target_list):
            random_cand_indices = set(random.sample(range(0, len(pred_target_list)), k))
            if i in random_cand_indices: random_cand_indices.remove(i)
            random_cand_indices = list(random_cand_indices)[:k-1]
            _compute_multi_cand_ranking_metrics(i, "random", random_cand_indices, k)
            knn_cand_indices = _add_knn_candidates(i, k-1)
            _compute_multi_cand_ranking_metrics(i, "knn", knn_cand_indices, k)
            bar.update(1)
        bar.close()
    for key in metrics:
        metrics[key] /= n_inst_dict[key]
    _compute_global_ranking_metrics()
    print("Ranking results: {}".format(metrics))
    metrics["input size"] = len(pred_target_list)
    return metrics


def build_nx_block_graph_v2(op_label_list, incoming_adj_lists):
    import networkx as nx
    graph = nx.DiGraph()
    for ni, node_label in enumerate(op_label_list):
        graph.add_node(node_label + "_{}".format(ni))
    for node_i, adj_list in enumerate(incoming_adj_lists):
        node_label = op_label_list[node_i] + "_{}".format(node_i)
        for neighbor_i, val in enumerate(adj_list):
            if neighbor_i < node_i and val > 0:
                neighbor_label = op_label_list[neighbor_i] + "_{}".format(neighbor_i)
                graph.add_edge(neighbor_label, node_label)
    return graph


def compute_block_edit_distance_v2(truth_op_label_list, truth_adj_lists, pred_op_label_list, pred_adj_lists):
    # This version is designed for AE, truth_adj_lists is incoming adj lists
    from networkx.algorithms.similarity import optimize_graph_edit_distance
    truth_graph = build_nx_block_graph_v2(truth_op_label_list, truth_adj_lists)
    pred_graph = build_nx_block_graph_v2(pred_op_label_list, pred_adj_lists)
    dist = float("inf")
    for v in optimize_graph_edit_distance(truth_graph, pred_graph):
        dist = min(dist, v)
    return dist


def build_nx_block_graph_v1(op_label_list, edge_lists):
    import networkx as nx
    graph = nx.DiGraph()
    for ni, node_label in enumerate(op_label_list):
        graph.add_node(node_label + "_{}".format(ni))
    for src, dst in zip(edge_lists[0], edge_lists[1]):
        if src == dst: continue
        src_label = op_label_list[src] + "_{}".format(src)
        dst_label = op_label_list[dst] + "_{}".format(dst)
        graph.add_edge(src_label, dst_label)
    return graph


def compute_block_edit_distance_v1(truth_op_label_list, truth_edge_lists, pred_op_label_list, pred_edge_lists):
    # This version is designed for torch geo edge lists
    from networkx.algorithms.similarity import optimize_graph_edit_distance
    truth_graph = build_nx_block_graph_v1(truth_op_label_list, truth_edge_lists)
    pred_graph = build_nx_block_graph_v1(pred_op_label_list, pred_edge_lists)
    dist = float("inf")
    for v in optimize_graph_edit_distance(truth_graph, pred_graph):
        dist = min(dist, v)
    return dist


def is_illegal_block(pred_adj_lists):
    # Input should not contain PAD
    if len(pred_adj_lists[0]) > 0:
        pred_adj_lists = [[]] + pred_adj_lists
    visited = set()
    def _dfs(n):
        neighbor_adj_list = pred_adj_lists[n]
        for neighbor_i, v in enumerate(neighbor_adj_list):
            if v == 1 and neighbor_i not in visited:
                visited.add(neighbor_i)
                _dfs(neighbor_i)
        visited.add(n)
    _dfs(len(pred_adj_lists)-1)
    return len(visited) != len(pred_adj_lists)


def report_block_matching_metrics(truth_op_labels, truth_adj_bin_labels, pred_op_labels, pred_adj_bin_labels,
                                  eos_idx=2, pad_op_idx=0):
    assert len(truth_op_labels) == len(truth_adj_bin_labels) == len(pred_op_labels) == len(pred_adj_bin_labels)
    assert all(len(lv) > 0 for lv in truth_op_labels)
    assert all(len(lv) > 0 for lv in truth_adj_bin_labels)
    assert all(len(lv) > 0 for lv in pred_op_labels)
    assert all(len(lv) > 0 for lv in pred_adj_bin_labels)
    metric_dict={
        "em_ratio": 0.,
        "avg_op_aligned_acc": 0.,
        "avg_adj_aligned_f1": 0.,
        "avg_op_overlap_ratio": 0.,
        "avg_edit_distance": 0.,
        "illegal_block_ratio": 0.,
    }
    n_instances = 0
    for i in tqdm(range(len(truth_op_labels)), desc="Computing block matching metrics", ascii=True):
        truth_op_list, pred_op_list = truth_op_labels[i], pred_op_labels[i][1:] # chop sos
        truth_adj_lists, pred_adj_lists = truth_adj_bin_labels[i], pred_adj_bin_labels[i][1:]
        assert len(truth_op_list) == len(truth_adj_lists)
        assert len(pred_op_list) == len(pred_adj_lists)
        while truth_op_list[-1] == pad_op_idx:
            truth_op_list.pop()
            truth_adj_lists.pop()
        while pred_op_list[-1] == pad_op_idx:
            pred_op_list.pop()
            pred_adj_lists.pop()
        edit_dist = compute_block_edit_distance_v2(truth_op_list, truth_adj_lists, pred_op_list, pred_adj_lists)
        metric_dict["avg_edit_distance"] += edit_dist
        if pred_op_list[-1] != eos_idx or is_illegal_block(pred_adj_lists):
            metric_dict["illegal_block_ratio"] += 1
        truth_adj_list = flatten_nested_lists(truth_adj_lists)
        pred_adj_list = flatten_nested_lists(pred_adj_lists)
        if all(list_element_wise_op(truth_op_list + truth_adj_list,
                                    pred_op_list + pred_adj_list,
                                    lambda v1,v2: v1==v2, pad_val=-1)):
            metric_dict["em_ratio"] += 1
        min_op_seq_len = min(len(truth_op_list), len(pred_op_list))
        min_adj_seq_len = min(len(truth_adj_list), len(pred_adj_list))
        metric_dict["avg_op_aligned_acc"] += sum(list_element_wise_op(truth_op_list[:min_op_seq_len],
                                                                      pred_op_list[:min_op_seq_len],
                                                                      lambda v1,v2: 1. if v1==v2 else 0.)) / min_op_seq_len

        metric_dict["avg_op_overlap_ratio"] += float(len(set(pred_op_list).intersection(set(truth_op_list)))) / len(set(truth_op_list))
        truth_adj_list = truth_adj_list[:min_adj_seq_len]
        pred_adj_list = pred_adj_list[:min_adj_seq_len]
        if all(v==0 for v in truth_adj_list): raise ValueError("Truth adj list should not be all zeroes")
        if not all(v == 0 for v in pred_adj_list):
            metric_dict["avg_adj_aligned_f1"] += f1_score(np.asarray(truth_adj_list), np.asarray(pred_adj_list))
        n_instances += 1
    for key in metric_dict:
        metric_dict[key] /= n_instances
    metric_dict["n_instances"] = n_instances
    return metric_dict


references = [[['<SOS>', 'Which', 'NFL', 'team', 'represented', 'the', 'AFC', 'at', 'Super', 'Bowl', '50', '?', '<EOS>']], [['<SOS>', 'Which', 'NFL', 'team', 'represented', 'the', 'NFC', 'at', 'Super', 'Bowl', '50', '?', '<EOS>']], [['<SOS>', 'Where', 'did', 'Super', 'Bowl', '50', 'take', 'place', '?', '<EOS>']], [['<SOS>', 'Where', 'did', 'Super', 'Bowl', '50', 'take', 'place', '?', '<EOS>']], [['<SOS>', 'Where', 'did', 'Super', 'Bowl', '50', 'take', 'place', '?', '<EOS>']], [['<SOS>', 'Which', 'NFL', 'team', 'won', 'Super', 'Bowl', '50', '?', '<EOS>']], [['<SOS>', 'What', 'color', 'was', 'used', 'to', 'emphasize', 'the', '50th', 'anniversary', 'of', 'the', 'Super', 'Bowl', '?', '<EOS>']], [['<SOS>', 'What', 'color', 'was', 'used', 'to', 'emphasize', 'the', '50th', 'anniversary', 'of', 'the', 'Super', 'Bowl', '?', '<EOS>']], [['<SOS>', 'What', 'was', 'the', 'theme', 'of', 'Super', 'Bowl', '50', '?', '<EOS>']], [['<SOS>', 'What', 'was', 'the', 'theme', 'of', 'Super', 'Bowl', '50', '?', '<EOS>']], [['<SOS>', 'What', 'was', 'the', 'theme', 'of', 'Super', 'Bowl', '50', '?', '<EOS>']], [['<SOS>', 'What', 'day', 'was', 'the', 'game', 'played', 'on', '?', '<EOS>']]]
candidates = [['defeated', '<OOV>', '<OOV>', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of'], ['Super', '<OOV>', 'title', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of'], ['Francisco', '<OOV>', '<OOV>', '?', '<OOV>', '<OOV>', '<OOV>', '<OOV>', 'game', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'Levi', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'Stadium', '<OOV>', '<OOV>', '<OOV>', '<OOV>', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of'], ['<OOV>', 'the', '<OOV>', '<OOV>', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of'], ['<OOV>', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of'], ['<OOV>', 'National', '<OOV>', '<OOV>', 'Denver', '<OOV>', 'Broncos', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', 'of', 'of', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', 'National', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>'], ['50th', '<OOV>', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', '50th', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of'], ['gold', '<OOV>', 'was', '<OOV>', '?', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'prominently', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>'], ['the', '<OOV>', 'the', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'could', '<OOV>', '<OOV>', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of'], ['suspending', '<OOV>', '<OOV>', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of'], ['<OOV>', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', '50th', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '?', '<OOV>', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of', 'of'], [',', '2016', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', ',', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>', '<OOV>']]
if __name__ == "__main__":
    # _pred_adj_lists = [[], [1], [1, 0], [0, 1, 1]]
    # print(is_illegal_block(_pred_adj_lists))
    # _pred_adj_lists = [[], [1], [1, 0], [0, 1, 0]]
    # print(is_illegal_block(_pred_adj_lists))
    # _pred_adj_lists = [[], [1], [0, 0], [1, 0, 0]]
    # print(is_illegal_block(_pred_adj_lists))
    # _pred_adj_lists = [[], [1], [1, 1], [0, 1, 0]]
    # print(is_illegal_block(_pred_adj_lists))
    _pred_adj_lists = [[1], [1, 1], [1, 1, 0], [0, 0, 0, 1], [0, 0, 0, 0, 1],[0, 0, 1, 0, 0, 1]]
    print(is_illegal_block(_pred_adj_lists))
    # rv = {}
    # for i, v in enumerate(references):
    #     gold_seg = v[0]
    #     pred_seg = candidates[i]
    #     multi_eval(pred_seg, gold_seg, result_dict=rv)
    # for k, v in rv.items():
    #     rv[k] = v / len(candidates)
    # print(rv)
    # print(post_evaluate_results_file("con_qa_test_results.txt"))
    # print(post_evaluate_results_file("con_crx_test_results.txt"))
