import numpy as np
from utils.misc_utils import *


def gen_block_visuals(blocks, group_dir_name, block_visual_dir,
                      model_name="model", prune_existing=True,
                      block_perfs=None, x_axis_idx=1, name_prefix="block"):
    directory = os.path.sep.join([block_visual_dir, group_dir_name])
    from utils.visual_utils import gviz_block_visualize
    if prune_existing and os.path.exists(directory):
        rmdir(directory)
    if not os.path.exists(directory):
        os.makedirs(directory)
    if block_perfs is not None:
        assert len(block_perfs) == len(blocks)
        tmp = list(zip(blocks, block_perfs))
        tmp.sort(key=lambda t: t[1][x_axis_idx])
        blocks = [t[0] for t in tmp]
        block_perfs = [t[1] for t in tmp]
    seen_blocks = set()
    block_index = 0
    for bi, block in enumerate(blocks):
        str_id = block.get_str_id()
        if str_id in seen_blocks: continue
        seen_blocks.add(str_id)
        filename = name_prefix + "_{}_{}".format(model_name, block_index)
        try:
            gviz_block_visualize(block.curr_node_label_list, block.curr_edge_list,
                                 filename=filename, directory=directory)
        except:
            pass
        block_index += 1
        if block_perfs is not None:
            write_line_to_file(filename + "|" + str(block_perfs[bi]), os.path.sep.join([directory, "metadata.txt"]))


def is_block_valid(state, output_node_label="output", input_node_idx=0):
    return get_num_block_violations(state, output_node_label, input_node_idx) == 0


def get_num_block_violations(state, output_node_label="output", input_node_idx=0):
    n_violations = 0
    if block_has_invalid_input(state):
        n_violations += 1
    if block_has_invalid_output(state, output_node_label):
        n_violations += 1
    if block_has_invalid_edges(state):
        n_violations += 1
    if len(get_output_node_indices(state.curr_edge_list)) > 1:
        n_violations += 1
    input_set = get_input_node_indices(state.curr_edge_list)
    if len(input_set) > 1 or input_node_idx not in input_set:
        n_violations += 1
    if block_has_loner_nodes(state):
        n_violations += 1
    return n_violations


def block_has_loner_nodes(state):
    connected_node_set = set(state.curr_edge_list[0]).union(set(state.curr_edge_list[1]))
    rv = len(state.curr_node_label_list) != len(connected_node_set)
    return rv


def block_has_invalid_edges(state):
    """
    Checks:
        1. If there is no valid edges
        2. If there is a node with no incoming or outgoing edges
    :param state:
    :return:
    """
    src_list, dst_list = state.curr_edge_list
    if len(src_list) == 0 or all(src == dst for src, dst in zip(src_list, dst_list)):
        return True
    nodes_with_edges = set(src_list).union(set(dst_list))
    if len(nodes_with_edges) != len(state.curr_node_label_list):
        return True
    return False


def block_has_invalid_input(state, input_node_idx=0):
    """
    Checks:
        1. If there is more than 1 input node
        2. If there is an edge going into an input node
    :param state:
    :param input_node_idx: by default, node 0 is always the input
    :return:
    """
    if len([n for n in state.curr_node_label_list if n == state.curr_node_label_list[input_node_idx]]) != 1:
        return True  # more than 1 input node
    edge_list = prune_self_edge_from_edge_list(state.curr_edge_list)
    return input_node_idx in edge_list[1]  # input node should never be a dst node


def block_has_invalid_output(state, output_node_label="output"):
    """
    Checks:
        1. If there is more than 1 output node
        2. If there is an edge coming from an output node
    :param state:
    :param output_node_label:
    :return:
    """
    nodes = [n for n in state.curr_node_label_list if n == output_node_label]
    if len(nodes) > 1:
        return True  # more than 1 output node
    if len(nodes) == 0:
        return True
    idx = None
    for i, n in enumerate(state.curr_node_label_list):
        if n == output_node_label:
            idx = i
            break
    edge_list = prune_self_edge_from_edge_list(state.curr_edge_list)
    return idx in edge_list[0]  # output node should never be a src node


def show_graph_with_labels(adj_matrix, node_labels):
    import matplotlib.pyplot as plt
    import networkx as nx
    if isinstance(adj_matrix, list):
        adj_matrix = np.asarray(adj_matrix)
    rows, cols = np.where(adj_matrix == 1)
    edges = list(zip(rows.tolist(), cols.tolist()))
    edges_with_labels = [(node_labels[ni1], node_labels[ni2]) for ni1, ni2 in edges]
    gr = nx.MultiDiGraph()
    gr.add_edges_from(edges_with_labels)
    pos = nx.spring_layout(gr)
    nx.draw_networkx_nodes(gr, pos, cmap=plt.get_cmap('jet'), node_size=500)
    nx.draw_networkx_labels(gr, pos)
    nx.draw_networkx_edges(gr, pos, edgelist=edges_with_labels, arrows=True)
    # gr.add_edges_from(edges)
    # nx.draw_networkx_edges(gr, node_size=500, labels=node_labels, with_labels=True)
    plt.show()


def truncate_batch_block_adj_lists(batch_adj_lists, base_idx_val=0,
                                   ignore_impossible_edges=False, flatten_inst=False):
    rv_batch = []
    for inst_adj_targets in batch_adj_lists:
        tmp_inst = []
        for ni, inst_adj_list in enumerate(inst_adj_targets):
            if not ignore_impossible_edges:
                assert all(v == 0 for v in inst_adj_list[ni + base_idx_val + 1:])
            tmp_inst.append(inst_adj_list[:ni + base_idx_val + 1])
        if flatten_inst:
            tmp_inst = flatten_nested_lists(tmp_inst)
        rv_batch.append(tmp_inst)
    return rv_batch


def edge_list_to_edge_matrix(edge_list, n_nodes):
    matrix = np.zeros((n_nodes, n_nodes), dtype=np.int8)
    for idx in range(len(edge_list[0])):
        matrix[edge_list[0][idx]][edge_list[1][idx]] = 1
    return matrix


def hash_module(matrix, labeling):
  """
  Computes a graph-invariance MD5 hash of the matrix and label pair.

  Args:
    matrix: np.ndarray square upper-triangular adjacency matrix.
    labeling: list of int labels of length equal to both dimensions of
      matrix.

  Returns:
    MD5 hash of the matrix and labeling.
  """
  import hashlib
  vertices = np.shape(matrix)[0]
  in_edges = np.sum(matrix, axis=0).tolist()
  out_edges = np.sum(matrix, axis=1).tolist()

  assert len(in_edges) == len(out_edges) == len(labeling)
  hashes = list(zip(out_edges, in_edges, labeling))
  hashes = [hashlib.md5(str(h).encode('utf-8')).hexdigest() for h in hashes]
  # Computing this up to the diameter is probably sufficient but since the
  # operation is fast, it is okay to repeat more times.
  for _ in range(vertices):
    new_hashes = []
    for v in range(vertices):
      in_neighbors = [hashes[w] for w in range(vertices) if matrix[w, v]]
      out_neighbors = [hashes[w] for w in range(vertices) if matrix[v, w]]
      new_hashes.append(hashlib.md5(
          (''.join(sorted(in_neighbors)) + '|' +
           ''.join(sorted(out_neighbors)) + '|' +
           hashes[v]).encode('utf-8')).hexdigest())
    hashes = new_hashes
  fingerprint = hashlib.md5(str(sorted(hashes)).encode('utf-8')).hexdigest()

  return fingerprint


def permute_graph(graph, label, permutation):
  """
  Permutes the graph and labels based on permutation.

  Args:
    graph: np.ndarray adjacency matrix.
    label: list of labels of same length as graph dimensions.
    permutation: a permutation list of ints of same length as graph dimensions.

  Returns:
    np.ndarray where vertex permutation[v] is vertex v from the original graph
  """
  # vertex permutation[v] in new graph is vertex v in the old graph
  forward_perm = zip(permutation, list(range(len(permutation))))
  inverse_perm = [x[1] for x in sorted(forward_perm)]
  edge_fn = lambda x, y: graph[inverse_perm[x], inverse_perm[y]] == 1
  new_matrix = np.fromfunction(np.vectorize(edge_fn),
                               (len(label), len(label)),
                               dtype=np.int8)
  new_label = [label[inverse_perm[i]] for i in range(len(label))]
  return new_matrix, new_label


def is_isomorphic(graph1, graph2):
  """Exhaustively checks if 2 graphs are isomorphic."""
  matrix1, label1 = np.array(graph1[0]), graph1[1]
  matrix2, label2 = np.array(graph2[0]), graph2[1]
  if np.shape(matrix1) != np.shape(matrix2):
      return False
  if len(label1) != len(label2):
      return False
  vertices = np.shape(matrix1)[0]
  # Note: input and output in our constrained graphs always map to themselves
  # but this script does not enforce that.
  for perm in itertools.permutations(range(0, vertices)):
    pmatrix1, plabel1 = permute_graph(matrix1, label1, perm)
    if np.array_equal(pmatrix1, matrix2) and plabel1 == label2:
      return True
  return False


def action_space_rand_sampling(searchable_ops, input_label, output_label, max_n_nodes=7,
                               n_samples=100000):
    blocks = []
    ops = list(searchable_ops) + [output_label]
    for _ in range(n_samples):
        curr_ops = [input_label]
        curr_edges = []
        for i in range(max_n_nodes - 1):
            op = random.choice(ops)
            op_adj = []
            for j in range(len(curr_ops)):
                conn = random.random() >= 0.5
                op_adj.append(conn)
            op_adj = op_adj[::-1]
            new_op_idx = len(curr_ops)
            for ni in range(len(curr_ops)):
                if op_adj[ni]:
                    curr_edges.append((ni, new_op_idx))
            curr_ops.append(op)
            if op == output_label:
                break
        curr_edges.sort()
        edge_list = [[t[0] for t in curr_edges], [t[1] for t in curr_edges]]
        blocks.append((curr_ops, edge_list))
    return blocks


if __name__ == "__main__":
    _samples = action_space_rand_sampling(("maxpool3x3", "conv1x1", "conv3x3"), "input", "output")
    print("done")

