import os
import numpy as np


def gviz_block_visualize(node_label_list, torch_geo_edge_list, filename, directory,
                         view=False, ignore_self_edges=True):
    """
    :param node_label_list: e.g. ["input", "maxpool3x3", "output"]
    :param torch_geo_edge_list: a list of two lists (src, dst), e.g. [ [0, 1, 0], [1, 2, 2] ]
    :param filename:
    :param directory:
    :param view: whether to display
    :param ignore_self_edges:
    :return:
    """
    if os.name == "nt":
        os.environ["PATH"] += os.pathsep + "C:\\Program Files (x86)\\Graphviz2.38\\bin\\"
    try:
        from graphviz import Digraph
    except ModuleNotFoundError:
        return
    dag = Digraph(name=filename, filename=filename, directory=directory)
    node_set = set()
    for ni, label in enumerate(node_label_list):
        dag.node(str(ni), label)
        node_set.add(str(ni))
    for src, dst in zip(torch_geo_edge_list[0], torch_geo_edge_list[1]):
        if ignore_self_edges and src == dst: continue
        assert str(src) in node_set, "node index {} not in node set {}".format(src, node_set)
        assert str(dst) in node_set, "node index {} not in node set {}".format(dst, node_set)
        dag.edge(str(src), str(dst))
    dag.render(view=view, format="png", cleanup=True)


class VisPlot:
    """
    Static version of the visdom visualizer
    """
    enable = False
    vis = None
    env = "VisPlot"
    index = {}

    def __init__(self):
        raise NotImplementedError("This class should be used statically")

    @staticmethod
    def plot(name, y, x, env=None, **kwargs):
        if not VisPlot.enable: return
        if VisPlot.vis is None:
            try:
                import visdom
                if env is None: env = VisPlot.env
                VisPlot.vis = visdom.Visdom(env=env)
            except ModuleNotFoundError:
                return
            except ConnectionRefusedError:
                return
        step = VisPlot.index.get(name, 0)
        VisPlot.vis.line(Y=np.array([y]), X=np.array([x]),
                         win=str(name), opts=dict(title=name),
                         update=None if step == 0 else 'append',
                         **kwargs)
        VisPlot.index[name] = step + 1

    @staticmethod
    def plot_ppo_loss(step, policy_loss, value_loss, entropy_loss, model_name="PPO", env=None):
        VisPlot.plot("{} train loss/policy loss".format(model_name), policy_loss, step, env=env)
        VisPlot.plot("{} train loss/value loss".format(model_name), value_loss, step, env=env)
        VisPlot.plot("{} train loss/entropy loss".format(model_name), entropy_loss, step, env=env)
        VisPlot.plot("{} train loss/total loss".format(model_name), policy_loss + value_loss + entropy_loss, step, env=env)

    @staticmethod
    def plot_ppo_final_reward(step, validity_reward, perf_reward, pareto_reward, model_name="PPO", env=None):
        VisPlot.plot("{} train reward/validity reward".format(model_name), validity_reward, step, env=env)
        VisPlot.plot("{} train reward/perf reward".format(model_name), perf_reward, step, env=env)
        VisPlot.plot("{} train reward/pareto reward".format(model_name), pareto_reward, step, env=env)
        VisPlot.plot("{} train reward/final reward".format(model_name), validity_reward + perf_reward + pareto_reward, step, env=env)

    @staticmethod
    def plot_ppo_step_reward(step, step_reward, model_name="PPO", env=None):
        VisPlot.plot("{} train reward/step reward".format(model_name), step_reward, step, env=env)

    @staticmethod
    def plot_discriminator_pred(step, truth_score, gen_score, model_name="PPO", env=None):
        VisPlot.plot("{} train dis pred/truth score".format(model_name), truth_score, step, env=env)
        VisPlot.plot("{} train dis pred/gen score".format(model_name), gen_score, step, env=env)

    @staticmethod
    def plot_perf_query(step, avg_acc, max_acc, avg_lat, min_lat, percent, model_name="PPO", env=None):
        VisPlot.plot("{} train perf query/avg_acc".format(model_name), avg_acc, step, env=env)
        VisPlot.plot("{} train perf query/max_acc".format(model_name), max_acc, step, env=env)
        VisPlot.plot("{} train perf query/avg_lat".format(model_name), avg_lat, step, env=env)
        # VisPlot.plot("{} train perf query/min_lat".format(model_name), min_lat, step, env=env)
        VisPlot.plot("{} train perf query/% predictor usage".format(model_name), percent, step, env=env)

    @staticmethod
    def plot_pareto_block_query(step, num_queried, model_name="PPO", env=None):
        VisPlot.plot("{} train pareto block query/num_queried".format(model_name), num_queried, step, env=env)

    @staticmethod
    def plot_block_properties(step, avg_num_nodes, avg_num_edges, model_name="PPO", env=None):
        VisPlot.plot("{} block properties/avg_num_nodes".format(model_name), avg_num_nodes, step, env=env)
        VisPlot.plot("{} block properties/avg_num_edges".format(model_name), avg_num_edges, step, env=env)

    @staticmethod
    def plot_block_shape_gen(step, n_times_generated, model_name="PPO", env=None):
        VisPlot.plot("{} desired block shape/n_times_generated".format(model_name), n_times_generated, step, env=env)

    @staticmethod
    def plot_block_node_gen(step, n_nodes_gen, avg_num_edges, model_name="PPO", env=None):
        VisPlot.plot("{} desired block nodes/n_times_generated".format(model_name), n_nodes_gen, step, env=env)
        VisPlot.plot("{} desired block nodes/avg_num_edges".format(model_name), avg_num_edges, step, env=env)

    @staticmethod
    def plot_block_edge_gen(step, n_edges_gen, model_name="PPO", env=None):
        VisPlot.plot("{} desired block edges/n_times_generated".format(model_name), n_edges_gen, step, env=env)


if  __name__ == "__main__":
    # from graphviz import Digraph
    # dot = Digraph(name="block1", filename="block2", directory="../logs/blocks")
    # dot.node('0', 'input')
    # dot.node('1', 'maxpool3x3')
    # dot.node('2', 'output')
    # dot.edges(['01', '12', '02'])
    # print(dot.source)
    # dot.render(view=True)
    print("done")
