import copy
import torch
import pickle
import random
import numpy as np
import torch_geometric
from params import *
from utils.model_utils import device, to_sorted_tensor, to_original_tensor, init_weights, fix_model_weights
from utils.nas_utils import hash_module, prune_self_edge_from_edge_list
from torch_geometric.data import Data, Batch
from utils.nas_utils import torch_geo_batch_to_data_list


class SimpleRNN(torch.nn.Module):

    def __init__(self, input_size, hidden_size,
                 n_layers=1, rnn_dir=2, dropout_prob=0.0, rnn_type="gru",
                 return_aggr_vector_only=False, return_output_vector_only=False):
        super(SimpleRNN, self).__init__()
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.n_layers = n_layers
        self.rnn_dir = rnn_dir
        self.rnn_type = rnn_type
        self.dropout_prob = dropout_prob
        self.return_aggr_vector_only = return_aggr_vector_only
        self.return_output_vector_only = return_output_vector_only
        if self.rnn_type.lower() == "lstm":
            self.rnn = torch.nn.LSTM(input_size=self.input_size,
                                     hidden_size=self.hidden_size,
                                     num_layers=self.n_layers,
                                     dropout=self.dropout_prob if self.n_layers > 1 else 0,
                                     batch_first=True,
                                     bidirectional=self.rnn_dir == 2)
        else:
            self.rnn = torch.nn.GRU(input_size=self.input_size,
                                    hidden_size=self.hidden_size,
                                    num_layers=self.n_layers,
                                    dropout=self.dropout_prob if self.n_layers > 1 else 0,
                                    batch_first=True,
                                    bidirectional=self.rnn_dir == 2)
        init_weights(self, base_model_type="rnn")

    def forward(self, embedded, hidden=None, input_seq_length_tsr=None, input_mask_tsr=None):
        if input_mask_tsr is not None and input_seq_length_tsr is None:
            input_seq_length_tsr = input_mask_tsr.sum(dim=-1).squeeze()
            if len(input_seq_length_tsr.shape) == 0:
                input_seq_length_tsr = input_seq_length_tsr.unsqueeze(0)
        self.rnn.flatten_parameters()
        if input_seq_length_tsr is None:
            outputs, hidden = self.rnn(embedded, hidden)
        else:
            embedded, sorted_lens, s_idx = to_sorted_tensor(embedded, input_seq_length_tsr, sort_dim=0)
            packed = torch.nn.utils.rnn.pack_padded_sequence(embedded, sorted_lens, batch_first=True)
            outputs, hidden = self.rnn(packed, hidden)
            outputs, sorted_lens = torch.nn.utils.rnn.pad_packed_sequence(outputs, batch_first=True)
            outputs = to_original_tensor(outputs, s_idx, sort_dim=0)

        if self.return_aggr_vector_only:
            if input_seq_length_tsr is not None:
                input_seq_length_tsr = input_seq_length_tsr.to(device())
                input_seq_length_tsr = input_seq_length_tsr - 1
                rv = torch.gather(outputs, 1, input_seq_length_tsr.view(-1, 1).unsqueeze(2).repeat(1, 1, outputs.size(-1)))
            else:
                rv = outputs[:, -1, :].unsqueeze(1)
            return rv
        elif self.return_output_vector_only:
            return outputs
        else:
            return outputs, hidden


class SimpleBlock:

    def __init__(self, node_label_list, node_feature_tsr, edge_list):
        self.curr_node_label_list = node_label_list
        self.curr_node_feature_tsr = node_feature_tsr
        self.curr_edge_list = copy.deepcopy(edge_list)

    def __len__(self):
        return len(self.curr_node_label_list)

    def __str__(self):
        return "SimpleBlock{}".format(str(self.curr_node_label_list)) + ", Edges: " + str(self.curr_edge_list)

    def __repr__(self):
        return str(self)

    def __deepcopy__(self, memodict={}):
        _state = SimpleBlock(
            node_label_list=[v for v in self.curr_node_label_list],
            node_feature_tsr=copy.deepcopy(self.curr_node_feature_tsr),
            edge_list=copy.deepcopy(self.curr_edge_list),
        )
        return _state

    def to(self, dev):
        self.curr_node_feature_tsr = self.curr_node_feature_tsr.to(dev)
        return self

    def get_str_id(self):
        from utils.nas_utils import edge_list_to_edge_matrix
        if isinstance(self.curr_node_feature_tsr, torch.Tensor):
            node_indices = self.curr_node_feature_tsr.tolist()
        else:
            node_indices = self.curr_node_feature_tsr
        src_list, dst_list = prune_self_edge_from_edge_list(self.curr_edge_list)
        matrix = edge_list_to_edge_matrix([src_list, dst_list], len(node_indices))
        fingerprint = hash_module(matrix, node_indices)
        return fingerprint


class SimpleBlockSampler:

    def __init__(self, searchable_ops, input_label, output_label, block_maker,
                 topo_file=P_SEP.join([DATA_DIR, "random_max_n7_topos.pkl"]),
                 ext_samples=None, max_unique_samples=None, n_nodes_set=None):
        self.input_label = input_label
        self.output_label = output_label
        self.block_maker = block_maker
        self.searchable_ops = searchable_ops
        with open(topo_file, "rb") as f:
            self.topos = pickle.load(f)
        self.ext_samples = ext_samples
        if n_nodes_set is not None:
            print("SimpleBlockSampler.n_nodes_set={}".format(n_nodes_set))
            filtered_topos = []
            for n_nodes, adj_list in self.topos:
                if n_nodes in n_nodes_set:
                    filtered_topos.append((n_nodes, adj_list))
            self.topos = filtered_topos
        if max_unique_samples is not None:
            print("SimpleBlockSampler.max_unique_samples={}".format(max_unique_samples))
        self.max_unique_samples = max_unique_samples
        self.past_sample_keys = set()
        self.past_samples = []

    def sample(self):
        if self.ext_samples is not None:
            return random.choice(self.ext_samples)

        if self.max_unique_samples is not None and len(self.past_samples) >= self.max_unique_samples:
            return random.choice(self.past_samples)

        n_nodes, edge_list = random.choice(self.topos)
        ops = [self.input_label] + \
              [random.choice(self.searchable_ops) for _ in range(n_nodes - 2)] + \
              [self.output_label]
        block = self.block_maker(ops, edge_list)

        if self.max_unique_samples is not None:
            key = block.get_str_id()
            if key not in self.past_sample_keys:
                self.past_sample_keys.add(key)
                self.past_samples.append(block)
        return block


class GraphEncoder(torch.nn.Module):

    def __init__(self, in_channels, hidden_size, out_channels, gnn_constructor, node_pre_embedding,
                 activ=torch.nn.Tanh(), n_layers=4, dropout_prob=0.0,
                 add_normal_prior=False, return_graph_embedding=True):
        super(GraphEncoder, self).__init__()
        self.node_pre_embedding = node_pre_embedding
        self.in_channels = in_channels
        self.hidden_size = hidden_size
        self.out_channels = out_channels
        self.gnn_layers = torch.nn.ModuleList()
        for i in range(n_layers):
            input_size, output_size = hidden_size, hidden_size
            if i == 0:
                input_size = in_channels
            if i == n_layers - 1:
                output_size = out_channels
            gnn_layer = gnn_constructor(input_size, output_size)
            self.gnn_layers.append(gnn_layer)
        self.activ = activ
        self.dropout = torch.nn.Dropout(dropout_prob)
        self.init_ff = torch.nn.Linear(2 * in_channels, in_channels)
        self.add_normal_prior = add_normal_prior
        self.return_graph_embedding = return_graph_embedding

    def add_prior(self, embedding):
        prior = np.random.normal(size=embedding.shape)
        prior = torch.from_numpy(prior).float().to(device())
        embedding = torch.cat([embedding, prior], dim=-1)
        return self.init_ff(embedding)

    def forward(self, batch_node_tsr, edge_tsr_list, batch_last_node_idx_list):
        node_embedding = self.node_pre_embedding(batch_node_tsr.to(device()))
        if self.add_normal_prior and node_embedding.shape[1] == 1:
            node_embedding = self.add_prior(node_embedding)
        data_list = [Data(x=node_embedding[i,:], edge_index=edge_tsr_list[i].to(device()), edge_attr=None)
                     for i in range(node_embedding.shape[0])]
        torch_geo_batch = Batch.from_data_list(data_list)
        edge_index_tsr = torch_geo_batch.edge_index
        curr_gnn_output = torch_geo_batch.x
        for li, gnn_layer in enumerate(self.gnn_layers):
            node_embedding_tsr = gnn_layer(curr_gnn_output, edge_index_tsr)
            curr_gnn_output = self.activ(node_embedding_tsr)
        curr_gnn_output = self.dropout(curr_gnn_output)
        batch_embedding_list = torch_geo_batch_to_data_list(curr_gnn_output, batch_last_node_idx_list)
        if self.return_graph_embedding:
            # Take output node embedding as the final embedding
            batch_embedding = torch.cat([t[-1, :].reshape(1, -1) for t in batch_embedding_list], dim=0)
        else:
            batch_embedding = torch.cat([t.unsqueeze(0) for t in batch_embedding_list], dim=0)
        return batch_embedding


class SiamesePairGraphEncoder(torch.nn.Module):

    def __init__(self, encoder):
        super(SiamesePairGraphEncoder, self).__init__()
        self.encoder = encoder

    def forward(self, g1, g2):
        embedding_1 = self.encoder(*g1)
        embedding_2 = self.encoder(*g2)
        return torch.cat([embedding_1, embedding_2], dim=-1)


def make_gnn_encoder(in_channels, hidden_size, out_channels, num_unique_nodes,
                     gnn_constructor, n_layers, dropout_prob):
    node_embedding_layer = torch.nn.Embedding(num_unique_nodes, in_channels)
    model = GraphEncoder(in_channels, hidden_size, out_channels,
                         gnn_constructor, node_embedding_layer,
                         n_layers=n_layers, dropout_prob=dropout_prob)
    return model


class GraphRegressor(torch.nn.Module):

    def __init__(self, encoder, hidden_size, activ=None, freeze_encoder=False):
        super(GraphRegressor, self).__init__()
        self.encoder = encoder
        self.regressor = torch.nn.Linear(hidden_size, 1)
        self.activ = activ
        self.freeze_encoder = freeze_encoder
        if self.freeze_encoder:
            fix_model_weights(self.encoder)

    def forward(self, batch_node_tsr, edge_tsr_list, batch_last_node_idx_list):
        batch_embedding = self.encoder(batch_node_tsr, edge_tsr_list, batch_last_node_idx_list)
        out = self.regressor(batch_embedding)
        if self.activ is not None:
            out = self.activ(out)
        return out


class GraphComparator(torch.nn.Module):

    def __init__(self, encoder, hidden_size, freeze_encoder=False):
        super(GraphComparator, self).__init__()
        self.encoder = encoder
        self.comparator = torch.nn.Linear(2 * hidden_size, 2)
        self.freeze_encoder = freeze_encoder
        if self.freeze_encoder:
            fix_model_weights(self.encoder)

    def forward(self, g1, g2):
        batch_node_tsr_1, edge_tsr_list_1, batch_last_node_idx_list_1 = g1
        batch_node_tsr_2, edge_tsr_list_2, batch_last_node_idx_list_2 = g2
        embedding_1 = self.encoder(batch_node_tsr_1, edge_tsr_list_1, batch_last_node_idx_list_1)
        embedding_2 = self.encoder(batch_node_tsr_2, edge_tsr_list_2, batch_last_node_idx_list_2)
        batch_embedding = torch.cat([embedding_1, embedding_2], dim=-1)
        logits = self.comparator(batch_embedding)
        return logits


class RNNIndexEncoder(torch.nn.Module):

    def __init__(self, n_unique_inds, input_size, hidden_size,
                 n_layers=1, rnn_dir=2, dropout_prob=0.0, rnn_type="gru",
                 return_aggr_vector_only=False, return_output_vector_only=False,
                 add_normal_prior=False):
        super(RNNIndexEncoder, self).__init__()
        self.pre_embedding = torch.nn.Embedding(n_unique_inds, input_size)
        self.rnn = SimpleRNN(input_size, hidden_size,
                             n_layers, rnn_dir=rnn_dir, dropout_prob=dropout_prob, rnn_type=rnn_type,
                             return_aggr_vector_only=return_aggr_vector_only,
                             return_output_vector_only=return_output_vector_only)
        self.add_normal_prior = add_normal_prior
        self.init_ff = torch.nn.Linear(input_size * 2, input_size)

    def add_prior(self, embedding):
        prior = np.random.normal(size=embedding.shape)
        prior = torch.from_numpy(prior).float().to(device())
        embedding = torch.cat([embedding, prior], dim=-1)
        return self.init_ff(embedding)

    def forward(self, x):
        embedding = self.pre_embedding(x.to(device()))
        if self.add_normal_prior and embedding.shape[1] == 1:
            embedding = self.add_prior(embedding)
        return self.rnn(embedding)


class GINEncoder(torch.nn.Module):

    def __init__(self, in_channels, hidden_size, out_channels,
                 inner_nn_constructor, node_pre_embedding,
                 eps=0., train_eps=False, activ=torch.nn.Tanh(), n_layers=4, dropout_prob=0.0,
                 add_normal_prior=False, return_graph_embedding=True):
        super(GINEncoder, self).__init__()
        self.node_pre_embedding = node_pre_embedding
        self.in_channels = in_channels
        self.hidden_size = hidden_size
        self.out_channels = out_channels
        self.gnn_layers = torch.nn.ModuleList()
        for i in range(n_layers):
            input_size, output_size = hidden_size, hidden_size
            if i == 0:
                input_size = in_channels
            if i == n_layers - 1:
                output_size = out_channels
            nn = inner_nn_constructor(input_size, output_size)
            gnn_layer = torch_geometric.nn.GINConv(nn, eps=eps, train_eps=train_eps)
            self.gnn_layers.append(gnn_layer)
        self.activ = activ
        self.dropout = torch.nn.Dropout(dropout_prob)
        self.init_ff = torch.nn.Linear(2 * in_channels, in_channels)
        self.add_normal_prior = add_normal_prior
        self.return_graph_embedding = return_graph_embedding

    def add_prior(self, embedding):
        prior = np.random.normal(size=embedding.shape)
        prior = torch.from_numpy(prior).float().to(device())
        embedding = torch.cat([embedding, prior], dim=-1)
        return self.init_ff(embedding)

    def forward(self, batch_node_tsr, edge_tsr_list, batch_last_node_idx_list):
        node_embedding = self.node_pre_embedding(batch_node_tsr.to(device()))
        if self.add_normal_prior and node_embedding.shape[1] == 1:
            node_embedding = self.add_prior(node_embedding)
        data_list = [Data(x=node_embedding[i,:], edge_index=edge_tsr_list[i].to(device()), edge_attr=None)
                     for i in range(node_embedding.shape[0])]
        torch_geo_batch = Batch.from_data_list(data_list)
        edge_index_tsr = torch_geo_batch.edge_index
        curr_gnn_output = torch_geo_batch.x
        for li, gnn_layer in enumerate(self.gnn_layers):
            node_embedding_tsr = gnn_layer(curr_gnn_output, edge_index_tsr)
            curr_gnn_output = self.activ(node_embedding_tsr)
        curr_gnn_output = self.dropout(curr_gnn_output)
        batch_embedding_list = torch_geo_batch_to_data_list(curr_gnn_output, batch_last_node_idx_list)
        if self.return_graph_embedding:
            # Take output node embedding as the final embedding
            batch_embedding = torch.cat([t[-1, :].reshape(1, -1) for t in batch_embedding_list], dim=0)
        else:
            batch_embedding = torch.cat([t.unsqueeze(0) for t in batch_embedding_list], dim=0)
        return batch_embedding


def make_gin_encoder(in_channels, hidden_size, out_channels, num_unique_nodes,
                     inner_nn_constructor, n_layers, dropout_prob, eps=0.):
    node_embedding_layer = torch.nn.Embedding(num_unique_nodes, in_channels)
    model = GINEncoder(in_channels, hidden_size, out_channels,
                       inner_nn_constructor, node_embedding_layer,
                       eps=eps, n_layers=n_layers, dropout_prob=dropout_prob)
    return model

