import torch
import torch.nn as nn


OPS = {
    'none'           : lambda C, stride, affine: Zero(stride),
    'avgpool3x3'     : lambda C, stride, affine: nn.Sequential(nn.BatchNorm2d(C, affine=affine),
                                                               nn.AvgPool2d(3, stride=stride, padding=1,
                                                                            count_include_pad=False)),
    'maxpool3x3'     : lambda C, stride, affine: nn.Sequential(nn.BatchNorm2d(C, affine=affine),
                                                               nn.MaxPool2d(3, stride=stride, padding=1)),
    'skip_connect'   : lambda C, stride, affine: BNIdentity(C, affine) \
                                                 if stride == 1 else FactorizedReduce(C, C, affine=affine),
    'scarlet'        : lambda C, stride, affine: nn.Conv2d(C, C, stride, stride=stride, bias=False),
    'sep_conv_3x3'   : lambda C, stride, affine: SepConv(C, C, 3, stride, 1, affine=affine),
    'sep_conv_5x5'   : lambda C, stride, affine: SepConv(C, C, 5, stride, 2, affine=affine),
    'sep_conv_7x7'   : lambda C, stride, affine: SepConv(C, C, 7, stride, 3, affine=affine),
    'dil_conv_3x3'   : lambda C, stride, affine: DilConv(C, C, 3, stride, 2, 2, affine=affine),
    'dil_conv_5x5'   : lambda C, stride, affine: DilConv(C, C, 5, stride, 4, 2, affine=affine),
    'conv_7x1_1x7'   : lambda C, stride, affine: nn.Sequential(
                        nn.ReLU(inplace=False),
                        nn.Conv2d(C, C, (1, 7), stride=(1, stride), padding=(0, 3), bias=False),
                        nn.Conv2d(C, C, (7, 1), stride=(stride, 1), padding=(3, 0), bias=False),
                        nn.BatchNorm2d(C, affine=affine)),
    'conv1x1-bn-relu': lambda C, stride, affine: ConvBNReLU(C, C, 1, stride, padding=0, affine=affine),
    'conv3x3-bn-relu': lambda C, stride, affine: ConvBNReLU(C, C, 3, stride, padding=1, affine=affine),
    # NB201
    'nor_conv_3x3'   : lambda C, stride, affine: ReLUConvBN(C, C, 3, stride, padding=1, affine=affine),
    'nor_conv_1x1'   : lambda C, stride, affine: ReLUConvBN(C, C, 1, stride, padding=0, affine=affine),
    'avg_pool_3x3'   : lambda C, stride, affine: nn.Sequential(nn.BatchNorm2d(C, affine=affine),
                                                               nn.AvgPool2d(3, stride=stride, padding=1,
                                                                            count_include_pad=False)),
}


class ConvBNReLU(nn.Module):

    def __init__(self, C_in, C_out, kernel_size, stride, padding=0, affine=True):
        super(ConvBNReLU, self).__init__()
        self.op = nn.Sequential(
            nn.Conv2d(C_in, C_out, kernel_size=kernel_size, stride=stride, padding=padding, bias=False),
            nn.BatchNorm2d(C_out, affine=affine),
            nn.ReLU(inplace=False)
        )

    def forward(self, x):
        return self.op(x)


class ConvBN(nn.Module):

    def __init__(self, C_in, C_out, kernel_size, stride, padding=0, affine=True):
        super(ConvBN, self).__init__()
        self.op = nn.Sequential(
            nn.Conv2d(C_in, C_out, kernel_size=kernel_size, stride=stride, padding=padding, bias=False),
            nn.BatchNorm2d(C_out, affine=affine)
        )

    def forward(self, x):
        return self.op(x)


class ReLUConvBN(nn.Module):

    def __init__(self, C_in, C_out, kernel_size, stride, padding, affine=True):
        super(ReLUConvBN, self).__init__()
        self.op = nn.Sequential(
            nn.ReLU(inplace=False),
            nn.Conv2d(C_in, C_out, kernel_size, stride=stride, padding=padding),
            nn.BatchNorm2d(C_out, affine=affine)
        )

    def forward(self, x):
        return self.op(x)


class DilConv(nn.Module):

    def __init__(self, C_in, C_out, kernel_size, stride, padding, dilation, affine=True):
        super(DilConv, self).__init__()
        self.op = nn.Sequential(
            nn.ReLU(inplace=False),
            nn.Conv2d(C_in, C_in, kernel_size=kernel_size, stride=stride, padding=padding, dilation=dilation,
                      groups=C_in, bias=False),
            nn.Conv2d(C_in, C_out, kernel_size=1, padding=0, bias=False),
            nn.BatchNorm2d(C_out, affine=affine),
        )

    def forward(self, x):
        return self.op(x)


class SepConv(nn.Module):

    def __init__(self, C_in, C_out, kernel_size, stride, padding, affine=True):
        super(SepConv, self).__init__()
        self.op = nn.Sequential(
            nn.ReLU(inplace=False),
            nn.Conv2d(C_in, C_in, kernel_size=kernel_size, stride=stride, padding=padding, groups=C_in, bias=False),
            nn.Conv2d(C_in, C_in, kernel_size=1, padding=0, bias=False),
            nn.BatchNorm2d(C_in, affine=affine),
            nn.ReLU(inplace=False),
            nn.Conv2d(C_in, C_in, kernel_size=kernel_size, stride=1, padding=padding, groups=C_in, bias=False),
            nn.Conv2d(C_in, C_out, kernel_size=1, padding=0, bias=False),
            nn.BatchNorm2d(C_out, affine=affine),
        )

    def forward(self, x):
        return self.op(x)


class Identity(nn.Module):

    def __init__(self):
        super(Identity, self).__init__()

    def forward(self, x, *args):
        return x


class BNIdentity(nn.Module):

    def __init__(self, C_in, affine=True):
        super(BNIdentity, self).__init__()
        self.bn = nn.BatchNorm2d(C_in, affine=affine)

    def forward(self, x, *args):
        return self.bn(x)


class Zero(nn.Module):

    def __init__(self, stride):
        super(Zero, self).__init__()
        self.stride = stride

    def forward(self, x):
        if self.stride == 1:
            return x.mul(0.)
        return x[:, :, ::self.stride, ::self.stride].mul(0.)


class FactorizedReduce(nn.Module):

    def __init__(self, C_in, C_out, affine=True):
        super(FactorizedReduce, self).__init__()
        assert C_out % 2 == 0
        self.relu = nn.ReLU(inplace=False)
        self.conv_1 = nn.Conv2d(C_in, C_out // 2, 1, stride=2, padding=0, bias=False)
        self.conv_2 = nn.Conv2d(C_in, C_out // 2, 1, stride=2, padding=0, bias=False)
        self.bn = nn.BatchNorm2d(C_out, affine=affine)

    def forward(self, x):
        x = self.relu(x)
        out = torch.cat([self.conv_1(x), self.conv_2(x[:, :, 1:, 1:])], dim=1)
        out = self.bn(out)
        return out


def benchmark_op_flops(C_in=16, W=32, H=32):
    from thop import profile
    from utils.model_utils import device
    rand_input = torch.randn(1, C_in, W, H).to(device())
    pairs = []
    for name, op in OPS.items():
        model = op(C_in, stride=1, affine=True).to(device())
        macs, params = profile(model, inputs=(rand_input,))
        pairs.append((name, macs))
    pairs.sort(key=lambda t: t[1], reverse=True)
    print(pairs)


if __name__ == "__main__":
    benchmark_op_flops()

