# Using ProxylessNAS supernet from:
    # ProxylessNAS: Direct Neural Architecture Search on Target Task and Hardware
    # Han Cai, Ligeng Zhu, Song Han
    # International Conference on Learning Representations (ICLR), 2019.

import argparse
import os

from proxyless_nas.search.models import ImagenetRunConfig
from proxyless_nas.search.nas_manager import *
from proxyless_nas.search.models.super_nets.super_proxyless import SuperProxylessNASNets
from model_ppo_gan_single_path import SPBlockState
from constants import *

from ast import literal_eval
import pickle
import torch
import re
import time
import copy
import random
import subprocess
from tqdm import tqdm
from model_ppo import BlockPerfEvaluator
from utils.efficientnet_utils import *
from utils.model_utils import device, can_parallel, ExponentialMovingAverage, recur_tsr_op
from utils.darts_utils import AverageMeter, accuracy
from utils.debug_utils import IntervalInfo as interval_info
from utils.misc_utils import rmdir

# ref values
ref_values = {
    'flops': {
        '0.35': 59 * 1e6,
        '0.50': 97 * 1e6,
        '0.75': 209 * 1e6,
        '1.00': 300 * 1e6,
        '1.30': 509 * 1e6,
        '1.40': 582 * 1e6,
    },
    # ms
    'mobile': {
        '1.00': 80,
    },
    'cpu': {},
    'gpu8': {},
}

parser = argparse.ArgumentParser()
parser.add_argument('--path', type=str, default=os.getcwd())
parser.add_argument('--gpu', help='gpu available', default='0,1,2,3,4,5,6,7')
parser.add_argument('--resume', action='store_true') #'store_true'
parser.add_argument('--debug', help='freeze the weight parameters', action='store_true')
parser.add_argument('--manual_seed', default=0, type=int)

""" run config """
parser.add_argument('--n_epochs', type=int, default=120)#120
parser.add_argument('--init_lr', type=float, default=0.025)
parser.add_argument('--lr_schedule_type', type=str, default='cosine')
# lr_schedule_param

parser.add_argument('--dataset', type=str, default='imagenet', choices=['imagenet'])
parser.add_argument('--train_batch_size', type=int, default=256)
parser.add_argument('--test_batch_size', type=int, default=1000)
parser.add_argument('--valid_size', type=int, default=50000)

parser.add_argument('--opt_type', type=str, default='sgd', choices=['sgd'])
parser.add_argument('--momentum', type=float, default=0.9)  # opt_param
parser.add_argument('--no_nesterov', action='store_true')  # opt_param
parser.add_argument('--weight_decay', type=float, default=4e-5)
parser.add_argument('--label_smoothing', type=float, default=0.1)
parser.add_argument('--no_decay_keys', type=str, default=None, choices=[None, 'bn', 'bn#bias'])

parser.add_argument('--model_init', type=str, default='he_fout', choices=['he_fin', 'he_fout'])
parser.add_argument('--init_div_groups', action='store_true')
parser.add_argument('--validation_frequency', type=int, default=1)
parser.add_argument('--print_frequency', type=int, default=10)

parser.add_argument('--n_worker', type=int, default=32)
parser.add_argument('--resize_scale', type=float, default=0.08)
parser.add_argument('--distort_color', type=str, default='normal', choices=['normal', 'strong', 'None'])

""" net config """
parser.add_argument('--width_stages', type=str, default='24,40,80,96,192,320')
parser.add_argument('--n_cell_stages', type=str, default='4,4,4,4,4,1')
parser.add_argument('--stride_stages', type=str, default='2,2,2,1,2,1')
parser.add_argument('--width_mult', type=float, default=1.0)
parser.add_argument('--bn_momentum', type=float, default=0.1)
parser.add_argument('--bn_eps', type=float, default=1e-3)
parser.add_argument('--dropout', type=float, default=0)

# architecture search config
""" arch search algo and warmup """
parser.add_argument('--arch_algo', type=str, default='rl', choices=['grad', 'rl']) #'grad'
parser.add_argument('--warmup_epochs', type=int, default=40)
""" shared hyper-parameters """
parser.add_argument('--arch_init_type', type=str, default='normal', choices=['normal', 'uniform'])
parser.add_argument('--arch_init_ratio', type=float, default=1e-3)
parser.add_argument('--arch_opt_type', type=str, default='adam', choices=['adam'])
parser.add_argument('--arch_lr', type=float, default=1e-3)
parser.add_argument('--arch_adam_beta1', type=float, default=0)  # arch_opt_param
parser.add_argument('--arch_adam_beta2', type=float, default=0.999)  # arch_opt_param
parser.add_argument('--arch_adam_eps', type=float, default=1e-8)  # arch_opt_param
parser.add_argument('--arch_weight_decay', type=float, default=0)
parser.add_argument('--target_hardware', type=str, default=None, choices=['mobile', 'cpu', 'gpu8', 'flops', None])
""" Grad hyper-parameters """
parser.add_argument('--grad_update_arch_param_every', type=int, default=5)
parser.add_argument('--grad_update_steps', type=int, default=1)
parser.add_argument('--grad_binary_mode', type=str, default='full_v2', choices=['full_v2', 'full', 'two'])
parser.add_argument('--grad_data_batch', type=int, default=None)
parser.add_argument('--grad_reg_loss_type', type=str, default='mul#log', choices=['add#linear', 'mul#log'])
parser.add_argument('--grad_reg_loss_lambda', type=float, default=1e-1)  # grad_reg_loss_params
parser.add_argument('--grad_reg_loss_alpha', type=float, default=0.2)  # grad_reg_loss_params
parser.add_argument('--grad_reg_loss_beta', type=float, default=0.3)  # grad_reg_loss_params
""" RL hyper-parameters """
parser.add_argument('--rl_batch_size', type=int, default=10)
parser.add_argument('--rl_update_per_epoch', action='store_true')
parser.add_argument('--rl_update_steps_per_epoch', type=int, default=300)
parser.add_argument('--rl_baseline_decay_weight', type=float, default=0.99)
parser.add_argument('--rl_tradeoff_ratio', type=float, default=0.1)

def get_mask_layers_with_shortcut(width_stages=[24,40,80,96,192,320], n_cell_stages=[4,4,4,4,4,1], stride_stages=[2,2,2,1,2,1]):
    ith_layer = 0
    list_layers_with_shortcut = []
    for width, n_cell, s in zip(width_stages, n_cell_stages, stride_stages):
        for i in range(n_cell):
            if i == 0:
                stride = s
            else:
                stride = 1
            # conv
            if stride == 1 and input_channel == width:
                list_layers_with_shortcut.append(ith_layer)
            input_channel = width
            ith_layer += 1
    mask_layers_with_shortcut = [i in list_layers_with_shortcut for i in range(sum(n_cell_stages))]
    return mask_layers_with_shortcut

def convert_SPBlockState_to_idxs(block: SPBlockState):
    labels = block.curr_node_label_list[1:-1]
    assert len(labels) == len(get_mask_layers_with_shortcut())
    idxs = [PN_OP2IDX_SUPERNET[lab] for lab in labels]
    return idxs

def is_SPBlockState_valid(block: SPBlockState):
    idxs = convert_SPBlockState_to_idxs(block)
    mask_layers_with_shortcut = get_mask_layers_with_shortcut()
    for i in range(len(mask_layers_with_shortcut)):
        # check if this layer cannot have shortcut
        if mask_layers_with_shortcut[i] == False:
            # if it uses the last candidate "Zero", than it is invalid
            if idxs[i] == (len(PN_OP2IDX_SUPERNET) - 1):
                return False
    # otherwise it is valid architecture
    return True

def make_proxylessnas_supernet(conv_candidates, model_fname):
    '''
    Example Usage:
    from ppo_gan_pn_model import make_proxylessnas_supernet
    conv_candidates = [
        '3x3_MBConv3', '3x3_MBConv6',
        '5x5_MBConv3', '5x5_MBConv6',
        '7x7_MBConv3', '7x7_MBConv6',
    ]
    super_net, arch_search_run_manager = make_proxylessnas_supernet(conv_candidates, "proxyless_nas/checkpoint/warmup.pth.tar",)
    valid_res, flops, latency = arch_search_run_manager.validate_with_idxs([0 for i in range(21)])
    valid_loss, valid_top1_acc, valid_top5_acc = valid_res
    '''
    args, unknown = parser.parse_known_args()
    torch.manual_seed(args.manual_seed)
    torch.cuda.manual_seed_all(args.manual_seed)
    np.random.seed(args.manual_seed)

    os.environ['CUDA_VISIBLE_DEVICES'] = args.gpu

    os.makedirs(args.path, exist_ok=True)

    # build run config from args
    args.lr_schedule_param = None
    args.opt_param = {
        'momentum': args.momentum,
        'nesterov': not args.no_nesterov,
    }
    run_config = ImagenetRunConfig(
        **args.__dict__
    )

    # debug, adjust run_config
    if args.debug:
        run_config.train_batch_size = 256
        run_config.test_batch_size = 256
        run_config.valid_size = 256
        run_config.n_worker = 0

    width_stages_str = '-'.join(args.width_stages.split(','))
    # build net from args
    args.width_stages = [int(val) for val in args.width_stages.split(',')]
    args.n_cell_stages = [int(val) for val in args.n_cell_stages.split(',')]
    args.stride_stages = [int(val) for val in args.stride_stages.split(',')]
    args.conv_candidates = conv_candidates
    super_net = SuperProxylessNASNets(
        width_stages=args.width_stages, n_cell_stages=args.n_cell_stages, stride_stages=args.stride_stages,
        conv_candidates=args.conv_candidates, n_classes=run_config.data_provider.n_classes, width_mult=args.width_mult,
        bn_param=(args.bn_momentum, args.bn_eps), dropout_rate=args.dropout
    )

    # build arch search config from args
    if args.arch_opt_type == 'adam':
        args.arch_opt_param = {
            'betas': (args.arch_adam_beta1, args.arch_adam_beta2),
            'eps': args.arch_adam_eps,
        }
    else:
        args.arch_opt_param = None
    if args.target_hardware is None:
        args.ref_value = None
    else:
        args.ref_value = ref_values[args.target_hardware]['%.2f' % args.width_mult]
    if args.arch_algo == 'grad':
        from proxyless_nas.search.nas_manager import GradientArchSearchConfig
        if args.grad_reg_loss_type == 'add#linear':
            args.grad_reg_loss_params = {'lambda': args.grad_reg_loss_lambda}
        elif args.grad_reg_loss_type == 'mul#log':
            args.grad_reg_loss_params = {
                'alpha': args.grad_reg_loss_alpha,
                'beta': args.grad_reg_loss_beta,
            }
        else:
            args.grad_reg_loss_params = None
        arch_search_config = GradientArchSearchConfig(**args.__dict__)
    elif args.arch_algo == 'rl':
        from proxyless_nas.search.nas_manager import RLArchSearchConfig
        arch_search_config = RLArchSearchConfig(**args.__dict__)
    else:
        raise NotImplementedError

    print('Run config:')
    for k, v in run_config.config.items():
        print('\t%s: %s' % (k, v))
    print('Architecture Search config:')
    for k, v in arch_search_config.config.items():
        print('\t%s: %s' % (k, v))

    # arch search run manager
    arch_search_run_manager = ArchSearchRunManager(args.path, super_net, run_config, arch_search_config)

    # load the warmed-up supernet weight
    arch_search_run_manager.load_model(model_fname=model_fname)
    
    return super_net, arch_search_run_manager

class ProxylessNASSupernetEvaluator(BlockPerfEvaluator):

    def __init__(self, model, arch_search_run_manager, train_loader, test_loader, lat_predictor, optim_maker, n_ops,
                 criterion=torch.nn.CrossEntropyLoss(), default_clue=1.0, enable_perf_memo=True):
        super(ProxylessNASSupernetEvaluator, self).__init__(default_clue=default_clue, enable_perf_memo=enable_perf_memo)
        self.model = model
        self.train_loader = train_loader
        self.test_loader = test_loader
        self.criterion = criterion.to(device())
        self.lat_predictor = lat_predictor
        self.optim_maker = optim_maker
        self.n_ops = n_ops
        self.arch_search_run_manager = arch_search_run_manager

    def get_perf_reward(self, state, model=None):
        raise NotImplementedError

    def get_perf_values(self, block: SPBlockState, model=None):
        if model is None:
            model = self.model
        key = block.get_str_id()
        idxs = convert_SPBlockState_to_idxs(block)
        if self.enable_perf_memo and key in self.perf_memo:
            return self.perf_memo[key]
        model = model.to(device())
        with torch.no_grad():
            valid_res, flops, latency = self.arch_search_run_manager.validate_with_idxs(idxs)
            valid_loss, valid_top1_acc, valid_top5_acc = valid_res
        pred_acc = valid_top1_acc
        pred_acc /= 100.
        pred_lat = self.lat_predictor(block) if self.lat_predictor is not None else 1.0
        self.perf_memo[key] = (pred_acc, pred_lat)
        interval_info.num_block_eval += 1
        return pred_acc, pred_lat

    def save_perf_memo(self, path_to_perf_memo):
        node_label2feat = {_k: torch.LongTensor([_v]) for _k, _v in PN_OP2IDX.items()}
        block2acc = {}
        for key, value in self.perf_memo.items():
            pred_acc, pred_lat = value
            labels = literal_eval(key)
            block = SPBlockState(labels, torch.cat([node_label2feat[lab] for lab in labels]))
            block2acc[key] = (block, pred_acc)

        with open(path_to_perf_memo, 'wb') as f:   
            pickle.dump(block2acc, f)
        if '.pkl' in path_to_perf_memo:
            with open(path_to_perf_memo.replace('.pkl', '.txt'), 'w') as f:
                for k, v in block2acc.items():
                    print("%s: %s"%(k, float(v[1])), file=f)

    def load_perf_memo(self, path_to_perf_memo):
        with open(path_to_perf_memo, 'rb') as f:
            block2acc = pickle.load(f)
        for key, value in block2acc.items():
            block, pred_acc = value
            pred_lat = self.lat_predictor(block) if self.lat_predictor is not None else 1.0
            self.perf_memo[key] = (pred_acc, pred_lat)

    def reset_perf_memo(self):
        # empty the perf memo dict
        self.perf_memo = {}