import os
import copy
import threading
from utils.misc_utils import write_line_to_file, get_tsr_value


class IntervalInfo:

    enable = True
    env = "main"
    _d = os.path.dirname(os.path.abspath(__file__))
    output_f = os.path.sep.join([_d, "../logs/{}_interval_info.txt"])
    lock = threading.RLock()

    # counters
    final_reward_i = 0
    step_reward_i = 0
    predictor_step_i = 0
    predictor_n_used = 0
    predictor_n_extreme_values = 0
    loss_i = 0
    perf_query_i = 0
    dis_pred_i = 0
    pareto_block_query_i = 0
    block_properties_i = 0
    block_shape_gen_i = 0
    block_shape_gen_times = 0
    block_tgt_node_i = 0
    block_tgt_edge_i = 0
    block_tgt_node_gen_count = 0
    block_tgt_edge_gen_count = 0
    num_block_eval = 0

    # interval constants
    final_reward_report_interval = 10000
    dis_pred_report_interval = 10000
    step_reward_report_interval = 10000
    predictor_usage_report_interval = 10000
    loss_report_interval = 5
    perf_query_report_interval = 10000
    pareto_block_query_report_interval = 10000
    block_properties_report_interval = 10000
    block_shape_gen_report_interval = 10000
    block_node_edge_gen_report_interval = 10000

    # values
    aggr_validity_reward = 0.
    aggr_pareto_reward = 0.
    aggr_perf_reward = 0.
    aggr_step_reward = 0.
    aggr_truth_block_score = 0.
    aggr_gen_block_score = 0.
    aggr_policy_loss = 0.
    aggr_value_loss = 0.
    aggr_entropy_loss = 0.
    aggr_acc = 0.
    aggr_lat = 0.
    max_seen_acc = 0.
    min_seen_lat = 0.
    aggr_num_edges = 0.
    aggr_num_nodes = 0.
    block_node_gen_num_edges = 0.

    pareto_block_queried_set = set()

    def __init__(self):
        raise NotImplementedError("This class should be used statically")

    @staticmethod
    def block_node_gen_update(has_desired_nodes, num_edges, plot_func=None, print_to_console=False):
        if not IntervalInfo.enable: return
        IntervalInfo.lock.acquire(blocking=True)
        IntervalInfo.block_tgt_node_i += 1
        if has_desired_nodes:
            IntervalInfo.block_tgt_node_gen_count += 1
            IntervalInfo.block_node_gen_num_edges += num_edges
        if IntervalInfo.block_tgt_node_i % IntervalInfo.block_node_edge_gen_report_interval == 0:
            if IntervalInfo.block_tgt_node_gen_count > 0:
                avg_num_edges = IntervalInfo.block_node_gen_num_edges / IntervalInfo.block_tgt_node_gen_count
            else:
                avg_num_edges = 0
            msg = "Times a block with desired nodes is generated at iter {}: {}, avg_num_edges:{} ".format(
                   IntervalInfo.block_tgt_node_i,
                   IntervalInfo.block_tgt_node_gen_count,
                   avg_num_edges)
            write_line_to_file(msg, IntervalInfo.output_f.format(IntervalInfo.env), verbose=print_to_console)
            if plot_func is not None:
                try:
                    plot_func(IntervalInfo.block_tgt_node_i, IntervalInfo.block_tgt_node_gen_count,
                              avg_num_edges, env=IntervalInfo.env)
                except ConnectionRefusedError:
                    pass
            IntervalInfo.block_tgt_node_gen_count = 0
            IntervalInfo.block_node_gen_num_edges = 0.
        IntervalInfo.lock.release()

    @staticmethod
    def block_edge_gen_update(has_desired_edges, plot_func=None, print_to_console=False):
        if not IntervalInfo.enable: return
        IntervalInfo.lock.acquire(blocking=True)
        IntervalInfo.block_tgt_edge_i += 1
        if has_desired_edges:
            IntervalInfo.block_tgt_edge_gen_count += 1
        if IntervalInfo.block_tgt_edge_i % IntervalInfo.block_node_edge_gen_report_interval == 0:
            msg = "times a block with desired edges is generated at iter {}: {}".format(IntervalInfo.block_tgt_edge_i,
                                                                                        IntervalInfo.block_tgt_edge_gen_count)
            write_line_to_file(msg, IntervalInfo.output_f.format(IntervalInfo.env), verbose=print_to_console)
            if plot_func is not None:
                try:
                    plot_func(IntervalInfo.block_tgt_edge_i, IntervalInfo.block_tgt_edge_gen_count, env=IntervalInfo.env)
                except ConnectionRefusedError:
                    pass
            IntervalInfo.block_tgt_edge_gen_count = 0
        IntervalInfo.lock.release()

    @staticmethod
    def block_shape_gen_update(is_desired_shape, plot_func=None, print_to_console=False):
        if not IntervalInfo.enable: return
        IntervalInfo.lock.acquire(blocking=True)
        IntervalInfo.block_shape_gen_i += 1
        if is_desired_shape:
            IntervalInfo.block_shape_gen_times += 1
        if IntervalInfo.block_shape_gen_i % IntervalInfo.block_shape_gen_report_interval == 0:
            msg = "Times a block of desired shape is generated at iter {}: {}".format(IntervalInfo.block_shape_gen_i,
                                                                                      IntervalInfo.block_shape_gen_times)
            write_line_to_file(msg, IntervalInfo.output_f.format(IntervalInfo.env), verbose=print_to_console)
            if plot_func is not None:
                try:
                    plot_func(IntervalInfo.block_shape_gen_i, IntervalInfo.block_shape_gen_times, env=IntervalInfo.env)
                except ConnectionError:
                    pass
            IntervalInfo.block_shape_gen_times = 0
        IntervalInfo.lock.release()

    @staticmethod
    def block_properties_update(num_nodes, num_edges, plot_func=None, print_to_console=False):
        if not IntervalInfo.enable: return
        IntervalInfo.lock.acquire(blocking=True)
        IntervalInfo.aggr_num_edges += num_edges
        IntervalInfo.aggr_num_nodes += num_nodes
        IntervalInfo.block_properties_i += 1
        if IntervalInfo.block_properties_i % IntervalInfo.block_properties_report_interval == 0:
            msg = "Block properties at iter {}, ".format(IntervalInfo.block_properties_i)
            avg_num_nodes = IntervalInfo.aggr_num_nodes / IntervalInfo.block_properties_report_interval
            avg_num_edges = IntervalInfo.aggr_num_edges / IntervalInfo.block_properties_report_interval
            msg += "avg num nodes: {}, ".format(avg_num_nodes)
            msg += "avg num edges: {}\n".format(avg_num_edges)
            write_line_to_file(msg, IntervalInfo.output_f.format(IntervalInfo.env), verbose=print_to_console)
            if plot_func is not None:
                try:
                    plot_func(IntervalInfo.block_properties_i, avg_num_nodes, avg_num_edges, env=IntervalInfo.env)
                except ConnectionRefusedError:
                    pass
            IntervalInfo._reset_block_properties()
        IntervalInfo.lock.release()

    @staticmethod
    def _reset_block_properties(reset_counter=False):
        if reset_counter: IntervalInfo.block_properties_i = 0
        IntervalInfo.aggr_num_edges = 0.
        IntervalInfo.aggr_num_nodes = 0.

    @staticmethod
    def pareto_block_query_update(key, is_pareto_block, plot_func=None, print_to_console=False):
        if not IntervalInfo.enable: return
        IntervalInfo.lock.acquire(blocking=True)
        IntervalInfo.pareto_block_query_i += 1
        if is_pareto_block and key not in IntervalInfo.pareto_block_queried_set:
            write_line_to_file("New pareto block {}: {}".format(len(IntervalInfo.pareto_block_queried_set) + 1, key),
                               IntervalInfo.output_f.format(IntervalInfo.env), verbose=print_to_console)
            IntervalInfo.pareto_block_queried_set.add(key)
        if IntervalInfo.pareto_block_query_i % IntervalInfo.pareto_block_query_report_interval == 0:
            msg = "Number of queried pareto blocks at iter {}: {}".format(IntervalInfo.pareto_block_query_i,
                                                                          len(IntervalInfo.pareto_block_queried_set))
            write_line_to_file(msg, IntervalInfo.output_f.format(IntervalInfo.env), verbose=print_to_console)
            if plot_func is not None:
                try:
                    plot_func(IntervalInfo.pareto_block_query_i, len(IntervalInfo.pareto_block_queried_set),
                              env=IntervalInfo.env)
                except ConnectionRefusedError:
                    pass
        IntervalInfo.lock.release()

    @staticmethod
    def dis_pred_update(truth_block_score, gen_block_score, plot_func=None, print_to_console=False):
        if not IntervalInfo.enable: return
        IntervalInfo.lock.acquire(blocking=True)
        truth_block_score = get_tsr_value(truth_block_score)
        gen_block_score = get_tsr_value(gen_block_score)
        IntervalInfo.aggr_truth_block_score += truth_block_score
        IntervalInfo.aggr_gen_block_score += gen_block_score
        IntervalInfo.dis_pred_i += 1
        if IntervalInfo.dis_pred_i % IntervalInfo.dis_pred_report_interval == 0:
            msg = "Discriminator output at iter {}, ".format(IntervalInfo.dis_pred_i)
            avg_truth_score = IntervalInfo.aggr_truth_block_score / IntervalInfo.dis_pred_report_interval
            avg_gen_score = IntervalInfo.aggr_gen_block_score / IntervalInfo.dis_pred_report_interval
            msg +="avg truth block score: {}, ".format(avg_truth_score)
            msg +="avg gen block score: {}\n".format(avg_gen_score)
            write_line_to_file(msg, IntervalInfo.output_f.format(IntervalInfo.env), verbose=print_to_console)
            if plot_func is not None:
                try:
                    plot_func(IntervalInfo.dis_pred_i, avg_truth_score, avg_gen_score, env=IntervalInfo.env)
                except ConnectionRefusedError:
                    pass
            IntervalInfo._reset_dis_pred()
        IntervalInfo.lock.release()

    @staticmethod
    def _reset_dis_pred(reset_counter=False):
        if reset_counter: IntervalInfo.dis_pred_i = 0
        IntervalInfo.aggr_truth_block_score = 0.
        IntervalInfo.aggr_gen_block_score = 0.

    @staticmethod
    def loss_update(policy_loss, value_loss, entropy_loss, plot_func=None, print_to_console=False):
        if not IntervalInfo.enable: return
        policy_loss = get_tsr_value(policy_loss)
        value_loss = get_tsr_value(value_loss)
        entropy_loss = get_tsr_value(entropy_loss)
        IntervalInfo.aggr_policy_loss += policy_loss
        IntervalInfo.aggr_value_loss += value_loss
        IntervalInfo.aggr_entropy_loss += entropy_loss
        IntervalInfo.loss_i += 1
        if IntervalInfo.loss_i % IntervalInfo.loss_report_interval == 0:
            msg = "Loss at iter {}, ".format(IntervalInfo.loss_i)
            avg_policy_loss = IntervalInfo.aggr_policy_loss / IntervalInfo.loss_report_interval
            avg_value_loss = IntervalInfo.aggr_value_loss / IntervalInfo.loss_report_interval
            avg_entropy_loss = IntervalInfo.aggr_entropy_loss / IntervalInfo.loss_report_interval
            msg += "avg policy loss: {}, ".format(avg_policy_loss)
            msg += "avg value loss: {}, ".format(avg_value_loss)
            msg += "avg entropy loss: {}\n".format(avg_entropy_loss)
            write_line_to_file(msg, IntervalInfo.output_f.format(IntervalInfo.env), verbose=print_to_console)
            if plot_func is not None:
                try:
                    plot_func(IntervalInfo.loss_i, avg_policy_loss, avg_value_loss, avg_entropy_loss,
                              env=IntervalInfo.env)
                except ConnectionRefusedError:
                    pass
            IntervalInfo._reset_loss()

    @staticmethod
    def _reset_loss(reset_counter=False):
        if reset_counter: IntervalInfo.loss_i = 0
        IntervalInfo.aggr_policy_loss = 0.
        IntervalInfo.aggr_value_loss = 0.
        IntervalInfo.aggr_entropy_loss = 0.

    @staticmethod
    def perf_query_update(acc, lat, plot_func=None, print_to_console=False):
        if not IntervalInfo.enable: return
        IntervalInfo.lock.acquire(blocking=True)
        acc = get_tsr_value(acc)
        lat = get_tsr_value(lat)
        IntervalInfo.aggr_acc += acc
        IntervalInfo.aggr_lat += lat
        IntervalInfo.max_seen_acc = max(IntervalInfo.max_seen_acc, acc)
        IntervalInfo.min_seen_lat = min(IntervalInfo.min_seen_lat, lat)
        IntervalInfo.perf_query_i += 1
        if IntervalInfo.perf_query_i % IntervalInfo.perf_query_report_interval == 0:
            msg = "Perf query at iter {}, ".format(IntervalInfo.perf_query_i)
            avg_acc = IntervalInfo.aggr_acc / IntervalInfo.perf_query_report_interval
            avg_lat = IntervalInfo.aggr_lat / IntervalInfo.perf_query_report_interval
            msg +="avg acc: {}, ".format(avg_acc)
            msg +="avg lat: {}, ".format(avg_lat)
            msg +="max acc: {}\n".format(IntervalInfo.max_seen_acc)
            msg +="min lat: {}\n".format(IntervalInfo.min_seen_lat)
            write_line_to_file(msg, IntervalInfo.output_f.format(IntervalInfo.env), verbose=print_to_console)
            if plot_func is not None:
                percent_usage = 100 * float(IntervalInfo.predictor_n_used) / IntervalInfo.perf_query_report_interval
                try:
                    plot_func(IntervalInfo.perf_query_i, avg_acc, IntervalInfo.max_seen_acc, avg_lat,
                              IntervalInfo.min_seen_lat, percent_usage, env=IntervalInfo.env)
                except ConnectionRefusedError:
                    pass
            IntervalInfo.predictor_n_used = 0
            IntervalInfo._reset_perf_query()
        IntervalInfo.lock.release()

    @staticmethod
    def _reset_perf_query(reset_counter=False):
        if reset_counter: IntervalInfo.perf_query_i = 0
        IntervalInfo.aggr_acc = 0.
        IntervalInfo.aggr_lat = 0.

    @staticmethod
    def predictor_n_used_update():
        if not IntervalInfo.enable: return
        IntervalInfo.lock.acquire(blocking=True)
        IntervalInfo.predictor_n_used += 1
        IntervalInfo.lock.release()

    @staticmethod
    def predictor_n_extreme_values_update():
        if not IntervalInfo.enable: return
        IntervalInfo.lock.acquire(blocking=True)
        IntervalInfo.predictor_n_extreme_values += 1
        IntervalInfo.lock.release()

    @staticmethod
    def predictor_step_update(print_to_console=False):
        if not IntervalInfo.enable: return
        IntervalInfo.lock.acquire(blocking=True)
        IntervalInfo.predictor_step_i += 1
        if IntervalInfo.predictor_step_i % IntervalInfo.predictor_usage_report_interval == 0:
            msg = "Perf predictor usage at iter {}, ".format(IntervalInfo.predictor_step_i)
            # msg +="num times called until now: {}, ".format(IntervalInfo.predictor_n_used)
            msg +="num times extreme values are predicted: {}\n".format(IntervalInfo.predictor_n_extreme_values)
            write_line_to_file(msg, IntervalInfo.output_f.format(IntervalInfo.env), verbose=print_to_console)
            IntervalInfo._reset_predictor_step()
        IntervalInfo.lock.release()

    @staticmethod
    def _reset_predictor_step():
        IntervalInfo.predictor_step_i = 0
        IntervalInfo.predictor_n_used = 0
        IntervalInfo.predictor_n_extreme_values = 0

    @staticmethod
    def block_step_reward_update(step_reward, plot_func=None, print_to_console=False):
        if not IntervalInfo.enable: return
        IntervalInfo.lock.acquire(blocking=True)
        step_reward = get_tsr_value(step_reward)
        IntervalInfo.aggr_step_reward += step_reward
        IntervalInfo.step_reward_i += 1
        if IntervalInfo.step_reward_i % IntervalInfo.step_reward_report_interval == 0:
            avg_step_reward = IntervalInfo.aggr_step_reward / IntervalInfo.step_reward_report_interval
            msg = "Step reward at iter {}, ".format(IntervalInfo.step_reward_i)
            msg +="avg step reward: {}\n".format(avg_step_reward)
            write_line_to_file(msg, IntervalInfo.output_f.format(IntervalInfo.env), verbose=print_to_console)
            if plot_func is not None:
                try:
                    plot_func(IntervalInfo.step_reward_i, avg_step_reward, env=IntervalInfo.env)
                except ConnectionRefusedError:
                    pass
            IntervalInfo._reset_block_step_reward()
        IntervalInfo.lock.release()

    @staticmethod
    def _reset_block_step_reward(reset_counter=False):
        if reset_counter: IntervalInfo.step_reward_i = 0
        IntervalInfo.aggr_step_reward = 0.

    @staticmethod
    def block_final_reward_update(validity_reward, pareto_reward, perf_reward,
                                  plot_func=None, print_to_console=False):
        if not IntervalInfo.enable: return
        IntervalInfo.lock.acquire(blocking=True)
        validity_reward = get_tsr_value(validity_reward)
        pareto_reward = get_tsr_value(pareto_reward)
        perf_reward = get_tsr_value(perf_reward)
        IntervalInfo.aggr_validity_reward += validity_reward
        IntervalInfo.aggr_pareto_reward += pareto_reward
        IntervalInfo.aggr_perf_reward += perf_reward
        IntervalInfo.final_reward_i += 1
        if IntervalInfo.final_reward_i % IntervalInfo.final_reward_report_interval == 0:
            avg_validity_reward = IntervalInfo.aggr_validity_reward / IntervalInfo.final_reward_report_interval
            avg_perf_reward = IntervalInfo.aggr_perf_reward / IntervalInfo.final_reward_report_interval
            avg_pareto_reward = IntervalInfo.aggr_pareto_reward / IntervalInfo.final_reward_report_interval
            msg = "Final reward at iter {}, ".format(IntervalInfo.final_reward_i)
            msg +="avg validity reward: {}, ".format(avg_validity_reward)
            msg +="avg perf reward: {}, ".format(avg_perf_reward)
            msg +="avg pareto reward: {}\n".format(avg_pareto_reward)
            write_line_to_file(msg, IntervalInfo.output_f.format(IntervalInfo.env), verbose=print_to_console)
            if plot_func is not None:
                try:
                    plot_func(IntervalInfo.final_reward_i, avg_validity_reward, avg_perf_reward, avg_pareto_reward,
                              env=IntervalInfo.env)
                except ConnectionRefusedError:
                    pass
            IntervalInfo._reset_block_final_reward()
        IntervalInfo.lock.release()

    @staticmethod
    def _reset_block_final_reward(reset_counter=False):
        if reset_counter: IntervalInfo.final_reward_i = 0
        IntervalInfo.aggr_validity_reward = 0.
        IntervalInfo.aggr_pareto_reward = 0.
        IntervalInfo.aggr_perf_reward = 0.


class DatasetInfo:

    max_acc = float("-inf")
    min_acc = float("inf")
    max_lat = float("-inf")
    min_lat = float("inf")
    avg_acc = 0.
    avg_lat = 0.
    n_instances = 0
    _truth_pareto_block_keys = set()

    def __init__(self):
        raise NotImplementedError("This class should be used statically")

    @staticmethod
    def num_truth_pareto_blocks():
        return len(DatasetInfo._truth_pareto_block_keys)

    @staticmethod
    def is_truth_pareto_block(str_id):
        return str_id in DatasetInfo._truth_pareto_block_keys

    @staticmethod
    def set_truth_pareto_block_keys(blocks):
        for block in blocks:
            str_id = block.get_str_id()
            DatasetInfo._truth_pareto_block_keys.add(str_id)

    @staticmethod
    def update_acc_lat(acc, lat):
        DatasetInfo.max_acc = max(acc, DatasetInfo.max_acc)
        DatasetInfo.min_acc = min(acc, DatasetInfo.min_acc)
        DatasetInfo.max_lat = max(lat, DatasetInfo.max_lat)
        DatasetInfo.min_lat = min(lat, DatasetInfo.min_lat)
        aggr_acc = DatasetInfo.avg_acc * DatasetInfo.n_instances + acc
        aggr_lat = DatasetInfo.avg_lat * DatasetInfo.n_instances + lat
        DatasetInfo.n_instances += 1
        DatasetInfo.avg_acc = aggr_acc / DatasetInfo.n_instances
        DatasetInfo.avg_lat = aggr_lat / DatasetInfo.n_instances

    @staticmethod
    def self_check():
        assert DatasetInfo.max_acc > float("-inf")
        assert DatasetInfo.min_acc < float("inf")
        assert DatasetInfo.max_lat > float("-inf")
        assert DatasetInfo.min_lat < float("inf")

    @staticmethod
    def is_extreme_acc(acc):
        DatasetInfo.self_check()
        return not DatasetInfo.min_acc <= acc <= DatasetInfo.max_acc

    @staticmethod
    def is_extreme_lat(lat):
        DatasetInfo.self_check()
        return not DatasetInfo.min_lat <= lat <= DatasetInfo.max_lat


class ModelUpdateCheck:

    pre_update_model = None

    def __init__(self):
        raise NotImplementedError("This class should be used statically")

    @staticmethod
    def set_pre_update_weights(model):
        ModelUpdateCheck.pre_update_model = copy.deepcopy(model).cpu()

    @staticmethod
    def check_weight_updated(new_model):
        if ModelUpdateCheck.pre_update_model is None:
            return
        new_model = copy.deepcopy(new_model).cpu()
        assert sum(1 for _ in new_model.parameters()) == sum(1 for _ in ModelUpdateCheck.pre_update_model.parameters())
        for p1, p2 in zip(new_model.parameters(), ModelUpdateCheck.pre_update_model.parameters()):
            if p1.data.ne(p2.data).sum() > 0:
                    return
        assert False, "Model not updated"


class Watchdog:

    curr_count = 0
    _timeout_count = 1e6

    def __init__(self):
        raise NotImplementedError("This class should be used statically")

    @staticmethod
    def reset(new_timeout=None):
        Watchdog.curr_count = 0
        if new_timeout is not None:
            Watchdog._timeout_count = new_timeout

    @staticmethod
    def update():
        Watchdog.curr_count += 1
        if Watchdog.curr_count >= Watchdog._timeout_count:
            raise TimeoutError("Watchdog timeout, threshold: {}".format(Watchdog._timeout_count))


if __name__ == "__main__":

    # from tensorboardX import SummaryWriter
    # import numpy as np
    #
    # writer = SummaryWriter()
    #
    # for n_iter in range(100):
    #     writer.add_scalar('Loss/train', np.random.random(), n_iter)
    #     writer.add_scalar('Loss/test', np.random.random(), n_iter)
    #     writer.add_scalar('Accuracy/train', np.random.random(), n_iter)
    #     writer.add_scalar('Accuracy/test', np.random.random(), n_iter)
    #
    # writer.close()

    # from model_ppo_gan import BookKeeper
    # bk = BookKeeper("tmp_log.txt", "dummy", "./")
    # bk.save_acc_lat_pareto_plot([0.87, 0.9, 0.92, 0.99, 0.73], [0.09, 0.1, 0.13, 0.22, 0.01], 1)
    # bk.save_acc_lat_pareto_plot([0.87, 0.9, 0.92, 0.99, 0.74], [0.09, 0.1, 0.13, 0.22, 0.01], 2)

    # import torch
    # net = torch.nn.Linear(100, 1)
    # ModelUpdateCheck.set_pre_update_weights(net)
    # net.bias.data -= 1
    # ModelUpdateCheck.check_weight_updated(net)

    print("done")
