import os
import torch
import numpy as np
from constants import *
from nltk.corpus import stopwords
from utils.model_utils import device
from sklearn.metrics.pairwise import cosine_similarity, euclidean_distances
from utils.misc_utils import get_np_sentence_embedding, mean


class W2VTrainableVocab:
    def __init__(self, w2c, w2v_dict,
                 embedding_dim=300, light_weight=True, rand_oov_embed=False,
                 special_tokens = [PAD_TOKEN,OOV_TOKEN]):
        self.embedding_dim = embedding_dim
        self.light_weight = light_weight
        self.vocab_size = 0
        self._w2v_original = w2v_dict
        self._w2c_input = w2c
        self.oov_count = 0
        self.w2v = {}
        self.w2i = {PAD_TOKEN:0, OOV_TOKEN:1}
        self.i2w = {}
        self.i2v = {}
        self.c2i = {}
        self.i2c = {}
        self.w2v_mat = None
        self.c2v_mat = None
        self.pad_token = PAD_TOKEN
        self.oov_token = OOV_TOKEN
        self.pad_idx = 0
        self.oov_idx = 1
        self.rand_oov_embed = rand_oov_embed
        # modify this to contain all special tokens
        self.special_tokens = special_tokens
        self._curr_word_idx = len(self.w2i)
        # self.trainable_oov_token_indices = []
        self.oov_w2i = {PAD_TOKEN:self.pad_idx, OOV_TOKEN:self.oov_idx}
        self.oov_i2w = {self.pad_idx:PAD_TOKEN, self.oov_idx:OOV_TOKEN}
        self.pos2i = {PAD_TOKEN:self.pad_idx, OOV_TOKEN:self.oov_idx}
        self.i2pos = {self.pad_idx:PAD_TOKEN, self.oov_idx:OOV_TOKEN}
        self._build_vocab()

    def _build_vocab(self):
        print("building vocab from custom w2c, vocab size: " + str(len(self._w2c_input)))
        for t in self.special_tokens:
            if len(t) == 0: continue
            if t in self.w2i: continue
            self.w2i[t] = self._curr_word_idx
            self._curr_word_idx += 1
        for w, c in self._w2c_input.items():
            if len(w) == 0: continue
            if w in self.w2i: continue
            self.w2i[w] = self._curr_word_idx
            self._curr_word_idx += 1
        self.i2w = {v: k for k, v in self.w2i.items()}
        assert len(self.w2i) == len(self.i2w)
        oov_tokens = set()
        for i, w in self.i2w.items():
            if i in self.i2v: continue
            if w in self._w2v_original:
                self.i2v[i] = self._w2v_original[w]
                self.w2v[w] = self._w2v_original[w]
            else:
                # if w == PAD_TOKEN:
                #     embed = np.zeros(self.embedding_dim)
                # elif self.rand_oov_embed:
                #     embed = np.random.randn(self.embedding_dim)
                # else:
                #     embed = np.zeros(self.embedding_dim)
                embed = np.zeros(self.embedding_dim)
                if w not in self.special_tokens:
                    self.oov_count += 1
                self.i2v[i] = embed
                self.w2v[w] = embed
                oov_tokens.add(w)

        oov_word_idx = max([i for w, i in self.oov_w2i.items()]) + 1
        for oov_w in list(oov_tokens):
            if oov_w in self.oov_w2i: continue
            self.oov_w2i[oov_w] = oov_word_idx
            oov_word_idx += 1
        self.oov_i2w = {v: k for k, v in self.oov_w2i.items()}
        assert len(self.oov_w2i) == len(self.oov_i2w)

        self._vocab_size = len(self.w2i)
        self.w2v_mat = build_i2v_mat(self._vocab_size, self.embedding_dim, self.i2v)
        self.w2v_mat = torch.from_numpy(self.w2v_mat).type(torch.FloatTensor)
        self._build_character_embedding()
        if self.light_weight:
            self.i2v = None
            self._w2v_original = None
            self._w2c_input = None
        print("{} words OOV when building vocab".format(self.oov_count))
        print("built vocab size {}".format(self._curr_word_idx))

    def _build_character_embedding(self, embedding_dim=200):
        # should be called at the end of build_vocab()
        self.c2i = {chr(i):i for i in range(128)} # TODO: for now only do ascii, and use 200d
        self.i2c = {v:k for k, v in self.c2i.items()}
        self.c2v_mat = np.random.randn(128, embedding_dim)
        self.c2v_mat = torch.from_numpy(self.c2v_mat).type(torch.FloatTensor)

    def build_pos_vocab(self, pos2c):
        self.pos2i = {PAD_TOKEN: self.pad_idx}
        self.i2pos = {self.pad_idx: PAD_TOKEN}
        idx = len(self.pos2i)
        for tag, count in pos2c.items():
            while idx in self.i2pos: idx += 1
            self.pos2i[tag] = idx
            self.i2pos[idx] = tag
            idx += 1

    def get_oov2wi_dict(self):
        rv = {}
        for oov_idx, w in self.oov_i2w.items():
            rv[oov_idx] = self.w2i[w]
        return rv

    def get_word_idx(self, word):
        return self.w2i if word in self.w2i else self.oov_idx

    def get_word_from_idx(self, idx):
        return self.i2w if idx in self.i2w else self.oov_token

    def items(self):
        return self.w2i.items()

    def __len__(self):
        return self._vocab_size

    def __repr__(self):
        return "Vocab with additional tokens like OOV:{}, PAD:{}".format(self.oov_token, self.pad_token)


class SegData:

    def __init__(self, raw_str, seg_list,
                 pos_list=None,
                 keyword_seg_list=None,
                 keyword_pos_list=None,
                 extras={},
                 is_light_weight=False):
        self.raw_str = raw_str
        self.seg_list = seg_list
        self.pos_list = pos_list
        self.keyword_seg_list = keyword_seg_list
        self.keyword_pos_list = keyword_pos_list
        self.extras = extras
        self.is_light_weight = is_light_weight
        if pos_list is not None:
            assert len(pos_list) == len(seg_list)
        if keyword_seg_list is not None and keyword_pos_list is not None:
            assert len(keyword_seg_list) == len(keyword_pos_list)
        if self.is_light_weight:
            self.pos_list = None
            self.keyword_pos_list = None
            self.extras = None

    def __str__(self):
        return self.raw_str

    def __repr__(self):
        return self.raw_str

    def __len__(self):
        return len(self.seg_list)

    def __hash__(self):
        return hash(self.raw_str)

    def __eq__(self, other):
        if not isinstance(other, SegData):
            return False
        return self.raw_str == other.raw_str

    def __ne__(self, other):
        return not (self == other)


class WrapperW2VDict(dict):

    def __init__(self, model, model_w2i, original_w2v, oov_idx):
        super(WrapperW2VDict, self).__init__()
        self._model = model
        self._oov_idx = oov_idx
        self._model_w2i = model_w2i
        self._original_w2v = original_w2v

    def __getitem__(self, word):
        with torch.no_grad():
            if word in self._model_w2i:
                word_idx = self._model_w2i[word]
                tsr = torch.ones(1,1,1).fill_(word_idx).type(torch.LongTensor).to(device())
                embedding = self._model(tsr).squeeze().cpu().detach().numpy()
            else:
                embedding = self._original_w2v[word]
        return embedding

    def __contains__(self, item):
        return item in self._model_w2i or item in self._original_w2v

    def __len__(self):
        return len(self._model_w2i)


class OnlineVerifier:

    def __init__(self, search_engine="google", pass_ratio=0.5, match_times=3):
        self.search_engine = search_engine
        self.pass_ratio = pass_ratio
        self.match_times = match_times

    def _count_match_times(self, text, data, word_delim=" ", in_between_words_count=2):
        import re
        em_count = data.count(text)
        pattern = "[\u4E00-\u9FA5\u0030-\u0039A-Za-z0-9_]{1,"+str(in_between_words_count)+"}".join(text.split(word_delim))
        pattern = re.compile(pattern)
        match = pattern.findall(data)
        re_count = len(match)
        return em_count + re_count

    def text_valid(self, text, word_delim=" ", topk_results=10, in_between_words_count=2):
        import requests
        from googlesearch import search
        from bs4 import BeautifulSoup
        urls = []
        for url in search(text, tld='com', lang='zh', num=topk_results, start=0, stop=None, pause=1.0):
            urls.append(url)
        rv = []
        for url in urls:
            r = requests.get(url)
            soup = BeautifulSoup(r.content, 'html5lib')
            parse_tree = soup.prettify(encoding="utf-8")
        return True


def load_gensim_word_vec(word_vec_file, cache_file=None):
    import gensim
    import pickle
    if cache_file is not None and os.path.isfile(cache_file):
        print("Loading word vec from cache: " + cache_file)
        with open(cache_file, "rb") as f:
            w2v = pickle.load(f)
        return w2v
    print("Loading word vec from source: " + word_vec_file)
    w2v = gensim.models.KeyedVectors.load_word2vec_format(word_vec_file, binary=False)
    if cache_file is not None:
        with open(cache_file, "wb") as f:
            pickle.dump(w2v, f, protocol=4)
    return w2v


def build_w2c_from_seg_word_lists(seg_word_lists, limit=None):
    w2c = {}
    for seg_list in seg_word_lists:
        for word in seg_list:
            if len(word) == 0: continue
            if word not in w2c: w2c[word] = 0
            w2c[word] += 1
    if limit is not None and len(w2c) > limit:
        tmp = sorted([(w,c) for w,c in w2c.items()],key=lambda t:t[1],reverse=True)[:limit]
        w2c = {t[0]:t[1] for t in tmp}
    return w2c


def pad_seg_list_in_place(seg_list, pad_len, pad_val=0):
    while len(seg_list) < pad_len:
        seg_list.append(pad_val)


def pad_seg_lists_in_place(seg_lists, pad_val=0, pad_len_override=None):
    if pad_len_override is not None:
        pad_len = pad_len_override
    else:
        pad_len = max([len(seg_list) for seg_list in seg_lists])
    for seg_list in seg_lists:
        while len(seg_list) < pad_len:
            seg_list.append(pad_val)


def mean_of_w2v(word_seg_list, w2v, to_torch_tensor=False):
    word_seg_list = [w for w in word_seg_list if w in w2v]
    if len(word_seg_list) == 0:
        raise ValueError("Input is empty")
    word_vecs = [w2v[w].reshape(1,-1) for w in word_seg_list]
    word_vec = np.concatenate(word_vecs, axis=0)
    word_vec = np.mean(word_vec, axis=0, keepdims=False)
    if to_torch_tensor:
        rv = torch.from_numpy(word_vec).type(torch.FloatTensor).to(device())
    else:
        rv = word_vec
    return rv


def cosine_sim_np(np_arr_1, np_arr_2):
    cos_sim = cosine_similarity(np_arr_1.reshape(1,-1), np_arr_2.reshape(1,-1))[0,0]
    return cos_sim


def euclidean_dist_np(np_arr_1, np_arr_2):
    euc_dist = euclidean_distances(np_arr_1.reshape(1,-1), np_arr_2.reshape(1,-1))[0,0]
    return euc_dist


def gen_cpy_np(src_word_seg_lists, tgt_word_seg_lists, max_tgt_len, w2i):
    assert len(src_word_seg_lists) == len(tgt_word_seg_lists)
    cpy_wids = np.zeros((len(src_word_seg_lists), max_tgt_len))
    cpy_gates = np.zeros((len(src_word_seg_lists), max_tgt_len))
    for bi, ctx_word_seg_list in enumerate(src_word_seg_lists):
        tgt_word_seg_list = tgt_word_seg_lists[bi]
        for ci, cw in enumerate(ctx_word_seg_list):
            for ai, aw in enumerate(tgt_word_seg_list):
                if aw in w2i: continue  # only allow copy for OOV words
                if cw == aw:
                    cpy_gates[bi,ai] = 1
                    cpy_wids[bi,ai] = ci
    return cpy_wids, cpy_gates


def w2c_selective_load_w2v(w2v_file, w2c, delim=" "):
    with open(w2v_file, "r", encoding="utf-8") as f:
        lines = f.readlines()
    header = lines[0].split(delim)
    full_vocab_size, embedding_dim = int(header[0]), int(header[1])
    lines = lines[1:]
    w2v_words = {line.rstrip().split(delim)[0]:i for i,line in enumerate(lines) if len(line.rstrip()) > 0}
    rv_w2v = {}
    for word, _ in w2c.items():
        if word in w2v_words and word not in rv_w2v:
            line = lines[w2v_words[word]].rstrip()
            vec_vals = [float(v) for v in line.split(delim)[1:] if len(v) > 0]
            assert len(vec_vals) == embedding_dim
            word_vec = np.asarray(vec_vals)
            rv_w2v[word] = word_vec
    return rv_w2v


def select_from_d_list_by_w2c(data_list, w2c, word_key, limit):
    if len(data_list) < limit:
        return data_list
    data = sorted([(word_key(d),w2c[word_key(d)],d) for d in data_list if word_key(d) in w2c], key=lambda t:t[1], reverse=True)
    data = data[:limit]
    rv = [d[2] for d in data]
    return rv


def select_from_d_dict_by_w2c(data_dict, w2c, word_key, limit):
    if len(data_dict) < limit:
        return data_dict
    data = sorted([(word_key(k,v),w2c[word_key(k,v)], k) for k,v in data_dict.items() if word_key(k,v) in w2c], key=lambda t:t[1], reverse=True)
    data = data[:limit]
    rv = {d[2]:data_dict[d[2]] for d in data}
    return rv


def build_i2v_mat(vocab_size, embedding_dim, i2v):
    rv = np.zeros((vocab_size, embedding_dim))
    assert len(i2v) == vocab_size
    for i, v in i2v.items():
        assert isinstance(i, int)
        if i >= vocab_size: assert False, "idx {} OOV".format(i)
        rv[i, :] = i2v[i]
    return rv


def get_char_ids_tensor(seg_lists, c2i, max_char_len=16, pad_idx=0, oov_idx=1):
    """
    return batch_size x seq_len x 16
    """
    max_seq_len = max([len(l) for l in seg_lists])
    rv = np.zeros((len(seg_lists), max_seq_len, max_char_len))
    rv.fill(pad_idx)
    for i, seg_list in enumerate(seg_lists):
        for j, word in enumerate(seg_list):
            for k, c in enumerate(word):
                if k >= rv.shape[2]:
                    break
                if c in c2i:
                    rv[i,j,k] = c2i[c]
                else:
                    rv[i,j,k] = oov_idx
    return torch.from_numpy(rv).type(torch.LongTensor)


def _read_qa_seg_cache_file(path, delim=","):
    with open(path, encoding='utf-8') as f:
        lines = f.readlines()
    seg_rv = []
    for line in lines:
        line = line.rstrip()
        if line:
            seg_rv.append(line.split(delim))
    return seg_rv


def read_qa_corpus_file(path, read_lines_limit=None):
    with open(path, encoding='utf-8') as f:
        lines = f.readlines()
    raw_msg = []
    raw_rsp = []
    fetch_msg = True
    if read_lines_limit:
        print('Read lines limit: ' + str(read_lines_limit))
    pairs_count = 0
    for line in lines:
        if read_lines_limit and pairs_count >= read_lines_limit:
            break
        line = line.rstrip()
        if line:
            if fetch_msg:
                raw_msg.append(line)
            else:
                pairs_count += 1
                raw_rsp.append(line)
            fetch_msg = not fetch_msg
    return raw_msg, raw_rsp


def load_word_vectors(w2v_file, w2v_cache_file=None, read_delim=" "):
    import gensim
    word2idx = {}
    import pickle
    if os.path.isfile(w2v_cache_file):
        fileObject = open(w2v_cache_file, 'rb')
        model = pickle.load(fileObject)
        fileObject.close()
    else:
        model = gensim.models.KeyedVectors.load_word2vec_format(
            w2v_file, binary=False)
        fileObject = open(w2v_cache_file, 'wb')
        pickle.dump(model, fileObject)
        fileObject.close()
    if hasattr(model, "index2word"):
        word2idx = {v: k for k, v in enumerate(model.index2word)}
    else:
        with open(w2v_file, encoding='utf-8') as f:
            lines = f.readlines()
        idx = 0
        for line in lines:
            tokens = line.split(read_delim)
            if len(tokens) < 10:
                continue
            word2idx[tokens[0]] = idx
            idx+=1
    idx2word = {v: k for k, v in word2idx.items()}
    return model, word2idx, idx2word


def pad_until_len(word_seg_list, target_len, pad_word="</s>"):
    for seg_list in word_seg_list:
        while len(seg_list) < target_len:
            seg_list.append(pad_word)


def one_hot_encode_word(word, word2idx, vocab_size):
    classes = np.zeros(vocab_size)
    if word in word2idx:
        idx = word2idx[word]
        classes[idx] = 1
    return classes


def _gen_vec_rep_for_tokenized_sent(word_list, max_sent_len, word_vec, embed_dim):
    rv = np.zeros((max_sent_len, embed_dim))
    i = 0
    for word in word_list:
        if i >= max_sent_len:
            break
        if word in word_vec:
            rv[i,:] = word_vec[word]
        i += 1
    return rv


def handle_OOV_for_words(words_list, dictionary, oov_token=None):
    rv = []
    for w in words_list:
        if w in dictionary:
            rv.append(w)
        elif oov_token is not None and oov_token in dictionary:
            rv.append(oov_token)
    return rv


def _read_sent_w2i_cache_file(file):
    npz_file = np.load(file)
    return npz_file['arr_0']


def remove_stop_words_en(seg_list, ):
    eng_stops = set(stopwords.words('english'))
    sent_list = [w for w in seg_list if w not in eng_stops]
    return sent_list


def _sent_word2idx_lists(seg_sent_list, word2idx, max_sent_len, oov_idx=None):
    sent_words2idx = []
    for seg_sent in seg_sent_list:
        j = 0
        tmp = []
        for word in seg_sent:
            if j >= max_sent_len:
                break
            if word in word2idx:
                tmp.append(word2idx[word])
            elif oov_idx is not None:
                tmp.append(oov_idx)
            j += 1
        sent_words2idx.append(tmp)
    return sent_words2idx


def _sent_word2idx_np(seg_sent_list, word2idx, max_sent_len, pad_idx=0, oov_idx=1):
    sent_words2idx = np.zeros((len(seg_sent_list), max_sent_len))
    if pad_idx is not None: sent_words2idx.fill(pad_idx)
    i = 0
    for seg_sent in seg_sent_list:
        j = 0
        for word in seg_sent:
            if j >= max_sent_len:
                break
            if word in word2idx:
                sent_words2idx[i,j] = word2idx[word]
            else:
                sent_words2idx[i,j] = oov_idx
            j += 1
        i += 1
    return sent_words2idx


def _sent_word2vec_np(seg_sent_list, w2v, embedding_size, max_sent_len, oov_token="<OOV>"):
    assert oov_token in w2v
    sent_words2idx = np.zeros((len(seg_sent_list), max_sent_len, embedding_size))
    i = 0
    for seg_sent in seg_sent_list:
        j = 0
        for word in seg_sent:
            if j >= max_sent_len:
                break
            if word in w2v:
                sent_words2idx[i,j,:] = w2v[word]
            else:
                sent_words2idx[i,j,:] = w2v[oov_token]
            j += 1
        i += 1
    return sent_words2idx


def sents_words2idx_to_text(w2i, w2v, eos_idx=0, oov_token="<OOV>", delim=" "):
    rv = []
    max_sent_len = get_valid_vec_rep_length(w2i.reshape(-1, 1))
    for i in range(max_sent_len):
        word_idx = int(w2i[i])
        if word_idx == eos_idx >= 0:
           break
        if 0 <= word_idx <= len(w2v.index2word):
            word = w2v.index2word[word_idx]
            rv.append(word)
        else:
            rv.append(oov_token)
    return delim.join(rv)


def sents_words2idx_to_one_hot(batch_size, w2i, vocab_size, max_sent_len_idx=1):
    max_sent_len = w2i.shape[max_sent_len_idx]
    rv = np.zeros((max_sent_len, batch_size, vocab_size))
    for i in range(batch_size):
        for j in range(max_sent_len):
            idx = int(w2i[i, j])
            if idx < 0:
                continue
            rv[j, i, idx] = 1
    return rv


def _word2idx_to_one_hot(word_idx, vocab_size):
    rv = np.zeros(vocab_size)
    rv[word_idx] = 1
    return rv


def append_to_seg_sents_list(seg_sents_list, token, to_front=True):
    for sent_word_list in seg_sents_list:
        if to_front:
            sent_word_list.insert(0, token)
        else:
            sent_word_list.append(token)


def handle_msg_rsp_OOV(msg_seg_list, rsp_seg_list, dictionary, oov_token=None):
    msg_seg_rv = []
    rsp_seg_rv = []
    for i in range(len(msg_seg_list)):
        msg_words = handle_OOV_for_words(msg_seg_list[i], dictionary, oov_token=oov_token)
        rsp_words = handle_OOV_for_words(rsp_seg_list[i], dictionary, oov_token=oov_token)
        if len(msg_words) == 0 or len(rsp_words) == 0:
            continue
        msg_seg_rv.append(msg_words)
        rsp_seg_rv.append(rsp_words)
    return msg_seg_rv, rsp_seg_rv


def get_valid_vec_rep_length(vec):
    """
    Input must be embedding_size
    """
    n_instances = vec.shape[0]
    rv = 0
    for i in range(n_instances):
        val = vec[i,:]
        if np.all(val==0) or np.all(val==-1):
            return rv
        rv += 1
    return rv


def gen_word_embedding_vec_rep(seg_sents_list, embedding_size, w2v, max_sent_len,
                               time_major=False,
                               word_embedding_vec_cache_file=None,):
    if word_embedding_vec_cache_file is not None and os.path.isfile(word_embedding_vec_cache_file):
        npz_file = np.load(word_embedding_vec_cache_file)
        return npz_file['arr_0']
    n_instances = len(seg_sents_list)
    if time_major:
        rv = np.zeros((max_sent_len, n_instances, embedding_size))
    else:
        rv = np.zeros((n_instances, max_sent_len, embedding_size))
    for i in range(n_instances):
        vec_rep = _gen_vec_rep_for_tokenized_sent(seg_sents_list[i], max_sent_len, w2v, embedding_size)
        if time_major:
            rv[:, i, :] = vec_rep
        else:
            rv[i, :, :] = vec_rep
    if word_embedding_vec_cache_file:
        np.savez(word_embedding_vec_cache_file, rv)
    return rv


def gen_word2idx_vec_rep(seg_sents_list, word2idx, max_sent_len, return_lists=False,
                         pad_idx=0, oov_idx=1):
    if return_lists:
        sent_word2idx = _sent_word2idx_lists(seg_sents_list, word2idx, max_sent_len, oov_idx)
    else:
        sent_word2idx = _sent_word2idx_np(seg_sents_list, word2idx, max_sent_len, pad_idx=pad_idx, oov_idx=oov_idx)
    return sent_word2idx


def gen_word2vec_np(seg_sents_list, w2v, embedding_size, max_sent_len, oov_token="<OOV>"):
    sent_word2vec = _sent_word2vec_np(seg_sents_list, w2v, embedding_size, max_sent_len, oov_token=oov_token)
    return sent_word2vec


def truncate_str_upto(s, upto_char, include=True):
    sp = s.split(upto_char)
    if len(sp) == 1:
        return s
    return sp[0]+upto_char if include else sp[0]


def compute_mowe(seg_list, w2v, word_embedding_dim=200,
                 ignore_oov=True, invalid_val=None, squeeze=False):
    embedded = get_np_sentence_embedding(w2v, seg_list, word_embedding_dim,
                                         return_np_mat=True, ignore_oov=ignore_oov)
    if np.all(embedded == 0):
        return invalid_val
    mowe = np.average(embedded, axis=0).reshape(1, -1)
    if squeeze:
        mowe = mowe.squeeze(0)
    return mowe


class LazyPOSDict:

    def __init__(self, nlp):
        self.nlp = nlp
        self._dict = {}

    def __getitem__(self, k):
        try:
            return self._dict[k]
        except KeyError:
            pos_set = set()
            tag_results = self.nlp.pos_tag(k)
            for new_word, tag in tag_results:
                pos_set.add(tag.lower())
            self._dict[k] = pos_set
            return pos_set


class POSWordFilter:

    def __init__(self, pos_dict, valid_pos_set={"vv", "nn", "nr"}):
        self.pos_dict = pos_dict
        self.valid_pos_set = valid_pos_set

    def filter_words_by_pos(self, word_set, valid_pos_set_override=None):
        valid_pos_set = self.valid_pos_set if valid_pos_set_override is None else valid_pos_set_override
        rv_set = set()
        for word in word_set:
            pos_tag_set = self.pos_dict[word]
            if len(pos_tag_set.intersection(valid_pos_set)) > 0:
                rv_set.add(word)
        return rv_set


class SpecialChineseWordTagger:

    def __init__(self, pos_dict,
                 default_special_word_set={"游戏","小游戏","手游","电影","电视剧","小说","漫画","笑话","故事","专辑",
                                           "动漫", "鬼片", "英雄联盟", "王者荣耀","男人","女人","爱好","发型","表现"},
                 special_pos_tag_set={"nr"},
                 log_file=None,
                 n_input_nr_allowed=1):
        self.pos_dict = pos_dict
        self.default_special_word_set = default_special_word_set
        self.special_pos_tag_set = special_pos_tag_set
        self.special_word_log_file = log_file
        self.logs = {}
        self.n_input_nr_allowed = n_input_nr_allowed

    def is_special_chinese_instance(self, word, override_special_word_set=None):
        word = word.rstrip().lower()
        if len(word) == 0: return False
        default_special_word_set = self.default_special_word_set
        if override_special_word_set is not None and isinstance(override_special_word_set, set):
            default_special_word_set = override_special_word_set
        if word in default_special_word_set:
            self._save_log(word, "default")
            return True
        if len(self.special_pos_tag_set) > 0:
            tag_set = self.pos_dict[word]
            rv = False
            for tag in tag_set:
                if tag.lower() in self.special_pos_tag_set:
                    rv = True
                    self._save_log(word, tag)
                    break
            return rv
        else:
            return False

    def _save_log(self, word, reason):
        if self.special_word_log_file is None: return
        if word not in self.logs:
            self.logs[word] = set()
        self.logs[word].add_instance(reason)

    def dump_logs(self):
        if self.special_word_log_file is None: return
        if len(self.logs) == 0: return
        with open(self.special_word_log_file, "w", encoding="utf-8") as f:
            for word, reason_set in self.logs.items():
                line = word + ": " + ",".join(reason_set) + "\n"
                f.write(line)

    def filter_words(self, words, input_word_set):
        rv = set()

        if self.n_input_nr_allowed > 0:
            input_nr_set = set()
            for word in input_word_set:
                tag_set = self.pos_dict[word]
                for tag in tag_set:
                    if tag.lower() == "nr":
                        self._save_log(word, tag)
                        input_nr_set.add(word)
                        break
            if len(input_nr_set) > self.n_input_nr_allowed:
                for word in words:
                    tag_set = self.pos_dict[word]
                    tag = "nn"
                    for tag in tag_set:
                        if tag.lower() == "nr":
                            self._save_log(word, tag)
                            break
                    if tag.lower() != "nr" and self.is_special_chinese_instance(word) and word in input_word_set:
                        rv.add(word)
                return rv

        for word in words:
            if self.is_special_chinese_instance(word):
                if word in input_word_set:
                    rv.add(word)
            else:
                rv.add(word)
        return rv


class TextCompareScore:

    def __init__(self, name, init_value):
        self.name = name
        self.init_value = init_value

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name

    def get_score(self, cand_1_seg_list, cand_2_seg_list):
        raise NotImplementedError("Should be implemented by sub-class")


class MOWECompareScore(TextCompareScore):

    def __init__(self, w2v, word_embedding_dim, name="MOWE cosine sim"):
        super(MOWECompareScore, self).__init__(name, 0.0)
        self.w2v = w2v
        self.word_embedding_dim = word_embedding_dim

    def get_score(self, cand_1_seg_list, cand_2_seg_list):
        cand_1_embedded = get_np_sentence_embedding(self.w2v, cand_1_seg_list, self.word_embedding_dim,
                                                    return_np_mat=True, ignore_oov=True)
        if np.all(cand_1_embedded == 0):
            return 0.0
        cand_2_embedded = get_np_sentence_embedding(self.w2v, cand_2_seg_list, self.word_embedding_dim,
                                                    return_np_mat=True, ignore_oov=True)
        if np.all(cand_2_embedded == 0):
            return 0.0
        cand_1_mowe = np.average(cand_1_embedded, axis=0).reshape(1, -1)
        cand_2_mowe = np.average(cand_2_embedded, axis=0).reshape(1, -1)
        cos_sim = cosine_similarity(cand_1_mowe, cand_2_mowe)[0, 0]
        return cos_sim


class MaskedMOWECompareScore(TextCompareScore):

    def __init__(self, w2v, word_embedding_dim, all_common_words_func,
                 name="Masked MOWE cosine sim"):
        super(MaskedMOWECompareScore, self).__init__(name, 0.0)
        self.w2v = w2v
        self.word_embedding_dim = word_embedding_dim
        self.all_common_words_func = all_common_words_func

    def get_score(self, cand_1_seg_list, cand_2_seg_list):
        cand_1_word_set = set(cand_1_seg_list)
        cand_2_word_set = set(cand_2_seg_list)
        common_words = cand_1_word_set.intersection(cand_2_word_set)
        cand_1_seg_list = [w for w in cand_1_seg_list if w not in common_words]
        cand_2_seg_list = [w for w in cand_2_seg_list if w not in common_words]

        if len(cand_1_seg_list) == 0 or len(cand_2_seg_list) == 0:
            return self.all_common_words_func(cand_1_seg_list, cand_2_seg_list)

        cand_1_embedded = get_np_sentence_embedding(self.w2v, cand_1_seg_list, self.word_embedding_dim,
                                                    return_np_mat=True, ignore_oov=True)
        if np.all(cand_1_embedded == 0):
            return 0.0
        cand_2_embedded = get_np_sentence_embedding(self.w2v, cand_2_seg_list, self.word_embedding_dim,
                                                    return_np_mat=True, ignore_oov=True)
        if np.all(cand_2_embedded == 0):
            return 0.0
        cand_1_mowe = np.average(cand_1_embedded, axis=0).reshape(1, -1)
        cand_2_mowe = np.average(cand_2_embedded, axis=0).reshape(1, -1)
        cos_sim = cosine_similarity(cand_1_mowe, cand_2_mowe)[0, 0]
        return cos_sim


class MaskedMOWE2WordCompareScore(TextCompareScore):

    def __init__(self, w2v, word_embedding_dim, all_common_words_func, rv_strategy="min",
                 name="Masked MOWE2word cosine sim"):
        super(MaskedMOWE2WordCompareScore, self).__init__(name, 0.0)
        self.w2v = w2v
        self.word_embedding_dim = word_embedding_dim
        self.all_common_words_func = all_common_words_func
        self.rv_strategy = rv_strategy

    def get_score(self, cand_1_seg_list, cand_2_seg_list):
        cand_1_word_set = set(cand_1_seg_list)
        cand_2_word_set = set(cand_2_seg_list)
        common_words = cand_1_word_set.intersection(cand_2_word_set)
        cand_1_seg_list = [w for w in cand_1_seg_list if w not in common_words]
        cand_2_seg_list = [w for w in cand_2_seg_list if w not in common_words]

        if len(cand_1_seg_list) == 0 or len(cand_2_seg_list) == 0:
            return self.all_common_words_func(cand_1_seg_list, cand_2_seg_list)

        cand_2_seg_list = [w for w in cand_2_seg_list if w in self.w2v]
        if len(cand_2_seg_list) == 0:
            return 0.0

        cand_1_embedded = get_np_sentence_embedding(self.w2v, cand_1_seg_list, self.word_embedding_dim,
                                                    return_np_mat=True, ignore_oov=True)
        if np.all(cand_1_embedded == 0):
            return 0.0
        cand_1_mowe = np.average(cand_1_embedded, axis=0).reshape(1, -1)
        word2sim = {}
        for word in cand_2_seg_list:
            assert word in self.w2v
            if word in word2sim:
                continue
            word_embedded = self.w2v[word].reshape(1,-1)
            if np.all(word_embedded == 0):
                continue
            cos_sim = cosine_similarity(cand_1_mowe, word_embedded)[0, 0]
            word2sim[word] = cos_sim
        if self.rv_strategy == "min":
            return min([v for w, v in word2sim.items()])
        elif self.rv_strategy == "max":
            return max([v for w, v in word2sim.items()])
        elif self.rv_strategy == "mean":
            return mean([v for w, v in word2sim.items()])
        elif self.rv_strategy == "all":
            return word2sim
        assert False, "Unknown strategy: {}".format(self.rv_strategy)


class MaskedPairwiseWordPivotScore(TextCompareScore):

    def __init__(self, w2v, word_embedding_dim, all_common_words_func,
                 match_strategy="max", rv_strategy="min", cand_2_as_pivot=True,
                 name="Masked pairwise word cosine sim"):
        super(MaskedPairwiseWordPivotScore, self).__init__(name, 0.0)
        self.w2v = w2v
        self.word_embedding_dim = word_embedding_dim
        self.all_common_words_func = all_common_words_func
        self.match_strategy = match_strategy
        self.rv_strategy = rv_strategy
        self.cand_2_as_pivot = cand_2_as_pivot

    def get_score(self, cand_1_seg_list, cand_2_seg_list):
        if not self.cand_2_as_pivot:
            tmp = cand_1_seg_list
            cand_1_seg_list = cand_2_seg_list
            cand_2_seg_list = tmp
        cand_1_word_set = set(cand_1_seg_list)
        cand_2_word_set = set(cand_2_seg_list)
        common_words = cand_1_word_set.intersection(cand_2_word_set)
        cand_1_seg_list = [w for w in cand_1_seg_list if w not in common_words]
        cand_2_seg_list = [w for w in cand_2_seg_list if w not in common_words]

        if len(cand_1_seg_list) == 0 or len(cand_2_seg_list) == 0:
            return self.all_common_words_func(cand_1_seg_list, cand_2_seg_list)

        cand_1_seg_list = [w for w in cand_1_seg_list if w in self.w2v]
        if len(cand_1_seg_list) == 0:
            return 0.0
        cand_2_seg_list = [w for w in cand_2_seg_list if w in self.w2v]
        if len(cand_2_seg_list) == 0:
            return 0.0

        word2sims = {}
        for cand_2_word in set(cand_2_seg_list):
            assert cand_2_word in self.w2v
            if cand_2_word in word2sims:
                continue
            cand_2_word_embedded = self.w2v[cand_2_word].reshape(1,-1)
            if np.all(cand_2_word_embedded == 0):
                continue
            word2sims[cand_2_word] = {}
            for cand_1_word in set(cand_1_seg_list):
                assert cand_1_word in self.w2v
                if cand_1_word in word2sims[cand_2_word]:
                    continue
                cand_1_word_embedded = self.w2v[cand_1_word].reshape(1,-1)
                cos_sim = cosine_similarity(cand_1_word_embedded, cand_2_word_embedded)[0, 0]
                word2sims[cand_2_word][cand_1_word] = cos_sim
        if self.match_strategy == "max":
            word2match_sim = {w: max([v for _, v in sim_dict.items()]) for w, sim_dict in word2sims.items()}
        elif self.match_strategy == "min":
            word2match_sim = {w: min([v for _, v in sim_dict.items()]) for w, sim_dict in word2sims.items()}
        elif self.match_strategy == "all":
            word2match_sim = word2sims
        else:
            assert False, "Unknown match strategy: {}".format(self.match_strategy)

        if self.rv_strategy == "min":
            return min([v for w, v in word2match_sim.items()])
        elif self.rv_strategy == "max":
            return max([v for w, v in word2match_sim.items()])
        elif self.rv_strategy == "mean":
            return mean([v for w, v in word2match_sim.items()])
        elif self.rv_strategy == "all":
            return word2match_sim
        assert False, "Unknown rv strategy: {}".format(self.rv_strategy)


class AvgWordCountScore(TextCompareScore):

    def __init__(self, w2c, no_available_words_score=0.5, cand_2_as_pivot=True,
                 name="Normalized average word count"):
        super(AvgWordCountScore, self).__init__(name, 0.0)
        self.w2c = w2c
        self.no_available_words_score = no_available_words_score
        self.cand_2_as_pivot = cand_2_as_pivot
        self.max_word_count = max([c for w, c in self.w2c.items()])

    def get_score(self, cand_1_seg_list, cand_2_seg_list):
        if not self.cand_2_as_pivot:
            tmp = cand_1_seg_list
            cand_1_seg_list = cand_2_seg_list
            cand_2_seg_list = tmp
        cand_1_word_set = set(cand_1_seg_list)

        cand_2_seg_list = [w for w in cand_2_seg_list if w in self.w2c and w not in cand_1_word_set]
        if len(cand_2_seg_list) == 0:
            return self.no_available_words_score

        return mean([float(self.w2c[word]) / self.max_word_count for word in cand_2_seg_list])


def percent_word_overlap(seg_list_1, seg_list_2):
    set_1 = set(seg_list_1)
    set_2 = set(seg_list_2)
    if len(set_1) + len(set_2) == 0:
        return 0
    return len(set_1.intersection(set_2)) / len(set_1.union(set_2))
