import gc
import os
import copy
import time
import math
import torch
import random
import itertools
import numpy as np
import collections
from itertools import chain, combinations, permutations
from collections import OrderedDict


def torch_geo_batch_to_data_list(batch_tsr, last_node_idx_list):
    """
    :param batch_tsr: PyTorch geo batch object
    :param last_node_idx_list: global batch indices of the last nodes of each graph
    :return: list of graphs recovered from the batch
    """
    rv = []
    batch_idx = 0
    graph_idx = 0
    while batch_idx < batch_tsr.shape[0]:
        rv.append(batch_tsr[batch_idx: last_node_idx_list[graph_idx] + 1, :])
        batch_idx = last_node_idx_list[graph_idx] + 1  # start of next batch
        graph_idx += 1
    return rv


def exp_scale(value, base=math.e, multiplier=10.,
              old_range_min=0., old_range_max=1.,
              new_range_min=0., new_range_max=1.):
    value = get_tsr_value(value)
    assert old_range_min <= value <= old_range_max
    old_range_min = base ** (old_range_min * multiplier)
    if abs(old_range_min) == float('inf'): raise OverflowError()
    old_range_max = base ** (old_range_max * multiplier)
    if abs(old_range_max) == float('inf'): raise OverflowError()
    new_value = base ** (value * multiplier)
    new_value = map_to_range(new_value, old_range_min, old_range_max, new_range_min, new_range_max)
    return new_value


def exp_scale_v2(value, p=10,
                 old_range_min=0., old_range_max=1.,
                 new_range_min=0., new_range_max=1.):
    value = get_tsr_value(value)
    assert old_range_min <= value <= old_range_max
    new_value = map_to_range(value, old_range_min, old_range_max, new_range_min, new_range_max)
    return new_value ** p


def get_tsr_value(t):
    if not isinstance(t, torch.Tensor):
        return t
    ts = t.squeeze().unsqueeze(0)
    if len(ts.shape) == 1 and ts.shape[0] == 1:
        return ts.squeeze().item()
    else:
        return ts.squeeze().tolist()


def map_to_range(value, old_min, old_max, new_min, new_max):
    leftSpan = old_max - old_min
    rightSpan = new_max - new_min
    scaled_value = float(value - old_min) / float(leftSpan)
    return new_min + (scaled_value * rightSpan)


def gen_log_scale_points(start, end, num_points, base=10.0):
    points = np.logspace(0, 1, num=num_points, base=base).tolist()
    rescaled_points = [map_to_range(p, min(points), max(points), start, end)  for p in points]
    return rescaled_points


def is_complete_dag(num_nodes, edge_list):
    edge_set = {(src, dst) for src, dst in zip(edge_list[0], edge_list[1])}
    for n1 in range(num_nodes):
        for n2 in range(num_nodes):
            if n2 <= n1: continue
            if (n1, n2) not in edge_set:
                return False
    return True


def get_complete_dag_edges(num_nodes):
    edge_pairs = set()
    for n1 in range(num_nodes):
        for n2 in range(num_nodes):
            if n2 <= n1: continue
            edge_pairs.add( (n1, n2) )
    return edge_pairs


def prune_self_edge_from_edge_list(torch_geo_edge_list):
    src_list, dst_list = [], []
    for src, dst in zip(torch_geo_edge_list[0], torch_geo_edge_list[1]):
        if src != dst:
            src_list.append(src)
            dst_list.append(dst)
    return [src_list, dst_list]


def get_output_node_indices(torch_geo_edge_list, output_node_idx=None):
    src_set, dst_set = set(), set()
    edge_list = torch_geo_edge_list
    for src, dst in zip(edge_list[0], edge_list[1]):
        if src == dst: continue  # ignore self edges
        if output_node_idx is not None and dst == output_node_idx: continue
        src_set.add(src)
        dst_set.add(dst)
    return dst_set - src_set


def get_input_node_indices(torch_geo_edge_list):
    src_set, dst_set = set(), set()
    edge_list = torch_geo_edge_list
    for src, dst in zip(edge_list[0], edge_list[1]):
        if src == dst: continue
        src_set.add(src)
        dst_set.add(dst)
    return src_set - dst_set


def edge_list_to_edge_pairs(edge_list):
    rv = list(zip(edge_list[0], edge_list[1]))
    return sorted(rv)


def random_truncate_txt_file(fn, n_lines):
    with open(fn, "r", encoding="utf-8") as f:
        lines = f.readlines()
    random.shuffle(lines)
    lines = lines[:n_lines]
    write_lines_to_file(lines, fn, new_file=True)


def write_dict_to_file(d, fn, col_delim="|"):
    n_vals_wrote = 0
    with open(fn, "w", encoding="utf-8") as f:
        for k, v in d.items():
            f.write(col_delim.join([k, str(v)]) + "\n")
            n_vals_wrote += 1
    print("{} dict values wrote to {}".format(n_vals_wrote, fn))


def get_project_root_dir():
    root_dir = os.path.sep.join( [os.path.dirname(os.path.abspath(__file__)),  ".."])
    return root_dir


def split_train_valid_test_files(fn, op_fn_prefix, valid_ratio=0.1, test_ratio=0.1):
    with open(fn, "r", encoding="utf-8") as f:
        lines = f.readlines()
    split_train_valid_test_lines(lines, op_fn_prefix, valid_ratio=valid_ratio, test_ratio=test_ratio)


def split_train_valid_test_lines(lines, op_fn_prefix, valid_ratio=0.1, test_ratio=0.1):
    random.shuffle(lines)
    assert len(lines) > 0, "No lines found"
    if valid_ratio > 0:
        assert valid_ratio < 1.0, "Invalid validation ratio, must be between 0 and 1"
        n_valid_lines = int(len(lines) * valid_ratio)
        valid_lines = lines[:n_valid_lines]
        lines = lines[n_valid_lines:]
        assert len(lines) > 0, "Ran out of lines"
        print("Writing {} lines of valid data to {}".format(n_valid_lines, op_fn_prefix + "_valid.txt"))
        write_lines_to_file(valid_lines, op_fn_prefix + "_valid.txt", new_file=True)
    if test_ratio > 0:
        assert test_ratio < 1.0, "Invalid test ratio, must be between 0 and 1"
        n_test_lines = int(len(lines) * test_ratio)
        test_lines = lines[:n_test_lines]
        lines = lines[n_test_lines:]
        assert len(lines) > 0, "Ran out of lines"
        print("Writing {} lines of test data to {}".format(n_test_lines, op_fn_prefix + "_test.txt"))
        write_lines_to_file(test_lines, op_fn_prefix + "_test.txt", new_file=True)
    print("Writing {} lines of train data to {}".format(len(lines), op_fn_prefix + "_train.txt"))
    write_lines_to_file(lines, op_fn_prefix + "_train.txt", new_file=True)


def split_bin_classify_train_valid_test_files(fn, op_fn_prefix,
                                              class_label_idx=0,
                                              col_delim="|",
                                              remove_header=True,
                                              valid_ratio=0.1, test_ratio=0.1):
    with open(fn, "r", encoding="utf-8") as f:
        lines = f.readlines()
    split_bin_classify_train_valid_test_lines(lines, op_fn_prefix=op_fn_prefix, class_label_idx=class_label_idx,
                                              col_delim=col_delim, remove_header=remove_header,
                                              valid_ratio=valid_ratio, test_ratio=test_ratio)


def split_bin_classify_train_valid_test_lines(lines, op_fn_prefix,
                                              class_label_idx=0,
                                              col_delim="|",
                                              remove_header=True,
                                              valid_ratio=0.1, test_ratio=0.1):
    if remove_header:
        del lines[0]
    label_to_insts = {}
    for line in lines:
        line = line.strip()
        if len(line) == 0: continue
        data = line.split(col_delim)
        label = data[class_label_idx]
        if label not in label_to_insts:
            label_to_insts[label] = set()
        label_to_insts[label].add(line)
    assert len(label_to_insts) == 2, "label is not binary: {}".format(str(label_to_insts.keys()))
    class_sizes = {k:len(v) for k, v in label_to_insts.items()}
    print("class sizes: {}".format(class_sizes))
    split_size = min([v for _, v in class_sizes.items()])
    valid_size = int(split_size * valid_ratio)
    test_size = int(split_size * test_ratio)
    train_size = split_size - valid_size - test_size
    print("train split size: {}, valid split size: {}, test split size: {}".format(train_size, valid_size, test_size))
    class_data = [(k, list(v)) for k, v in label_to_insts.items()]
    class_train = [(k, v[:train_size]) for k, v in class_data]
    class_valid = [(k, v[train_size:train_size + valid_size]) for k, v in class_data]
    class_test = [(k, v[train_size+valid_size:]) for k, v in class_data]
    train_data, valid_data, test_data = [], [], []
    for _, v in class_train: train_data += v
    for _, v in class_valid: valid_data += v
    for _, v in class_test: test_data += v
    print("train final size: {}, valid split size: {}, test split size: {}".format(len(train_data), len(valid_data), len(test_data)))
    random.shuffle(train_data)
    random.shuffle(valid_data)
    random.shuffle(test_data)
    with open(op_fn_prefix + "_train.txt", "w", encoding='utf-8') as f:
        for line in train_data:
            f.write(line + "\n")
    if valid_ratio > 0:
        with open(op_fn_prefix + "_valid.txt", "w", encoding='utf-8') as f:
            for line in valid_data:
                f.write(line + "\n")
    if test_ratio > 0:
        with open(op_fn_prefix + "_test.txt", "w", encoding='utf-8') as f:
            for line in test_data:
                f.write(line + "\n")


def is_number(s):
    try:
        float(s)
        return True
    except ValueError:
        return False


def merge_batch_word_seg_lists(b_seg_list_1, b_seg_list_2, remove_duplicate_words=True):
    assert len(b_seg_list_1) == len(b_seg_list_2), "Batch size must equal"
    rv = []
    for bi in range(len(b_seg_list_1)):
        if remove_duplicate_words:
            src_words = set(b_seg_list_1[bi])
            val = b_seg_list_1[bi] + [w for w in b_seg_list_2[bi] if w not in src_words]
        else:
            val = b_seg_list_1[bi] + b_seg_list_2[bi]
        rv.append(val)
    return rv


def label_tsr_to_one_hot_tsr(label_tsr, vocab_size, mask_pad=False):
    batch_size = label_tsr.shape[0]
    y_onehot = torch.zeros(batch_size, vocab_size).float()
    y_onehot.scatter_(1, label_tsr, 1)
    if mask_pad:
        y_onehot[:, 0] = 0 # mask out <PAD>
    # print(y_onehot)
    # print(y_onehot.sum())
    # print(label_tsr)
    return y_onehot


def idx_list_to_one_hot_lists(label_list, label_dim):
    rv = []
    for idx in label_list:
        rv.append(idx_to_one_hot_list(idx, label_dim))
    return rv


def idx_to_one_hot_list(label, label_dim):
    vec = [0] * label_dim
    vec[label] = 1
    return vec


def one_hot_lists_to_idx_list(one_hot_lists):
    rv = []
    for one_hot_list in one_hot_lists:
        rv.append(one_hot_list_to_idx(one_hot_list))
    return rv


def one_hot_list_to_idx(one_hot_list):
    idx_list = [i for i, v in enumerate(one_hot_list) if v > 0]
    assert len(idx_list) == 1, "Zero or more than one number of 1s found"
    return idx_list[0]


def time_checkpoint(prev_chkpt=None, chkpt_name="", verbose=True):
    curr_chkpt = time.time()
    if prev_chkpt is None:
        return curr_chkpt
    elapsed = curr_chkpt - prev_chkpt
    if verbose:
        print(chkpt_name + " elapsed {}".format(elapsed))
    return curr_chkpt


def mean(list_val, fallback_val=None):
    if len(list_val) == 0:
        if fallback_val is not None:
            return fallback_val
        else:
            raise ZeroDivisionError()
    return sum(list_val) / len(list_val)


def variance(list_val, fallback_val=0):
    if len(list_val) == 0:
        if fallback_val is not None:
            return fallback_val
        else:
            raise ValueError()
    v = np.var(np.asarray(list_val))
    return v

def median(list_val, fallback_val=0):
    if len(list_val) == 0:
        if fallback_val is not None:
            return fallback_val
        else:
            raise ValueError()
    v = np.median(np.asarray(list_val))
    return v

# By @ smth
def mem_report():
    for obj in gc.get_objects():
        if torch.is_tensor(obj) or (hasattr(obj, 'data') and torch.is_tensor(obj.data)):
            print(type(obj), obj.size())


def write_line_to_file(s, f_path="progress.txt", new_file=False, verbose=False):
    code = "w" if new_file else "a"
    if verbose: print(s)
    with open(f_path, code, encoding='utf-8') as f:
        f.write(s)
        f.write("\n")


def write_lines_to_file(lines, f_path="progress.txt", new_file=False, verbose=False):
    code = "w" if new_file else "a"
    if verbose:
        print("Writing {} lines to {}".format(len(lines), f_path))
    with open(f_path, code, encoding='utf-8') as f:
        for s in lines:
            f.write(s.strip())
            f.write("\n")


def file_len(fname):
    i = -1
    with open(fname, "r", encoding="utf-8") as f:
        for i, l in enumerate(f):
            pass
    return i + 1


def read_file_by_line(fname, skip_lines=0):
    lines = []
    with open(fname) as f:
        for i, l in enumerate(f):
            if i < skip_lines: continue
            lines.append(l)
    return lines


def read_lines_in_file(fname, lines_set):
    rv = []
    with open(fname) as f:
        for i, line in enumerate(f):
            if i in lines_set:
                rv.append((i, line.strip()))
    return sorted(rv, key=lambda t:t[0])


def split_read_process_file(in_fn, split_line_count, process_func, func_args_list, ignore_header=False):
    total_n_lines = file_len(in_fn)
    processed_lines_count = 0
    while processed_lines_count < total_n_lines:
        lines = read_file_by_line(in_fn, skip_lines=split_line_count)
        if processed_lines_count == 0 and ignore_header:
            lines = lines[1:]
        process_func(lines, *func_args_list)
        processed_lines_count += len(lines)


def flatten_nested_lists(lists):
    def _recur_flat(val):
        if isinstance(val, list):
            for v in val:
                _recur_flat(v)
        else:
            rv.append(val)
    rv = []
    _recur_flat(lists)
    return rv


def global_max(lists):
    def _recur_max(val):
        if isinstance(val, list):
            return max(_recur_max(v) for v in val)
        else:
            return val
    return _recur_max(lists)


def global_min(lists):
    def _recur_min(val):
        if isinstance(val, list):
            return min(_recur_min(v) for v in val)
        else:
            return val
    return _recur_min(lists)


def inflate_to_inc_nested_lists(flat_list, allow_remainder=False):
    rv = []
    curr_idx, curr_size = 0, 1
    while curr_idx < len(flat_list):
        chunk = flat_list[curr_idx: curr_idx + curr_size]
        if not allow_remainder and len(chunk) < curr_size:
            raise ValueError("Unable to divide input list into size-incrementing partitions")
        rv.append(chunk)
        curr_idx += curr_size
        curr_size += 1
    return rv


class UnbufferedStdOut:
    # sys.stdout=UnbufferedStdOut(sys.stdout)
    def __init__(self, stream, filename=None):
        self.stream = stream
        self.filename = filename

    def write(self, data):
        self.stream.write(data)
        self.stream.flush()
        if self.filename is not None:
            write_line_to_file(str(data), self.filename)


class UniqueDict(dict):
    def __init__(self, inp=None):
        self._no_dups = True
        if isinstance(inp, dict):
            super(UniqueDict,self).__init__(inp)
        else:
            super(UniqueDict,self).__init__()
            if isinstance(inp, (collections.Mapping, collections.Iterable)):
                si = self.__setitem__
                for k,v in inp:
                    si(k,v)
        self._no_dups = False

    def __setitem__(self, k, v):
        try:
            self.__getitem__(k)
            if self._no_dups:
                raise ValueError("duplicate key '{0}' found".format(k))
            else:
                super(UniqueDict, self).__setitem__(k, v)
        except KeyError:
            super(UniqueDict,self).__setitem__(k,v)


class TrieNode:

    def __init__(self, child_nodes=OrderedDict(), is_end_of_instance=False):
        self.child_nodes = child_nodes
        self.is_end_of_instance = is_end_of_instance

    def n_candidates(self):
        return len([wi for wi, val in self.child_nodes.items() if val is not None])

    def __str__(self):
        return "TrieNode with {} child nodes, is_end_of_instance={}".format(len(self.child_nodes), str(self.is_end_of_instance))

    def __repr__(self):
        return str(self)


class WordIndexTrie:

    def __init__(self, add_special_pad_wi=True):
        self.unique_words = set()
        self.unique_instances = set()
        self.n_instances = 0
        self.add_special_pad_wi = add_special_pad_wi
        self.root = self._init_trie_node()

    def _init_trie_node(self):
        if self.add_special_pad_wi:
            n = TrieNode(child_nodes=OrderedDict([(0, None)]))
        else:
            n = TrieNode(child_nodes=OrderedDict())
        return n

    def iterative_insert(self, word_idx_list):
        curr = self.root
        for i, wi in enumerate(word_idx_list):
            self.unique_words.add(wi)
            if wi not in curr.child_nodes:
                curr.child_nodes[wi] = self._init_trie_node()
            curr = curr.child_nodes[wi]
        curr.is_end_of_instance = True
        self.unique_instances.add(" ".join([str(wi) for wi in word_idx_list]))

    @staticmethod
    def get_batch_cand_word_idx_tsr(nodes, pad_idx=0, max_n_cand_words_override=None):
        # Avoid inputting nodes that have no child_nodes
        # Return tensor need to make sure there is at least 2 words to choose from
        assert all([len(node.child_nodes) > 0 for node in nodes])
        rv = []
        max_n_cand_words = max([len(node.child_nodes) for node in nodes])
        if max_n_cand_words == 0: return None
        if max_n_cand_words_override is not None:
            max_n_cand_words = min(max_n_cand_words, max_n_cand_words_override)
        for bi, node in enumerate(nodes):
            inst_rv = []
            for wi, _ in node.child_nodes.items():
                inst_rv.append(wi)
            assert len(inst_rv) > 0
            while len(inst_rv) < max_n_cand_words:
                inst_rv.append(pad_idx)
            rv.append(inst_rv)
        return torch.LongTensor(rv)

    def get_cand_child_nodes(self, wi_list):
        curr = self.root
        rv = []
        for wi in wi_list:
            assert wi in curr.child_nodes
            cand_nodes = list(curr.child_nodes.keys())
            rv.append(cand_nodes)
            curr = curr.child_nodes[wi]
        return rv

    def get_truth_wi_idx_in_nodes(self, wi_list):
        curr = self.root
        rv = []
        for wi in wi_list:
            assert wi in curr.child_nodes
            i = 0
            for key, _ in curr.child_nodes.items():
                if key == wi:
                    rv.append(i)
                    break
                i += 1
            assert i < len(curr.child_nodes)
            curr = curr.child_nodes[wi]
        return rv

    def _build_leveled_traversal_nodes_for_inst(self, wi_list, rv,
                                                batch_idx=0, inst_idx=0,
                                                curr_input_wi=2):
        curr_node = self.root
        curr_level = 0
        curr_path_key = ""
        for wi in wi_list:
            assert wi in curr_node.child_nodes, "curr_level: {}, {} not found among {} child_nodes".format(curr_level, wi,
                                                                                                           len(curr_node.child_nodes))
            if curr_node.n_candidates() == 1:
                # No need to train if there is only one choice!
                curr_input_wi = wi
                curr_node = curr_node.child_nodes[wi]
                continue
            # Find truth_idx
            i = 0
            truth_idx = None
            cand_wi_list = [k for k, _ in curr_node.child_nodes.items()]
            for key, _ in curr_node.child_nodes.items():
                if key == wi:
                    truth_idx = i
                    break
                i += 1
            assert truth_idx is not None
            # Build node
            t_node = WITraversalNode(curr_level, batch_idx, inst_idx,
                                     curr_input_wi, cand_wi_list, truth_idx, wi, curr_path_key)
            while len(rv) < curr_level + 1: rv.append([])
            rv[curr_level].append(t_node)
            curr_level += 1
            curr_input_wi = wi
            curr_path_key += " " + str(wi) if len(curr_path_key) > 0 else str(wi)
            curr_node = curr_node.child_nodes[wi]
        return rv

    def get_leveled_batch_traversal_nodes(self, b_tgt_insts, sos_idx=2):
        rv = []
        for bi, insts in enumerate(b_tgt_insts):
            for inst_idx, inst_seg_list in enumerate(insts):
                self._build_leveled_traversal_nodes_for_inst(inst_seg_list, rv, batch_idx=bi, inst_idx=inst_idx,
                                                             curr_input_wi=sos_idx)
        return rv

    def __str__(self):
        return "WordIndexTrie with {} unique words and {} instances".format(len(self.unique_words), len(self.unique_instances))

    def __repr__(self):
        return str(self)


class WITraversalNode:

    def __init__(self, level, batch_idx, inst_idx, input_wi, cand_wi_list, truth_idx, truth_wi, path_key):
        self.level = level
        self.batch_idx = batch_idx
        self.inst_idx = inst_idx
        self.input_wi = input_wi
        self.cand_wi_list = cand_wi_list
        self.truth_idx = truth_idx
        self.truth_wi = truth_wi
        self.path_key = path_key
        self.cand_wi_key = " ".join([str(wi) for wi in cand_wi_list])

    def __str__(self):
        return "WITraversalNode(level:{}, batch_idx:{}, inst_idx:{}, input_wi:{}, truth_wi:{})".format(self.level,
                                                                                                       self.batch_idx,
                                                                                                       self.inst_idx,
                                                                                                       self.input_wi,
                                                                                                       self.truth_wi)

    def __repr__(self):
        return str(self)


def get_np_sentence_embedding(w2v, words_list, embed_dim,
                              ignore_oov=True, return_np_mat=False):
    if ignore_oov:
        iv_words = [w for w in words_list if w in w2v]
        if len(iv_words) == 0: return np.zeros( (1, embed_dim) ) if return_np_mat else [np.zeros(embed_dim)]
    else:
        iv_words = words_list
    rv = np.zeros( (len(iv_words), embed_dim) ) if return_np_mat else []
    for i, w in enumerate(iv_words):
        if return_np_mat:
            if w in w2v:
                rv[i,:] = w2v[w]
        else:
            if w in w2v:
                rv.append(w2v[w])
            elif not ignore_oov:
                rv.append(np.zeros(embed_dim))
    return rv


def get_compact_str_for_level_seg_lists(level_seg_lists, level_delim="#", inst_delim="|", value_delim=" "):
    return level_delim.join([get_compact_str_for_seg_lists(seg_lists, inst_delim=inst_delim, value_delim=value_delim)
                             for seg_lists in level_seg_lists])


def get_compact_str_for_seg_lists(seg_lists, inst_delim="|", value_delim=" "):
    return inst_delim.join([get_compact_str_for_seg_list(seg_list, value_delim=value_delim) for seg_list in seg_lists])


def get_compact_str_for_seg_list(seg_list, value_delim=" ", packing_func=str):
    return value_delim.join([packing_func(val) for val in seg_list])


def get_seg_list_from_compact_str(text, value_delim=" ", d_type_func=int):
    return [d_type_func(val) for val in text.split(value_delim)]


def get_seg_lists_from_compact_str(text, inst_delim="|", value_delim=" ", d_type_func=int):
    return [get_seg_list_from_compact_str(inner_text, value_delim=value_delim, d_type_func=d_type_func)
            for inner_text in text.split(inst_delim)]


def get_level_seg_lists_from_compact_str(text, level_delim="#", inst_delim="|", value_delim=" ", d_type_func=int):
    return [get_seg_lists_from_compact_str(inner_text, inst_delim=inst_delim, value_delim=value_delim, d_type_func=d_type_func)
            for inner_text in text.split(level_delim)]


def get_torch_tensor_from_batch_seg_lists(batch_seg_lists, tsr_func=torch.LongTensor):
    return tsr_func(batch_seg_lists)


def make_batch_pair_2d_mask_tsr(cand_1_mask, cand_2_mask):
    assert len(cand_1_mask.shape) == len(cand_2_mask.shape) == 3
    cand_1_mask_2d = cand_1_mask.repeat(1, cand_2_mask.shape[2], 1).transpose(1, 2)
    cand_2_mask_2d = cand_2_mask.repeat(1, cand_1_mask.shape[2], 1)
    assert cand_1_mask_2d.shape == cand_2_mask_2d.shape
    rv = cand_1_mask_2d & cand_2_mask_2d
    return rv


def cumulative_moving_average(prev_avg, prev_n, new_val):
    rv = (new_val + prev_n * prev_avg) / (prev_n + 1)
    return rv


class RelevantWordGroup:

    def __init__(self, default_score_list,
                 score_delim=",", word_scores_delim=":", inst_delim=" "):
        self.word2scores = {}
        self.word2count = {}
        self.default_score_list = default_score_list
        self.score_delim = score_delim
        self.word_scores_delim = word_scores_delim
        self.inst_delim = inst_delim

    def add_new_word(self, word, score_list=None, word2count_value=1):
        if score_list is None:
            score_list = self.default_score_list
        if word not in self.word2scores:
            self.word2scores[word] = score_list
            self.word2count[word] = word2count_value
        else:
            new_score_list = [
                max(self.word2scores[word][i], score_list[i])
                for i in range(len(self.default_score_list))
            ]
            self.word2scores[word] = new_score_list
            self.word2count[word] += word2count_value

    @staticmethod
    def build_from_str(str_rep, default_score_list,
                       inst_delim=" ", word_scores_delim=":", score_delim=","):
        rv = RelevantWordGroup(default_score_list,
                               inst_delim=inst_delim, word_scores_delim=word_scores_delim,
                               score_delim=score_delim)
        word_scores_list = [val.split(word_scores_delim) for val in str_rep.split(inst_delim)]
        for word, count, scores_str in word_scores_list:
            score_list = [float(score) for score in scores_str.split(score_delim)]
            assert len(score_list) == len(default_score_list)
            rv.add_new_word(word, score_list, word2count_value=int(count))
        return rv

    def merge_group(self, group):
        if not isinstance(group, RelevantWordGroup):
            return
        for word, score_list in group.word2scores.items():
            self.add_new_word(word, score_list, word2count_value=group.word2count[word])

    def __len__(self):
        return len(self.word2scores)

    def __str__(self):
        sorted_word_scores = [(word, scores) for word, scores in self.word2scores.items()]
        sorted_word_scores.sort(key=lambda t: t[1], reverse=True)
        rv = self.inst_delim.join([
            self.word_scores_delim.join([word, str(self.word2count[word]), self.score_delim.join([str(score) for score in scores])])
            for word, scores in sorted_word_scores
        ])
        return rv

    def __repr__(self):
        return str(self)


def get_subset_lists(seg_list, return_all_permutations=True, include_empty=False, include_self=False):
    subsets = []
    from itertools import permutations

    def _backtrack(curr_list, start_idx):
        subsets.append([v for v in curr_list])
        for i in range(start_idx, len(seg_list)):
            curr_list.append(seg_list[i])
            _backtrack(curr_list, start_idx + 1)
            curr_list.pop()

    _backtrack([], 0)
    subsets.sort(key=lambda v:len(v))
    if not include_empty:
        while len(subsets) > 0 and len(subsets[0]) == 0:
            del subsets[0]
    if not include_self:
        while len(subsets) > 0 and len(subsets[-1]) == len(seg_list):
            subsets.pop()

    if return_all_permutations:
        rv = []
        for vals in subsets:
            rv.extend(list(permutations(vals)))
    else:
        rv = subsets
    return rv


def get_files_by_extension(extension, dir_path=None):
    import os
    from os import listdir
    from os.path import isfile, join
    if len(extension) > 0 and extension[0] != ".":
        extension = "." + extension
    if dir_path is None:
        curr_dir = os.path.dirname(os.path.realpath(__file__))
        files_dir = curr_dir + os.path.sep
    else:
        files_dir = dir_path + os.path.sep
    onlyfiles = [f for f in listdir(files_dir) if isfile(join(files_dir, f))]
    rv = []
    for path in onlyfiles:
        full_path = files_dir + path
        filename, file_extension = os.path.splitext(full_path)
        if file_extension != extension: continue # includes dot
        rv.append(full_path)
    return rv


def recur_get_files_by_extension(extension, parent_dir):
    from os import listdir
    from os.path import isdir, join
    rv = get_files_by_extension(extension, parent_dir)
    folders = [f for f in listdir(parent_dir) if isdir(join(parent_dir, f))]
    for folder in folders:
        rv += recur_get_files_by_extension(extension, os.path.sep.join([parent_dir, folder]))
    return rv


def powerset(iterable, exclude_empty=False, exclude_full=False):
    s = list(iterable)
    start = 0 if not exclude_empty else 1
    end = len(s) + 1 if not exclude_full else len(s)
    return chain.from_iterable(combinations(s, r) for r in range(start, end))


def list_permutations(iterable):
    perm = permutations(iterable)
    return list(perm)


def generate_subsets_size_n(list_vals, n):
    return list(itertools.combinations(list_vals, n))


def list_element_wise_op(lst_1, lst_2, op, pad_val=None):
    if pad_val is not None:
        while len(lst_1) < len(lst_2):
            lst_1.append(pad_val)
        while len(lst_2) < len(lst_1):
            lst_2.append(pad_val)
    else:
        assert len(lst_1) == len(lst_2)
    return [op(v1, v2) for v1, v2 in zip(lst_1, lst_2)]


def str2bool(v):
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise TypeError('Boolean value expected.')


def pjoin(str_list):
    return os.path.join(str_list)


def flip_discrete_probs(probs):
    indexed_probs = [[i, p] for i, p in enumerate(probs)]
    indexed_probs.sort(key=lambda t: t[1])
    start, end = 0, len(indexed_probs) - 1
    while start < end:
        indexed_probs[start][1], indexed_probs[end][1] = indexed_probs[end][1], indexed_probs[start][1]
        start += 1
        end -= 1
    rv = [-1] * len(probs)
    for i, t in enumerate(indexed_probs):
        rv[i] = t[1]
    assert all(v >= 0. for v in rv) and abs(sum(rv) - 1.) < 1e-4
    return rv

def _is_cyclic_util(v, adj_list, visited, recStack):

    # Mark current node as visited and
    # adds to recursion stack
    visited[v] = True
    recStack[v] = True

    # Recur for all neighbours
    # if any neighbour is visited and in
    # recStack then graph is cyclic
    for neighbor in adj_list[v]:
        if not visited[neighbor]:
            if _is_cyclic_util(neighbor, adj_list, visited, recStack):
                return True
        elif recStack[v]:
            return True

    # The node needs to be popped from
    # recursion stack before function ends
    recStack[v] = False
    return False


def is_cyclic(adj_list):
    visited = [False] * len(adj_list)
    recStack = [False] * len(adj_list)
    for node in range(len(adj_list)):
        if not visited[node]:
            if _is_cyclic_util(node, adj_list, visited, recStack):
                return True
    return False


def edge_list_to_adj_dict(src_inds, dst_inds, ignore_self_edge=True):
    adj_dict = collections.defaultdict(set)
    for src, dst in zip(src_inds, dst_inds):
        if ignore_self_edge and src == dst:
            continue
        adj_dict[src].add(dst)
    return adj_dict


def edge_list_to_adj_mat(src_inds, dst_inds, ignore_self_edge=True):
    edge_pairs = edge_list_to_edge_pairs([src_inds, dst_inds])
    n_nodes = max(max(p) for p in edge_pairs) + 1 # Input must be zero indexed
    mat = [[0 for i in range(n_nodes)] for j in range(n_nodes)]
    for src, dst in edge_pairs:
        if ignore_self_edge and src == dst: continue
        mat[src][dst] = 1
    return mat


def get_incoming_nodes_dict(adj_dict):
    rv = collections.defaultdict(set)
    for src, dst_set in adj_dict.items():
        for dst in dst_set:
            rv[dst].add(src)
    return rv


def dfs_path(adj_dict, curr_src=0, seen=None, path=None):
    if seen is None: seen = []
    if path is None: path = [curr_src]
    seen.append(curr_src)
    paths = []
    for t in list(adj_dict[curr_src]):
        if t not in seen:
            t_path = path + [t]
            paths.append(tuple(t_path))
            paths.extend(dfs_path(adj_dict, t, seen[:], t_path))
    return paths


def rmdir(directory):
    from pathlib import Path
    directory = Path(directory)
    for item in directory.iterdir():
        if item.is_dir():
            rmdir(item)
        else:
            item.unlink()
    directory.rmdir()


class UniqueList:

    def __init__(self, key_func=str):
        self._i = 0
        self._list = []
        self._set = set()
        self._key_func = key_func

    def tolist(self):
        return [v for v in self._list]

    def keys(self):
        return copy.deepcopy(self._set)

    def clear(self):
        self._i = 0
        self._list = []
        self._set = set()

    def append(self, val):
        key = self._key_func(val)
        if key not in self._set:
            self._set.add(key)
            self._list.append(val)
            return True
        return False

    def extend(self, vals):
        n_added = 0
        for item in vals:
            if self.append(item):
                n_added += 1
        return n_added

    def next(self):
        if self._i >= len(self._list):
            self._i  = 0
            raise StopIteration()
        item = self._list[self._i]
        self._i += 1
        return item

    def __contains__(self, item):
        key = self._key_func(item)
        return key in self._set

    def __iter__(self):
        self._i = 0
        return self

    def __next__(self):
        return self.next()

    def __len__(self):
        return len(self._list)

    def __str__(self):
        return str(self._list)

    def __repr__(self):
        return str(self)


class MultiCompMetric:

    def __init__(self, vals):
        self.vals = vals

    def __str__(self):
        return "({})".format(" ".join([str(v) for v in self.vals]))

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        if isinstance(other, MultiCompMetric):
            return all(v1 == v2 for v1, v2 in zip(self.vals, other.vals))
        else:
            return all(v == other for v in self.vals)

    def __ge__(self, other):
        if isinstance(other, MultiCompMetric):
            return all(v1 >= v2 for v1, v2 in zip(self.vals, other.vals))
        else:
            return all(v >= other for v in self.vals)


def greedy_core_set_selection(unique_points, size, dist_func, verbose=True):
    """
    Naive implementation of algorithm 1 in SmallGAN
    """
    assert len(unique_points) > size, \
        "Specified core set size: {} larger than number of points: {}".format(size, len(unique_points))
    bar = None
    if verbose:
        from tqdm import tqdm
        bar = tqdm(total=size, desc="Greedy core set selection", ascii=True)
    remaining_points = [v for v in unique_points]
    random.shuffle(remaining_points)
    core_set = [remaining_points[0]]
    remaining_points = remaining_points[1:]
    memo = {}
    while len(core_set) < size:
        max_dist = None
        cand_point_idx = None
        for pi, p1 in enumerate(remaining_points):
            min_dist = min(dist_func(p1, p2, memo) for p2 in core_set)
            if max_dist is None or min_dist > max_dist:
                max_dist = min_dist
                cand_point_idx = pi
        assert cand_point_idx is not None
        core_set.append(remaining_points[cand_point_idx])
        if bar is not None: bar.update(1)
        del remaining_points[cand_point_idx]
    if bar is not None: bar.close()
    return core_set


def ranked_select(data_list, key_func, n):
    assert len(data_list[1:-1]) > n > 2
    data_list.sort(key=key_func, reverse=True)
    sel_interval = math.ceil(len(data_list) / n)
    remaining, selected = [], []
    for i, v in enumerate(data_list[1:-1]):
        if i % sel_interval == 0:
            selected.append(v)
        else:
            remaining.append(v)
    while len(selected) < n:
        random.shuffle(remaining)
        selected.append(remaining[-1])
        del remaining[-1]
    while len(selected) > n:
        random.shuffle(selected)
        remaining.append(selected[-1])
        del selected[-1]
    remaining.append(data_list[0])
    remaining.append(data_list[-1])
    assert len(selected) == n
    assert len(selected) + len(remaining) == len(data_list)
    random.shuffle(selected)
    random.shuffle(remaining)
    return selected, remaining


class RunningStatMeter(object):

    def __init__(self):
        self.avg = 0.
        self.max = float("-inf")
        self.min = float("inf")
        self.sum = 0.
        self.cnt = 0

    def reset(self):
        self.avg = 0
        self.sum = 0
        self.cnt = 0
        return self

    def update(self, val, n=1):
        self.sum += val * n
        self.cnt += n
        self.avg = self.sum / self.cnt
        self.max = max(self.max, val)
        self.min = min(self.min, val)


if __name__ == "__main__":
    # gen_log_scale_points(0, math.pi, 10)
    # print(get_output_node_indices([[0,0,1,0,1,2,2,3,0,1,3,4,0,3,4], [1,2,2,3,3,3,4,4,4,5,5,5,5,6,6,6]]))
    # print(get_input_node_indices([[0,2,3], [1,3,4]]))
    # print(has_cycle([0,0,0,1,4,3], [0,1,2,3,1,4]))
    # print(is_complete_dag(6, [[0, 1, 2, 3, 4], [1, 2, 3, 4, 5]]))
    # print(is_complete_dag(4, [[0, 1, 2, 1, 0, 0], [1, 2, 3, 3, 3, 2]]))

    # _s, _r = ranked_select(list(range(773)), None, 50)

    # print("Before: {}, after: {}".format(0.9431, exp_scale(0.9431, old_range_min=0., old_range_max=0.9431)))
    # print("Before: {}, after: {}".format(0.9423, exp_scale(0.9423, old_range_min=0., old_range_max=0.9431)))
    # print("Before: {}, after: {}".format(0.9422, exp_scale(0.9422, old_range_min=0., old_range_max=0.9431)))
    # print("Before: {}, after: {}".format(0.9401, exp_scale(0.9401, old_range_min=0., old_range_max=0.9431)))
    # print("Before: {}, after: {}".format(0.9404, exp_scale(0.9404, old_range_min=0., old_range_max=0.9431)))
    # print("Before: {}, after: {}".format(0.94, exp_scale(0.94, old_range_min=0., old_range_max=0.9431)))
    # print("Before: {}, after: {}".format(0.939, exp_scale(0.939, old_range_min=0., old_range_max=0.9431)))
    # print("Before: {}, after: {}".format(0.936, exp_scale(0.936, old_range_min=0., old_range_max=0.9431)))
    # print("Before: {}, after: {}".format(0.933, exp_scale(0.933, old_range_min=0., old_range_max=0.9431)))
    # print("Before: {}, after: {}".format(0.92, exp_scale(0.92, old_range_min=0., old_range_max=0.9431)))
    # print("Before: {}, after: {}".format(0.90, exp_scale(0.90, old_range_min=0., old_range_max=0.9431)))
    # print("Before: {}, after: {}".format(0.88, exp_scale(0.88, old_range_min=0., old_range_max=0.9431)))

    # print(global_max([0,[1,2,3],1,2,[4,[[[5,6,7]]]]]))
    # print(global_min([0,[1,2,3],1,2,[4,[[[5,6,-7]]]]]))

    # pics = recur_get_files_by_extension("png", "../logs")

    multiplier = 100
    print("dst=0.949, src=0.95, p_accept={}"
          .format(exp_scale(0.949, multiplier=multiplier)/exp_scale(0.95, multiplier=multiplier)))
    print("dst=0.945, src=0.95, p_accept={}"
          .format(exp_scale(0.945, multiplier=multiplier)/exp_scale(0.95, multiplier=multiplier)))
    print("dst=0.943, src=0.95, p_accept={}"
          .format(exp_scale(0.943, multiplier=multiplier)/exp_scale(0.95, multiplier=multiplier)))
    print("dst=0.94, src=0.95, p_accept={}"
          .format(exp_scale(0.94, multiplier=multiplier)/exp_scale(0.95, multiplier=multiplier)))
    print("dst=0.93, src=0.95, p_accept={}"
          .format(exp_scale(0.93, multiplier=multiplier)/exp_scale(0.95, multiplier=multiplier)))
    print("dst=0.92, src=0.95, p_accept={}"
          .format(exp_scale(0.92, multiplier=multiplier)/exp_scale(0.95, multiplier=multiplier)))
    print("dst=0.89, src=0.95, p_accept={}"
          .format(exp_scale(0.89, multiplier=multiplier)/exp_scale(0.95, multiplier=multiplier)))

    print("done")
