import copy
import math
import torch
import random
import torch.nn as nn
import numpy as np
from constants import *
from utils.misc_utils import write_line_to_file


DEVICE_STR_OVERRIDE = None
CUDA_DEVICE_ID = "cuda"


def device(ref_tensor=None):
    if ref_tensor is not None:
        return ref_tensor.get_device()
    if DEVICE_STR_OVERRIDE is None:
        return torch.device(CUDA_DEVICE_ID if torch.cuda.is_available() else "cpu")
    else:
        return torch.device(DEVICE_STR_OVERRIDE)


def can_parallel(device_str=None):
    if not torch.cuda.is_available():
        return False
    if torch.cuda.device_count() <= 1:
        return False
    return device_str is None or device_str.lower() == "cuda"


def get_flops(model, C=3, W=32, H=32):
    from thop import profile
    rand_input = torch.randn(1, C, W, H)
    macs, params = profile(model, inputs=(rand_input, ))
    return macs


def config_bn(model):
    def _recur_config(op):
        if isinstance(op, torch.nn.BatchNorm2d):
            # op.momentum = 0.1 # default
            op.momentum = (1 - 0.99) # 0.99 momentum in tf corresponds to 1 - 0.99 in torch
            op.track_running_stats = False
    recur_tsr_op(model, _recur_config)


def set_full_train_bn(model):
    def _recur_config(op):
        if isinstance(op, torch.nn.BatchNorm2d):
            op.momentum = 0.01 # 0.99 momentum in tf corresponds to 1 - 0.99 in torch
            op.track_running_stats = True
    recur_tsr_op(model, _recur_config)


def recur_tsr_op(model, op):
    for child in model.children():
        op(child)
        recur_tsr_op(child, op)


def set_random_seed(seed, log_f=print):
    torch.manual_seed(seed)
    log_f('My seed is {}'.format(torch.initial_seed()))
    if torch.cuda.is_available():
        torch.cuda.manual_seed(seed)
        log_f('My cuda seed is {}'.format(torch.cuda.initial_seed()))
    np.random.seed(seed)
    random.seed(seed)


def get_column_wise_batched_tensors(batch_list):
    """
    Input: [(tsr1, tsr2, tsr3), (tsr4, tsr5, tsr6)]
    Return: [batch(tsr1, tsr4), batch(tsr2, tsr5), batch(tsr3, tsr6)]
    """
    tmp = []
    for i in range(len(batch_list[0])):
        tmp.append([v[i] for v in batch_list])
    return [torch.cat([t for t in v], dim=0) for v in tmp]


def to_sorted_tensor(tensor, lengths, sort_dim):
    """
    Sort tensor according to sequence lengths.
    This is used before applying pack_padded_sequence.
    :param tensor: input tensor
    :param lengths: 1D tensor of sequence lengths
    :param sort_dim: the dimension to sort in input tensor
    :param device: calculation device
    :return: sorted tensor, sorted lengths, and sorted index.
    """
    # rng = list(range(lengths.size(0)))
    # original_idx = torch.LongTensor(rng).unsqueeze(1).to(device())
    # lengths = lengths.unsqueeze(1)
    # lengths = torch.cat([original_idx, lengths], dim=1)
    sorted_lengths, sorted_idx = torch.sort(lengths.long(), dim=0, descending=True)
    sorted_idx = sorted_idx.to(device())
    sorted_tensor = tensor.index_select(dim=sort_dim, index=sorted_idx)
    return sorted_tensor, sorted_lengths, sorted_idx


def to_original_tensor(sorted_tensor, sorted_idx, sort_dim):
    """
    Restore tensor to its original order.
    This is used after applying pad_packed_sequence.
    :param sorted_tensor: a sorted tensor
    :param sorted_idx: sorted index of the sorted_tensor
    :param sort_dim: the dimension of sorted_tensor where it is sorted
    :device: calculation device
    :return: the original unsorted tensor
    """
    rng = range(sorted_idx.size(0))
    tmp = sorted(rng, key=lambda x: sorted_idx[x])
    original_idx = torch.LongTensor(tmp).to(device())
    tensor = sorted_tensor.index_select(dim=sort_dim, index=original_idx)
    return tensor


def get_masked_nll_criterion(vocab_size, pad_idx=0, reduction="sum"):
    weight = torch.ones(vocab_size)
    weight[pad_idx] = 0
    crit = nn.NLLLoss(reduction=reduction, weight=weight)
    return crit


def get_masked_ce_criterion(vocab_size, pad_idx=0, reduction="sum"):
    weight = torch.ones(vocab_size)
    weight[pad_idx] = 0
    crit = nn.CrossEntropyLoss(reduction=reduction, weight=weight)
    return crit


def fix_model_weights(model):
    for param in model.parameters():
        param.requires_grad = False


def unfix_model_weights(model):
    for param in model.parameters():
        param.requires_grad = True


def init_weights(model, base_model_type=None):
    if base_model_type == "rnn":
        for p in filter(lambda pa: pa.requires_grad, model.parameters()):
            if p.dim() == 1:
                p.data.normal_(0, math.sqrt(6 / (1 + p.size(0))))
            else:
                nn.init.xavier_normal_(p, math.sqrt(3))
    elif base_model_type == "transformer":
        for p in model.parameters():
            if p.dim() > 1:
                nn.init.xavier_uniform_(p)


def parallel(model):
    if hasattr(model, "avoid_parallel") and model.avoid_parallel:
        return model
    return nn.DataParallel(model)


def _save_model(file_path, epoch, model, optimizer, metadata={}):
    sv = {
        CHKPT_COMPLETED_EPOCHS: epoch,
        CHKPT_MODEL: model.state_dict(),
        CHKPT_OPTIMIZER: optimizer.state_dict(),
        CHKPT_METADATA: metadata
    }
    torch.save(sv, file_path)


def load_transformer_dict(file_path):
    return torch.load(file_path, map_location=lambda storage,loc:storage)


def model_checkpoint(file_path, epoch, model, optimizer, params,
                     past_eval_results, best_eval_result, best_eval_epoch, metadata={}):
    print("saving model to {}, please do not terminate".format(file_path))
    sv = {
        CHKPT_COMPLETED_EPOCHS: epoch,
        CHKPT_MODEL: model.state_dict(),
        CHKPT_OPTIMIZER: optimizer.state_dict(),
        CHKPT_PARAMS: params,
        CHKPT_PAST_EVAL_RESULTS: past_eval_results,
        CHKPT_BEST_EVAL_RESULT: best_eval_result,
        CHKPT_BEST_EVAL_EPOCH: best_eval_epoch,
        CHKPT_METADATA: metadata
    }
    model_save(file_path, sv)
    print("checkpoint done")


def model_save(file_path, data):
    torch.save(data, file_path)


def model_load(file_path):
    return torch.load(file_path, map_location=lambda storage,loc:storage)


def check_nan(tsr, chkpt="0", terminate=True):
    if not isinstance(tsr, torch.Tensor): return False
    if torch.isnan(tsr).any():
        if terminate:
            assert False, "nan chkpt {}: nan detected in tensor".format(chkpt)
        else:
            print("nan chkpt {}: nan detected in tensor".format(chkpt))
        return True
    return False


def inspect_tensor(tsr, name="tensor", logs_dir="logs/", verbose=False):
    opf = logs_dir + name + "_report.txt"
    torch.set_printoptions(threshold=5000)
    _output_tensor_stats(tsr, opf, verbose=verbose)
    check_nan(tsr, chkpt=name)


def _output_tensor_stats(tsr, opf, verbose=False):
    write_line_to_file(str(tsr), opf, verbose=verbose)
    max_idx, max_val = tsr.topk(1)
    write_line_to_file("max idx in tsr: {}".format(max_idx.item()), opf, verbose=verbose)
    write_line_to_file("max val in tsr: {}".format(max_val.item()), opf, verbose=verbose)
    min_val = tsr.min(1)
    write_line_to_file("min val in tsr: {}".format(min_val.item()), opf, verbose=verbose)
    unique_vals = torch.unique(tsr, sorted=True).tolist()
    write_line_to_file("unique values in tensor: {}".format(str(unique_vals)), verbose=verbose)
    n_zeros = torch.sum((tsr == 0).int()).item()
    write_line_to_file("number of zeros in tensor: {}".format(n_zeros), verbose=verbose)


def inspect_tensors(named_tuples, fn="tensors", logs_dir="logs/", verbose=False):
    opf = logs_dir + fn + "_report.txt"
    torch.set_printoptions(threshold=5000)
    for name, tsr in named_tuples:
        write_line_to_file(name, opf, verbose=verbose)
        _output_tensor_stats(tsr, opf, verbose=verbose)
        write_line_to_file("", opf, verbose=verbose)
        check_nan(tsr, chkpt=name)


def count_zero_weights(model):
    zeros = 0
    for param in model.parameters():
        if param is not None:
            zeros += torch.sum((param == 0).int()).data[0]
    return zeros


def inspect_model(model, name="model", logs_dir="logs/", verbose=False):
    opf = logs_dir + name + "_report.txt"
    write_line_to_file("model name: {}".format(name), opf, verbose=verbose)
    _output_component_stats(model, opf, verbose=verbose)


def _output_component_stats(net, opf, verbose=False):
    write_line_to_file(str(net), opf, verbose=verbose)
    n_params = sum(p.numel() for p in net.parameters() if p.requires_grad)
    write_line_to_file("number of trainable parameters: {}".format(n_params), opf, verbose=verbose)
    zero_weights = count_zero_weights(net)
    write_line_to_file("current zeros weights: {}".format(zero_weights), opf, verbose=verbose)


def step_with_check(step_func, model, terminate=True, report_dict={}):
    a = copy.deepcopy(list(filter(lambda pa: pa[1].requires_grad, model.named_parameters())))
    step_func()
    b = copy.deepcopy(list(filter(lambda pa: pa[1].requires_grad, model.named_parameters())))
    not_changed = []
    named_comps = []
    for i, tup in enumerate(a):
        name1, p1 = tup
        name2, p2 = b[i]
        assert name1 == name2
        same = torch.equal(p1.data, p2.data)
        not_changed.append(same)
        named_comps.append([name1, name2, same])
        if same:
            if name1 not in report_dict: report_dict[name1] = 0
            report_dict[name1] += 1
    if all(not_changed):
        if terminate: assert False, "error: all weights not updated after step()"
        else: print("warning: all weights not updated after step()")
    return named_comps


def report_torch_tensor_info(tensor, name="", breakpoint_id="0", busy_wait=True,
                             terminate=False):
    print("Tensor report breakpoint {} reached".format(breakpoint_id))
    print("Tensor name: {}".format(name))
    print("Tensor shape: {}".format(tensor.shape))
    print("Tensor type: {}".format(str(tensor.dtype)))
    print("Tensor device: {}".format(str(tensor.device)))
    print("Tensor requires_grad: {}".format(str(tensor.requires_grad)))
    print("Tensor raw str value:\n {}".format(str(tensor)))
    if terminate:
        exit(-1)
    if busy_wait:
        val = input("Press Enter to continue, Type q to terminate...\n")
        if val.lower() == "q": exit(1)


class RandomDataLoader:

    def __init__(self, n_instances, batch_size, feature_size, n_classes,
                 seq_len=1, regression_target=False, std_dev=0.5, multiplier=1):
        self.n_instances = n_instances
        self.batch_size = batch_size
        self.n_classes = n_classes
        self.regression_target = regression_target
        self.feature_size = feature_size
        self.seq_len = seq_len
        self.curr_batch_idx = 0
        self.multiplier = multiplier
        self._std_dev = std_dev
        self.batches = []
        self.build_batches()

    def _rand_inst_feature(self, center):
        rv = []
        for si in range(self.seq_len):
            f = torch.FloatTensor(np.random.normal(center, self._std_dev, (1, self.feature_size)))
            rv.append(f)
        return torch.cat(rv, dim=0).unsqueeze(0)

    def _rand_inst_target(self, center):
        if self.regression_target:
            tgt = torch.FloatTensor(np.random.normal(center, self._std_dev, (1, 1))).squeeze(0)
            tgt /= self.n_classes - 1
        else:
            return torch.LongTensor([center])

    def build_batches(self):
        max_f = 0
        instances = []
        inst_per_class = self.n_instances // self.n_classes
        for c in range(self.n_classes):
            for i in range(inst_per_class):
                f = self._rand_inst_feature(c * self.multiplier)
                max_f = max(torch.max(f).item(), max_f)
                t = self._rand_inst_target(c)
                instances.append( (f, t) )
        random.shuffle(instances)
        b_feature, b_tgt = [], []
        for f, t in instances:
            if max_f > 0: f /= max_f
            b_feature.append(f)
            b_tgt.append(t)
            if len(b_feature) == self.batch_size:
                self.batches.append( (torch.cat(b_feature, dim=0), torch.cat(b_tgt, dim=0)) )
                b_feature, b_tgt = [], []

    def shuffle(self):
        random.shuffle(self.batches)

    @property
    def num_batches(self):
        return len(self)

    @property
    def num_instances(self):
        return self.n_instances

    def __repr__(self):
        return str(self)

    def __str__(self):
        return "RandomDataLoader with {} batches".format(self.num_batches)

    def __len__(self):
        return len(self.batches)

    def __iter__(self):
        self.curr_batch_idx = 0
        return self

    def __next__(self):
        if self.curr_batch_idx < self.num_batches:
            batch = self.batches[self.curr_batch_idx]
            self.curr_batch_idx += 1
            return batch
        else:
            self.curr_batch_idx = 0
            raise StopIteration


def add_weight_decay(model, weight_decay=1e-5, skip_list=()):
    decay = []
    no_decay = []
    for name, param in model.named_parameters():
        if not param.requires_grad:
            continue
        if len(param.shape) == 1 or name in skip_list:
            no_decay.append(param)
        else:
            decay.append(param)
    return [
        {'params': no_decay, 'weight_decay': 0.},
        {'params': decay, 'weight_decay': weight_decay}]


def weight_init(m):
    import torch.nn.init as init
    '''
    https://gist.github.com/jeasinema/ed9236ce743c8efaf30fa2ff732749f5
    Usage:
        model = Model()
        model.apply(weight_init)
    '''
    if isinstance(m, nn.Conv1d):
        init.normal_(m.weight.data)
        if m.bias is not None:
            init.normal_(m.bias.data)
    elif isinstance(m, nn.Conv2d):
        init.xavier_normal_(m.weight.data)
        if m.bias is not None:
            init.normal_(m.bias.data)
    elif isinstance(m, nn.Conv3d):
        init.xavier_normal_(m.weight.data)
        if m.bias is not None:
            init.normal_(m.bias.data)
    elif isinstance(m, nn.ConvTranspose1d):
        init.normal_(m.weight.data)
        if m.bias is not None:
            init.normal_(m.bias.data)
    elif isinstance(m, nn.ConvTranspose2d):
        init.xavier_normal_(m.weight.data)
        if m.bias is not None:
            init.normal_(m.bias.data)
    elif isinstance(m, nn.ConvTranspose3d):
        init.xavier_normal_(m.weight.data)
        if m.bias is not None:
            init.normal_(m.bias.data)
    elif isinstance(m, nn.BatchNorm1d):
        if m.weight is not None:
            init.normal_(m.weight.data, mean=1, std=0.02)
            init.constant_(m.bias.data, 0)
    elif isinstance(m, nn.BatchNorm2d):
        if m.weight is not None:
            init.normal_(m.weight.data, mean=1, std=0.02)
            init.constant_(m.bias.data, 0)
    elif isinstance(m, nn.BatchNorm3d):
        if m.weight is not None:
            init.normal_(m.weight.data, mean=1, std=0.02)
            init.constant_(m.bias.data, 0)
    elif isinstance(m, nn.Linear):
        init.xavier_normal_(m.weight.data)
        init.normal_(m.bias.data)
    elif isinstance(m, nn.LSTM):
        for param in m.parameters():
            if len(param.shape) >= 2:
                init.orthogonal_(param.data)
            else:
                init.normal_(param.data)
    elif isinstance(m, nn.LSTMCell):
        for param in m.parameters():
            if len(param.shape) >= 2:
                init.orthogonal_(param.data)
            else:
                init.normal_(param.data)
    elif isinstance(m, nn.GRU):
        for param in m.parameters():
            if len(param.shape) >= 2:
                init.orthogonal_(param.data)
            else:
                init.normal_(param.data)
    elif isinstance(m, nn.GRUCell):
        for param in m.parameters():
            if len(param.shape) >= 2:
                init.orthogonal_(param.data)
            else:
                init.normal_(param.data)
    else:
        return False
    return True


class ExponentialMovingAverage:
    """
    Maintains (exponential) moving average of a set of parameters.
    https://github.com/fadel/pytorch_ema/blob/master/torch_ema/ema.py
    """
    def __init__(self, parameters, decay, use_num_updates=True):
        """
        Args:
          parameters: Iterable of `torch.nn.Parameter`; usually the result of
            `model.parameters()`.
          decay: The exponential decay.
          use_num_updates: Whether to use number of updates when computing
            averages.
        """
        if decay < 0.0 or decay > 1.0:
            raise ValueError('Decay must be between 0 and 1')
        self.decay = decay
        self.num_updates = 0 if use_num_updates else None
        self.shadow_params = [p.clone().detach()
                              for p in parameters if p.requires_grad]

    def update(self, parameters):
        """
        Update currently maintained parameters.
        Call this every time the parameters are updated, such as the result of
        the `optimizer.step()` call.
        Args:
          parameters: Iterable of `torch.nn.Parameter`; usually the same set of
            parameters used to initialize this object.
        """
        decay = self.decay
        if self.num_updates is not None:
            self.num_updates += 1
            decay = min(decay, (1 + self.num_updates) / (10 + self.num_updates))
        one_minus_decay = 1.0 - decay
        with torch.no_grad():
            parameters = [p for p in parameters if p.requires_grad]
            for s_param, param in zip(self.shadow_params, parameters):
                s_param.sub_(one_minus_decay * (s_param - param))

    def copy_to(self, parameters):
        """
        Copies current parameters into given collection of parameters.
        Args:
          parameters: Iterable of `torch.nn.Parameter`; the parameters to be
            updated with the stored moving averages.
        """
        for s_param, param in zip(self.shadow_params, parameters):
            if param.requires_grad:
                param.data.copy_(s_param.data)


if __name__ == "__main__":
    loader = RandomDataLoader(50000, 512, 100, 10)
    print("done")
